<?php

namespace App\Http\Controllers;

use App\Custom\RestInterface;
use App\Custom\ElasticsearchInterface;
use App\Providers\RestServiceProvider;
use Illuminate\Http\Request;
use App\Corpus;
use App\Document;
use App\Annotation;
use App\CorpusProject;
use App\CorpusFile;
use App\Format;
use App\Custom\GitRepoInterface;
use App\Custom\LaudatioUtilsInterface;
use GrahamCampbell\Flysystem\FlysystemManager;
use Response;
use JavaScript;
use Cache;
use Log;
use League\Flysystem\Cached\Storage;
use Illuminate\Support\Facades\Artisan;

use Illuminate\Database\Eloquent\ModelNotFoundException;
/**
 * @group Corpus data Endpoints
 *
 * APIs for Corpus data
 *
 * Class ApiController
 * @package App\Http\Controllers
 */
class ApiController extends Controller
{
    protected $LaudatioUtilService;
    protected $GitRepoService;
    protected $RestService;
    protected $flysystem;
    protected $Elasticservice;
    protected $current_corpus_index;
    protected $current_document_index;
    protected $current_annotation_index;
    protected $current_guideline_index;

    /**
     * ApiController constructor.
     * @param LaudatioUtilsInterface $laudatioUtils
     * @param GitRepoInterface $GitRepoService
     * @param FlysystemManager $flysystem
     */
    public function __construct(LaudatioUtilsInterface $laudatioUtils,GitRepoInterface $GitRepoService, RestInterface $restService, FlysystemManager $flysystem,ElasticsearchInterface $Elasticservice)
    {
        $this->LaudatioUtilService = $laudatioUtils;
        $this->GitRepoService = $GitRepoService;
        $this->flysystem = $flysystem;
        $this->RestService = $restService;
        $this->Elasticservice = $Elasticservice;
        $this->current_corpus_index =  config("laudatio.elasticsearch_corpus_index","corpora");
        $this->current_document_index = config("laudatio.elasticsearch_document_index","documents");
        $this->current_annotation_index =  config("laudatio.elasticsearch_annotation_index","annotations");
        $this->current_guideline_index = config("laudatio.elasticsearch_guideline_index","guidelines");
    }


    /** ELOQUENT MODEL REST API **/

    public function checkElasticSearchIds($corpusId) {
        $data = $this->LaudatioUtilService->checkForElasticsearchIds($corpusId);
        return $this->RestService->sendResponse($data, 'Data retrieved successfully.');
    }

    public function fixElasticSearchIds($corpusId) {
        $data = $this->Elasticservice->fixElasticsearchIds($corpusId,$this->current_corpus_index,$this->current_document_index,$this->current_annotation_index);
        return $this->RestService->sendResponse($data, 'Data fixed successfully.');
    }



    /**
     *
     * /api/corpusapi/corpusprojects
     * @return mixed
     */
    public function getCorpusProjects(){
        $projects = array();
        if (Cache::store('redis')->has('corpusprojects')) {
            $projects = Cache::store('redis')->get('corpusprojects');
        }
        else{
            $projects = CorpusProject::all();
            Cache::store('redis')->forever('corpusprojects', $projects);
        }

        return $this->RestService->sendResponse($projects->toArray(), count($projects->toArray()).' Projects retrieved successfully.');
    }


    /**
     * /api/corpusapi/corpusproject/{id}
     * @param  $id String required 
     * @return mixed
     */
    public function getCorpusProject($id){
        if (Cache::store('redis')->has('corpusproject_'.$id)) {
            $project = Cache::store('redis')->get('corpusproject_'.$id);
            return $this->RestService->sendResponse($project->toArray(), 'Project retrieved successfully.');
        }
        else{
            try {
                $project = CorpusProject::findOrFail($id);
                Cache::store('redis')->forever('corpusproject_'.$id, $project);
                return $this->RestService->sendResponse($project->toArray(), 'Project retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                 return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }


    /**
     * /api/corpusapi/corpusproject/{id}/corpora
     * @param  $id String required 
     * @return mixed
     */
    public function getCorporaByCorpusproject($id){
        if (Cache::store('redis')->has('corpora_by_corpusproject_'.$id)) {
            $corpora = Cache::store('redis')->get('corpora_by_corpusproject_'.$id);
            return $this->RestService->sendResponse($corpora->toArray(), count($corpora->toArray()).' Project corpora retrieved successfully.');
        }
        else{
            try {
                $corpora = CorpusProject::with("corpora")->find($id);
                Cache::store('redis')->forever('corpora_by_corpusproject_'.$id, $corpora);
                return $this->RestService->sendResponse($corpora->toArray(), count($corpora->toArray()).' Project corpora retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }

    /**
     *
     * /api/corpusapi/corpora
     * @return mixed
     */
    public function getCorpora() {
        $corpora = array();
        if (Cache::store('redis')->has('corpora')) {
            $corpora = Cache::store('redis')->get('corpora');
        }
        else{
            $corpora = Corpus::all();
            Cache::store('redis')->forever('corpora', $corpora);
        }

        return $this->RestService->sendResponse($corpora->toArray(), count($corpora->toArray()).' Corpora retrieved successfully.');
    }

    /**
     *
     * /api/corpusapi/corpora/{id}
     * @param  $id String required 
     * @return mixed
     */
    public function getCorpus($id) {
        if (Cache::store('redis')->has('corpora_'.$id)) {
            $corpus = Cache::store('redis')->get('corpora_'.$id);
            return $this->RestService->sendResponse($corpus->toArray(), 'Corpus retrieved successfully.');
        }
        else{
            try {
                $corpus = Corpus::findOrFail($id);
                Cache::store('redis')->forever('corpora_'.$id, $corpus);
                return $this->RestService->sendResponse($corpus->toArray(), 'Corpus retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }


    /**
     * /api/corpusapi/corpusproject/{id}/corpora
     * @param  $id String required 
     * @return mixed
     */
    public function getDocumentsByCorpus($id){
        if (Cache::store('redis')->has('documents_by_corpora_'.$id)) {
            $documents = Cache::store('redis')->get('documents_by_corpora_'.$id);
            return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Corpus documents retrieved successfully.');
        }
        else{
            try {
                $documents = Corpus::with("documents")->find($id);
                Cache::store('redis')->forever('documents_by_corpora_'.$id, $documents);
                return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Corpus documents retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }

    /**
     * /api/corpusapi/corpusproject/{id}/corpora
     * @param  $id String required 
     * @return mixed
     */
    public function getAnnotationsByCorpus($id){
        if (Cache::store('redis')->has('annotations_by_corpora_'.$id)) {
            $annotations = Cache::store('redis')->get('annotations_by_corpora_'.$id);
            return $this->RestService->sendResponse($annotations->toArray(), count($annotations->toArray()).' Corpus annotations  retrieved successfully.');
        }
        else{
            try {
                $annotations = Corpus::with("annotations")->find($id);
                Cache::store('redis')->forever('annotations_by_corpora_'.$id, $annotations);
                return $this->RestService->sendResponse($annotations->toArray(), count($annotations->toArray()).' Corpus annotations retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }

    /**
     *
     * /api/corpusapi/documents
     *
     * @return mixed
     */
    public function getDocuments(){
        $documents = array();
        if (Cache::store('redis')->has('documents')) {
            $documents = Cache::store('redis')->get('documents');
        }
        else{
            $documents = Document::all();
            Cache::store('redis')->forever('documents', $documents);
        }

        return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Documents retrieved successfully.');
    }

    /**
     *
     * /api/corpusapi/documents/{id}
     * @param  $id String required 
     * @return mixed
     */
    public function getDocument($id) {
        if (Cache::store('redis')->has('document_'.$id)) {
            $document = Cache::store('redis')->get('document_'.$id);
            return $this->RestService->sendResponse($document->toArray(), 'Document retrieved successfully.');
        }
        else{
            try {
                $document = Document::findOrFail($id);
                Cache::store('redis')->forever('document_'.$id, $document);
                return $this->RestService->sendResponse($document->toArray(), 'Document retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }


    /**
     * /api/corpusapi/documents/{id}/annotations
     * @param  $id String required 
     * @return mixed
     */
    public function getAnnotationsByDocument($id){
        if (Cache::store('redis')->has('annotations_by_document_'.$id)) {
            $documents = Cache::store('redis')->get('annotations_by_document_'.$id);
            return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Document annotations retrieved successfully.');
        }
        else{
            try {
                $documents = Document::with("annotations")->find($id);
                Cache::store('redis')->forever('annotations_by_document_'.$id, $documents);
                return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Document annotations retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }


    /**
     *
     * /api/corpusapi/annotations
     *
     * @return mixed
     */
    public function getAnnotations(){
        $annotations = array();
        if (Cache::store('redis')->has('annotations')) {
            $annotations = Cache::store('redis')->get('annotations');
        }
        else{
            $annotations = Annotation::all();
            Cache::store('redis')->forever('annotations', $annotations);
        }

        return $this->RestService->sendResponse($annotations->toArray(), count($annotations->toArray()).' Annotations retrieved successfully.');
    }



    /**
     *
     * /api/corpusapi/documents/{id}
     * @param  $id String required 
     * @return mixed
     */
    public function getAnnotation($id) {
        if (Cache::store('redis')->has('annotation_'.$id)) {
            $annotation = Cache::store('redis')->get('annotation_'.$id);
            return $this->RestService->sendResponse($annotation->toArray(), 'Annotation retrieved successfully.');
        }
        else{
            try {
                $annotation = Annotation::findOrFail($id);
                Cache::store('redis')->forever('annotation_'.$id, $annotation);
                return $this->RestService->sendResponse($annotation->toArray(), 'Annotation retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }


    /**
     * /api/corpusapi/annotations/{id}/documents
     * @param  $id String required 
     * @return mixed
     */
    public function getDocumentsByAnnotation($id){
        if (Cache::store('redis')->has('documents_by_annotation_'.$id)) {
            $documents = Cache::store('redis')->get('documents_by_annotation_'.$id);
            return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Annotation documents retrieved successfully.');
        }
        else{
            try {
                $documents = Annotation::with("documents")->find($id);
                Cache::store('redis')->forever('documents_by_annotation_'.$id, $documents);
                return $this->RestService->sendResponse($documents->toArray(), count($documents->toArray()).' Annotation documents retrieved successfully.');
            } catch (ModelNotFoundException $exception) {
                return $this->RestService->sendError('REST API Error.', $exception->getMessage(),404);
            }
        }
    }

    /*
    public function getFormats(){
        $formats = array();
        if (Cache::store('redis')->has('formats')) {
            $formats = Cache::store('redis')->get('formats');
        }
        else{
            $formats = Format::all();
            Cache::store('redis')->forever('formats', $formats);
        }

        return $this->RestService->sendResponse($formats->toArray(), count($formats->toArray()).' Formats retrieved successfully.');
    }
*/

    /** END ELOQUENT MODEL REST API **/



    /** START ELASTICSEARCH REST API **/

    /**
     * /api/elasticapi/v1/corpora
     *
     * @responseFile responses/corpora.json
     *
     * @return mixed
     */
    public function getElasticCorpora() {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $updatedcorpora = array();

            $corpora  = $this->Elasticservice->getAllPublishedCorpora();
            foreach($corpora['hits']['hits'] as $corpus) {
                if(isset($corpus['_index']) && isset($corpus['_id'])) {
                    $corpus['_source']['doi'] = $this->LaudatioUtilService->getDOIByCorpusElasticsearchId($corpus['_id'],$corpus['_index']);
                    array_push($updatedcorpora,$corpus);
                }
            }


            return $this->RestService->sendResponse($updatedcorpora, count($updatedcorpora).' Corpora retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/corpora/latest/count
     * @responseFile responses/corpora_count.json
     *
     * @return mixed
     */
    public function getCorporaCount(){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $index = config("laudatio.elasticsearch_corpus_index","corpora");

            $query = array(
                "query" => array(
                    "bool" => array(
                        "must" => array(
                            "exists" => array(
                                "field" => "is_latest_publication"
                            )
                        ),
                        "filter" => array(
                            "term" => array(
                                "is_latest_publication" => true
                            )
                        )
                    )
                )
            );

            $result = $this->Elasticservice->getHitsCount($index,json_encode($query));
            return $this->RestService->sendResponse($result['resultdata'], ' Latest corpora count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }



    /**
     *
     * /api/elasticapi/v1/documents/latest/count
     * @responseFile responses/document_count.json
     * @return mixed
     */
    public function getDocumentsCount(){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }
        try {
            //$indexes = $this->LaudatioUtilService->pickLatestDocuments();
            $indexes = config("laudatio.elasticsearch_document_index","documents");
            $query = array(
                "query" => array(
                    "bool" => array(
                        "must" => array(
                            "exists" => array(
                                "field" => "document_title"
                            )
                        ),
                        "filter" => array(
                            "term" => array(
                                "is_latest_publication" => true
                            )
                        )
                    )
                )
            );
            $result = $this->Elasticservice->getHitsCount($indexes,json_encode($query));
            return $this->RestService->sendResponse($result['resultdata'], ' Latest documents count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }

    /**
     *
     * /api/elasticapi/v1/annotations/latest/count
     * @responseFile responses/annotation_count.json
     * @return mixed
     */
    public function getAnnotationsCount(){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }
        try {
            //$indexes = $this->LaudatioUtilService->pickLatestAnnotations();
            $indexes = config("laudatio.elasticsearch_annotation_index","annotations");
            $query = array(
                "query" => array(
                    "bool" => array(
                        "must" => array(
                            "exists" => array(
                                "field" => "preparation_title"
                            )
                        ),
                        "filter" => array(
                            "term" => array(
                                "is_latest_publication" => true
                            )
                        )
                    )
                )
            );
            $result = $this->Elasticservice->getHitsCount($indexes,json_encode($query));
            return $this->RestService->sendResponse($result['resultdata'], ' Latest annotations count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }

    /**
     * /api/elasticapi/v1/corpora/latest/{from}/{size}
     *
     * @responseFile responses/latest_corpora_from_limit.json
     * @queryParam from required  The offset to start the result from
     * @queryParam size required  The limit of returned results
     *
     * @return mixed
     */
    public function getLatestElasticCorpora($from,$size) {

        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        if (Cache::store('redis')->has('latest_published_corpora_'.$from.'_'.$size)) {
            $resultData = Cache::store('redis')->get('latest_published_corpora_'.$from.'_'.$size);
            return $this->RestService->sendResponse($resultData, count($resultData).' Latest version of published Corpora retrieved successfully.');
        }
        else {
            try {
                $indexes = config("laudatio.elasticsearch_corpus_index","corpora");
                $source = [
                    "laudatio_corpus_id",
                    "corpus_title",
                    "corpus_editor_forename",
                    "corpus_editor_surname",
                    "corpus_publication_publisher",
                    "corpus_documents",
                    "corpus_merged_formats",
                    "corpus_encoding_tool",
                    "corpus_encoding_project_description",
                    "corpus_annotator_forename",
                    "corpus_annotator_surname",
                    "corpus_publication_license",
                    "has_documents",
                    "has_annotations",
                    "annotation_id",
                    "corpus_languages_language",
                    "corpus_merged_languages",
                    "corpus_languages_iso_code",
                    "corpus_publication_publication_date",
                    "corpus_size_type",
                    "corpus_size_value",
                    "corpus_document_genre",
                    "publication_status",
                    "publication_version"
                ];


                $corpora = $this->Elasticservice->getLatestPublishedCorpora($indexes,$source,$from,$size);

                $data = $corpora['hits']['hits'];

                for ($i = 0; $i < count($data); $i++) {
                    $doc = $data[$i];

                    if(count($doc) > 0){
                        $projectPath = $this->LaudatioUtilService->getCorpusProjectPathByCorpusId($doc['_id'],$doc["_index"]);
                        $corpusPath = $this->LaudatioUtilService->getCorpusPathByCorpusId($doc["_id"],$doc["_index"]);
                        $corpusLogo = $this->LaudatioUtilService->getCorpusLogoByCorpusId($doc["_id"],$doc["_index"]);
                        $corpusFiles = $this->LaudatioUtilService->getFormatFilesByCorpusId($doc['_source']['laudatio_corpus_id']);

                        $corpora['hits']['hits'][$i]['_source']['projectpath'] = $projectPath;
                        $corpora['hits']['hits'][$i]['_source']['corpuspath'] = $corpusPath;
                        $corpora['hits']['hits'][$i]['_source']['corpuslogo'] = $corpusLogo;
                        $corpora['hits']['hits'][$i]['_source']['corpusfiles'] = $corpusFiles;

                        $corpora['hits']['hits'][$i]['_source']['doi'] = $this->LaudatioUtilService->getDOIByCorpusElasticsearchId($doc['_id'],$doc["_index"]);
                        $documentResult = array();
                        if(isset($doc['_source']['has_documents'])){
                            $documentResult = $this->Elasticservice->getDocumentByCorpus(
                                $doc['_source']['has_documents'],
                                array($doc['_source']['laudatio_corpus_id']),
                                array("document_title","document_publication_publishing_date","document_publication_place","document_list_of_annotations_name","in_corpora","document_size_extent"),
                                $this->current_document_index
                            );
                        }

                        $docdata = array("result" => $doc['_source']);
                        $document_range = $this->LaudatioUtilService->getDocumentRange($docdata,$documentResult);
                        $corpora['hits']['hits'][$i]['_source']['documentrange'] = $document_range;

                        if(!isset($doc['_source']['corpus_document_genre'])) {
                            $corpora['hits']['hits'][$i]['_source']['corpus_document_genre'] = "NA";
                        }

                        if(isset($doc['_source']['corpus_merged_languages'])) {
                            $mergedLanguages = $doc['_source']['corpus_merged_languages'];
                            $corpora['hits']['hits'][$i]['_source']['corpus_merged_languages'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedLanguages));
                        }
                        /*
                        if(isset($doc['_source']['corpus_merged_formats'])) {
                            $mergedFormats = $doc['_source']['corpus_merged_formats'];
                            $corpora['hits']['hits'][$i]['_source']['corpus_merged_formats'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedFormats));
                        }
                        */

                        $corpora['hits']['hits'][$i]['_source']['visibility'] = 1;
                    }//end if _index
                }
                $resultData = $corpora['hits']['hits'];
                Cache::store('redis')->forever('latest_published_corpora_'.$from.'_'.$size, $resultData);
                return $this->RestService->sendResponse($resultData, count($resultData).' Latest version of published Corpora retrieved successfully.');
            } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
                return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
            }
        }

    }



    /**
     *
     * /api/elasticapi/v1/corpora/{id}/{index}
     *
     * @responseFile responses/corpus_by_id_and_index.json
     * @param  $id String required 
     * @param  $index String required 
     * @queryParam id required  The identifier of the document
     * @queryParam index required  The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticCorpus($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $corpus  = $this->Elasticservice->getPublishedCorpus($id,$index);
            $corpus['_source']['doi'] = $this->LaudatioUtilService->getDOIByCorpusElasticsearchId($corpus['_id'],$corpus['_index']);
            return $this->RestService->sendResponse($corpus, 'Corpus retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     * /api/elasticapi/v1/corpora/{id}/{index}/documents
     * @responseFile responses/corpusdocuments_by_id_and_index.json
     * @param  $id String required 
     * @param  $index String required 
     * @queryParam id required  The identifier of the document
     * @queryParam index required  The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticDocumentsByCorpus($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

        $corpus = $this->getElasticCorpus($id,$index);
        $corpus_id = $corpus->getData()->data->_source->corpus_id[0];

        $indexes = $this->LaudatioUtilService->transformElasticSearchIndexes($index);
        $documentIndex = config("laudatio.elasticsearch_document_index","documents");


        try {
            $documents = $this->Elasticservice->getAllPublishedDocumentsByCorpus($corpus_id,$documentIndex);
            return $this->RestService->sendResponse($documents, $documents['hits']['total'] . ' documents retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     * /api/elasticapi/v1/corpora/{id}/{index}/annotations
     * @responseFile responses/corpusannotations_by_id_and_index.json
     * @param  $id String required 
     * @param  $index String required 
     * @queryParam id required  The identifier of the document
     * @queryParam index required  The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticAnnotationsByCorpus($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

        $corpus = $this->getElasticCorpus($id,$index);
        $corpus_id = $corpus->getData()->data->_source->corpus_id[0];

        //$indexes = $this->LaudatioUtilService->transformElasticSearchIndexes($index);
        $annotationIndex = config("laudatio.elasticsearch_annotation_index","annotations");


        try {
            $annotations = $this->Elasticservice->getAllPublishedAnnotationsByCorpus($corpus_id,$annotationIndex);
            return $this->RestService->sendResponse($annotations, $annotations['hits']['total'] .' annotations retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }

    /**
     *
     * /api/elasticapi/v1/documents
     * @responseFile responses/documents.json
     * @return mixed
     */
    public function getElasticDocuments() {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

        try {
            $documents  = $this->Elasticservice->getAllPublishedDocuments();
            return $this->RestService->sendResponse($documents['hits']['hits'], count($documents['hits']['hits']).' Documents retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     * /api/elasticapi/v1/documents/latest/{from}/{size}
     * @responseFile responses/latest_documents_from_limit.json
     * @param $from
     * @param $size
     * @queryParam from required  The offset to start the result from
     * @queryParam size required  The limit of returned results
     *
     * @return mixed
     */
    public function getLatestElasticDocuments($from,$size) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }
        if (Cache::store('redis')->has('latest_published_documents_'.$from.'_'.$size)) {
            $resultData = Cache::store('redis')->get('latest_published_documents_'.$from.'_'.$size);
            return $this->RestService->sendResponse($resultData, count($resultData).' Latest version of published Documents retrieved successfully.');
        }
        else {
            try {
                $foundCorpusNames = array();
                $indexes =  config("laudatio.elasticsearch_document_index","documents");
                $source = [
                    "document_title",
                    "document_author_forename",
                    "document_author_surname",
                    "document_merged_authors",
                    "document_editor_forename",
                    "document_editor_surname",
                    "document_history_not_before",
                    "document_history_not_after",
                    "document_merged_languages",
                    "document_languages_iso_code",
                    "document_publication_place",
                    "document_publication_publishing_date",
                    "document_list_of_annotations_id",
                    "document_history_original_date",
                    "document_history_original_place",
                    "document_size_type",
                    "document_size_extent",
                    "document_genre",
                    "publication_status",
                    "laudatio_corpus_id"];


                $documents = $this->Elasticservice->getLatestPublishedDocuments($indexes,$source,$from,$size);

                $data = $documents['hits']['hits'];
                for ($i = 0; $i < count($data); $i++) {
                    $doc = $data[$i];
                    if(count($doc) > 0  && isset($doc["_index"])){
                        if(strpos($doc["_index"],"document") !== false) {
                            $corpusName = $this->LaudatioUtilService->getCorpusNameByObjectElasticsearchId('document',$doc['_id']);
                            $corpusElasticSearchId = $this->LaudatioUtilService->getElasticSearchIdByCorpusId($doc['_source']['laudatio_corpus_id']);
                            if(!empty($corpusName)){
                                $documents['hits']['hits'][$i]['_source']['corpus_name'] = $corpusName;
                                $documents['hits']['hits'][$i]['_source']['corpus_elasticsearch_id'] = $corpusElasticSearchId;
                                $documents['hits']['hits'][$i]['_source']['corpus_elasticsearch_index'] = $this->current_corpus_index;
                            }

                            if(isset($doc['_source']['document_merged_languages'])) {
                                $mergedLanguages = $doc['_source']['document_merged_languages'];
                                $documents['hits']['hits'][$i]['_source']['document_merged_languages'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedLanguages));
                            }
                        }

                        $documents['hits']['hits'][$i]['_source']['visibility'] = 1;
                    }//end if _index

                }
                $resultData = $documents['hits']['hits'];
                Cache::store('redis')->forever('latest_published_documents_'.$from.'_'.$size, $resultData);
                return $this->RestService->sendResponse($resultData, count($resultData).' Latest version of published Documents retrieved successfully.');
            } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
                return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
            }
        }

    }

    /**
     *
     * /api/elasticapi/v1/documents/{id}/{index}
     *
     * @param $id
     * @param  $index String required
     * @responseFile responses/document_by_id_and_index.json
     * @queryParam id required The identifier of the document
     * @queryParam index required The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticDocument($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $document  = $this->Elasticservice->getPublishedDocument($id,$index);
            return $this->RestService->sendResponse($document, 'Document retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/documents/{id}/{index}/annotations
     *
     * @responseFile responses/documentannotations_by_id_and_index.json
     * @param  $id String required 
     * @param  $index String required 
     * @queryParam id required  The identifier of the document
     * @queryParam index required  The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticAnnotationsByDocument($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $annotations  = $this->Elasticservice->getAllPublishedAnnotationsByDocument($id,$index);
            return $this->RestService->sendResponse($annotations, count($annotations). ' annotations retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/annotations
     * @responseFile responses/annotations.json
     * @return mixed
     */
    public function getElasticAnnotations() {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

        try {
            $annotations  = $this->Elasticservice->getAllPublishedAnnotations();
            return $this->RestService->sendResponse($annotations['hits']['hits'], count($annotations['hits']['hits']).' annotations retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/annotations/latest/{from}/{size}
     *
     * @param $from
     * @param $size
     * @responseFile responses/latest_annotations_from_limit.json
     * @queryParam from required  The offset to start the result from
     * @queryParam size required  The limit of returned results
     *
     * @return mixed
     */
    public function getLatestElasticAnnotations($from,$size) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        if (Cache::store('redis')->has('latest_published_annotations_'.$from.'_'.$size)) {
            $resultData = Cache::store('redis')->get('latest_published_annotations_'.$from.'_'.$size);
            return $this->RestService->sendResponse($resultData, count($resultData).' Latest version of published Annotations retrieved successfully.');
        }
        else {

            try {
                $foundCorpusNames = array();
                //$indexes = $this->LaudatioUtilService->pickLatestAnnotations();
                $indexes =  config("laudatio.elasticsearch_annotation_index","annotations");

                $source = [
                    "preparation_title",
                    "preparation_annotation_id",
                    "preparation_encoding_annotation_group",
                    "preparation_encoding_annotation_sub_group",
                    "preparation_encoding_full_name",
                    "annotation_merged_formats",
                    "in_documents",
                    "publication_status",
                    "laudatio_corpus_id"];


                $annotations = $this->Elasticservice->getLatestPublishedAnnotations($indexes,$source,$from,$size);

                $data = $annotations['hits']['hits'];

                for ($i = 0; $i < count($data); $i++) {
                    $doc = $data[$i];
                    if(count($doc) > 0  && isset($doc["_index"])){
                        if(strpos($doc["_index"],"annotation") !== false) {

                            $corpusName = $this->LaudatioUtilService->getCorpusNameByObjectElasticsearchId('annotation',$doc['_id']);
                            $corpusElasticSearchId = $this->LaudatioUtilService->getElasticSearchIdByCorpusId($doc['_source']['laudatio_corpus_id']);
                            if(!empty($corpusName)) {
                                $annotations['hits']['hits'][$i]['_source']['corpus_name'] = $corpusName;
                                $annotations['hits']['hits'][$i]['_source']['corpus_elasticsearch_id'] = $corpusElasticSearchId;
                                $annotations['hits']['hits'][$i]['_source']['corpus_elasticsearch_index'] = $this->current_corpus_index;
                            }

                            /*
                            if(isset($doc['_source']['annotation_merged_formats'])) {
                              $mergedFormats = $doc['_source']['annotation_merged_formats'];
                              $annotations['hits']['hits'][$i]['_source']['annotation_merged_formats'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedFormats));
                            }
                            */
                        }

                        $annotations['hits']['hits'][$i]['_source']['visibility'] = 1;
                    }//end if _index

                }
                //Log::info("ANNOS? ".print_r($data,1));
                $resultData = $annotations['hits']['hits'];
                Cache::store('redis')->forever('latest_published_annotations_'.$from.'_'.$size, $resultData);
                return $this->RestService->sendResponse($resultData, count($resultData).' Latest version of published Annotations retrieved successfully.');
            } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
                return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
            }
        }

    }


    /**
     *
     * /api/elasticapi/v1/corpora/latest/searchMain
     *
     * @param Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @responseFile responses/latest_corpus_searchMain.json
     * @bodyParam Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @bodyParam  from String required  The offset to start the result from
     * @bodyParam  size String required  The limit of returned results
     * @bodyParam  query String required  The search query string
     *
     * @return mixed
     */
    public function corpusSearchMain(Request $request){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {

            $searchData = array();
            $hasWildCards = false;

            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            if(strpos($searchData['query'],'?') > -1 || strpos($searchData['query'],'*') > -1){
                $hasWildCards = true;
            }

            $indexes = config("laudatio.elasticsearch_corpus_index","corpora");

            $source = [
                "corpus_title",
                "corpus_editor_forename",
                "corpus_editor_surname",
                "corpus_annotator_forename",
                "corpus_annotator_surname",
                "corpus_publication_publisher",
                "corpus_merged_formats",
                "annotation_encoding_tool",
                "corpus_encoding_project_description",
                "corpus_publication_license",
                "corpus_merged_languages",
                "corpus_languages_iso_identifier",
                "corpus_publication_publication_date.string",
                //"corpus_size_value",
                "corpus_size_type",
                "corpus_document_genre"
            ];

            $searchData["fields"] = $source;

            //re-add annotation_id
            if(!array_key_exists("corpus_publication_publication_date",$source)){
                array_push($source,"corpus_publication_publication_date");
            }

            if(!array_key_exists("corpus_size_value",$source)){
                array_push($source,"corpus_size_value");
            }

            //re-add annotation_id
            if(!array_key_exists("annotation_id",$source)){
                array_push($source,"annotation_id");
            }

            //re-add corpus_id
            if(!array_key_exists("corpus_id",$source)){
                array_push($source,"corpus_id");
            }

            //re-add has_documents
            if(!array_key_exists("has_documents",$source)){
                array_push($source,"has_documents");
            }

            //re-add has_annotations
            if(!array_key_exists("has_annotations",$source)){
                array_push($source,"has_annotations");
            }

            //re-add laudatio_corpus_id
            if(!array_key_exists("laudatio_corpus_id",$source)){
                array_push($source,"laudatio_corpus_id");
            }

            //re-add publication_status
            if(!array_key_exists("publication_status",$source)){
                array_push($source,"publication_status");
            }

            //re-add corpus_documents
            if(!array_key_exists("corpus_documents",$source)){
                array_push($source,"corpus_documents");
            }

            if($hasWildCards){
                $result  = $this->Elasticservice->searchMainWildCards($searchData,$indexes,$source,$searchData["from"],$searchData["size"]);
            }
            else {
                $result  = $this->Elasticservice->searchMain($searchData,$indexes,$source,$searchData["from"],$searchData["size"]);
            }


            $data = $result['hits']['hits'];

            for ($i = 0; $i < count($data); $i++) {
                $doc = $data[$i];
                if(count($doc) > 0){

                    $projectPath = $this->LaudatioUtilService->getCorpusProjectPathByCorpusId($doc['_id'],$doc["_index"]);
                    $corpusPath = $this->LaudatioUtilService->getCorpusPathByCorpusId($doc["_index"],$doc["_index"]);
                    $corpusLogo = $this->LaudatioUtilService->getCorpusLogoByCorpusId($doc["_index"],$doc["_index"]);

                    $result['hits']['hits'][$i]['_source']['projectpath'] = $projectPath;
                    $result['hits']['hits'][$i]['_source']['corpuspath'] = $corpusPath;
                    $result['hits']['hits'][$i]['_source']['corpuslogo'] = $corpusLogo;


                    if(isset($doc['highlight'])){
                        $newHighlights = array();
                        foreach ($doc['highlight'] as $highlightName => $highlightData){
                            $newHighlightName = str_replace(".string", "",$highlightName);
                            $newHighlights[$newHighlightName] = $highlightData;
                        }

                        $result['hits']['hits'][$i]['highlight'] = $newHighlights;
                    }

                    $documentResult = array();

                    if(isset($doc['_source']['has_documents'])){
                        $documentResult = $this->Elasticservice->getDocumentByCorpus(
                            $doc['_source']['has_documents'],
                            array($doc['_source']['laudatio_corpus_id']),
                            array("document_title","document_publication_publishing_date","document_publication_place","document_list_of_annotations_name","in_corpora","document_size_extent"),
                            $this->current_document_index
                        );
                    }


                    $docdata = array("result" => $doc['_source']);
                    $document_range = $this->LaudatioUtilService->getDocumentRange($docdata,$documentResult);

                    $result['hits']['hits'][$i]['_source']['documentrange'] = $document_range;

                    if(!isset($doc['_source']['corpus_document_genre'])) {
                        $result['hits']['hits'][$i]['_source']['corpus_document_genre'] = "NA";
                    }

                    if(isset($doc['_source']['corpus_merged_languages'])) {
                        $mergedLanguages = $doc['_source']['corpus_merged_languages'];
                        $result['hits']['hits'][$i]['_source']['corpus_merged_languages'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedLanguages));
                    }
                    /*
                    if(isset($doc['_source']['corpus_merged_formats'])) {
                        $mergedFormats = $doc['_source']['corpus_merged_formats'];
                        $result['hits']['hits'][$i]['_source']['corpus_merged_formats'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedFormats));
                    }
                    */


                    $result['hits']['hits'][$i]['_source']['visibility'] = 1;

                }//end if _index

            }
            //Log::info("searchMain result ? ".print_r($result['hits']['hits'],1));
            return $this->RestService->sendResponse($result['hits']['hits'], count($result['hits']['hits']).' corpus results retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }

    /**
     * searchMainSuggest
     *
     * @param Request $request
     * @return mixed
     */

    public function searchMainSuggest(Request $request) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        //try {

            $searchData = array();
            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            $corpus_index = config("laudatio.elasticsearch_corpus_index","corpora");
            $document_index = config("laudatio.elasticsearch_document_index","documents");
            $annotation_index = config("laudatio.elasticsearch_annotation_index","annotations");

            $searchFields = array(
                "corpora" => array(
                    "fields" => array(
                        "corpus_document_genre_suggest",
                        "corpus_annotator_forename_suggest",
                        "corpus_annotator_surname_suggest",
                        "corpus_author_infrastructure_forename_suggest",
                        "corpus_author_infrastructure_surname_suggest",
                        "corpus_editor_forename_suggest",
                        "corpus_editor_surname_suggest",
                        "annotation_encoding_format_suggest",
                        "annotation_encoding_tool_suggest",
                        "corpus_languages_language_suggest",
                        "corpus_publication_publisher_suggest",
                        "corpus_title_suggest"
                    ),
                    "source" => array(
                        "corpus_document_genre",
                        "corpus_annotator_forename",
                        "corpus_annotator_surname",
                        "corpus_author_infrastructure_forename",
                        "corpus_author_infrastructure_surname",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "annotation_encoding_format",
                        "annotation_encoding_tool",
                        "corpus_languages_language",
                        "corpus_publication_publisher",
                        "corpus_title",
                        "corpus_publication_publication_date",
                        "publication_version",
                        "laudatio_corpus_id",
                        "is_latest_publication"
                    ),
                    "icons" => array(
                        "corpus_document_genre" => "<i class=\"fas fa-th-list align-baseline fa-lg text-wine \" title=\"Corpus Document Genre\"></i>",
                        "corpus_annotator_forename" => "<i class=\"fas fa-user-edit align-baseline fa-lg text-wine \" title=\"Corpus Annotator Forename\"></i>",
                        "corpus_annotator_surname" => "<i class=\"fas fa-user-edit align-baseline fa-lg text-wine \" title=\"Corpus Annotator Surname\"></i>",
                        "corpus_author_infrastructure_forename" => "<i class=\"fas fa-user-cog align-baseline fa-lg text-wine \" title=\"Corpus Author Infrastructure Forename\"></i>",
                        "corpus_author_infrastructure_surname" => "<i class=\"fas fa-user-cog align-baseline fa-lg text-wine \" title=\"Corpus Author Infrastructure Surname\"></i>",
                        "corpus_editor_forename"  => "<i class=\"fas fa-user-edit align-baseline fa-lg text-wine \" title=\"Corpus Editor Forname\"></i>",
                        "corpus_editor_surname"  => "<i class=\"fas fa-user-edit align-baseline fa-lg text-wine \" title=\"Corpus Editor Surname\"></i>",
                        "annotation_encoding_format" => "<i class=\"fas fa-text-width align-baseline fa-lg text-wine \" title=\"Corpus Annotation Encoding Format\"></i>",
                        "annotation_encoding_tool" => "<i class=\"fas fa-text-width align-baseline fa-lg text-wine \" title=\"Corpus Annotation Encoding Tool\"></i>",
                        "corpus_languages_language" => "<i class=\"fas fa-globe  align-baseline fa-lg text-wine \" title=\"Corpus Language\"></i>",
                        "corpus_publication_publisher" => "<i class=\"fas fa-university align-baseline fa-lg text-wine \" title=\"Corpus Publisher\"></i>",
                        "corpus_title" => "<i class=\"fas fa-book align-baseline fa-lg text-wine\" title=\"Corpus\"></i>",
                        "corpus_publication_publication_date" => "<i class=\"fas fa-clock-o align-baseline fa-lg text-wine \" title=\"Corpus Publication Date\"></i>"
                    )                                                                                                                                        
                ),
                "documents" => array(
                    "fields" => array(
                        "document_author_forename_suggest",
                        "document_author_forename_suggest",
                        "document_editor_forename_suggest",
                        "document_editor_surname_suggest",
                        "document_genre_suggest",
                        "document_history_original_place_suggest",
                        "document_history_original_title_suggest",
                        "document_history_title_suggest",
                        "document_languages_language_suggest",
                        "document_publication_place_suggest",
                        "document_publication_publisher_suggest",
                        "document_title_suggest"
                    ),
                    "source" => array(
                        "document_title",
                        "document_author_forename",
                        "document_author_forename",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_genre",
                        "document_history_original_place",
                        "document_history_original_title",
                        "document_history_title",
                        "document_languages_language",
                        "document_publication_place",
                        "document_publication_publisher",
                        "publication_version",
                        "laudatio_corpus_id",
                        "is_latest_publication"
                    ),
                    "icons" => array(
                        "document_title" => "<i class=\"fas fa-file-alt align-baseline fa-lg text-wine\" title=\"Document\"></i>",
                        "document_author_forename" => "<i class=\"fas fa-user-edit align-baseline fa-lg text-wine \" title=\"Document Author Forename\"></i>",
                        "document_author_surname" =>  "<i class=\"fas fa-user-edit align-baseline fa-lg text-wine \" title=\"Document Author Surname\"></i>",
                        "document_editor_forename" => "<i class=\"fas fa-user-cog align-baseline fa-lg text-wine \" title=\"Document Editor Forename\"></i>",
                        "document_editor_surname" => "<i class=\"fas fa-user-cog align-baseline fa-lg text-wine \" title=\"Document Editor Surname\"></i>",
                        "document_genre" => "<i class=\"fas fa-th-list align-baseline fa-lg text-wine \" title=\"Document Genre\"></i>",
                        "document_history_original_place" => "<i class=\"fas fa-map-marker-alt align-baseline fa-lg text-wine \" title=\"Document Historical Location\"></i>",
                        "document_history_original_title" => "<i class=\"fas fa-file-signature align-baseline fa-lg text-wine \" title=\"Document Original Title\"></i>",
                        "document_history_title" => "<i class=\"fas fa-file-signature align-baseline fa-lg text-wine \" title=\"Document Historical Title\"></i>",
                        "document_languages_language" => "<i class=\"fas fa-globe align-baseline fa-lg text-wine \" title=\"Document Language\"></i>",
                        "document_publication_place"  => "<i class=\"fas fa-map-marker-alt align-baseline fa-lg text-wine \" title=\"Document Publication Location\"></i>",
                        "document_publication_publisher" => "<i class=\"fas fa-university align-baseline fa-lg text-wine \" title=\"Document Publisher\"></i>"
                    )
                ),
                "annotations" => array(
                    "fields" => array(
                        "preparation_encoding_annotation_group_suggest",
                        "preparation_encoding_annotation_sub_group_suggest",
                        "annotation_merged_formats_suggest",
                        "preparation_encoding_tool_suggest",
                        "preparation_title_suggest"
                    ),
                    "source" => array(
                        "preparation_title",
                        "preparation_encoding_annotation_group",
                        "preparation_encoding_annotation_sub_group",
                        "annotation_merged_formats",
                        "preparation_encoding_tool",
                        "publication_version",
                        "laudatio_corpus_id",
                        "is_latest_publication"
                    ),
                    "icons" => array(
                        "preparation_title" => "<i class=\"fas fa-edit align-text-middle fa-lg text-wine\" title=\"Annotation\"></i>",
                        "preparation_encoding_annotation_group" => "<i class=\"fas fa-text-width align-baseline fa-lg text-wine \" title=\"Annotation Encoding Group\"></i>",
                        "preparation_encoding_annotation_sub_group" => "<i class=\"fas fa-text-width align-baseline fa-lg text-wine \" title=\"Annotation Encoding Sub Group\"></i>",
                        "annotation_merged_formats" => "<i class=\"fas fa-text-width align-baseline fa-lg text-wine \" title=\"Annotation Format\"></i>",
                        "preparation_encoding_tool" => "<i class=\"fas fa-text-width align-baseline fa-lg text-wine \" title=\"Annotation Encoding Tool\"></i>"
                    )
                )
            );


            $queryBody = '';
        //  Log::info("FAPP PURNE");
            foreach ($searchFields as $header => $fieldData) {
                if($header == "corpora") {

                    foreach ($fieldData["fields"] as $field) {
                        $queryBody .= '{"index" : "'.$corpus_index.'"}'."\n";
                        $queryBody .= '{"_source" : ["'.join("\",\"",$fieldData["source"]).'"],';
                        $queryBody .= ' "suggest" : {"corpus-suggest" : {"prefix" : "'.$searchData["query"].'",';
                        $queryBody .= ' "completion" : {"field" : "'.$field.'", "fuzzy" : false, "size": 100, "analyzer": "standard","contexts": {"latest_publication": [ "true" ]}}}}}';
                        $queryBody .= "\n";
                    }

                }
                else if($header == "documents") {

                    foreach ($fieldData["fields"] as $field) {
                        $queryBody .= '{"index" : "'.$document_index.'"}'."\n";
                        $queryBody .= '{"_source" : ["'.join("\",\"",$fieldData["source"]).'"],';
                        $queryBody .= ' "suggest" : {"document-suggest" : {"prefix" : "'.$searchData["query"].'",';
                        $queryBody .= ' "completion" : {"field" : "'.$field.'", "fuzzy" : false, "size": 100, "analyzer": "standard", "contexts": {"latest_publication": [ "true" ]}}}}}';
                        $queryBody .= "\n";
                    }
                }
                else if($header == "annotations") {
                    foreach ($fieldData["fields"] as $field) {
                        $queryBody .= '{"index" : "'.$annotation_index.'"}'."\n";
                        $queryBody .= '{"_source" : ["'.join("\",\"",$fieldData["source"]).'"],';
                        $queryBody .= ' "suggest" : {"annotation-suggest" : {"prefix" : "'.$searchData["query"].'",';
                        $queryBody .= ' "completion" : {"field" : "'.$field.'", "fuzzy" : false, "size": 100, "analyzer": "standard", "contexts": {"latest_publication": [ "true" ]}}}}}';
                        $queryBody .= "\n";
                    }
                }

            }
            //Log::info("suggeQveri: ".print_r($queryBody,1));
            $results = $this->Elasticservice->getSuggestions($queryBody);
            //Log::info("results ?: ".print_r($results,1));
            $items = array("items" => array());
            $foundItems = array();

            foreach ($results as $header => $resultDataArray) {
                $item = array();

                if(count($resultDataArray) > 0 ) {
                    $corpusSuggestCounter = 0;
                    $documentSuggestCounter = 0;
                    $annotationSuggestCounter = 0;
                    foreach ($resultDataArray as $resultData) {
                        if(isset($resultData['_source']['is_latest_publication']) && $resultData['_source']['is_latest_publication']){

                            $item['_id'] = $resultData['_id'];
                            $item['_index'] = $resultData['_index'];
                            $item['corpus'] = $this->LaudatioUtilService->getCorpusNameByCorpusId($resultData['_source']['laudatio_corpus_id'],$resultData['_index']);

                            //Log::info("ITEMCORPUS ?: ".print_r($resultData,1));
                            if(substr_count(strtolower($resultData['text']),",") >= 1){
                                $subArray = explode(",",$resultData['text']);
                                $resultData['text'] = $subArray[0];
                            }

                            if($header == 'corpus_suggest') {
                                $item['suggestcounter'] = "corpus_".$corpusSuggestCounter;

                                foreach( $searchFields['corpora']['source'] as $source) {
                                    if (($source != "publication_version" && $source != "laudatio_corpus_id" && $source != "is_latest_publication") &&
                                        array_key_exists($source, $resultData['_source'])) {
                                        $stringArray = array();

                                        if (is_array($resultData['_source'][$source])) {
                                            $stringArray = array_map('trim', array_map('strtolower', $resultData['_source'][$source]));
                                        } else {
                                            $stringArray = array_map('trim', array_map('strtolower', explode(",", $resultData['_source'][$source])));
                                        }


                                        if (in_array(strtolower($resultData['text']), $stringArray)  || $this->LaudatioUtilService->strPosInArray($resultData['text'],$stringArray) != -1) {
                                            //Log::info("corps. safa ?: ".print_r($stringArray,1));
                                            $foundIndex = $this->LaudatioUtilService->strPosInArray($resultData['text'], $stringArray);
                                            if ($foundIndex != -1) {
                                                $item['icon'] = $searchFields['corpora']['icons'][$source];
                                                $item['text'] = $stringArray[$foundIndex];
                                                $item['realtitle'] = $resultData['_source']['corpus_title'][0];
                                                break;
                                            }
                                        }
                                        else{
                                            // hack for now... maybe better to reset elasticsearch stringlimit to 50,
                                            //  split long strings at index time, and then search all parts here afterwards?
                                            if(!empty($item['text']) && strlen($item['text']) > 100){
                                                $item['text']  = substr($item['text'],0,100);
                                                $item['realtitle'] = $resultData['_source']['corpus_title'][0];
                                                $item['icon'] = $searchFields['corpora']['icons'][$source];
                                                break;
                                            }
                                        }
                                    }
                                }
                                //Log::info("CRAPPUS: itemtext ".print_r($item['text'],1));
                                //$item['text'] = $resultData['_source']['corpus_title'][0];
                                $item['version'] = $resultData['_source']['publication_version'];
                                end($resultData['_source']['corpus_publication_publication_date']);
                                $last_date = key($resultData['_source']['corpus_publication_publication_date']);
                                $item['date'] = $resultData['_source']['corpus_publication_publication_date'][$last_date];
                                $corpusSuggestCounter++;
                                //Log::info($header. " " .$searchData["query"]." ITEM ?: ".print_r($item,1));
                            }
                            else if($header == 'document_suggest'){
                                $item['suggestcounter'] = "document_".$documentSuggestCounter;
                                foreach( $searchFields['documents']['source'] as $source) {

                                    if(($source != "publication_version" && $source != "laudatio_corpus_id" && $source != "is_latest_publication") &&
                                        array_key_exists($source,$resultData['_source'])) {
                                        $stringArray = array();
                                        if(is_array($resultData['_source'][$source])) {
                                            $stringArray = array_map('trim',array_map('strtolower', $resultData['_source'][$source]));
                                        }
                                        else{
                                            $stringArray  = array_map('trim',array_map('strtolower',explode(",",$resultData['_source'][$source])));
                                        }

                                        if(in_array(strtolower($resultData['text']),$stringArray) || $this->LaudatioUtilService->strPosInArray($resultData['text'],$stringArray) != -1) {

                                            $foundIndex = $this->LaudatioUtilService->strPosInArray($resultData['text'],$stringArray);
                                            if($foundIndex != -1) {
                                                $item['text'] = $stringArray[$foundIndex];
                                                $item['realtitle'] = $resultData['_source']['document_title'][0];
                                                $item['icon'] = $searchFields['documents']['icons'][$source];
                                                break;
                                            }
                                        }
                                        else {
                                            // hack for now... maybe better to reset elasticsearch stringlimit to 50,
                                            //  split long strings at index time, and then search all parts here afterwards?
                                            if(!empty($item['text']) && strlen($item['text']) > 100){
                                                $item['text']  = substr($item['text'],0,100);
                                                $item['realtitle'] = $resultData['_source']['document_title'][0];
                                                $item['icon'] = $searchFields['documents']['icons'][$source];
                                                break;
                                            }
                                        }
                                    }
                                }
                                //Log::info("DOCUS: ".print_r($item['text'],1));
                                //$item['text'] = $resultData['_source']['document_title'][0];
                                $item['version'] = $resultData['_source']['publication_version'];
                                $documentSuggestCounter++;

                            }
                            else if($header == 'annotation_suggest'){
                                $item['suggestcounter'] = "annotation_".$annotationSuggestCounter;
                                foreach( $searchFields['annotations']['source'] as $source) {
                                    if(($source != "publication_version" && $source != "laudatio_corpus_id" && $source != "is_latest_publication") &&
                                        array_key_exists($source,$resultData['_source'])) {
                                        $stringArray = array();
                                        if(is_array($resultData['_source'][$source])) {
                                            $stringArray = array_map('trim',array_map('strtolower', $resultData['_source'][$source]));
                                        }
                                        else{
                                            $stringArray = array_map('trim',array_map('strtolower',explode(",",$resultData['_source'][$source])));
                                        }

                                        if(in_array(strtolower($resultData['text']),$stringArray) || $this->LaudatioUtilService->strPosInArray($resultData['text'],$stringArray) != -1) {
                                            $foundIndex = $this->LaudatioUtilService->strPosInArray($resultData['text'],$stringArray);
                                            if($foundIndex != -1) {
                                                $item['text'] = $stringArray[$foundIndex];
                                                $item['realtitle'] = $resultData['_source']['preparation_title'][0];
                                                $item['icon'] = $searchFields['annotations']['icons'][$source];
                                                break;
                                            }
                                        }
                                        else{
                                            // hack for now... maybe better to reset elasticsearch stringlimit to 50,
                                            //  split long strings at index time, and then search all parts here afterwards?
                                            if(!empty($item['text']) && strlen($item['text']) > 100){
                                                $item['text']  = substr($item['text'],0,100);
                                                $item['realtitle'] = $resultData['_source']['preparation_title'][0];
                                                $item['icon'] = $searchFields['annotations']['icons'][$source];
                                                break;
                                            }
                                        }
                                    }
                                }
                                //Log::info("ANNUS: ".print_r($item['text'],1));
                                $item['version'] = $resultData['_source']['publication_version'];
                                $annotationSuggestCounter++;
                            }
                            if(!in_array($item['text'],$foundItems)) {
                                array_push($items['items'], $item);
                            }
                            array_push($foundItems, $item['text']);
                        }//end if islatestpublication
                    }
                }
            }
            //Log::info("BAPP ?: ".print_r($items,1));
            return $this->RestService->sendResponse($items, ' Search suggestions retrieved successfully.');
       // } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
         //   return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        //}

    }


    public function corpusSearchMainFilter(Request $request) {

        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        $isDateSearch = false;
        $isTokenSearch = false;
        $isMixedSearch = false;
        $searchData = array();
        $hasWildCards = false;
        //Log::info("corpusSeqarchMainFilter: request->searchData: ".print_r($request->searchData,1));

        if(is_array($request->searchData)){
            $searchData = $request->searchData;
        }
        else {
            $searchData = json_decode($request->searchData,true);
        }
        //Log::info("corpusSeqarchMainFilter: searchData: ".print_r($searchData,1));
        if((!empty($searchData['filterData']['corpus_publication_publication_date']) ||
            !empty($searchData['filterData']['corpusYearTo'])) &&
            (empty($searchData['filterData']['corpus_title']) &&
             empty($searchData['filterData']['corpus_publication_publisher']) &&
             empty($searchData['filterData']['corpus_size_value']) &&
             empty($searchData['filterData']['corpusSizeTo']) &&
             empty($searchData['filterData']['corpus_merged_languages']) &&
             count($searchData['filterData']['corpus_merged_formats']) == 0 &&
             empty($searchData['filterData']['corpus_publication_license'])
            )
            ){
            $isDateSearch = true;
        }
        else if((empty($searchData['filterData']['corpus_publication_publication_date']) &&
                empty($searchData['filterData']['corpusYearTo'])) &&
            (empty($searchData['filterData']['corpus_title']) &&
                empty($searchData['filterData']['corpus_publication_publisher']) &&
                (!empty($searchData['filterData']['corpus_size_value']) ||
                !empty($searchData['filterData']['corpusSizeTo'])) &&
                empty($searchData['filterData']['corpus_merged_languages']) &&
                count($searchData['filterData']['corpus_merged_formats']) == 0 &&
                empty($searchData['filterData']['corpus_publication_license'])
            )
        ){
            $isTokenSearch = true;
        }
        else if(
            ((!empty($searchData['filterData']['corpus_publication_publication_date']) ||
            !empty($searchData['filterData']['corpusYearTo']))
            &&
            (!empty($searchData['filterData']['corpus_title']) ||
                !empty($searchData['filterData']['corpus_publication_publisher']) ||
                !empty($searchData['filterData']['corpus_size_value']) ||
                !empty($searchData['filterData']['corpusSizeTo']) ||
                !empty($searchData['filterData']['corpus_merged_languages']) ||
                count($searchData['filterData']['corpus_merged_formats']) > 0 ||
                !empty($searchData['filterData']['corpus_publication_license'])
            )) ||
            ((!empty($searchData['filterData']['corpus_size_value']) ||
                    !empty($searchData['filterData']['corpusSizeTo']))
                &&
                (!empty($searchData['filterData']['corpus_title']) ||
                    !empty($searchData['filterData']['corpus_publication_publisher']) ||
                    !empty($searchData['filterData']['corpus_publication_publication_date']) ||
                    !empty($searchData['filterData']['corpusYearTo']) ||
                    count($searchData['filterData']['corpus_merged_languages']) > 0 ||
                    !empty($searchData['filterData']['corpus_merged_formats']) ||
                    !empty($searchData['filterData']['corpus_publication_license'])
                ))


        ) {
            $isMixedSearch = true;
        }


        $source = [
            "laudatio_corpus_id",
            "corpus_title",
            "corpus_editor_forename",
            "corpus_editor_surname",
            "corpus_publication_publisher",
            "corpus_documents",
            "corpus_merged_formats",
            "corpus_encoding_tool",
            "corpus_encoding_project_description",
            "corpus_annotator_forename",
            "corpus_annotator_surname",
            "corpus_publication_license",
            "has_documents",
            "has_annotations",
            "annotation_id",
            "corpus_languages_language",
            "corpus_merged_languages",
            "corpus_languages_iso_code",
            "corpus_publication_publication_date",
            "corpus_size_type",
            "corpus_size_value",
            "corpus_document_genre",
            "publication_status"
        ];


        if($isDateSearch) {
            $searchData["fields"] = array(
                "corpus_publication_publication_date"
            );
        }
        else if($isTokenSearch){
            $searchData["fields"] = array(
                "corpus_size_value"
            );
        }
        else if($isMixedSearch) {
            $searchData["fields"] = array(
                "corpus_title",
                "corpus_editor_forename",
                "corpus_editor_surname",
                "corpus_publication_publisher",
                "corpus_merged_formats",
                "corpus_encoding_tool",
                "corpus_encoding_project_description",
                "corpus_publication_license",
                "corpus_merged_languages",
                "corpus_languages_iso_code",
                "corpus_document_genre",
                "publication_status",
                "corpus_size_value",
                "corpus_publication_publication_date"
            );
        }
        else{
            $searchData["fields"] = array(
                "corpus_title",
                "corpus_editor_forename",
                "corpus_editor_surname",
                "corpus_publication_publisher",
                "corpus_merged_formats",
                "corpus_encoding_tool",
                "corpus_encoding_project_description",
                "corpus_publication_license",
                "corpus_merged_languages",
                "corpus_languages_iso_code",
                "corpus_document_genre",
                "corpus_publication_publication_date.string",
                "publication_status"
            );
        }



        $index = config("laudatio.elasticsearch_corpus_index","corpora");

        try{
            $fullTextQuery = "";


            $result = array();

            if(isset($searchData['mainSearchTerm'])) {
                $fullTextQuery = $searchData['mainSearchTerm'];
            }
            else if(isset($searchData['frontpagequery'])) {
                $fullTextQuery = $searchData['frontpagequery'];
            }

            if(strpos($fullTextQuery,'?') > -1 || strpos($fullTextQuery,'*') > -1 ){
                $hasWildCards = true;
            }

            if($isDateSearch && !$isMixedSearch && !$isTokenSearch) {
                $searchData["fulltextFields"] = array();
                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "corpus_title",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "corpus_publication_publisher",
                        "corpus_merged_formats",
                        "corpus_encoding_tool",
                        "corpus_encoding_project_description",
                        "corpus_publication_license",
                        "corpus_merged_languages",
                        "corpus_languages_iso_code",
                        "corpus_document_genre",
                        "corpus_publication_publication_date.string",
                        "publication_status"
                    );


                }

                $result = $this->Elasticservice->corpusRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],$searchData["size"],$hasWildCards);
            }
            else if($isTokenSearch && !$isDateSearch && !$isMixedSearch) {

                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "corpus_title",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "corpus_publication_publisher",
                        "corpus_merged_formats",
                        "corpus_encoding_tool",
                        "corpus_encoding_project_description",
                        "corpus_publication_license",
                        "corpus_merged_languages",
                        "corpus_languages_iso_code",
                        "corpus_document_genre",
                        "corpus_publication_publication_date.string",
                        "publication_status"
                    );

                }

                $result = $this->Elasticservice->corpusRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],$searchData["size"], $hasWildCards);
            }
            else if(!$isDateSearch && !$isTokenSearch && $isMixedSearch) {
                $searchData["fulltextFields"] = array();
                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "corpus_title",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "corpus_publication_publisher",
                        "corpus_merged_formats",
                        "corpus_encoding_tool",
                        "corpus_encoding_project_description",
                        "corpus_publication_license",
                        "corpus_merged_languages",
                        "corpus_languages_iso_code",
                        "corpus_document_genre",
                        "corpus_publication_publication_date.string",
                        "corpus_size_value.string",
                        "publication_status"
                    );
                }

                $result = $this->Elasticservice->corpusMixSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"] ,array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],$searchData["size"], $hasWildCards);

            }
            else if(!$isDateSearch && !$isMixedSearch && !$isTokenSearch) {

                $result = $this->Elasticservice->filterSearch($searchData, $fullTextQuery ,$index,$searchData["from"],$searchData["size"], $source, $hasWildCards);
            }



            $data = $result['hits']['hits'];
            //Log::info("corpusSeqarchMainFilter: data: ".print_r($data,1));
            for ($i = 0; $i < count($data); $i++) {
                $doc = $data[$i];
                if(count($doc) > 0){
                    $projectPath = $this->LaudatioUtilService->getCorpusProjectPathByCorpusId($doc['_id'],$doc["_index"]);
                    $corpusPath = $this->LaudatioUtilService->getCorpusPathByCorpusId($doc["_id"],$doc["_index"]);
                    $corpusLogo = $this->LaudatioUtilService->getCorpusLogoByCorpusId($doc["_id"],$doc["_index"]);

                    $result['hits']['hits'][$i]['_source']['projectpath'] = $projectPath;
                    $result['hits']['hits'][$i]['_source']['corpuspath'] = $corpusPath;
                    $result['hits']['hits'][$i]['_source']['corpuslogo'] = $corpusLogo;

                    $result['hits']['hits'][$i]['_source']['doi'] = $this->LaudatioUtilService->getDOIByCorpusElasticsearchId($doc['_id'],$doc["_index"]);

                    /*
                     * if(isset($doc['highlight'])){
                    foreach ($doc['highlight'] as $highlightName => $highlightData){
                        if(strpos($highlightName,".string") > -1) {
                            $newHighlightName = str_replace(".string", "",$highlightName);
                            unset($result['hits']['hits'][$i]['highlight'][$highlightName]);
                            $result['hits']['hits'][$i]['highlight'][$newHighlightName] = $highlightData;
                        }
                    }
                }
                     */
                    if(isset($doc['highlight'])){
                        $newHighlights = array();
                        foreach ($doc['highlight'] as $highlightName => $highlightData){
                            if(strpos($highlightName,".string") > -1) {
                                $newHighlightName = str_replace(".string", "",$highlightName);
                                unset($result['hits']['hits'][$i]['highlight'][$highlightName]);
                                $result['hits']['hits'][$i]['highlight'][$newHighlightName] = $highlightData;
                            }
                        }

                    }

                    if((!empty($searchData['filterData']['corpus_publication_publication_date']) ||
                        !empty($searchData['filterData']['corpusYearTo'])) && !empty($doc['_source']['corpus_publication_publication_date'])) {
                        $dateCount = count($doc['_source']['corpus_publication_publication_date']);
                        $last_date = $doc['_source']['corpus_publication_publication_date'][$dateCount -1];
                        $result['hits']['hits'][$i]['highlight']['corpus_publication_publication_date'] = array(array('<span class="laudatiofilterhighlight">'.$last_date.'</span>'));
                    }


                    if(
                        !empty($searchData['filterData']['corpus_size_value']) ||
                        !empty($searchData['filterData']['corpusSizeTo']) && !empty($doc['_source']['corpus_size_value'])) {

                        $extentFrom = -1;
                        $extentTo = -1;
                        if(!empty($searchData['filterData']['corpus_size_value'])) {
                            $extentFrom = intval($searchData['filterData']['corpus_size_value']);
                        }

                        if(!empty($searchData['filterData']['corpusSizeTo'])) {
                            $extentTo = intval($searchData['filterData']['corpusSizeTo']);
                        }

                        $sourceVal = intval($doc['_source']['corpus_size_value'][0]);
                        if($extentFrom > -1 && $extentTo == -1){
                            if($extentFrom <= $sourceVal){
                                $result['hits']['hits'][$i]['highlight']['corpus_size_value'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['corpus_size_value'][0].'</span>'));
                            }
                        }
                        else if($extentFrom == -1 && $extentTo > -1){
                            if($extentTo >= $sourceVal){
                                $result['hits']['hits'][$i]['highlight']['corpus_size_value'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['corpus_size_value'][0].'</span>'));
                            }
                        }
                        else if($extentFrom > -1 && $extentTo > -1){
                            if($extentFrom <= $sourceVal && $sourceVal <= $extentTo){
                                $result['hits']['hits'][$i]['highlight']['corpus_size_value'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['corpus_size_value'][0].'</span>'));
                            }
                        }

                        //$result['hits']['hits'][$i]['highlight']['corpus_size_value'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['corpus_size_value'][0].'</span>'));
                    }

                    $documentResult = array();
                    if(isset($doc['_source']['has_documents'])){
                        $documentResult = $this->Elasticservice->getDocumentByCorpus(
                            $doc['_source']['has_documents'],
                            array($doc['_source']['laudatio_corpus_id']),
                            array("document_title","document_publication_publishing_date","document_publication_place","document_list_of_annotations_name","in_corpora","document_size_extent"),
                            $this->current_document_index
                        );
                    }

                    $docdata = array("result" => $doc['_source']);
                    $document_range = $this->LaudatioUtilService->getDocumentRange($docdata,$documentResult);
                    $result['hits']['hits'][$i]['_source']['documentrange'] = $document_range;

                    if(!isset($doc['_source']['corpus_document_genre'])) {
                        $result['hits']['hits'][$i]['_source']['corpus_document_genre'] = "NA";
                    }

                    if(isset($doc['_source']['corpus_merged_languages'])) {
                        $mergedLanguages = $doc['_source']['corpus_merged_languages'];
                        $result['hits']['hits'][$i]['_source']['corpus_merged_languages'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedLanguages));
                    }
                    /*
                    if(isset($doc['_source']['corpus_merged_formats'])) {
                        $mergedFormats = $doc['_source']['corpus_merged_formats'];
                        $result['hits']['hits'][$i]['_source']['corpus_merged_formats'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedFormats));
                    }
                    */
                    $result['hits']['hits'][$i]['_source']['visibility'] = 1;
                }//end if doc

            }

            return $this->RestService->sendResponse($result['hits']['hits'], count($result['hits']['hits']).' Corpus filter results retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }
    public function corpusSearchMainFilterCount(Request $request) {

        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        $isDateSearch = false;
        $isTokenSearch = false;
        $isMixedSearch = false;
        $searchData = array();
        $hasWildCards = false;
        //Log::info("corpusSeqarchMainFilter: request->searchData: ".print_r($request->searchData,1));

        if(is_array($request->searchData)){
            $searchData = $request->searchData;
        }
        else {
            $searchData = json_decode($request->searchData,true);
        }
        //Log::info("corpusSeqarchMainFilter: searchData: ".print_r($searchData,1));
        if((!empty($searchData['filterData']['corpus_publication_publication_date']) ||
                !empty($searchData['filterData']['corpusYearTo'])) &&
            (empty($searchData['filterData']['corpus_title']) &&
                empty($searchData['filterData']['corpus_publication_publisher']) &&
                empty($searchData['filterData']['corpus_size_value']) &&
                empty($searchData['filterData']['corpusSizeTo']) &&
                empty($searchData['filterData']['corpus_merged_languages']) &&
                count($searchData['filterData']['corpus_merged_formats']) == 0 &&
                empty($searchData['filterData']['corpus_publication_license'])
            )
        ){
            $isDateSearch = true;
        }
        else if((empty($searchData['filterData']['corpus_publication_publication_date']) &&
                empty($searchData['filterData']['corpusYearTo'])) &&
            (empty($searchData['filterData']['corpus_title']) &&
                empty($searchData['filterData']['corpus_publication_publisher']) &&
                (!empty($searchData['filterData']['corpus_size_value']) ||
                    !empty($searchData['filterData']['corpusSizeTo'])) &&
                empty($searchData['filterData']['corpus_merged_languages']) &&
                count($searchData['filterData']['corpus_merged_formats']) == 0 &&
                empty($searchData['filterData']['corpus_publication_license'])
            )
        ){
            $isTokenSearch = true;
        }
        else if(
            ((!empty($searchData['filterData']['corpus_publication_publication_date']) ||
                    !empty($searchData['filterData']['corpusYearTo']))
                &&
                (!empty($searchData['filterData']['corpus_title']) ||
                    !empty($searchData['filterData']['corpus_publication_publisher']) ||
                    !empty($searchData['filterData']['corpus_size_value']) ||
                    !empty($searchData['filterData']['corpusSizeTo']) ||
                    !empty($searchData['filterData']['corpus_merged_languages']) ||
                    count($searchData['filterData']['corpus_merged_formats']) > 0 ||
                    !empty($searchData['filterData']['corpus_publication_license'])
                )) ||
            ((!empty($searchData['filterData']['corpus_size_value']) ||
                    !empty($searchData['filterData']['corpusSizeTo']))
                &&
                (!empty($searchData['filterData']['corpus_title']) ||
                    !empty($searchData['filterData']['corpus_publication_publisher']) ||
                    !empty($searchData['filterData']['corpus_publication_publication_date']) ||
                    !empty($searchData['filterData']['corpusYearTo']) ||
                    count($searchData['filterData']['corpus_merged_languages']) > 0 ||
                    !empty($searchData['filterData']['corpus_merged_formats']) ||
                    !empty($searchData['filterData']['corpus_publication_license'])
                ))


        ) {
            $isMixedSearch = true;
        }


        $source = [
            "laudatio_corpus_id",
            "corpus_title",
            "corpus_editor_forename",
            "corpus_editor_surname",
            "corpus_publication_publisher",
            "corpus_documents",
            "corpus_merged_formats",
            "corpus_encoding_tool",
            "corpus_encoding_project_description",
            "corpus_annotator_forename",
            "corpus_annotator_surname",
            "corpus_publication_license",
            "has_documents",
            "has_annotations",
            "annotation_id",
            "corpus_languages_language",
            "corpus_merged_languages",
            "corpus_languages_iso_code",
            "corpus_publication_publication_date",
            "corpus_size_type",
            "corpus_size_value",
            "corpus_document_genre",
            "publication_status"
        ];


        if($isDateSearch) {
            $searchData["fields"] = array(
                "corpus_publication_publication_date"
            );
        }
        else if($isTokenSearch){
            $searchData["fields"] = array(
                "corpus_size_value"
            );
        }
        else if($isMixedSearch) {
            $searchData["fields"] = array(
                "corpus_title",
                "corpus_editor_forename",
                "corpus_editor_surname",
                "corpus_publication_publisher",
                "corpus_merged_formats",
                "corpus_encoding_tool",
                "corpus_encoding_project_description",
                "corpus_publication_license",
                "corpus_merged_languages",
                "corpus_languages_iso_code",
                "corpus_document_genre",
                "publication_status",
                "corpus_size_value",
                "corpus_publication_publication_date"
            );
        }
        else{
            $searchData["fields"] = array(
                "corpus_title",
                "corpus_editor_forename",
                "corpus_editor_surname",
                "corpus_publication_publisher",
                "corpus_merged_formats",
                "corpus_encoding_tool",
                "corpus_encoding_project_description",
                "corpus_publication_license",
                "corpus_merged_languages",
                "corpus_languages_iso_code",
                "corpus_document_genre",
                "corpus_publication_publication_date.string",
                "publication_status"
            );
        }



        $index = config("laudatio.elasticsearch_corpus_index","corpora");

        try{
            $fullTextQuery = "";


            $result = array();

            if(isset($searchData['mainSearchTerm'])) {
                $fullTextQuery = $searchData['mainSearchTerm'];
            }
            else if(isset($searchData['frontpagequery'])) {
                $fullTextQuery = $searchData['frontpagequery'];
            }

            if(strpos($fullTextQuery,'?') > -1 || strpos($fullTextQuery,'*') > -1 ){
                $hasWildCards = true;
            }

            if($isDateSearch && !$isMixedSearch && !$isTokenSearch) {
                $searchData["fulltextFields"] = array();
                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "corpus_title",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "corpus_publication_publisher",
                        "corpus_merged_formats",
                        "corpus_encoding_tool",
                        "corpus_encoding_project_description",
                        "corpus_publication_license",
                        "corpus_merged_languages",
                        "corpus_languages_iso_code",
                        "corpus_document_genre",
                        "corpus_publication_publication_date.string",
                        "publication_status"
                    );


                }

                $result = $this->Elasticservice->corpusRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],10000,$hasWildCards);
            }
            else if($isTokenSearch && !$isDateSearch && !$isMixedSearch) {

                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "corpus_title",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "corpus_publication_publisher",
                        "corpus_merged_formats",
                        "corpus_encoding_tool",
                        "corpus_encoding_project_description",
                        "corpus_publication_license",
                        "corpus_merged_languages",
                        "corpus_languages_iso_code",
                        "corpus_document_genre",
                        "corpus_publication_publication_date.string",
                        "publication_status"
                    );

                }

                $result = $this->Elasticservice->corpusRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],10000, $hasWildCards);
            }
            else if(!$isDateSearch && !$isTokenSearch && $isMixedSearch) {
                $searchData["fulltextFields"] = array();
                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "corpus_title",
                        "corpus_editor_forename",
                        "corpus_editor_surname",
                        "corpus_publication_publisher",
                        "corpus_merged_formats",
                        "corpus_encoding_tool",
                        "corpus_encoding_project_description",
                        "corpus_publication_license",
                        "corpus_merged_languages",
                        "corpus_languages_iso_code",
                        "corpus_document_genre",
                        "corpus_publication_publication_date.string",
                        "corpus_size_value.string",
                        "publication_status"
                    );
                }

                $result = $this->Elasticservice->corpusMixSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"] ,array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],10000, $hasWildCards);

            }
            else if(!$isDateSearch && !$isMixedSearch && !$isTokenSearch) {

                $result = $this->Elasticservice->filterSearch($searchData, $fullTextQuery ,$index,$searchData["from"],10000, $source, $hasWildCards);
            }

            return $this->RestService->sendResponse($result['hits']['total']['value'], $result['hits']['total']['value'].' Corpus filter result count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }
    public function documentSearchMainFilter(Request $request) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $isDateSearch = false;
            $isMixedSearch = false;
            $isTokenSearch = false;
            $searchData = array();
            $hasWildCards = false;


            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            $index = config("laudatio.elasticsearch_document_index","documents");

            if((!empty($searchData['filterData']['document_publication_publishing_date']) ||
                    !empty($searchData['filterData']['document_publication_publishing_date_to'])) &&
                (empty($searchData['filterData']['document_title']) &&
                    empty($searchData['filterData']['document_author_forename']) &&
                    empty($searchData['filterData']['document_author_surname']) &&
                    empty($searchData['filterData']['document_merged_authors']) &&
                    empty($searchData['filterData']['document_publication_place']) &&
                    empty($searchData['filterData']['document_size_extent']) &&
                    empty($searchData['filterData']['document_size_extent_to']) &&
                    empty($searchData['filterData']['document_merged_languages'])
                )
            ){
                $isDateSearch = true;
            }
            else if((empty($searchData['filterData']['document_publication_publishing_date']) ||
                    empty($searchData['filterData']['document_publication_publishing_date_to'])) &&
                (empty($searchData['filterData']['document_title']) &&
                    empty($searchData['filterData']['document_author_forename']) &&
                    empty($searchData['filterData']['document_author_surname']) &&
                    empty($searchData['filterData']['document_merged_authors']) &&
                    empty($searchData['filterData']['document_publication_place']) &&
                    (!empty($searchData['filterData']['document_size_extent']) ||
                    !empty($searchData['filterData']['document_size_extent_to']) ) &&
                    empty($searchData['filterData']['document_merged_languages'])
                )
            ){
                $isTokenSearch = true;
            }
            else if(((!empty($searchData['filterData']['document_publication_publishing_date']) ||
                    !empty($searchData['filterData']['document_publication_publishing_date_to'])) &&
                (!empty($searchData['filterData']['document_title']) ||
                    !empty($searchData['filterData']['document_author_forename']) ||
                    !empty($searchData['filterData']['document_author_surname']) ||
                    !empty($searchData['filterData']['document_merged_authors']) ||
                    !empty($searchData['filterData']['document_publication_place']) ||
                    !empty($searchData['filterData']['document_size_extent']) ||
                    !empty($searchData['filterData']['document_size_extent_to']) ||
                    !empty($searchData['filterData']['document_merged_languages'])
                )) ||
                ((!empty($searchData['filterData']['document_size_extent']) ||
                        !empty($searchData['filterData']['document_size_extent_to'])) &&
                    (!empty($searchData['filterData']['document_title']) ||
                        !empty($searchData['filterData']['document_author_forename']) ||
                        !empty($searchData['filterData']['document_author_surname']) ||
                        !empty($searchData['filterData']['document_merged_authors']) ||
                        !empty($searchData['filterData']['document_publication_place']) ||
                        !empty($searchData['filterData']['document_publication_publishing_date']) ||
                        !empty($searchData['filterData']['document_publication_publishing_date_to']) ||
                        !empty($searchData['filterData']['document_merged_languages'])
                    ))

            ) {
                $isMixedSearch = true;
            }

            $source = [
                "document_title",
                "document_author_forename",
                "document_author_surname",
                "document_merged_authors",
                "document_editor_forename",
                "document_editor_surname",
                //"document_languages_language",
                "document_merged_languages",
                "document_languages_iso_code",
                "document_publication_place",
                //@todo: not possible to search in date if the format is just a string "document_publication_publishing_date",
                //"document_list_of_annotations_id"
                "document_list_of_annotations_id",
                "document_history_not_before",
                "document_history_not_after",
                "document_size_type",
                "document_size_extent",
                "document_genre",
                'document_publication_publishing_date',
                "publication_status",
                "laudatio_corpus_id"
                ];


            if($isDateSearch){
                $searchData["fields"] = array(
                    "document_history_not_before",
                    "document_history_not_after",
                    "document_publication_publishing_date"
                );
            }
            else if($isTokenSearch){
                $searchData["fields"] = array(
                    "document_size_extent"
                );
            }
            else if($isMixedSearch){
                $searchData["fields"] = array(
                    "document_publication_publishing_date.string",
                    "document_history_not_before.string",
                    "document_history_not_after.string",
                    "document_title",
                    "document_author_forename",
                    "document_author_surname",
                    "document_merged_authors",
                    "document_editor_forename",
                    "document_editor_surname",
                    "document_languages_language",
                    "document_merged_languages",
                    "document_languages_iso_code",
                    "document_publication_place",
                    "document_size_type",
                    "document_size_extent",
                    "document_genre",
                    "publication_status"
                );
            }
            else{
                $searchData["fields"] = array(
                    "document_title",
                    "document_author_forename",
                    "document_author_surname",
                    "document_merged_authors",
                    "document_editor_forename",
                    "document_editor_surname",
                    "document_languages_language",
                    "document_merged_languages",
                    "document_languages_iso_code",
                    "document_publication_place",
                    "document_size_type",
                    "document_size_extent",
                    "document_genre",
                    "publication_status",
                    'document_publication_publishing_date.string',
                    "document_history_not_before.string",
                    "document_history_not_after.string",
                );
            }


            $fullTextQuery = "";
            if(isset($searchData['mainSearchTerm'])) {
                $fullTextQuery = $searchData['mainSearchTerm'];
            }
            else if(isset($searchData['frontpagequery'])) {
                $fullTextQuery = $searchData['frontpagequery'];
            }

            if(strpos($fullTextQuery,'?') > -1 || strpos($fullTextQuery,'*') > -1 ){
                $hasWildCards = true;
            }


            if($isDateSearch && !$isMixedSearch && !$isTokenSearch) {
                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "document_title",
                        "document_author_forename",
                        "document_author_surname",
                        "document_merged_authors",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_languages_language",
                        "document_merged_languages",
                        "document_languages_iso_code",
                        "document_publication_place",
                        "document_size_type",
                        "document_size_extent.string",
                        "document_genre",
                        "publication_status",
                        'document_publication_publishing_date.string',
                        "document_history_not_before.string",
                        "document_history_not_after.string",
                    );
                }

                $result = $this->Elasticservice->documentDateRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],$searchData["size"],$hasWildCards);
            }
            else if(!$isDateSearch && $isTokenSearch && !$isMixedSearch) {
                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "document_title",
                        "document_author_forename",
                        "document_author_surname",
                        "document_merged_authors",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_languages_language",
                        "document_merged_languages",
                        "document_languages_iso_code",
                        "document_publication_place",
                        "document_size_type",
                        "document_size_extent.string",
                        "document_genre",
                        "publication_status",
                        'document_publication_publishing_date.string',
                        "document_history_not_before.string",
                        "document_history_not_after.string",
                    );
                }
                $result = $this->Elasticservice->documentRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],$searchData["size"],$hasWildCards);
            }
            else if(!$isDateSearch && !$isTokenSearch && $isMixedSearch) {
                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "document_title",
                        "document_author_forename",
                        "document_author_surname",
                        "document_merged_authors",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_languages_language",
                        "document_merged_languages",
                        "document_languages_iso_code",
                        "document_publication_place",
                        "document_size_type",
                        "document_size_extent.string",
                        "document_genre",
                        "publication_status",
                        'document_publication_publishing_date.string',
                        "document_history_not_before.string",
                        "document_history_not_after.string",
                    );
                }

                $result = $this->Elasticservice->documentMixSearch($searchData['filterData'],$fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],$searchData["size"],$hasWildCards);
                //Log::info("REZ: ".print_r($result,1));
            }
            else if(!$isDateSearch && !$isMixedSearch && !$isTokenSearch){
                $result =  $this->Elasticservice->filterSearch($searchData, $fullTextQuery ,$index,$searchData["from"],$searchData["size"], $source, $hasWildCards);
            }



            $data = $result['hits']['hits'];

            $foundCorpusNames = array();
            for ($i = 0; $i < count($data); $i++) {
                $doc = $data[$i];

                if(count($doc) > 0  && isset($doc["_index"])){
                    $corpusName = $this->LaudatioUtilService->getCorpusNameByObjectElasticsearchId('document',$doc['_id']);
                    $corpusElasticSearchId = $this->LaudatioUtilService->getElasticSearchIdByCorpusId($doc['_source']['laudatio_corpus_id']);
                    if(!empty($corpusName)){
                        $result['hits']['hits'][$i]['_source']['corpus_name'] = $corpusName;
                        $result['hits']['hits'][$i]['_source']['corpus_elasticsearch_id'] = $corpusElasticSearchId;
                        $result['hits']['hits'][$i]['_source']['corpus_elasticsearch_index'] = $this->current_corpus_index;
                    }


                    if((!empty($searchData['filterData']['document_publication_publishing_date']) ||
                            !empty($searchData['filterData']['document_publication_publishing_date_to'])) && (
                            !empty($doc['_source']['document_publication_publishing_date']) ||
                            !empty($doc['_source']['document_history_not_before']) ||
                            !empty($doc['_source']['document_history_not_after'])
                        )) {
                        $dateFrom = -1;
                        $dateTo = -1;

                        if(!empty($searchData['filterData']['document_publication_publishing_date'])) {
                            $dateFrom = strtotime($searchData['filterData']['document_publication_publishing_date']."-01-01");
                        }

                        if(!empty($searchData['filterData']['document_publication_publishing_date_to'])) {
                            $dateTo = strtotime($searchData['filterData']['document_publication_publishing_date_to']."-01-01"   );
                        }


                        $dateSourceVal = -1;
                        $notBeforeSourceVal = -1;
                        $notAfterSourceVal = -1;
                        if(isset($doc['_source']['document_publication_publishing_date'][0])){
                            if(strpos($doc['_source']['document_publication_publishing_date'][0],"-") === false && strpos($doc['_source']['document_publication_publishing_date'][0],"/") === false){
                                $dateSourceVal = strtotime($doc['_source']['document_publication_publishing_date'][0]."-01-01");
                            }
                            else{
                                $dateSourceVal = strtotime($doc['_source']['document_publication_publishing_date'][0]);
                            }

                        }

                        if(isset($doc['_source']['document_history_not_before']) && count($doc['_source']['document_history_not_before']) > 0){
                            if($doc['_source']['document_history_not_before'][0] != "NA"){
                                if(strpos($doc['_source']['document_history_not_before'][0],"-") === false && strpos($doc['_source']['document_history_not_before'][0],"/") === false){
                                    if(strlen($doc['_source']['document_history_not_before'][0]) == 3){
                                        $notBeforeSourceVal = strtotime("0".trim($doc['_source']['document_history_not_before'][0])."-01-01");
                                    }
                                    else{
                                        $notBeforeSourceVal = strtotime(trim($doc['_source']['document_history_not_before'][0])."-01-01");
                                    }
                                }
                                else{
                                    $notBeforeSourceVal = strtotime(trim($doc['_source']['document_history_not_before'][0]));
                                }
                            }
                            else{
                                $notBeforeSourceVal = -1;
                            }

                        }


                        if(isset($doc['_source']['document_history_not_after']) && count($doc['_source']['document_history_not_after']) > 0){
                            if($doc['_source']['document_history_not_before'][0] != "NA") {
                                if(strpos($doc['_source']['document_history_not_after'][0],"-") === false && strpos($doc['_source']['document_history_not_after'][0],"/") === false){
                                    if(strlen($doc['_source']['document_history_not_after'][0]) == 3){
                                        $notAfterSourceVal = strtotime("0".trim($doc['_source']['document_history_not_after'][0])."-01-01");
                                    }
                                    else{
                                        $notAfterSourceVal = strtotime(trim($doc['_source']['document_history_not_after'][0])."-01-01");
                                    }

                                }
                                else{
                                    $notAfterSourceVal = strtotime(trim($doc['_source']['document_history_not_after'][0]));
                                }
                            }
                            else{
                                $notAfterSourceVal = -1;
                            }

                        }


                        if($dateFrom != -1 && $dateTo == -1){
                            if($dateSourceVal != -1 && $dateFrom <= $dateSourceVal){
                                $result['hits']['hits'][$i]['highlight']['document_publication_publishing_date'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_publication_publishing_date'][0].'</span>'));
                            }

                            if($notAfterSourceVal != -1 && $dateFrom <= $notAfterSourceVal){
                                $result['hits']['hits'][$i]['highlight']['document_history_not_after'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_after'][0].'</span>'));
                            }

                            if($notBeforeSourceVal != -1 && $dateFrom == $notBeforeSourceVal){
                                $result['hits']['hits'][$i]['highlight']['document_history_not_before'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_before'][0].'</span>'));
                            }
                        }
                        else if($dateFrom == -1 && $dateTo != -1){
                            if($dateSourceVal != -1 && $dateTo >= $dateSourceVal){
                                $result['hits']['hits'][$i]['highlight']['document_publication_publishing_date'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_publication_publishing_date'][0].'</span>'));
                            }

                            if($notAfterSourceVal != -1 && $notAfterSourceVal == $dateTo){
                                $result['hits']['hits'][$i]['highlight']['document_history_not_after'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_after'][0].'</span>'));
                            }

                            if($notBeforeSourceVal != -1 && $dateTo >= $notBeforeSourceVal){
                                $result['hits']['hits'][$i]['highlight']['document_history_not_before'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_before'][0].'</span>'));
                            }
                        }
                        else if($dateFrom != -1 && $dateTo != -1){

                            if($dateFrom <= $dateSourceVal && $dateSourceVal <= $dateTo){
                                $result['hits']['hits'][$i]['highlight']['document_publication_publishing_date'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_publication_publishing_date'][0].'</span>'));
                            }

                            if($notAfterSourceVal != -1 && $notBeforeSourceVal == -1 && ($notAfterSourceVal <= $dateTo && $notAfterSourceVal >= $dateFrom)){
                                $result['hits']['hits'][$i]['highlight']['document_history_not_after'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_after'][0].'</span>'));
                            }
                            else if($notBeforeSourceVal != -1 && $notAfterSourceVal == -1 && ($notBeforeSourceVal <= $dateTo && $notBeforeSourceVal >= $dateFrom)){
                                $result['hits']['hits'][$i]['highlight']['document_history_not_before'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_before'][0].'</span>'));
                            }
                            else if($notAfterSourceVal != -1 && $notBeforeSourceVal != -1){
                                if(($notBeforeSourceVal <= $dateTo && $notBeforeSourceVal >= $dateFrom)
                                && ($notAfterSourceVal <= $dateTo && $notAfterSourceVal >= $dateFrom)){
                                    $result['hits']['hits'][$i]['highlight']['document_history_not_after'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_history_not_after'][0].'</span>'));
                                }

                            }
                        }

                    }

                    if((!empty($searchData['filterData']['document_size_extent']) ||
                        !empty($searchData['filterData']['document_size_extent_to'])) && !empty($doc['_source']['document_size_extent'])) {

                        $extentFrom = -1;
                        $extentTo = -1;
                         if(!empty($searchData['filterData']['document_size_extent'])) {
                             $extentFrom = intval($searchData['filterData']['document_size_extent']);
                         }

                         if(!empty($searchData['filterData']['document_size_extent_to'])) {
                             $extentTo = intval($searchData['filterData']['document_size_extent_to']);
                         }

                         $sourceVal = intval($doc['_source']['document_size_extent'][0]);
                         if($extentFrom > -1 && $extentTo == -1){
                             if($extentFrom <= $sourceVal){
                                 $result['hits']['hits'][$i]['highlight']['document_size_extent'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_size_extent'][0].'</span>'));
                             }
                         }
                         else if($extentFrom == -1 && $extentTo > -1){
                             if($extentTo >= $sourceVal){
                                 $result['hits']['hits'][$i]['highlight']['document_size_extent'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_size_extent'][0].'</span>'));
                             }
                         }
                         else if($extentFrom > -1 && $extentTo > -1){
                             if($extentFrom <= $sourceVal && $sourceVal <= $extentTo){
                                 $result['hits']['hits'][$i]['highlight']['document_size_extent'] = array(array('<span class="laudatiofilterhighlight">'.$doc['_source']['document_size_extent'][0].'</span>'));
                             }
                         }

                    }

                    if(isset($doc['_source']['document_merged_languages'])) {
                        $mergedLanguages = $doc['_source']['document_merged_languages'];
                        $result['hits']['hits'][$i]['_source']['document_merged_languages'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedLanguages));
                    }

                    $result['hits']['hits'][$i]['_source']['visibility'] = 1;
                }//end if _index

                if(isset($doc['highlight'])){
                    foreach ($doc['highlight'] as $highlightName => $highlightData){
                        if(strpos($highlightName,".string") > -1) {
                            $newHighlightName = str_replace(".string", "",$highlightName);
                            unset($result['hits']['hits'][$i]['highlight'][$highlightName]);
                            $result['hits']['hits'][$i]['highlight'][$newHighlightName] = $highlightData;
                        }
                    }
                }

            }

            return $this->RestService->sendResponse($result['hits']['hits'], count($result['hits']['hits']).' document filter results retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }
    public function documentSearchMainFilterCount(Request $request) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $isDateSearch = false;
            $isMixedSearch = false;
            $isTokenSearch = false;
            $searchData = array();
            $hasWildCards = false;


            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            $index = config("laudatio.elasticsearch_document_index","documents");

            if((!empty($searchData['filterData']['document_publication_publishing_date']) ||
                    !empty($searchData['filterData']['document_publication_publishing_date_to'])) &&
                (empty($searchData['filterData']['document_title']) &&
                    empty($searchData['filterData']['document_author_forename']) &&
                    empty($searchData['filterData']['document_author_surname']) &&
                    empty($searchData['filterData']['document_merged_authors']) &&
                    empty($searchData['filterData']['document_publication_place']) &&
                    empty($searchData['filterData']['document_size_extent']) &&
                    empty($searchData['filterData']['document_size_extent_to']) &&
                    empty($searchData['filterData']['document_merged_languages'])
                )
            ){
                $isDateSearch = true;
            }
            else if((empty($searchData['filterData']['document_publication_publishing_date']) ||
                    empty($searchData['filterData']['document_publication_publishing_date_to'])) &&
                (empty($searchData['filterData']['document_title']) &&
                    empty($searchData['filterData']['document_author_forename']) &&
                    empty($searchData['filterData']['document_author_surname']) &&
                    empty($searchData['filterData']['document_merged_authors']) &&
                    empty($searchData['filterData']['document_publication_place']) &&
                    (!empty($searchData['filterData']['document_size_extent']) ||
                        !empty($searchData['filterData']['document_size_extent_to']) ) &&
                    empty($searchData['filterData']['document_merged_languages'])
                )
            ){
                $isTokenSearch = true;
            }
            else if(((!empty($searchData['filterData']['document_publication_publishing_date']) ||
                        !empty($searchData['filterData']['document_publication_publishing_date_to'])) &&
                    (!empty($searchData['filterData']['document_title']) ||
                        !empty($searchData['filterData']['document_author_forename']) ||
                        !empty($searchData['filterData']['document_author_surname']) ||
                        !empty($searchData['filterData']['document_merged_authors']) ||
                        !empty($searchData['filterData']['document_publication_place']) ||
                        !empty($searchData['filterData']['document_size_extent']) ||
                        !empty($searchData['filterData']['document_size_extent_to']) ||
                        !empty($searchData['filterData']['document_merged_languages'])
                    )) ||
                ((!empty($searchData['filterData']['document_size_extent']) ||
                        !empty($searchData['filterData']['document_size_extent_to'])) &&
                    (!empty($searchData['filterData']['document_title']) ||
                        !empty($searchData['filterData']['document_author_forename']) ||
                        !empty($searchData['filterData']['document_author_surname']) ||
                        !empty($searchData['filterData']['document_merged_authors']) ||
                        !empty($searchData['filterData']['document_publication_place']) ||
                        !empty($searchData['filterData']['document_publication_publishing_date']) ||
                        !empty($searchData['filterData']['document_publication_publishing_date_to']) ||
                        !empty($searchData['filterData']['document_merged_languages'])
                    ))

            ) {
                $isMixedSearch = true;
            }

            $source = [
                "document_title",
                "document_author_forename",
                "document_author_surname",
                "document_merged_authors",
                "document_editor_forename",
                "document_editor_surname",
                //"document_languages_language",
                "document_merged_languages",
                "document_languages_iso_code",
                "document_publication_place",
                //@todo: not possible to search in date if the format is just a string "document_publication_publishing_date",
                //"document_list_of_annotations_id"
                "document_list_of_annotations_id",
                "document_history_not_before",
                "document_history_not_after",
                "document_size_type",
                "document_size_extent",
                "document_genre",
                'document_publication_publishing_date',
                "publication_status",
                "laudatio_corpus_id"
            ];


            if($isDateSearch){
                $searchData["fields"] = array(
                    "document_history_not_before",
                    "document_history_not_after",
                    "document_publication_publishing_date"
                );
            }
            else if($isTokenSearch){
                $searchData["fields"] = array(
                    "document_size_extent"
                );
            }
            else if($isMixedSearch){
                $searchData["fields"] = array(
                    "document_publication_publishing_date.string",
                    "document_history_not_before.string",
                    "document_history_not_after.string",
                    "document_title",
                    "document_author_forename",
                    "document_author_surname",
                    "document_merged_authors",
                    "document_editor_forename",
                    "document_editor_surname",
                    "document_languages_language",
                    "document_merged_languages",
                    "document_languages_iso_code",
                    "document_publication_place",
                    "document_size_type",
                    "document_size_extent",
                    "document_genre",
                    "publication_status"
                );
            }
            else{
                $searchData["fields"] = array(
                    "document_title",
                    "document_author_forename",
                    "document_author_surname",
                    "document_merged_authors",
                    "document_editor_forename",
                    "document_editor_surname",
                    "document_languages_language",
                    "document_merged_languages",
                    "document_languages_iso_code",
                    "document_publication_place",
                    "document_size_type",
                    "document_size_extent",
                    "document_genre",
                    "publication_status",
                    'document_publication_publishing_date.string',
                    "document_history_not_before.string",
                    "document_history_not_after.string",
                );
            }


            $fullTextQuery = "";
            if(isset($searchData['mainSearchTerm'])) {
                $fullTextQuery = $searchData['mainSearchTerm'];
            }
            else if(isset($searchData['frontpagequery'])) {
                $fullTextQuery = $searchData['frontpagequery'];
            }

            if(strpos($fullTextQuery,'?') > -1 || strpos($fullTextQuery,'*') > -1 ){
                $hasWildCards = true;
            }


            if($isDateSearch && !$isMixedSearch && !$isTokenSearch) {
                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "document_title",
                        "document_author_forename",
                        "document_author_surname",
                        "document_merged_authors",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_languages_language",
                        "document_merged_languages",
                        "document_languages_iso_code",
                        "document_publication_place",
                        "document_size_type",
                        "document_size_extent.string",
                        "document_genre",
                        "publication_status",
                        'document_publication_publishing_date.string',
                        "document_history_not_before.string",
                        "document_history_not_after.string",
                    );
                }

                $result = $this->Elasticservice->documentDateRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],10000,$hasWildCards);
            }
            else if(!$isDateSearch && $isTokenSearch && !$isMixedSearch) {
                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "document_title",
                        "document_author_forename",
                        "document_author_surname",
                        "document_merged_authors",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_languages_language",
                        "document_merged_languages",
                        "document_languages_iso_code",
                        "document_publication_place",
                        "document_size_type",
                        "document_size_extent.string",
                        "document_genre",
                        "publication_status",
                        'document_publication_publishing_date.string',
                        "document_history_not_before.string",
                        "document_history_not_after.string",
                    );
                }

                $result = $this->Elasticservice->documentRangeSearch($searchData['filterData'], $fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],10000,$hasWildCards);
            }
            else if(!$isDateSearch && !$isTokenSearch && $isMixedSearch) {
                $searchData["fulltextFields"] = array();

                if(!empty($fullTextQuery)) {
                    $searchData["fulltextFields"] = array(
                        "document_title",
                        "document_author_forename",
                        "document_author_surname",
                        "document_merged_authors",
                        "document_editor_forename",
                        "document_editor_surname",
                        "document_languages_language",
                        "document_merged_languages",
                        "document_languages_iso_code",
                        "document_publication_place",
                        "document_size_type",
                        "document_size_extent.string",
                        "document_genre",
                        "publication_status",
                        'document_publication_publishing_date.string',
                        "document_history_not_before.string",
                        "document_history_not_after.string",
                    );
                }

                $result = $this->Elasticservice->documentMixSearch($searchData['filterData'],$fullTextQuery, $searchData["fulltextFields"], array(
                    "is_latest_publication" => true
                ),$searchData["fields"], $searchData["from"],10000,$hasWildCards);
                //Log::info("REZ: ".print_r($result,1));
            }
            else if(!$isDateSearch && !$isMixedSearch && !$isTokenSearch){
                $result =  $this->Elasticservice->filterSearch($searchData, $fullTextQuery ,$index,$searchData["from"],10000, $source, $hasWildCards);
            }

            return $this->RestService->sendResponse($result['hits']['total']['value'], $result['hits']['total']['value'].' document filter result count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }

    public function annotationSearchMainFilter(Request $request) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }
        //  Log::info("annotationSearchMainFilter: request->searchData: ".print_r($request->searchData,1));
        try {
            $index = config("laudatio.elasticsearch_annotation_index","annotations");
            $hasWildCards = false;

            $source = [
                "preparation_title",
                "preparation_annotation_id",
                "preparation_encoding_annotation_group",
                "preparation_encoding_annotation_sub_group",
                "preparation_encoding_full_name",
                "annotation_merged_formats",
                "in_documents",
                "publication_status",
                "laudatio_corpus_id"
			];

            $searchData = array();
            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            $searchData["fields"] = $source;
            $fullTextQuery = "";

            if(isset($searchData['mainSearchTerm'])) {
                $fullTextQuery = $searchData['mainSearchTerm'];
            }
            else if(isset($searchData['frontpagequery'])) {
                $fullTextQuery = $searchData['frontpagequery'];
            }

            if(strpos($fullTextQuery,'?') > -1 || strpos($fullTextQuery,'*') > -1){
                $hasWildCards = true;
            }

            $result =  $this->Elasticservice->filterSearch($searchData, $fullTextQuery ,$index,$searchData["from"],$searchData["size"], $source, $hasWildCards);

            $data = $result['hits']['hits'];

            $foundCorpusNames = array();
            for ($i = 0; $i < count($data); $i++) {
                $doc = $data[$i];
				           
                if(count($doc) > 0){
                    $corpusName = $this->LaudatioUtilService->getCorpusNameByObjectElasticsearchId('annotation',$doc['_id']);
                    $corpusElasticSearchId = $this->LaudatioUtilService->getElasticSearchIdByCorpusId($doc['_source']['laudatio_corpus_id']);
                    if(!empty($corpusName)) {
                        $result['hits']['hits'][$i]['_source']['corpus_name'] = $corpusName;
                        $result['hits']['hits'][$i]['_source']['corpus_elasticsearch_id'] = $corpusElasticSearchId;
                        $result['hits']['hits'][$i]['_source']['corpus_elasticsearch_index'] = $this->current_corpus_index;
                    }
                    $result['hits']['hits'][$i]['_source']['visibility'] = 1;
                }//end if _index

            }

            return $this->RestService->sendResponse($result['hits']['hits'], count($result['hits']['hits']).' annotation results retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }

    public function annotationSearchMainFilterCount(Request $request) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }
        //  Log::info("annotationSearchMainFilter: request->searchData: ".print_r($request->searchData,1));
        try {
            $index = config("laudatio.elasticsearch_annotation_index","annotations");
            $hasWildCards = false;

            $source = [
                "preparation_title",
                "preparation_annotation_id",
                "preparation_encoding_annotation_group",
                "preparation_encoding_annotation_sub_group",
                "preparation_encoding_full_name",
                "annotation_merged_formats",
                "in_documents",
                "publication_status",
                "laudatio_corpus_id"
            ];

            $searchData = array();
            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            $searchData["fields"] = $source;
            $fullTextQuery = "";

            if(isset($searchData['mainSearchTerm'])) {
                $fullTextQuery = $searchData['mainSearchTerm'];
            }
            else if(isset($searchData['frontpagequery'])) {
                $fullTextQuery = $searchData['frontpagequery'];
            }

            if(strpos($fullTextQuery,'?') > -1 || strpos($fullTextQuery,'*') > -1){
                $hasWildCards = true;
            }

            $result =  $this->Elasticservice->filterSearch($searchData, $fullTextQuery ,$index,$searchData["from"],10000, $source, $hasWildCards);

            return $this->RestService->sendResponse($result['hits']['total']['value'], $result['hits']['total']['value'].' annotation filter results count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }
    /**
     * /api/elasticapi/v1/corpora/latest/searchMain/count
     *
     * @param Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @responseFile responses/latest_corpus_searchMain_count.json
     * @bodyParam Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @bodyParam  from String required  The offset to start the result from
     * @bodyParam  size String required  The limit of returned results
     * @bodyParam  query String required  The search query string
     *
     * @return mixed
     */
    public function corpusSearchMainCount(Request $request){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {

            $searchData = array();
            $hasWildCards = false;
            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            if(strpos($searchData['query'],'?') > -1 || strpos($searchData['query'],'*') > -1){
                $hasWildCards = true;
            }

            $indexes = config("laudatio.elasticsearch_corpus_index","corpora");//$this->LaudatioUtilService->pickLatestCorpora();

            $source = [
                "corpus_title",
                "corpus_editor_forename",
                "corpus_editor_surname",
                "corpus_annotator_forename",
                "corpus_annotator_surname",
                "corpus_publication_publisher",
                "corpus_merged_formats",
                "annotation_encoding_tool",
                "corpus_encoding_project_description",
                "corpus_publication_license",
                "corpus_merged_languages",
                "corpus_languages_iso_identifier",
                "corpus_publication_publication_date.string",
                //"corpus_size_value",
                "corpus_size_type",
                "corpus_document_genre"
            ];

            $searchData["fields"] = $source;

            //re-add annotation_id
            if(!array_key_exists("corpus_publication_publication_date",$source)){
                array_push($source,"corpus_publication_publication_date");
            }

            if(!array_key_exists("corpus_size_value",$source)){
                array_push($source,"corpus_size_value");
            }

            //re-add annotation_id
            if(!array_key_exists("annotation_id",$source)){
                array_push($source,"annotation_id");
            }

            //re-add corpus_id
            if(!array_key_exists("corpus_id",$source)){
                array_push($source,"corpus_id");
            }

            //re-add has_documents
            if(!array_key_exists("has_documents",$source)){
                array_push($source,"has_documents");
            }

            //re-add has_annotations
            if(!array_key_exists("has_annotations",$source)){
                array_push($source,"has_annotations");
            }

            //re-add laudatio_corpus_id
            if(!array_key_exists("laudatio_corpus_id",$source)){
                array_push($source,"laudatio_corpus_id");
            }

            //re-add publication_status
            if(!array_key_exists("publication_status",$source)){
                array_push($source,"publication_status");
            }

            //re-add corpus_documents
            if(!array_key_exists("corpus_documents",$source)){
                array_push($source,"corpus_documents");
            }

            if($hasWildCards) {
                $result  = $this->Elasticservice->searchMainWildCards($searchData,$indexes,$source,$searchData["from"],10000);
            }
            else {
                $result  = $this->Elasticservice->searchMain($searchData,$indexes,$source,$searchData["from"],10000);
            }

            Log::info("corpusSearchMainCount: ".$result['hits']['total']['value']);
            return $this->RestService->sendResponse($result['hits']['total']['value'], ' Latest corpora count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }

    }


    /**
     *
     * /api/elasticapi/v1/documents/latest/searchMain
     *
     * @param Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @responseFile responses/latest_document_searchMain.json
     * @bodyParam Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @bodyParam  from String required  The offset to start the result from
     * @bodyParam  size String required  The limit of returned results
     * @bodyParam  query String required  The search query string
     *
     * @return mixed
     */
    public function documentSearchMain(Request $request){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $hasWildCards = false;
            $searchData = array();

            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }


            if(strpos($searchData['query'],'?') > -1 || strpos($searchData['query'],'*') > -1){
                $hasWildCards = true;
            }

            $indexes = config("laudatio.elasticsearch_document_index","documents");//$this->LaudatioUtilService->pickLatestDocuments();

            $source = [
                "document_title",
                "document_author_forename",
                "document_author_surname",
                "document_editor_forename",
                "document_editor_surname",
                "document_merged_authors",
                "document_merged_editors",
                "document_history_original_place",
                //"document_history_original_title",
                "document_merged_languages",
                "document_languages_iso_code",
                "document_publication_place",
                "document_publication_publishing_date.string",
                "document_history_not_before.string",
                "document_history_not_after.string",
                //"document_list_of_annotations_id",
                "document_size_type",
                //"document_size_extent",
                "document_genre"
            ];

            $searchData["fields"] = $source;

            //@todo:  Do we have dates in our searchstring?, if so add the date_field until fix for the index mapping
            $dates = $this->LaudatioUtilService->convert_date_format(array($searchData["query"]));
            if(count($dates) > 0) {
                array_push($source,"document_publication_publishing_date");
            }


            //re-add publishing date
            if(!array_key_exists('document_publication_publishing_date',$source)){
                array_push($source,'document_publication_publishing_date');
            }

            if(!array_key_exists("publication_status",$source)){
                array_push($source,"publication_status");
            }


            //re-add original publishing date
            if(!array_key_exists('document_history_original_date',$source)){
                array_push($source,'document_history_original_date');
            }

            //re-add document_list_of_annotations_id
            if(!array_key_exists("document_list_of_annotations_id",$source)){
                array_push($source,"document_list_of_annotations_id");
            }

            //re-add document_size_extent
            if(!array_key_exists("document_size_extent",$source)){
                array_push($source,"document_size_extent");
            }

            if($hasWildCards) {
                $result  = $this->Elasticservice->searchMainWildCards($searchData,$indexes,$source,$searchData["from"],$searchData["size"]);
            }
            else {
                $result  = $this->Elasticservice->searchMain($searchData,$indexes,$source,$searchData["from"],$searchData["size"]);
            }

            $data = $result['hits']['hits'];

            $foundCorpusNames = array();
            for ($i = 0; $i < count($data); $i++) {
                $doc = $data[$i];
                if(count($doc) > 0  && isset($doc["_index"])){
                    $corpusName = $this->LaudatioUtilService->getCorpusNameByObjectElasticsearchId('document',$doc['_id']);
                    if(!empty($corpusName)) {
                        $result['hits']['hits'][$i]['_source']['corpus_name'] = $corpusName;
                    }

                    if(isset($doc['highlight'])){
                        $newHighlights = array();
                        foreach ($doc['highlight'] as $highlightName => $highlightData){
                            $newHighlightName = str_replace(".string", "",$highlightName);
                            $newHighlights[$newHighlightName] = $highlightData;
                        }

                        $result['hits']['hits'][$i]['highlight'] = $newHighlights;
                    }

                    if(isset($doc['_source']['document_merged_languages'])) {
                        $mergedLanguages = $doc['_source']['document_merged_languages'];
                        $result['hits']['hits'][$i]['_source']['document_merged_languages'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedLanguages));
                    }

                    $result['hits']['hits'][$i]['_source']['visibility'] = 1;
                }//end if _index

            }

            return $this->RestService->sendResponse($result['hits']['hits'], count($result['hits']['hits']).' document results retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/documents/latest/searchMain/count
     *
     * @param Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @responseFile responses/latest_document_searchMain_count.json
     * @bodyParam Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @bodyParam  from String required  The offset to start the result from
     * @bodyParam  size String required  The limit of returned results
     * @bodyParam  query String required  The search query string
     *
     * @return mixed
     */
    public function documentSearchMainCount(Request $request){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $searchData = array();
            $hasWildCards = false;

            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            if(strpos($searchData['query'],'?') > -1 || strpos($searchData['query'],'*') > -1){
                $hasWildCards = true;
            }

            $indexes = config("laudatio.elasticsearch_document_index","documents");

            $source = [
                "document_title",
                "document_author_forename",
                "document_author_surname",
                "document_editor_forename",
                "document_editor_surname",
                "document_merged_authors",
                "document_merged_editors",
                "document_history_original_place",
                //"document_history_original_title",
                "document_merged_languages",
                "document_languages_iso_code",
                "document_publication_place",
                "document_publication_publishing_date.string",
                "document_history_not_before.string",
                "document_history_not_after.string",
                "document_size_type",
                "document_genre"];


            $searchData["fields"] = $source;

            //@todo:  Do we have dates in our searchstring?, if so add the date_field until fix for the index mapping
            $dates = $this->LaudatioUtilService->convert_date_format(array($searchData["query"]));
            if(count($dates) > 0) {
                array_push($source,"document_publication_publishing_date");
            }



            //re-add publishing date
            if(!array_key_exists('document_publication_publishing_date',$source)){
                array_push($source,'document_publication_publishing_date');
            }

            //re-add original publishing date
            if(!array_key_exists('document_history_original_date',$source)){
                array_push($source,'document_history_original_date');
            }

            //re-add document_list_of_annotations_id
            if(!array_key_exists("document_list_of_annotations_id",$source)){
                array_push($source,"document_list_of_annotations_id");
            }

            //re-add document_size_extent
            if(!array_key_exists("document_size_extent",$source)){
                array_push($source,"document_size_extent");
            }

            if($hasWildCards) {
                $result  = $this->Elasticservice->searchMainWildCards($searchData,$indexes,$source,$searchData["from"],10000);
            }
            else {
                $result  = $this->Elasticservice->searchMain($searchData,$indexes,$source,$searchData["from"],10000);
            }

            Log::info("documentSearchMainCount: ".$result['hits']['total']['value']);

            return $this->RestService->sendResponse($result['hits']['total']['value'], ' Latest document result count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/documents/latest/searchMain
     *
     * @param Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @responseFile responses/latest_annotation_searchMain.json
     * @bodyParam Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @bodyParam  from String required  The offset to start the result from
     * @bodyParam  size String required  The limit of returned results
     * @bodyParam  query String required  The search query string
     *
     * @return mixed
     */
    public function annotationSearchMain(Request $request){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $indexes = config("laudatio.elasticsearch_annotation_index","annotations");//$this->LaudatioUtilService->pickLatestAnnotations();
            $hasWildCards = false;

            $source = [
                "preparation_title",
                "preparation_encoding_annotation_group",
                "preparation_encoding_annotation_sub_group",
                "preparation_encoding_full_name",
                "annotation_merged_formats",
                "in_documents",
                "publication_status",
                "laudatio_corpus_id"
            ];

            $searchData = array();
            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            if(strpos($searchData['query'],'?') > -1 || strpos($searchData['query'],'*') > -1){
                $hasWildCards = true;
            }

            $searchData["fields"] = array(
                "preparation_title",
                "preparation_encoding_annotation_group",
                "preparation_encoding_annotation_sub_group",
                "annotation_merged_formats",
            );


            if($hasWildCards) {
                $result  = $this->Elasticservice->searchMainWildCards($searchData,$indexes,$source,$searchData["from"],$searchData["size"]);
            }
            else {
                $result  = $this->Elasticservice->searchMain($searchData,$indexes,$source,$searchData["from"],$searchData["size"]);
            }

            $data = $result['hits']['hits'];
            $foundCorpusNames = array();
            for ($i = 0; $i < count($data); $i++) {
                $doc = $data[$i];
                if(count($doc) > 0  && isset($doc["_index"])){
                    $corpusName = $this->LaudatioUtilService->getCorpusNameByObjectElasticsearchId('annotation',$doc['_id']);
                    if(!empty($corpusName)) {
                        $result['hits']['hits'][$i]['_source']['corpus_name'] = $corpusName;
                    }

                    /*
                    if(isset($doc['_source']['annotation_merged_formats'])) {
                        $mergedFormats = $doc['_source']['annotation_merged_formats'];
                        $result['hits']['hits'][$i]['_source']['annotation_merged_formats'] = $this->LaudatioUtilService->removeMergedDuplicates($this->LaudatioUtilService->removeMergedNA($mergedFormats));
                    }
                    */

                    $result['hits']['hits'][$i]['_source']['visibility'] = 1;
                }//end if _index

            }

            return $this->RestService->sendResponse($result['hits']['hits'], count($result['hits']['hits']).' annotation results retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }

    /**
     *
     * /api/elasticapi/v1/documents/latest/searchMain/count
     *
     * @param Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @responseFile responses/latest_annotation_searchMain_count.json
     * @bodyParam Request $request required  {"searchData": {"from": '0', "size" : '10', "query": 'FOO'}}
     * @bodyParam  from String required  The offset to start the result from
     * @bodyParam  size String required  The limit of returned results
     * @bodyParam  query String required  The search query string
     *
     * @return mixed
     */
    public function annotationSearchMainCount(Request $request){
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $indexes = config("laudatio.elasticsearch_annotation_index","annotations");//$this->LaudatioUtilService->pickLatestAnnotations();


            $source = [
                "preparation_title",
                "preparation_annotation_id",
                "preparation_encoding_annotation_group",
                "preparation_encoding_annotation_sub_group",
                "preparation_encoding_full_name",
                "annotation_merged_formats",
                "in_documents",
                "publication_status"];

            $searchData = array();
            $hasWildCards = false;

            if(is_array($request->searchData)){
                $searchData = $request->searchData;
            }
            else {
                $searchData = json_decode($request->searchData,true);
            }

            if(strpos($searchData['query'],'?') > -1 || strpos($searchData['query'],'*') > -1){
                $hasWildCards = true;
            }

            $searchData["fields"] = array(
                "preparation_title",
                "preparation_encoding_annotation_group",
                "preparation_encoding_annotation_sub_group",
                "annotation_merged_formats",
            );

            if($hasWildCards) {
                $result  = $this->Elasticservice->searchMainWildCards($searchData,$indexes,$source,$searchData["from"],10000);
            }
            else {
                $result  = $this->Elasticservice->searchMain($searchData,$indexes,$source,$searchData["from"],10000);
            }

            Log::info("annotationSearchMainCount: ".$result['hits']['total']['value']);
            return $this->RestService->sendResponse($result['hits']['total']['value'], ' Latest annotation result count retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }

    /**
     *
     * /api/elasticapi/v1/annotations/{id}/{index}
     *
     * @param  $id String required 
     * @param  $index String required
     *
     * @queryParam id required  The identifier of the document
     * @queryParam index required  The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticAnnotation($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $annotation  = $this->Elasticservice->getPublishedAnnotation($id,$index);
            return $this->RestService->sendResponse($annotation, 'Annotation retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /**
     *
     * /api/elasticapi/v1/annotations/{id}/{index}/documents
     *
     * @param  $id String required 
     * @param  $index String required
     * @responseFile responses/annotations_by_id_and_index.json
     * @queryParam id required  The identifier of the document
     * @queryParam index required  The elastic search index of the document
     *
     * @return mixed
     */
    public function getElasticDocumentsByAnnotation($id,$index) {
        try {
            $pingresponse = $this->Elasticservice->pingServer(true);
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),503);
        }

        try {
            $documents  = $this->Elasticservice->getAllPublishedDocumentsByAnnotation($id,$index);
            return $this->RestService->sendResponse($documents, count($documents). ' documents retrieved successfully.');
        } catch (LaudatioElasticsearchException $laudatioElasticsearchException) {
            return $this->RestService->sendError('REST API Error.', $laudatioElasticsearchException->getMessage(),404);
        }
    }


    /** END ELASTICSEARCH REST API **/


    /**
     *
     * api: /api/adminapi/deleteFormatContent
     *
     * destroyFormatFileContent: delete corpus data files
     *
     * @param Request $request
     * @return mixed JSON
     */
    public function destroyFormatFileContent(Request $request) {
        $result = array();
        $status = "success";

        try{
            $corpusid = $request->input('corpusid');
            $corpusPath = $request->input('path');
            $auth_user_name = $request->input('auth_user_name');
            $auth_user_id = $request->input('auth_user_id');
            $auth_user_email = $request->input('auth_user_email');
            $toBeDeletedCollection = $request->input('tobedeleted');
            Log::info("destroyFormatFileContent: ".print_r($toBeDeletedCollection,1));
            foreach ($toBeDeletedCollection as $toBeDeleted) {
                if(!is_array($toBeDeleted)){
                    $toBeDeleted = json_decode($toBeDeleted);
                }

                $corpusFile = CorpusFile::findOrFail($toBeDeleted['databaseId']);
                if($corpusFile->format != 0) {
                    $format = Format::findOrFail($corpusFile->format);
                    $fileDeleteResult = $this->GitRepoService->deleteFormatFile($this->flysystem,$corpusPath."/".$format->name."/".$toBeDeleted['fileName'],$auth_user_name,$auth_user_email);
                    Log::info("fileDeleteResult: ".print_r($fileDeleteResult,1));
                }

            }

            if($fileDeleteResult) {
                if(count($corpusFile->corpus()->get()) > 0) {
                    $corpusFile->corpus()->dissociate();
                    $corpusFile->delete();
                    $pushResult = $this->GitRepoService->pushFiles($corpusPath,$corpusid,$auth_user_name);
                }

                $corpusPath = $this->LaudatioUtilService->getCorpusAndProjectPathByCorpusId($corpusid);
                $dataFilePath = $corpusPath."/CORPUS-DATA";
                Log::info("DELETING CACHE: : ".print_r($dataFilePath,1));
                $this->LaudatioUtilService->emptyCorpusDataFileCache($dataFilePath,$corpusid);
                //hmmmmmmmm, works, but not a good idea..

                //$this->LaudatioUtilService->emptyFlySystemFileCache();
                $result['delete_content_response']  = "Corpus file content was successfully deleted";
            }
            else{
                $status = "error";
                $result['delete_content_response']  = "There was a problem deleting the Corpus file content. The error was: ($e) A message has been sent to the site administrator. Please try again later";
            }
        }
        catch (\Exception $e) {
            $status = "error";
            $result['delete_content_response']  = "There was a problem deleting the Corpus file content. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }


        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }


    /**
     *
     * api: /api/adminapi/flushFileCache
     *
     * flushFileCache
     *
     * THis method flushes the Flysystem file metadata cache for a given corpus
     *
     * @param Request $request
     * @return mixed JSON
     */
    public function flushFileCache(Request $request){
        $result = array();
        $status = "success";
        try {

            $corpusId = $request->input('corpusid');
            $filePath = $request->input('filepath');

            $this->LaudatioUtilService->emptyCorpusFileInfoCache($filePath,$corpusId);
            $this->LaudatioUtilService->emptyCorpusFileCache($filePath,$corpusId);


            if(strpos($filePath,"corpus") !== false) {
                $result['delete_filecache_response']  = "The corpus file cache was successfully deleted";
            }
            else if(strpos($filePath,"document") !== false) {
                $result['delete_filecache_response']  = "The document file cache was successfully deleted";
            }
            else if(strpos($filePath,"annotation") !== false) {
                $result['delete_filecache_response']  = "The annotation file cache was successfully deleted";
            }
            else if(strpos($filePath,"CORPUS-DATA") !== false) {

                Log::info("DELETING CACHE FOR CORPUS-DATA: ".$filePath."".$corpusId);
                $this->LaudatioUtilService->emptyCorpusDataFileCache($filePath,$corpusId);
                $result['delete_filecache_response']  = "The CORPUS-DATA file cache was successfully deleted";
            }

        }
        catch(\Exception $e) {
            $status = "error";
            $result['delete_filecache_response']  = "There was a problem flushing the Corpus file cache. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }

        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }


    public function clearCache(){
        $this->LaudatioUtilService->emptySearchCache();
    }

    public function flushCacheTest(){
        Cache::forget("getCorpusDataFiles_8whatnow/fuerstinnenkorrespondenz_1-0/CORPUS-DATA");
    }
    /**
     *
     * api: /api/dbapi/getdatabaseid
     *
     * getDatabaseIdByFilenameAndCorpusId
     * @param Request $request
     * @return mixed
     */
    public function getDatabaseIdByFilenameAndCorpusId(Request $request){
        $result = array();
        $status = "success";

        try{
            $dataArray = $request->input('dataArray');

            foreach($dataArray as $item) {
                $type = $item['type'];
                if(strpos($item['type'],'format') !== false) {
                    $type = 'CORPUS-DATA';
                }
                $model = $this->LaudatioUtilService->getModelByFileName($item['filename'], $type, false ,$item['corpusid']);
                $modelarray = array();
                if($model) {
                    if($type == 'CORPUS-DATA'){
                        $modelarray['format'] = $model[0]->formatName;
                    }
                    else {
                        $modelarray['format'] = '';
                    }
                    $modelarray['type'] = $type;
                    $modelarray['database_id'] = $model[0]->id;
                    $modelarray['file_name'] = $model[0]->file_name;
                    $modelarray['created_at'] = $model[0]->created_at;
                    $modelarray['content_response']  = "The document id was successfully fetched";
                    array_push($result,$modelarray);
                }
            }
        }
        catch (\Exception $e) {
            $status = "error";
            $result['content_response']  = "There was a problem fetching the database id. The error was: ($e)";
        }


        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }


    /**
     *
     * /api/corpusapi/formats
     *
     * @param Request $request
     * @return mixed
     */
    public function getAllFormats(Request $request) {
        $result = array();
        $status = "success";

        try{
            $formats =  Format::all();

            $result['format_content_response']  = "The format category were successfully fetched";

            $foundFormats = array();
            foreach ($formats as $format) {
                $foundFormats[$format->id] = $format->name;
            }

            $result['formats'] = $foundFormats;
        }
        catch(\Exception $e){
            $status = "error";
            $result['format_content_response']  = "There was a problem fetching the format categories. The error was: ($e)";
        }
        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }

    /**
     *
     * /api/adminpoi/createFormat
     *
     * @param Request $request
     * @return mixed
     */
    public function createFormat(Request $request) {
        $result = array();
        $status = "success";

        try{
            $format_name = strtolower($request->input('formattitle'));
            $format = new Format();
            $format->name = $format_name;
            $format->save();

            $result['format_id'] = $format->id;
            $result['format_name'] = $format->name;
            $result['format_create_response']  = "The format category <em>$format_name</em> was successfully created";
        }
        catch(\Exception $e){
            $status = "error";
            $result['format_content_response']  = "There was a problem creating the format category. The error was: ($e)";
        }
        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }

    public function importAnnisLinks(Request $request){
        $this->LaudatioUtilService->importAnnisLinks();
    }


    public function testCache(Request $request){
        $corpusId = $request->input('corpusId');
        $path = $request->input('path');
        Log::info("EMPTYING: getCorpusDataFiles_".$corpusId.$path);
        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $listAllCorpusDatafiles = $redis->keys('*getCorpusDataFiles*');
        Log::info("testCache: listAllCorpusDatafiles: ".print_r($listAllCorpusDatafiles,1));
        foreach ($listAllCorpusDatafiles as $lkey){
            $result = $redis->del($lkey);
        }
        ;
        Log::info("redis: : ".print_r($redis->keys('*'),1));

        //Log::info("storage: : ".print_r(Storage::disk('redis')->getAdapter()->getCache(),1));
        Log::info("FLYBOY: ".print_r($this->flysystem->listContents('jojjemeg/germanc_1-0/CORPUS-DATA/THICK'),1));
        if(count($this->flysystem->listContents('jojjemeg/germanc_1-0/CORPUS-DATA/THICK')) < 1){
            Log::info("COUNT WAS : ".count($this->flysystem->listContents('jojjemeg/germanc_1-0/CORPUS-DATA/THICK'))." SO WE DELETE THE FORMAT FOLDER");
            foreach ($this->flysystem->listContents('jojjemeg/germanc_1-0/CORPUS-DATA') as $content){
                if($content['basename'] == "THICK"){
                    $deletedir = $this->flysystem->deleteDir($content['path']);
                    Log::info("deletedir: ".print_r($deletedir,1));
                    $delete = $this->flysystem->delete($content['path']);
                    Log::info("delete: ".print_r($delete,1));
                }

            }


            Log::info("FLYBOY AFTER DELETE: ".print_r($this->flysystem->listContents('jojjemeg/germanc_1-0'),1));
        }
    }
}
