<?php

namespace App\Http\Controllers;

use App\Exceptions\CorpusRepositoryFileSystemException;
use App\ExternalEnvironment;
use Illuminate\Http\Request;
use App\Corpus;
use App\CorpusProject;
use App\MessageBoard;
use App\User;
use App\Role;
use App\Format;
use App\Invite;
use App\Custom\GitRepoInterface;
use App\Custom\LaudatioUtilsInterface;
use GrahamCampbell\Flysystem\FlysystemManager;
use App\Custom\ElasticsearchInterface;
use App\Custom\GitLabInterface;
use App\Exceptions\GitInitException;
use Response;
use JavaScript;
use Log;

/**
 * Class CorpusController
 * @package App\Http\Controllers
 */
class CorpusController extends Controller
{


    protected $GitRepoService;
    protected $LaudatioUtilService;
    protected $elasticService;
    protected $flysystem;
    protected $GitLabService;

    /**
     * CorpusController constructor.
     * @param GitRepoInterface $GitRepoService
     * @param FlysystemManager $flysystem
     * @param LaudatioUtilsInterface $laudatioUtils
     * @param ElasticsearchInterface $elasticService
     */
    public function __construct(GitRepoInterface $GitRepoService, FlysystemManager $flysystem,LaudatioUtilsInterface $laudatioUtils,ElasticsearchInterface $elasticService,GitLabInterface $GitLabService)
    {
        $this->GitRepoService = $GitRepoService;
        $this->LaudatioUtilService = $laudatioUtils;
        $this->flysystem = $flysystem;
        $this->elasticService = $elasticService;
        $this->GitLabService = $GitLabService;
    }


    /**
     *
     * index: Display a listing of the resource.
     *
     * @return \Illuminate\Http\Response
     */
    public function index()
    {
        $isLoggedIn = \Auth::check();
        $user = \Auth::user();
        $corpora = Corpus::latest()->get();

        $corpusProjects = array();

        foreach ($corpora as $corpus){
            $corpusProjectsTemp = $corpus->corpusprojects()->get();
            $cp = isset($corpusProjectsTemp[0]) ? $corpusProjectsTemp[0] : false;
            if($cp) {
                $corpusProjects[$corpus->id] = $corpusProjectsTemp[0]->directory_path;
            }
        }

        return view('/project.corpus.index', compact('corpora'))
            ->with('isLoggedIn', $isLoggedIn)
            ->with('corpusProjects', $corpusProjects)
            ->with('user',$user);
    }


    /**
     * create a Corpus
     * @param CorpusProject $corpusproject
     * @return \Illuminate\Http\RedirectResponse
     */
    public function create(CorpusProject $corpusproject)
    {
        $isLoggedIn = \Auth::check();
        $user = \Auth::user();
        $corpusProjectPath = $corpusproject->directory_path;
        $corpusProjectId = $corpusproject->gitlab_id;

        // Create the directory structure for the Corpus
        $corpusProjectPath = $corpusproject->directory_path;
        $corpusCount = count($corpusproject->corpora()->get());
        $corpus_name = "Untitled_".$corpusProjectId."_".($corpusCount++)."_".time();

        try{
            $corpusPath = $this->GitRepoService->createCorpusFileStructure($this->flysystem,$corpusProjectPath,$corpus_name,$user);
        }
        catch (GitInitException $gitInitException) {
            //dd($gitInitException);
            $message = $gitInitException->createUserFacingMessage($gitInitException->getMessage());
            //Log the error to file
            $gitInitException->logErrorMessage($gitInitException->getMessage()." : corpus directory path: ".$corpusProjectPath."/".$corpus_name,['file' => __FILE__, 'line' => __LINE__]);

            //roll back the creation of the filestructure
            $corpusPath = $this->GitRepoService->normalizeString($corpus_name);
            $deleted = $this->GitRepoService->deleteUntrackedCorpusFileStructure($this->flysystem, $corpusProjectPath."/".$corpusPath);
            if($deleted) {
                Log::stack(['daily', 'errorlog'])->error("Deleted : corpus directory path: ".$corpusProjectPath."/".$corpus_name." after unsuccessful GIT INIT");
            }
            return back()->withError($message);
        }


        $corpus = null;
        $user_role = array();
        if($corpusPath){
            $corpus = Corpus::create([
                "name" => $corpus_name,
                "description" => "",
                'directory_path' => $corpusPath
            ]);

            $corpus->corpusprojects()->attach($corpusproject);


            $corpusAdminRole = Role::findById(3);
            //$user->roles()->sync($corpusAdminRole);
            if($user) {
                if(!$user->roles->contains($corpusAdminRole)){
                    $user->roles()->attach($corpusAdminRole);
                }

                $corpus->users()->save($user,['role_id' => 3]);
                $user_role['user_name'] = $user->name;
                $user_role['user_id'] = $user->id;
                $user_role['role_name'] = $corpusAdminRole->name;
            }
        }


        $filePath = $corpusProjectPath.'/'.$corpus->directory_path;

        $corpus_data = array(
            'name' => $corpus->name,
            'project_name' => $corpusproject->name,
            'filepath' => $filePath,
            'admin' => $user_role

        );
/*
        $this->LaudatioUtilService->emptyCorpusFileCache($filePath,$corpus->id);
        $this->LaudatioUtilService->emptyCorpusFileInfoCache($filePath,$corpus->id);
        $this->LaudatioUtilService->emptyCorpusFileCache($filePath.'/TEI-HEADERS/corpus',$corpus->id);
        $this->LaudatioUtilService->emptyCorpusFileInfoCache($filePath.'/TEI-HEADERS/corpus',$corpus->id);
        */
        return redirect()->route('corpus.edit', compact('corpus'))
            ->with('isLoggedIn', $isLoggedIn)
            ->with('corpus_data', $corpus_data)
            ->with('user',$user);


    }

    /**
     *
     * store:Store a newly created resource in storage.
     *
     * @param Request $request
     * @return \Illuminate\Http\RedirectResponse
     * @throws \Illuminate\Validation\ValidationException
     */
    public function store(Request $request)
    {
        $this->validate(request(), [
            //'corpus_name' => 'required',
            'corpus_description' => 'required',
            'corpusProjectId' => 'required'
        ]);


        $corpusProject = CorpusProject::find(request('corpusProjectId'));
        $corpusProjectPath = $corpusProject->directory_path;
        $corpusProjectId = $corpusProject->gitlab_id;

        // Create the directory structure for the Corpus
        $corpusProjectPath = $corpusProject->directory_path;
        $corpusCount = count($corpusProject->corpora()->get());
        $corpus_name = "Untitled_".$corpusProjectId."_".($corpusCount++);
        $corpusPath = $this->GitRepoService->createCorpusFileStructure($this->flysystem,$corpusProjectPath,$corpus_name);


        if($corpusPath){
            $corpus = Corpus::create([
                "name" => $corpus_name,
                "description" => request('corpus_description'),
                'directory_path' => $corpusPath
            ]);

            $corpus->corpusprojects()->attach($corpusProject);
        }

        return redirect()->route('project.corpora.index');
    }

    /**
     *
     * store_untitled: Store a newly created resource in storage.
     *
     * @param Request $request
     * @return \Illuminate\Http\RedirectResponse
     * @throws \Illuminate\Validation\ValidationException
     */
    public function store_untitled(Request $request)
    {
        $this->validate(request(), [
            //'corpus_name' => 'required',
            'corpus_description' => 'required',
            'corpusProjectId' => 'required'
        ]);

        $corpusProject = CorpusProject::find(request('corpusProjectId'));

        // Create the directory structure for the Corpus
        $corpusProjectPath = $corpusProject->directory_path;
        $corpusProjectId = $corpusProject->gitlab_id;
        $corpusPath = null;
        /**
        if(request('corpus_name')){
            $corpusPath = $this->GitRepoService->createCorpusFileStructure($this->flysystem,$corpusProjectPath,request('corpus_name'));
        }
        else{

        }
         * */

        $corpusPath = "Untitled".$corpusProjectId;
        if($corpusPath){
            $corpus = Corpus::create([
                "name" => "Untitled",
                "description" => request('corpus_description'),
                'directory_path' => $corpusPath
            ]);
            $corpus->corpusprojects()->attach($corpusProject);
        }

        return redirect()->route('project.corpora.index');
    }

    /**
     *
     * show
     *
     * @param Corpus $corpus
     * @param string $path
     * @param string $show
     * @return $this
     */
    public function show(Corpus $corpus,$path = "",$show = "")
    {
        $isLoggedIn = \Auth::check();
        $user = \Auth::user();

        $corpusProjects = $corpus->corpusprojects()->get();
        $corpusProject_directory_path = '';

        $pathArray = explode("/",$path);
        end($pathArray);
        $last_id=key($pathArray);

        if(count($corpusProjects) == 1) {
            $corpusProject_directory_path = $corpusProjects->first()->directory_path;
        }
        else{
            // what to do when we can assign corpora to many projects?
        }

        $corpusPath = "";
        $corpus_directory_path = $corpus->directory_path;
        if($path == ""){
            $corpusPath = $corpusProject_directory_path."/".$corpus_directory_path;
        }
        else{
            $corpusPath = $path;
        }


        $fileData = array(
            "corpusData" => array(
                'path' => $corpusPath.'/CORPUS-DATA',
                'hasdir' => false
            ),
            "corpusDataFolder" => "",
            "headerData" => array(
                'path' => $corpusPath.'/TEI-HEADERS',
                'hasdir' => false
            ),
            "headerDataFolder" => "",
            "folderType" => ""
        );

        $folder = "";
        $folderType = "";
        $user_roles = array();
        if(strpos($corpusPath,"Untitled") === false){
            $corpusBasePath = $pathArray[0]."/".$pathArray[1];
            if(strpos($corpusPath,"CORPUS-DATA") !== false && strpos($corpusPath,"TEI-HEADERS") === false){
                $corpusData = $this->GitRepoService->getCorpusDataFiles($this->flysystem,$corpusPath,$corpus->id);
                dd($corpusData);
                $headerData = $this->GitRepoService->getCorpusFileInfo($this->flysystem,$corpusBasePath.'/TEI-HEADERS',$corpus->id);
                $folderType = "CORPUS-DATA";

            }
            else if(strpos($corpusPath,"TEI-HEADERS") !== false && strpos($corpusPath,"CORPUS-DATA") === false){
                $corpusData = $this->GitRepoService->getCorpusDataFiles($this->flysystem,$corpusBasePath.'/CORPUS-DATA',$corpus->id);
                $headerData = $this->GitRepoService->getCorpusFileInfo($this->flysystem,$corpusPath,$corpus->id);
                $folderType = "TEI-HEADERS";
            }
            else{
                $corpusData = $this->GitRepoService->getCorpusDataFiles($this->flysystem,$corpusPath.'/CORPUS-DATA',$corpus->id);
                $headerData = $this->GitRepoService->getCorpusFileInfo($this->flysystem,$corpusPath.'/TEI-HEADERS',$corpus->id);
            }
            $corpusDataFolder = substr($corpusData['path'],strrpos($corpusData['path'],"/")+1);
            $headerDataFolder = substr($headerData['path'],strrpos($headerData['path'],"/")+1);
            $fileData = array(
                "corpusData" => $corpusData,
                "corpusDataFolder" => $corpusDataFolder,
                "headerData" => $headerData,
                "headerDataFolder" => $headerDataFolder,
                "folderType" => $folderType
            );


            $corpusUsers = $corpus->users()->get();

            foreach ($corpusUsers as $corpusUser){
                if(!isset($user_roles[$corpusUser->id])){
                    $user_roles[$corpusUser->id] = array();
                }

                $role = Role::find($corpusUser->pivot->role_id);
                if($role){
                    array_push($user_roles[$corpusUser->id],$role->name);
                }

            }
        }
    }

    /**
     *
     * edit
     *
     * @param Corpus $corpus
     * @return $this
     */
    public function edit(Corpus $corpus)
    {
        $isLoggedIn = \Auth::check();
        $user = \Auth::user();
        $corpusProjects = $corpus->corpusprojects()->get();
        $corpusproject = null;
        $corpusProject_directory_path = '';

        if(count($corpusProjects) == 1) {
            $corpusproject = $corpusProjects->first();
            $corpusProject_directory_path = $corpusproject->directory_path;
        }

        $path = $corpusProject_directory_path.'/'.$corpus->directory_path;

        $corpusUsers = $corpus->users()->get();
        $corpusProjectUsers = $corpusproject->users()->get();

        $admin_roles = array();
        $user_roles = array();

        foreach($corpusProjectUsers as $corpusProjectUser) {
            if(!isset($admin_roles[$corpusProjectUser->id])){
                $admin_roles[$corpusProjectUser->id] = array();
            }

            if(!isset($user_roles[$corpusProjectUser->id])){
                $user_roles[$corpusProjectUser->id] = array();
            }

            $corpusProjectUserRole = array();
            $corpusProjectRole = Role::find($corpusProjectUser->pivot->role_id);
            $corpusProjectUserRole['user_name'] = $corpusProjectUser->name;
            $corpusProjectUserRole['user_affiliation'] = $corpusProjectUser->affiliation;
            $corpusProjectUserRole['user_id'] = $corpusProjectUser->id;
            $corpusProjectUserRole['role_name'] = $corpusProjectRole->name;
            $corpusProjectUserRole['role_id'] = $corpusProjectRole->id;

            array_push($admin_roles[$corpusProjectUser->id],$corpusProjectUserRole);
            array_push($user_roles[$corpusProjectUser->id],$corpusProjectUserRole);
        }//end foreach


        // get all roles
        $roleCollection = Role::all();
        $roles = array();
        foreach ($roleCollection as $role){
            if($role->id != 1) {
                $roles[$role->id] = $role->name;
            }

        }
        ksort($roles);

        //dd($corpusUsers);

        foreach ($corpusUsers as $corpusUser){
            $user_role = array();
            $admin_role = array();

            $corpusRole = Role::find($corpusUser->pivot->role_id);

            if($corpusRole){
                if($corpusRole->hasPermissionTo('Can create corpus')){
                    $admin_role['user_name'] = $corpusUser->name;
                    $admin_role['user_id'] = $corpusUser->id;
                    $admin_role['user_affiliation'] = $corpusUser->affiliation;
                    $admin_role['role_name'] = $corpusRole->name;
                    $admin_role['role_id'] = $corpusRole->id;
                }

                if(null != $admin_role) {
                    if(!array_key_exists($corpusUser->id,$admin_roles)) {
                        $admin_roles[$corpusUser->id] = array();
                    }
                    array_push($admin_roles[$corpusUser->id], $admin_role);
                }

                $user_role['user_name'] = $corpusUser->name;
                $user_role['user_affiliation'] = $corpusUser->affiliation;
                $user_role['user_id'] = $corpusUser->id;
                $user_role['role_name'] = $corpusRole->name;
                $user_role['role_id'] = $corpusRole->id;


                if(null != $user_role) {
                    if(!array_key_exists($corpusUser->id,$user_roles)) {
                        $user_roles[$corpusUser->id] = array();
                    }
                    array_push($user_roles[$corpusUser->id],$user_role);
                }

            }
        }


        $admin_roles = $this->LaudatioUtilService->determineAdminRole($admin_roles);
        $user_roles = $this->LaudatioUtilService->determineUserAdminRole($user_roles);
        $role_mappings = $this->LaudatioUtilService->getRoleMappingByUser($user);


        //@todo: add try catch for validateheaders
        try{
            $checkResult = json_decode($this->GitRepoService->checkForCorpusFiles($path."/TEI-HEADERS"), true);
            $checkResult['corpusheader'] = ($checkResult['corpusheader'] == "") ? 0 : 1;
        }
        catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException){
            $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());
            //Log the error to file
            $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException,['file' => __FILE__, 'line' => __LINE__]);

            //roll back the creation of the filestructure
            /*
            $corpusPath = $this->GitRepoService->normalizeString($corpus_name);
            $deleted = $this->GitRepoService->deleteUntrackedCorpusFileStructure($this->flysystem, $corpusProjectPath."/".$corpusPath);
            if($deleted) {
                Log::stack(['daily', 'errorlog'])->error("Deleted : corpus directory path: ".$corpusProjectPath."/".$corpus_name." after unsuccessful GIT INIT");
            }
            */
            //$corpus_data = array();
            //return redirect()->route('corpus.edit', compact('corpus'))->with('user',$user)->withError($message);
            return back()->withError($message);
        }



        //get filedata $flysystem,$corpus,$path = ""

        $folderData = $this->GitRepoService->getCorpusFiles($this->flysystem,$corpus->id, $path."/TEI-HEADERS");
        $corpusFileData = $this->GitRepoService->getCorpusFiles($this->flysystem,$corpus->id, $path."/TEI-HEADERS/corpus");
        $documentFileData = $this->GitRepoService->getCorpusFiles($this->flysystem,$corpus->id, $path."/TEI-HEADERS/document");
        $annotationFileData = $this->GitRepoService->getCorpusFiles($this->flysystem,$corpus->id, $path."/TEI-HEADERS/annotation");
//dd($annotationFileData);
        $corpusFormatData = $this->GitRepoService->getCorpusDataFiles($this->flysystem,$path."/CORPUS-DATA",$corpus->id);
        //dd($corpusFormatData);
        //get userinfo for the fileData

        $corpusnewelements = $this->GitRepoService->getUploader($corpusFileData['headerData']['elements'],'corpus');
        $corpusFileData['headerData']['elements'] = $corpusnewelements;

        $documentnewelements = $this->GitRepoService->getUploader($documentFileData['headerData']['elements'],'document');
        $documentFileData['headerData']['elements'] = $documentnewelements;

        $annotationnewelements = $this->GitRepoService->getUploader($annotationFileData['headerData']['elements'],'annotation');
        $annotationFileData['headerData']['elements'] = $annotationnewelements;

        $formatfilenewelements = $this->GitRepoService->getFormatFileUploader($corpusFormatData['projects']);
        $corpusFormatData['projects'] = $formatfilenewelements;
        //Log::info("PROJECTS: ".print_r($corpusFormatData['projects'],1));

        $corpusUpload = false;
        if($corpus->gitlab_id == ""){
            $corpusUpload = true;
        }

        $documentUpload = false;
        $annotationUpload = false;

        if(count($corpus->annotations) == 0){
            $annotationUpload = true;
        }

        if(count($corpus->documents) == 0){
            $documentUpload = true;
        }

        $formatUpload = false;
        if(count($corpus->corpusfiles) == 0) {
            $formatUpload = true;
        }

        $logoUpload = false;
        if( null == $corpus->corpus_logo) {
            $logoUpload = true;
        }

        // Get the messageboard for the CorpusProject this corpus is assigned to
        $messageboard = MessageBoard::where(['corpus_project_id' => $corpusproject->id])->get();
        $allMessages = array();
        $corpusMessages = array();
        $messagecount = 0;
        if(count($messageboard) > 0){
            $boardmessages = $messageboard[0]->boardmessages()->get();
            foreach ($boardmessages as $boardmessage) {

                $messageuser = User::findOrFail($boardmessage->user_id);
                $messageCorpus = Corpus::findOrFail($boardmessage->corpus_id);
                $message = array(
                    'user_name' => $messageuser->name,
                    'user_id' => $messageuser->id,
                    'corpus_id' => $boardmessage->corpus_id,
                    'corpus_name' => $messageCorpus->name,
                    'message' => $boardmessage->message,
                    'messageid' => $boardmessage->id,
                    'last_updated' => $boardmessage->updated_at,
                    'status' => $boardmessage->getStatus($boardmessage->status),
                    'status_id' => $boardmessage->status
                );

                if($corpus->id == $boardmessage->corpus_id){
                    $messagecount++;
                    array_push($corpusMessages,$message);
                }

                array_push($allMessages,$message);
            }
        }

        $invited = $this->LaudatioUtilService->getInvitedByCorpus($corpus->id);
        $invitedUsers = array();
        foreach($invited as $invite) {
            $invitedUser = User::findOrFail($invite->user_id);
            $invitedRole = Role::findOrFail($invite->role_id);
            if(!array_key_exists($invite->user_id,$invitedUsers)){
                $invitedUsers[$invite->user_id] = array();
            }
            $invitedUsers[$invite->user_id] = array(
                "user_name" => $invitedUser->name,
                "user_id" => $invitedUser->id,
                "user_affiliation" => $invitedUser->affiliation,
                "user_role_name" => $invitedRole->name,
                "user_role_id" => $invitedRole->id,
            );
        }
        //dd($invitedUsers);

        $externalEnvironments = array();
        foreach ($corpus->externalenvironments()->get() as $externalenvironment){
            $envuser = User::find($externalenvironment->uid);
            $externalenvironment->user_name = $envuser->name;
            array_push($externalEnvironments, $externalenvironment);
        }
        
        // GET all formats

        $formats =  Format::all();

        $result['format_content_response']  = "The format categories were successfully fetched";

        $foundFormats = array();
        foreach ($formats as $format) {
            $foundFormats[$format->id] = $format->name;
        }


        $corpus_data = array(
            'name' => $corpus->name,
            'version' => $corpus->publication_version,
            'project_name' => $corpusproject->name,
            'project_directorypath' => $corpusproject->directory_path,
            'project_id' => $corpusproject->id,
            'filepath' => $path,
            'user_roles' => $user_roles,
            'roles' => $roles,
            'corpus_admin' => $admin_roles,
            'authuser_role_mappings' => $role_mappings,
            'headerdata' => $checkResult,
            'filedata' => array(
                'folderData' => $folderData,
                'corpusFileData' => $corpusFileData,
                'corpusUpload' => $corpusUpload,
                'documentFileData' => $documentFileData,
                'documentUpload' => $documentUpload,
                'annotationFileData' => $annotationFileData,
                'annotationUpload' => $annotationUpload,
                "formatUpload" => $formatUpload,
                "logoUpload" => $logoUpload
            ),
            'corpusFormatData' => $corpusFormatData,
            'boardmessages' => $corpusMessages,
            'allBoardMessages' => $allMessages,
            'messagecount' => $messagecount,
            'invitedusers' => $invitedUsers,
            'externalenvironments' => $externalEnvironments,
            'formats' => $foundFormats

        );

        //dd($corpus_data);
        JavaScript::put([
            'corpusUpload' => $corpusUpload,
            'documentUpload' => $documentUpload,
            'annotationUpload' => $annotationUpload,
            'formatUpload' => $formatUpload,
            'logoUpload' => $logoUpload,
            'corpus_id' => $corpus->id,
            'corpus_path' => $path,
            'auth_user_name' => $user->name,
            'auth_user_email' => $user->email,
            'auth_user_affiliation' => $user->affiliation
        ]);

        return view('project.corpus.edit', compact('corpus'))
            ->with('isLoggedIn', $isLoggedIn)
            ->with('corpus_data', $corpus_data)
            ->with('user',$user);
    }

    /**
     *
     * update
     *
     * @param Request $request
     * @param Corpus $corpus
     * @return $this
     */
    public function update(Request $request, Corpus $corpus)
    {
        $isLoggedIn = \Auth::check();
        $user = \Auth::user();

        $corpus->update([
            "name" => request('corpus_name'),
            "description" => request('corpus_description')
        ]);

        $corpusProjects = $corpus->corpusprojects()->get();
        $corpusproject = null;
        $corpusProject_directory_path = '';
        if(count($corpusProjects) == 1) {
            $corpusproject = $corpusProjects->first();
            $corpusProject_directory_path = $corpusproject->directory_path;
        }
        else{
            // what to do when we can assign corpora to many projects?
        }

        $corpora = $corpusproject->corpora()->get();

        $user_roles = array();
        $corpusProjectUsers = $corpusProjects->first()->users()->get();
        foreach ($corpusProjectUsers as $corpusProjectUser){
            if(!isset($user_roles[$corpusProjectUser->id])){
                $user_roles[$corpusProjectUser->id]['roles'] = array();
            }
            $user_roles[$corpusProjectUser->id]['user_name'] = $corpusProjectUser->name;

            $role = Role::find($corpusProjectUser->pivot->role_id);
            array_push($user_roles[$corpusProjectUser->id]['roles'],$role->name);
        }


        return view('project.corpusproject.show', compact('corpusproject'))
            ->with('isLoggedIn', $isLoggedIn)
            ->with('corpora',$corpora)
            ->with('user_roles',$user_roles)
            ->with('user',$user);
    }



    /**
     *
     * destroy
     *
     * @param Request $request
     * @return $this
     */
    public function destroy(Request $request)
    {

        $result = array();
        $status = "";

        try{
            $corpusid = $request->input('corpusid');
            $projectid = $request->input('projectid');
            $corpusPath = $request->input('corpuspath');
            $corpusName = $request->input('corpusname');
            $userName = $request->input('authusername');
            $userEmail = $request->input('authuseremail');



            $corpus = Corpus::findOrFail($corpusid);

            $gitLabProjectId = $corpus->gitlab_id;

            $corpus_elasticsearch_index = $corpus->elasticsearch_index;
            $corpus_guidelines_elasticsearch_index = $corpus->guidelines_elasticsearch_index;
            $corpus_document_elasticsearch_index = str_replace("corpus","document",$corpus_elasticsearch_index);
            $corpus_annotation_elasticsearch_index = str_replace("corpus","annotation",$corpus_elasticsearch_index);

            $deleteAnnotationsParams = ['index' => $corpus_annotation_elasticsearch_index];
            $deleteDocumentsParams = ['index' => $corpus_document_elasticsearch_index];
            $deleteGuidelinesParams = ['index' => $corpus_guidelines_elasticsearch_index];
            $deleteCorpusParams = ['index' => $corpus_elasticsearch_index];

            $guidelineDeleteResponse = $this->elasticService->deleteIndex($deleteGuidelinesParams);
            $annotationDeleteResponse = $this->elasticService->deleteIndex($deleteAnnotationsParams);
            $documentDeleteResponse = $this->elasticService->deleteIndex($deleteDocumentsParams);
            $corpusDeleteResponse = $this->elasticService->deleteIndex($deleteCorpusParams);


            if(count($corpus->corpusprojects) > 0) {
                $corpus->corpusprojects()->detach();
            }

            if(count($corpus->users) > 0) {
                $corpus->users()->detach();
            }


            if(count($corpus->boardmessages) > 0) {
                $corpus->boardmessages()->delete();
            }

            if(count($corpus->documents) > 0){
                foreach ($corpus->documents as $document){

                    if(count($document->annotations) > 0){
                        foreach ($document->annotations as $annotation){
                            if(count($annotation->documents) > 0) {
                                $annotation->documents()->detach();
                            }

                        }//end for annotations
                        $document->annotations()->delete();
                    }//end if annotations
                }
                $corpus->documents()->delete();
            }

            if(count($corpus->annotations) > 0){
                $corpus->annotations()->delete();
            }


            $filesDeleted = $this->GitRepoService->deleteCorpusFileStructure($this->flysystem,$corpusPath,$userName,$userEmail);

            $this->LaudatioUtilService->emptyCorpusFileInfoCache($corpusPath,$corpusid);
            $this->LaudatioUtilService->emptyCorpusDataFileCache($corpusPath,$corpusid);
            $this->LaudatioUtilService->emptyCorpusFileCache($corpusPath,$corpusid);


            if($gitLabProjectId != ""){
                $gitlabProjectResult = $this->GitLabService->deleteGitLabProject($gitLabProjectId);
            }

            $corpus->delete();


            $status = "success";
            $result['corpus_delete_response']  = "The '".$corpusName."'' corpus was successfully deleted";

        }
        catch (\Exception $e) {
            $status = "error";
            $result['corpus_delete_response']  = "There was a problem deleting the '".$corpusName."'' Corpus. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }


        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }


    /**
     *
     * destroyCorpusContent
     *
     * @param Request $request
     * @return mixed
     */
    public function destroyCorpusContent(Request $request)
    {

        $result = array();
        $status = "success";

        try{
            $corpusid = $request->input('corpusid');
            $corpusPath = $request->input('path');
            $auth_user_name = $request->input('auth_user_name');
            $auth_user_id = $request->input('auth_user_id');
            $auth_user_email = $request->input('auth_user_email');
            $toBeDeletedCollection = $request->input('tobedeleted');

            $corpus = Corpus::findOrFail($corpusid);

            foreach ($toBeDeletedCollection as $toBeDeleted) {
                if(!is_array($toBeDeleted)){
                    $toBeDeleted = json_decode($toBeDeleted);
                }

                $fileDeleteResult = $this->GitRepoService->deleteFile($this->flysystem,$corpusPath."/".$toBeDeleted['fileName'],$auth_user_name,$auth_user_email);
                if($fileDeleteResult) {
                    $corpus = Corpus::findOrFail($toBeDeleted['databaseId']);


                    $searchData = array();
                    array_push($searchData,array(
                        "_id" => $corpus->elasticsearch_id,
                        "laudatio_corpus_id" => $corpusid
                    ));


                    $elasticSearchDeleteResult = $this->elasticService->deleteIndexedObject($corpus->elasticsearch_index,$searchData);

                    if(!$elasticSearchDeleteResult['error']) {
                        if(count($corpus->annotations()) > 0) {
                            $corpus->annotations()->detach();
                        }

                        if(count($corpus->documents()) > 0) {
                            $corpus->documents()->detach();
                        }

                        $pushResult = $this->GitRepoService->pushFiles($corpusPath,$corpusid,$auth_user_name);
                    }
                    /*
                    $this->LaudatioUtilService->emptyCorpusCache($corpus->elasticsearch_id,$corpus->elasticsearch_index);
                    */
                }
            }

            $result['delete_content_response']  = "Corpus content was successfully deleted";

        }
        catch (\Exception $e) {
            $status = "error";
            $result['delete_content_response']  = "There was a problem deleting the Corpus content. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }


        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }


    /**
     *
     * usersAndRolesForAssignment
     *
     * @param Request $request
     * @return mixed
     */
    public function usersAndRolesForAssignment(Request $request){
        $result = array();
        $status = "success";
        try {

            $corpusId = $request->input('corpusid');
            $corpus = Corpus::findOrFail($corpusId);

            $users = User::all();
            $filteredList = array();

            foreach ($users as $user){
                if(!$corpus->users->contains($user)){
                    $invite = Invite::where([
                        ['corpus_id', '=', $corpusId],
                        ['user_id', '=', $user->id]])->get();


                    if(count($invite) < 1){
                        array_push($filteredList,$user);
                    }
                }
            }

            $roles = Role::all();
            $filteredRoles = array();
            foreach ($roles as $role){
                if(
                    !$role->hasPermissionTo('Administer the application') &&
                    !$role->hasPermissionTo('Can create corpus project') &&
                    ($role->hasPermissionTo('Can create corpus')
                        || $role->hasPermissionTo('Can edit corpus')
                    )
                ){

                    array_push($filteredRoles,$role);
                }
            }

            $result['users'] = $filteredList;
            $result['roles'] = $filteredRoles;
            $result['users_and_roles_for_assignment_response']  = "Users and Roles were successfully fetched";
        }
        catch(\Exception $e) {
            $status = "error";
            $result['users_and_roles_for_assignment_response']  = "There was a problem fetching the Users and Roles. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }

        $response = array(
            'status' => $status,
            'msg' => $result,
        );

        return Response::json($response);
    }

    public function storeExternalEnvironmentByCorpus(Request $request) {
        $result = array();
        $status = "success";

        try {
            $corpusId = $request->input('external_corpus_id');
            $corpus = Corpus::findOrFail($corpusId);
            $userId = $request->input('external_user_id');
            $external_environment_name = $request->input('external_environment_name');
            $external_environment_uri = $request->input('external_environment_uri');
            $external_environment_description =$request->input('external_environment_description');
            $externalEnvironment = new ExternalEnvironment();

            $externalEnvironment->uid = $userId;
            $externalEnvironment->corpus_id = $corpusId;
            $externalEnvironment->external_environment_name = $external_environment_name;
            $externalEnvironment->external_environment_uri = $external_environment_uri;
            $externalEnvironment->external_environment_description = $external_environment_description;

            $externalEnvironment->save();

            $envuser = User::find($externalEnvironment->uid);
            $envuser->name;

            $result['message']  = "The link to the External Environment was successfully created.";
            $result['externalEnvironmentId']  = $externalEnvironment->id;
            $result['externalEnvironmentUserName']  = $envuser->name;

        }
        catch(\Exception $e) {
            $status = "error";
            $result['message']  = "There was a problem creating a link to the External Environment. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }

        $response = array(
            'status' => $status,
            'data' => $result,
        );

        return Response::json($response);
    }

    public function getExternalEnvironmentByCorpus(Request $request) {
        $result = array();
        $status = "success";

        try {
            $corpusId = $request->input('external_corpus_id');
            $corpus = Corpus::findOrFail($corpusId);
            $externalEnvironments = $corpus->externalenvironments()->get();
            $result['environments']  = $externalEnvironments;
        }
        catch(\Exception $e) {
            $status = "error";
            $result['message']  = "There was a problem creating a link to the External Environment. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }
        $response = array(
            'status' => $status,
            'data' => $result,
        );

        return Response::json($response);
    }

    public function deleteExternalEnvironment(Request $request) {
        $result = array();
        $status = "success";

        try {
            $external_environment_id = $request->input('external_environment_id');
            $external_corpus_id = $request->input('external_corpus_id');
            $external_environment = ExternalEnvironment::findOrFail($external_environment_id);
            if($external_environment->corpus_id == $external_corpus_id){
                $external_environment->delete();
                $result['message']  = "The link to the External Environment was successfully deleted.";
            }
            else{
                $status = "error";
                $result['message']  = "There was a problem deleting the link to the External Environment. The error was: The link does not seem to match the corpus";
            }

        }
        catch(\Exception $e) {
            $status = "error";
            $result['message']  = "There was a problem deleting the link to the External Environment. The error was: ($e) A message has been sent to the site administrator. Please try again later";
        }
        $response = array(
            'status' => $status,
            'data' => $result,
        );

        return Response::json($response);
    }


    /**
     *
     * destroyCorpusUser
     *
     * @param $corpusId
     * @param $userId
     * @return \Illuminate\Contracts\View\Factory|\Illuminate\View\View
     */
    public function  destroyCorpusUser($corpusId,$userId){
        $isLoggedIn = \Auth::check();
        $user = \Auth::user();

        $corpus = Corpus::find($corpusId);
        $corpus->users()->detach($userId);

        $roles = Role::latest()->get();
        return view('project.useradmin.roles.index', compact('roles'))
            ->with('isLoggedIn', $isLoggedIn)
            ->with('user',$user);
    }


    /**
     * @param Corpus $corpus
     * @return $this
     */
    public function assignCorpusUsers(Corpus $corpus) {

        $isLoggedIn = \Auth::check();
        $loggedInUser = \Auth::user();
        $users = User::latest()->get();
        $filteredList = array();

        foreach ($users as $user){
            if(!$corpus->users->contains($user)){
                array_push($filteredList,$user);
            }
        }

        $user_roles = array();
        $corpusUsers = $corpus->users()->get();


        foreach ($corpusUsers as $corpusUser){
            $role = Role::find($corpusUser->pivot->role_id);

            if(!isset($user_roles[$corpusUser->pivot->role_id])){
                $user_roles[$corpusUser->pivot->role_id] = array();
            }
            array_push($user_roles[$corpusUser->pivot->role_id],$corpusUser->id);
        }

        /*
        $projects = $corpus->corpusprojects()->get();
        $project_user_roles = array();

        foreach ($projects as $project){
            $projectUsers = $project->users()->get();
            foreach ($projectUsers as $projectUser){
                $project_user_roles[$projectUser->id] = $projectUser->pivot->role_id;
            }
        }
*/
        $roles = Role::all();
        $filteredRoles = array();
        foreach ($roles as $role){
            if(
                !$role->hasPermissionTo('Administer the application') &&
                !$role->hasPermissionTo('Can create corpus project') &&
                ($role->hasPermissionTo('Can create corpus')
                    || $role->hasPermissionTo('Can edit corpus')
                )
            ){
                array_push($filteredRoles,$role);
            }
        }

        //dd($filteredRoles);

        return view('project.corpus.assign_corpusroles_to_user')
            ->with('corpus', $corpus)
            ->with('users', $users)
            ->with('user',$loggedInUser)
            ->with('roles', $filteredRoles)
            ->with('user_roles',$user_roles)
            ->with('isLoggedIn', $isLoggedIn)
            ->with('loggedInUser',$loggedInUser);

    }


}
