<?php

namespace App\Http\Controllers;

use App\Corpus;
use App\CorpusFile;
use App\CorpusProject;
use App\Document;
use App\Annotation;
use App\Exceptions\GitNetworkException;
use GrahamCampbell\Flysystem\FlysystemManager;
use DB;
use Illuminate\Http\Request;
use App\Custom\LaudatioUtilsInterface;
use App\Custom\GitRepoInterface;
use App\Custom\GitLabInterface;
use App\Custom\ElasticsearchInterface;
use App\Laudatio\GitLab\GitFunction;
use App\Custom\ValidatorInterface;
use App\Exceptions\XMLNotWellformedException;
use App\Exceptions\XMLNotValidException;
use App\Exceptions\CorpusNameAlreadyExistsException;
use App\Exceptions\CorpusRepositoryFileSystemException;
use App\Exceptions\CreateGitLabResourceException;
use Elasticsearch\Common\Exceptions\NoNodesAvailableException;
use App\Exceptions\LaudatioElasticsearchException;
use Http\Client\Exception\NetworkException;
use App\Exceptions\GitAddException;
use App\Exceptions\GitCommitException;
use App\Exceptions\GitPushException;
use App\Exceptions\GitDeleteException;
use App\Exceptions\GitInitException;
use App\Exceptions\FileUploadException;
use League\Flysystem\FileExistsException;
use Log;
use Flow\JSONPath\JSONPath;
use Cache;
use Session;
//use Illuminate\Http\Response;
use Response;

/**
 * Class UploadController
 * @package App\Http\Controllers
 */
class UploadController extends Controller
{

    protected $flysystem;
    protected $laudatioUtilsService;
    protected $connection;
    protected $blacklist = array('.git','README.md');
    protected $basePath;
    protected $GitLabService;
    protected $GitRepoService;
    protected $elasticService;
    protected $indexMappingPath;
    protected $schemaBasePath;
    protected $validationService;
    protected $sshAlias;
    protected $corpusIndexName;
    protected $documentIndexName;
    protected $annotationIndexName;
    protected $guidelineIndexName;


    /**
     * UploadController constructor.
     * @param FlysystemManager $flysystem
     * @param LaudatioUtilsInterface $laudatioUtilsService
     * @param GitLabInterface $GitLabService
     * @param GitRepoInterface $Gitservice
     * @param ElasticsearchInterface $elasticService
     * @param ValidatorInterface $validationService
     */
    public function __construct(FlysystemManager $flysystem, LaudatioUtilsInterface $laudatioUtilsService, GitLabInterface $GitLabService, GitRepoInterface $Gitservice, ElasticsearchInterface $elasticService, ValidatorInterface $validationService)
    {
        $this->flysystem = $flysystem;
        $this->GitRepoService = $Gitservice;
        $this->basePath = config('laudatio.basePath');
        $this->indexMappingPath = config('laudatio.indexMappingPath');
        $this->schemaBasePath = config('laudatio.schemaPath');
        $this->laudatioUtilsService = $laudatioUtilsService;
        $this->connection = $this->flysystem->getDefaultConnection();
        $this->GitLabService = $GitLabService;
        $this->elasticService = $elasticService;
        $this->validationService = $validationService;
        $this->sshAlias = config('laudatio.ssh-alias');
        $this->corpusIndexName = config("laudatio.elasticsearch_corpus_index","corpora");
        $this->documentIndexName = config("laudatio.elasticsearch_document_index","documents");
        $this->annotationIndexName = config("laudatio.elasticsearch_annotation_index","annotations");
        $this->guidelineIndexName = config("laudatio.elasticsearch_guideline_index","guidelines");
    }

    /**
     * uploadSubmit: upload corpus headers
     *
     * @param Request $request
     * @return \Illuminate\Http\RedirectResponse
     * @throws \Flow\JSONPath\JSONPathException
     *
     */

    public function uploadSubmit(Request $request)
    {
        $user = \Auth::user();

        /** assign variables **/

        $dirPath = $request->directorypath;;
        $dirPathArray = explode("/", $dirPath);
        end($dirPathArray);
        $last_id = key($dirPathArray);
        $corpusId = $request->corpusid;

        $corpusProjectPath = $dirPathArray[0];
        $corpusPath = $dirPathArray[1];
        $headerPath = $dirPathArray[$last_id];

        $fileName = $request->formats->getClientOriginalName();
        $pathname = $request->formats->getPathName();
        $xmlpath = $request->formats->getRealPath();
        $directoryPath = $this->laudatioUtilsService->getDirectoryPath(array($fileName), $fileName);

        $gitlab_id = null;

        if (empty($corpusProjectPath) || empty($corpusPath)) {
            return redirect()->route('corpus.edit', ['corpus' => $corpusId]);
        }

        $corpusProjectDB = DB::table('corpus_projects')->where('directory_path', $corpusProjectPath)->get();
        $corpusProjectId = $corpusProjectDB[0]->gitlab_id;

        $currentFilePath = "";

        /** initialize variables  **/

        // DB = ELOQUENT
        $document = null;
        $annotation = null;
        $documents = array();
        $annotations = array();

        //XML => JSONPATH
        $json = null;
        $jsonPath = null;
        $xmlNode = null;

        //Fetch XML by file path and parse it into JSONPATH
        if (!empty($xmlpath)) {
            $xmlfile = preg_replace_callback('/\\\\([0-7]{1,3})/', array($this, 'convertOctalToCharacter'), $xmlpath);
            $xmlfile = str_replace('"',"",$xmlfile);

            $this->validationService->setXml($xmlfile);

            //validate xml
            $response = new Response();
            $responsearray = array();

            try {
                //wellformed?
                $isWellFormed = $this->validationService->isWellFormed(true);

                //valid?
                Log::info("STTING SCHEMA: ".$this->schemaBasePath.'/'.$headerPath.'/'.$headerPath.'.rng');
                $this->validationService->setRelaxNGSchema($this->schemaBasePath.'/'.$headerPath.'/'.$headerPath.'.rng');
                $isValid = $this->validationService->isValidByRNG(true);

                //parse xml
                $xmlNode = simplexml_load_file($xmlpath);

                if($xmlNode) {
                    $json = $this->laudatioUtilsService->parseXMLToJson($xmlNode, array());

                }
                if(isset($json)){
                    $jsonPath = new JSONPath($json, JSONPath::ALLOW_MAGIC);
                }

            } catch (XMLNotWellformedException $exception) {
                $notification = $exception->createUserFacingMessage($exception);
                array_push($responsearray,$notification);
                return Response::json($responsearray, 400);

            }
            catch (XMLNotValidException $valid_exception) {
                $notification = $valid_exception->createUserFacingMessage($valid_exception);
                array_push($responsearray,$notification);
                return Response::json($responsearray, 400);
            }
        }

        //GIT
        $addPath = "";
        $commitPath = "";
        $pushPath = $corpusProjectPath.'/'.$corpusPath;
        $initialPushPath = "";
        $gitLabId = "";
        $gitLabTagName = "";
        $gitLabCorpusPath = "";
        $filePath = "";
        $remoteRepoUrl = "";


        /** booleans **/
        $corpusIsVersioned = false;
        $canUpload = true;
        $pushCorpusStructure = false;

        /** pipeline **/
        $lastIndexes = array();

        if (isset($request->corpusid)) {
            $corpus = Corpus::find($corpusId);
            $currentFilePath = $corpus->directory_path;
            $gitLabId = $corpus->gitlab_id;
        }

        $gitFunction = new GitFunction();
        // Is the Corpus Header uploaded ? We know that if the Corpus has an Eloquent version set
        $corpusIsVersioned = $this->laudatioUtilsService->corpusIsVersioned($corpusId);
        Log::info("CORPUSISVERSIONED: ".$corpusIsVersioned);
        Log::info("HEADERPATH: ".print_r($headerPath,1));

        $latestGitTag = $this->laudatioUtilsService->getLatestCorpusTagByCorpusAndGitlabId($corpusId,$gitLabId);
        if($corpusIsVersioned) {
            //empty the lastIndexId json array
            $resetResult = $gitFunction->resetJSONIdArray($pushPath,$latestGitTag);

            //set variables and paths
            if($headerPath != 'corpus'){

                $addPath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/'.$dirPathArray[$last_id];
                $commitPath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/'.$dirPathArray[$last_id];
                $commitDataPath = "";

                if($headerPath == 'document') {
                    //write to db
                    $document = $this->laudatioUtilsService->setDocumentAttributes($json,$corpusId,$user->id,$fileName,false);
                    $filePath = $corpusProjectPath . '/' . $corpus->directory_path . '/TEI-HEADERS/document/' . $fileName;
                    $commitDataPath = 'TEI-HEADERS/document/'.$fileName;
                }
                else if($headerPath == 'annotation'){
                    //write to db
                    $annotation = $this->laudatioUtilsService->setAnnotationAttributes($json,$corpusId,$user->id,$fileName,false);
                    $filePath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/annotation/'.$fileName;
                    $commitDataPath = 'TEI-HEADERS/annotation/'.$fileName;
                }

            }
            else{
                $filePath = $corpusProjectPath."/".$corpus->directory_path.'/TEI-HEADERS/corpus/'.$fileName;
                $addPath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/corpus/'.$fileName;
                $commitPath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/corpus/';
                $commitDataPath = 'TEI-HEADERS/corpus/'.$fileName;
            }

            //upload

            try {
                $createdPaths = $gitFunction->writeFiles($dirPath,array($fileName), $this->flysystem,$request->formats->getRealPath(),$directoryPath);
            }
            catch (FileExistsException $fileExistsException) {
                $fileUploadException = new FileUploadException($fileExistsException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FileNotFoundException $fileNotFoundException) {
                $fileUploadException = new FileUploadException($fileNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FilesystemNotFoundException $filesystemNotFoundException) {
                $fileUploadException = new FileUploadException($filesystemNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }
            catch (NotSupportedException $notSupportedException) {
                $fileUploadException = new FileUploadException($notSupportedException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (RootViolationException $rootViolationException) {
                $fileUploadException = new FileUploadException($rootViolationException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);


                return Response::json($responsearray, 400);

            }
            catch (UnreadableFileException $unreadableFileException) {
                $fileUploadException = new FileUploadException($unreadableFileException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }

            //add
            $isAdded = $this->GitRepoService->addFiles($this->flysystem, $filePath);
            //commit
            $isPushed = false;
            $commitData = null;

            if($isAdded) {
                $commitData = $this->GitRepoService->commitFile($commitPath, $fileName,"Adding ",$corpusId,$user->name,$user->email);

                if(!empty($commitData)){

                    $lastIndexes = $this->GitRepoService->getLastELasticsearchIndex($commitPath, $latestGitTag);

                    Log::info("ALREADZ VERSIONED: lastIndexes: ".print_r($lastIndexes,1));
                    //push
                    $isPushedSucessfully = $this->GitRepoService->pushFiles($pushPath,$corpusId,$user);

                    if(is_array($isPushedSucessfully)){
                        $exception = $isPushedSucessfully["exception"];
                        $message = $exception->createUserFacingJSONMessage($exception->getMessage());
                        $responsearray = array();
                        array_push($responsearray,$message);

                        $resetCorpusPath = "";
                        try {
                            $isDeleted = $gitFunction->deleteFiles($filePath,$user->name,$user->email);
                        }
                        catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                            $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());
                            $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException->getMessage()." : directory path: ".$filePath,['file' => __FILE__, 'line' => __LINE__]);
                            return back()->withError($message);
                        }

                        if($isDeleted) {
                            $exception->logErrorMessage("Deleted uploaded file: ".$filePath." after unsuccessful push to git repository ",['file' => __FILE__, 'line' => __LINE__]);
                        }
                        return Response::json($responsearray, 400);
                    }
                    else{
                        $isPushed = true;
                    }
                }
            }

            if($isPushed){
                /*
                 * @todo: update last-updated in order that upload date and time is updated in the interface
                 */
                if($headerPath == 'document') {
                    if(null != $document && null != $commitData){

                        if(!empty($lastIndexes)) {
                            foreach($lastIndexes['lastIndexedId'] as $indexDocumentFilePath => $documentElasticSearchId) {
                                if(strpos($indexDocumentFilePath,"/document/")) {
                                    $documentFilename = substr($indexDocumentFilePath,strrpos($indexDocumentFilePath,"/") +1);
                                    if($documentFilename == $fileName){
                                        $commitData['elasticsearch_id'] = $documentElasticSearchId;
                                        break;
                                    }
                                }
                            }
                        }

                        $setData = $this->laudatioUtilsService->setCommitData(array($commitDataPath => $commitData),$corpusId);

                        if ($setData) {
                            $document->elasticsearch_index = $this->documentIndexName;
                            $document->directory_path = $corpus->directory_path;
                            $document->publication_version = $corpus->publication_version;
                            $document->workflow_status = $corpus->workflow_status;
                            $document->gitlab_version_tag = $latestGitTag;
                            $document->save();
                        }
                    }


                    //empty cache
                    $this->laudatioUtilsService->emptyDocumentCacheByDocumentElasticsearchId($document->elasticsearch_id,$this->documentIndexName);
                    $this->laudatioUtilsService->emptyDocumentCacheByCorpusId($corpus->corpus_id,$this->documentIndexName);
                    $this->laudatioUtilsService->emptyDocumentCacheByDocumentIndex($this->documentIndexName);
                }
                else if($headerPath == 'annotation'){

                    if(null != $annotation && null != $commitData){
                        if(!empty($lastIndexes)) {
                            foreach($lastIndexes['lastIndexedId'] as $indexAnnotationFilePath => $annotationElasticSearchId) {
                                if(strpos($indexAnnotationFilePath,"/annotation/")) {
                                    $annotationFilename = substr($indexAnnotationFilePath,strrpos($indexAnnotationFilePath,"/") +1);
                                    if($annotationFilename == $fileName){
                                        $commitData['elasticsearch_id'] = $annotationElasticSearchId;
                                        break;
                                    }
                                }
                            }
                        }

                        $setData = $this->laudatioUtilsService->setCommitData(array($commitDataPath => $commitData),$corpusId);
                        if($setData){
                            $annotation->elasticsearch_index = $this->annotationIndexName;
                            $annotation->directory_path = $corpus->directory_path;
                            $annotation->publication_version = $corpus->publication_version;
                            $annotation->workflow_status = $corpus->workflow_status;
                            $annotation->gitlab_version_tag = $latestGitTag;
                            $annotation->save();
                        }
                    }

                    //empty cache
                    $this->laudatioUtilsService->emptyAnnotationCacheByNameAndCorpusId($annotation->annotation_id, $corpusId, $this->annotationIndexName);
                    $this->laudatioUtilsService->emptyAnnotationCacheByCorpusId($corpus->corpus_id,$this->annotationIndexName);
                    $this->laudatioUtilsService->emptyAnnotationCacheByAnnotationIndex($this->annotationIndexName);

                }
                else if($headerPath == 'corpus'){
                    $idParams = array();
                    array_push($idParams,array(
                        "laudatio_corpus_id" => $corpus->corpus_id
                    ));

                    $corpusObject[$corpus->corpus_id] = $idParams;

                    if(!isset($corpus->id)){
                        $elasticIds = $this->elasticService->getElasticIdByObjectId($this->corpusIndexName,$corpusObject);
                        Log::info("ALREADZ VERSIONED: ELASTICIDS: ".print_r($elasticIds,1));
                        foreach ($elasticIds as $ecorpusId => $elasticId){
                            $corpus->elasticsearch_id = $elasticIds[$ecorpusId]['elasticsearchid'];
                            $corpus->elasticsearch_index = $this->corpusIndexName;
                            $corpus->save();
                        }
                    }

                    if(isset($jsonPath)){
                        $corpusTitle = $jsonPath->find('$.TEI.teiHeader.fileDesc.titleStmt.title')->data();

                        $corpusDescription = $jsonPath->find('$.TEI.teiHeader.encodingDesc[0].projectDesc.p.text')->data();
                        //there can be a more than one encodingdescs
                        if(count($corpusDescription) < 1){
                            $corpusDescription = $jsonPath->find('$.TEI.teiHeader.encodingDesc[*].p.text')->data();
                        }


                        $corpusPublicationVersions = $jsonPath->find('$.TEI.teiHeader.revisionDesc.change.n')->data();

                        if(empty($corpusPublicationVersions)) {
                            $corpusPublicationVersions = $jsonPath->find('$.TEI.teiHeader.revisionDesc.change[*].n')->data();
                        }
                        $corpusPublicationVersion = max(array_values($corpusPublicationVersions));

                        if(isset($corpusTitle[0])) {
                            $corpus->name = $corpusTitle[0];

                            if(isset($corpusDescription[0])) {
                                $corpus->description = $corpusDescription[0];
                            }

                            if(isset($corpusPublicationVersion)) {
                                $corpus->publication_version = $corpusPublicationVersion;
                            }

                            $corpus->save();

                        }
                    }
                    //empty cache
                    $this->laudatioUtilsService->emptyCorpusCache($corpus->elasticsearch_id,$this->corpusIndexName);
                }
            }

        }
        else {
            //the corpus header is not yet uploaded
            $directoryPath = $this->laudatioUtilsService->getDirectoryPath(array($fileName),$fileName);
            $corpusDescription = null;
            //upload

            try {
                $createdPaths = $gitFunction->writeFiles($dirPath,array($fileName), $this->flysystem,$request->formats->getRealPath(),$directoryPath);
            }
            catch (FileExistsException $fileExistsException) {
                $fileUploadException = new FileUploadException($fileExistsException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FileNotFoundException $fileNotFoundException) {
                $fileUploadException = new FileUploadException($fileNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FilesystemNotFoundException $filesystemNotFoundException) {
                $fileUploadException = new FileUploadException($filesystemNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }
            catch (NotSupportedException $notSupportedException) {
                $fileUploadException = new FileUploadException($notSupportedException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (RootViolationException $rootViolationException) {
                $fileUploadException = new FileUploadException($rootViolationException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);


                return Response::json($responsearray, 400);

            }
            catch (UnreadableFileException $unreadableFileException) {
                $fileUploadException = new FileUploadException($unreadableFileException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }

            Log::info("CREATED FILEE: ".print_r($fileName,1)." CREATED PATHS: ".print_r($createdPaths,1));

            $corpusPublicationVersion = "";
            $addStructurePath = "";
            //parse
            if($headerPath == 'corpus'){
                $now = time();
                $gitLabResponse = null;

                //read data
                $corpusIsVersioned = $this->laudatioUtilsService->corpusIsVersioned($corpusId);

                if(isset($jsonPath)){
                    $corpusTitle = $jsonPath->find('$.TEI.teiHeader.fileDesc.titleStmt.title')->data();

                    $corpusDescription = $jsonPath->find('$.TEI.teiHeader.encodingDesc[0].projectDesc.p.text')->data();
                    //there can be a more than one encodingdescs
                    if(count($corpusDescription) < 1){
                        $corpusDescription = $jsonPath->find('$.TEI.teiHeader.encodingDesc.projectDesc.p.text')->data();
                    }

                    $corpusPublicationVersions = $jsonPath->find('$.TEI.teiHeader.revisionDesc.change.n')->data();

                    if(empty($corpusPublicationVersions)) {
                        $corpusPublicationVersions = $jsonPath->find('$.TEI.teiHeader.revisionDesc.change[*].n')->data();
                    }
                    $corpusPublicationVersion = max(array_values($corpusPublicationVersions));
                }

                if(!empty($corpusTitle[0])){
                    $corpusName = $corpusTitle[0]."_".$corpusPublicationVersion;

                    $updatedCorpusPath = "";

                    try {

                        $updatedCorpusPath = $this->GitRepoService->updateCorpusFileStructure($this->flysystem,$corpusProjectPath,$corpus->directory_path,$corpusName);
                        //Delete cached untitled-paths:
                        $this->laudatioUtilsService->emptyFileCaches($corpus->id);
                        $updatePathAndNameParams = array(
                            'name' => $corpusName,
                            'directory_path' => substr($updatedCorpusPath,strrpos($updatedCorpusPath,"/")+1),
                        );

                        $corpus = $this->laudatioUtilsService->updateCorpusAttributes($updatePathAndNameParams,$corpus->id);

                    }
                    catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                        $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());

                        //Log the error to file
                        $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException->getMessage()." : corpus directory path: ".$corpusProjectPath."/".$corpusName,['file' => __FILE__, 'line' => __LINE__]);

                        return back()->withError($message);
                    }


                    if(!empty($updatedCorpusPath)){

                        $gitLabCorpusPath = substr($updatedCorpusPath,strrpos($updatedCorpusPath,"/")+1);
                        $gitLabCorpusPath = substr($gitLabCorpusPath,0,strrpos($gitLabCorpusPath,"_"));

                        $normalizedCorpusName = $this->GitRepoService->normalizeString($corpusName);

                        //set new path for what ? where is it used ?
                        $dirPath =  $updatedCorpusPath.'/TEI-HEADERS/corpus';


                        try{
                            $gitLabResponse = $this->GitLabService->createGitLabProject(
                                $this->GitRepoService->normalizeTitle($corpusTitle[0]),
                                array(
                                    'namespace_id' => $corpusProjectId,
                                    'path' => $gitLabCorpusPath,
                                    'description' => $corpusDescription[0],
                                    'visibility' => 'private'
                                ));

                        }
                        catch (NetworkException $networkException){
                            $gitNetworkException = new GitNetworkException($networkException->getMessage(),0,null);
                            $message = $gitNetworkException->createUserFacingJSONMessage($gitNetworkException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);

                            $resetCorpusPath = "";
                            try {
                                $resetCorpusPath = $this->GitRepoService->updateCorpusFileStructure($this->flysystem,$corpusProjectPath,$this->GitRepoService->normalizeString($corpusName),$currentFilePath);
                                $isDeleted = $gitFunction->deleteUntrackedFiles($resetCorpusPath."/TEI-HEADERS/corpus/".$fileName);
                            }
                            catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                                $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());
                                $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException->getMessage()." : corpus directory path: ".$corpusProjectPath."/".$corpusName,['file' => __FILE__, 'line' => __LINE__]);
                                return back()->withError($message);
                            }

                            if(!empty($resetCorpusPath) && $isDeleted) {
                                $corpus->directory_path = $currentFilePath;
                                $corpus->name = 'U'.substr($currentFilePath,1);
                                $corpus->save();
                                $gitNetworkException->logErrorMessage("Reverted: corpus directory path: ".$corpusProjectPath."/".$currentFilePath." after unsuccessful creation of git repository ",['file' => __FILE__, 'line' => __LINE__]);
                                $gitNetworkException->logErrorMessage("Deleted uploaded file: ".$resetCorpusPath."/TEI-HEADERS/corpus/".$fileName." after unsuccessful creation of git repository ",['file' => __FILE__, 'line' => __LINE__]);
                            }

                            return Response::json($responsearray, 400);
                        }

                        if( null != $gitLabResponse &&
                            isset($gitLabResponse['id'])
                        ){
                            $remoteRepoUrl = $gitLabResponse['ssh_url_to_repo'];
                            $gitlab_id = $gitLabResponse['id'];

                            $params = array(
                                'corpusId' => $corpusId,
                                'corpus_id' => $this->corpusIndexName,
                                "uid" => $user->id,
                                'corpus_path' => $normalizedCorpusName,
                                'publication_version' => $corpusPublicationVersion,
                                'workflow_status' => 0,
                                'gitlab_group_id' => $corpusProjectId,
                                'gitlab_id' => $gitLabResponse['id'],
                                'gitlab_web_url' => $gitLabResponse['web_url'],
                                'gitlab_ssh_url' => $remoteRepoUrl,
                                'gitlab_name_with_namespace' => $gitLabResponse['name_with_namespace'],
                                'elasticsearch_index' => $this->corpusIndexName,
                                'guidelines_elasticsearch_index' => $this->guidelineIndexName,
                                'file_name' => $fileName
                            );

                            $corpus = $this->laudatioUtilsService->setCorpusAttributes($json,$params);
                            $pushCorpusStructure = true;
                            $pushPath = $corpusProjectPath.'/'.$corpus->directory_path;

                            // create git corpus project
                            $this->laudatioUtilsService->updateDirectoryPaths($normalizedCorpusName,$corpusId);
                        }//end if gitlabResponse

                    }//end updatedcorpusPath
                }//end if corpustitle


            }
            else if($headerPath == 'document') {
                //write to db
                $document = $this->laudatioUtilsService->setDocumentAttributes($json,$corpusId,$user->id,$fileName,false);
                //empty cache
                $this->laudatioUtilsService->emptyDocumentCacheByCorpusId($corpus->corpus_id,$this->documentIndexName);
                $this->laudatioUtilsService->emptyDocumentCacheByDocumentIndex($this->documentIndexName);
                $this->laudatioUtilsService->emptyDocumentCacheByDocumentElasticsearchId($document->elasticsearch_id,$this->documentIndexName);
            }
            else if($headerPath == 'annotation') {
                //write to db
                $annotation = $this->laudatioUtilsService->setAnnotationAttributes($json,$corpusId,$user->id,$fileName,false);
                //empty cache
                $this->laudatioUtilsService->emptyAnnotationCacheByCorpusId($corpus->corpus_id,$this->annotationIndexName);
                $this->laudatioUtilsService->emptyAnnotationCacheByAnnotationIndex($this->annotationIndexName);
            }

            //set paths
            if($headerPath != 'corpus'){
                $addPath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/'.$dirPathArray[$last_id];
                $commitPath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/'.$dirPathArray[$last_id];

                if($headerPath == 'document') {
                    $filePath = $corpusProjectPath . '/' . $corpus->directory_path . '/TEI-HEADERS/document/' . $fileName;
                }
                else if($headerPath == 'annotation'){
                    $filePath = $corpusProjectPath.'/'.$corpus->directory_path.'/TEI-HEADERS/annotation/'.$fileName;
                }
            }
            else{
                if(isset($updatedCorpusPath)){
                    $filePath = $updatedCorpusPath.'/TEI-HEADERS/corpus/'.$fileName;
                    $addPath = $updatedCorpusPath.'/TEI-HEADERS';
                    $addStructurePath = $updatedCorpusPath;
                    $initialCommitPath = $updatedCorpusPath;
                    $corpusCommitpath = $updatedCorpusPath.'/TEI-HEADERS';
                    $initialPushPath = $updatedCorpusPath;
                }

            }

            Log::info("FILEPATH: ".print_r($filePath,1));

            $pushPath = $corpusProjectPath.'/'.$corpus->directory_path;

            if($pushCorpusStructure && !empty($initialPushPath) && $remoteRepoUrl
            && isset($initialPushPath)
            && isset($corpusCommitpath)
            && isset($initialCommitPath)
            && isset($addPath)
            && isset($filePath)
            ){
                Log::info("THIS ONLY WITH CORPUS RIGHT-0?");
                $newRemoteRepoUrl = str_replace("scm.cms.hu-berlin.de",$this->sshAlias,$remoteRepoUrl);
                $addRemote = $this->GitRepoService->addRemote($newRemoteRepoUrl,$initialPushPath);

                $hooksAdded = $this->GitRepoService->addHooks($initialPushPath, $user->name, $user->email);
                $isReset = $this->GitRepoService->resetAdd($initialPushPath,array("TEI-HEADERS", "CORPUS-DATA", "CORPUS-IMAGES"));

                if($isReset) {

                    $hookCommitdata = $this->GitRepoService->commitStructureFiles($initialCommitPath.'/githooks', "Adding githooks",$corpusId, $user->name, $user->email);
                    $isInitiallyPushed = $this->GitRepoService->initialPush($initialPushPath,$user);
                    if($isInitiallyPushed) {
                        $tagMessage = array(
                            "laudatioCorpusid" => $corpusId,
                            "gitlabId" => $gitlab_id,
                            "username" => $user->name,
                            "useremail" =>  $user->email
                        );

                        $hexedJson = str_replace('"','\"',json_encode( $tagMessage));
                        $tag = $this->GitRepoService->setCorpusVersionTag($initialPushPath,$hexedJson,0.0,$user->name,$user->email, false);
                        if($tag) {
                            $params = array(
                                'gitlab_version_tag' => 'v0',
                            );

                            $corpus = $this->laudatioUtilsService->updateCorpusAttributes($params,$corpusId);
                        }
                        $createdPaths = "";

                        try {
                            $createdPaths = $gitFunction->writeFiles($dirPath,array($fileName), $this->flysystem,$request->formats->getRealPath(),$directoryPath);
                        }
                        catch (FileExistsException $fileExistsException) {
                            $fileUploadException = new FileUploadException($fileExistsException->getMessage(),0,null);

                            $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);

                            return Response::json($responsearray, 400);

                        }
                        catch (FileNotFoundException $fileNotFoundException) {
                            $fileUploadException = new FileUploadException($fileNotFoundException->getMessage(),0,null);

                            $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);

                            return Response::json($responsearray, 400);

                        }
                        catch (FilesystemNotFoundException $filesystemNotFoundException) {
                            $fileUploadException = new FileUploadException($filesystemNotFoundException->getMessage(),0,null);

                            $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);

                            return Response::json($responsearray, 400);
                        }
                        catch (NotSupportedException $notSupportedException) {
                            $fileUploadException = new FileUploadException($notSupportedException->getMessage(),0,null);

                            $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);

                            return Response::json($responsearray, 400);

                        }
                        catch (RootViolationException $rootViolationException) {
                            $fileUploadException = new FileUploadException($rootViolationException->getMessage(),0,null);

                            $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);


                            return Response::json($responsearray, 400);

                        }
                        catch (UnreadableFileException $unreadableFileException) {
                            $fileUploadException = new FileUploadException($unreadableFileException->getMessage(),0,null);

                            $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                            $responsearray = array();
                            array_push($responsearray,$message);

                            return Response::json($responsearray, 400);
                        }

                        //add files
                        Log::info("CREATED FILEE AGAIN?: ".print_r($fileName,1)." CREATED PATHS: ".print_r($createdPaths,1));

                        //add .gitignore
                        $gitIgnoreIsAdded = $gitFunction->addUntracked($addStructurePath,".gitignore");
                        $isAdded = $this->GitRepoService->addFiles($this->flysystem, $addStructurePath);
                        if($isAdded) {
                            Log::info("ISADDED : ".print_r($isAdded,1));
                            $corpusCommitdata = $this->GitRepoService->commitFiles($corpusCommitpath, "Adding files for ", $corpusId, $user->name, $user->email);
                            Log::info("CORPUSCOMMITDATA : ".print_r($corpusCommitdata,1));
                            if(!empty($corpusCommitdata)){
                                $lastIndexes = $this->GitRepoService->getLastELasticsearchIndex($corpusCommitpath,'v0');

                                if(!empty($lastIndexes)) {
                                    $corpusIndexedElasticsearchId = '';
                                    if(empty($lastIndexes['corpusElasticId']) && !empty($lastIndexes['lastIndexedId']['TEI-HEADERS/corpus/corpusHeader.xml'])){
                                        $corpusIndexedElasticsearchId = $lastIndexes['lastIndexedId']['TEI-HEADERS/corpus/corpusHeader.xml'];
                                    }
                                    else{
                                        $corpusIndexedElasticsearchId = $lastIndexes['corpusElasticId'];
                                    }
                                    $corpusCommitdata['TEI-HEADERS/corpus/corpusHeader.xml']['elasticsearch_id'] = $corpusIndexedElasticsearchId;
                                }

                                Log::info("CORPUSCOMMITDAT2 : ".print_r($corpusCommitdata,1));
                                $setData = $this->laudatioUtilsService->setCommitData($corpusCommitdata,$corpusId);
                                Log::info("SETDATA : ".print_r($setData,1));
                                $isPushedSucessfully = $this->GitRepoService->pushFiles($pushPath,$corpusId,$user);
                                Log::info("isPushedSucessfully : ".print_r($isPushedSucessfully,1));
                                if(is_array($isPushedSucessfully)){
                                    $exception = $isPushedSucessfully["exception"];
                                    $message = $exception->createUserFacingJSONMessage($exception->getMessage());
                                    $responsearray = array();
                                    array_push($responsearray,$message);

                                    $resetCorpusPath = "";
                                    try {
                                        $isDeleted = $gitFunction->deleteFiles($filePath,$user->name,$user->email);
                                    }
                                    catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                                        $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());
                                        $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException->getMessage()." : directory path: ".$filePath,['file' => __FILE__, 'line' => __LINE__]);
                                        return back()->withError($message);
                                    }

                                    if($isDeleted) {
                                        $exception->logErrorMessage("Deleted uploaded file: ".$filePath." after unsuccessful push to git repository ",['file' => __FILE__, 'line' => __LINE__]);
                                    }
                                    return Response::json($responsearray, 400);
                                }
                                else{
                                    $isPushed = true;
                                }

                                if($isPushed) {
                                    $params = array(
                                        'elasticsearch_index' => $this->corpusIndexName,
                                        'guidelines_elasticsearch_index' => $this->guidelineIndexName,
                                        'file_name' => $fileName
                                    );

                                    $corpus = $this->laudatioUtilsService->updateCorpusAttributes($params,$corpusId);
                                    $corpusFiles = $this->laudatioUtilsService->updateCorpusFileAttributes($corpus);
                                    $idParams = array();
                                    array_push($idParams,array(
                                        "laudatio_corpus_id" => $corpus->id
                                    ));

                                    $corpusObject[$corpus->corpus_id] = $idParams;

                                    $elasticIds = $this->elasticService->getElasticIdByObjectId($this->corpusIndexName,$corpusObject);
                                    foreach ($elasticIds as $ecorpusId => $elasticId){
                                        $corpus->elasticsearch_id = $elasticIds[$ecorpusId]['elasticsearchid'];
                                        $corpus->elasticsearch_index = $this->corpusIndexName;
                                        $corpus->save();
                                    }

                                    $this->laudatioUtilsService->emptyCorpusCache($corpus->elasticsearch_id,$this->corpusIndexName);


                                    // set elasticsearch ids for documents and annotations
                                    if(!empty($lastIndexes)) {

                                        foreach($corpus->annotations as $annotationelm) {
                                            foreach($lastIndexes['lastIndexedId'] as $indexFilePath => $elasticSearchId) {
                                                if(strpos($indexFilePath,"/annotation/")) {
                                                    $annotationFilename = substr($indexFilePath,strrpos($indexFilePath,"/") +1);
                                                    if($annotationFilename == $annotationelm->file_name){
                                                        $annotationelm->elasticsearch_id = $elasticSearchId;
                                                        $annotationelm->elasticsearch_index = $this->annotationIndexName;
                                                        $annotationelm->directory_path = $updatedCorpusPath;
                                                        $annotationelm->publication_version = $corpus->publication_version;
                                                        $annotationelm->workflow_status = $corpus->workflow_status;
                                                        $annotationelm->gitlab_version_tag = $corpus->gitlab_version_tag;
                                                        $annotationelm->save();
                                                        $this->laudatioUtilsService->emptyAnnotationCacheByNameAndCorpusId($annotationelm->annotation_id,$corpusId,$this->annotationIndexName);
                                                    }
                                                }
                                            }
                                        }

                                        $this->laudatioUtilsService->emptyAnnotationCacheByCorpusId($corpus->corpus_id,$this->annotationIndexName);
                                        $this->laudatioUtilsService->emptyAnnotationCacheByAnnotationIndex($this->annotationIndexName);

                                        $documentParams = array();
                                        foreach($corpus->documents as $documentelm) {
                                            foreach($lastIndexes['lastIndexedId'] as $indexDocumentFilePath => $documentElasticSearchId) {
                                                if(strpos($indexDocumentFilePath,"/document/")) {
                                                    $documentFilename = substr($indexDocumentFilePath,strrpos($indexDocumentFilePath,"/") +1);
                                                    if($documentFilename == $documentelm->file_name){
                                                        $documentelm->elasticsearch_id = $documentElasticSearchId;
                                                        $documentelm->elasticsearch_index = $this->documentIndexName;
                                                        $documentelm->directory_path = $updatedCorpusPath;
                                                        $documentelm->publication_version = $corpus->publication_version;
                                                        $documentelm->workflow_status = $corpus->workflow_status;
                                                        $documentelm->gitlab_version_tag = $corpus->gitlab_version_tag;
                                                        $documentelm->save();
                                                        $this->laudatioUtilsService->emptyDocumentCacheByDocumentElasticsearchId($documentelm->elasticsearch_id,$this->documentIndexName);
                                                    }
                                                }
                                            }
                                        }

                                        $this->laudatioUtilsService->emptyDocumentCacheByDocumentIndex($this->documentIndexName);
                                        $this->laudatioUtilsService->emptyDocumentCacheByCorpusId($corpus->corpus_id,$this->documentIndexName);
                                    }

                                }//end if pushed
                            }//end if returnpath
                        }//end if added
                    }//end if initiallypushed
                }//end if reset
            }//end if pushcorpus
        }//end if corpus is versioned
        Log::info("REDIRECTING: ");
        return redirect()->route('corpus.edit', ['corpus' => $corpusId]);
    }

    /**
     *
     * uploadSubmitFiles: upload corpus data files
     *
     * @param Request $request
     * @return \Illuminate\Http\RedirectResponse
     */
    public function uploadSubmitFiles(Request $request)
    {
        $user = \Auth::user();
        $dirPath = $request->directorypath;
        $currentFormatId  = $request->currentformatid;
        $currentFormatName  = $request->currentformatname;
        $dirPathArray = explode("/",$dirPath);
        $corpusId = $request->corpusid;

        $corpusIsVersioned = null;

        $corpusProjectPath = $dirPathArray[0];
        $corpusPath = $dirPathArray[1];
        $uploadFolder = $dirPathArray[2];

        $corpus = null;
        $addformat = false;

        if(isset($request->corpusid)) {
            $corpus = Corpus::find($corpusId);
            $corpusIsVersioned = $this->laudatioUtilsService->corpusIsVersioned($corpusId);

        }
        $realPath = realpath($this->basePath.'/'.$corpusProjectPath.'/'.$corpusPath."/CORPUS-DATA/".$this->GitRepoService->normalizeString($currentFormatName));
        if(!$realPath && isset($currentFormatName)) {

            try {
                $formatResponse = $this->GitRepoService->createFormatDirectory($this->flysystem,$corpusProjectPath,$corpusPath,$currentFormatName);
                if($formatResponse) {
                    $dirPath .= "/".$this->GitRepoService->normalizeString($currentFormatName);
                }
            }
            catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                $fileUploadException = new FileUploadException($corpusRepositoryFileSystemException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }

        }
        else{
            $dirPath .= "/".$this->GitRepoService->normalizeString($currentFormatName);
        }

        if(isset($corpus->id)){
            $fileName = $request->formats->getClientOriginalName();
            $fileNameParts = pathinfo($fileName);
            $corpusVersion = $this->GitRepoService->normalizeString($corpus->publication_version);
            $fileName = $fileNameParts["filename"]."_".$corpusVersion.".".$fileNameParts["extension"];
            $directoryPath = $this->laudatioUtilsService->getDirectoryPath(array($fileName),$fileName);
            $gitFunction = new GitFunction();


            try {
                $createdPaths = $gitFunction->writeFiles($dirPath,array($fileName), $this->flysystem,$request->formats->getRealPath(),$directoryPath);
            }
            catch (FileExistsException $fileExistsException) {
                $fileUploadException = new FileUploadException($fileExistsException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FileNotFoundException $fileNotFoundException) {
                $fileUploadException = new FileUploadException($fileNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FilesystemNotFoundException $filesystemNotFoundException) {
                $fileUploadException = new FileUploadException($filesystemNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }
            catch (NotSupportedException $notSupportedException) {
                $fileUploadException = new FileUploadException($notSupportedException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (RootViolationException $rootViolationException) {
                $fileUploadException = new FileUploadException($rootViolationException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);


                return Response::json($responsearray, 400);

            }
            catch (UnreadableFileException $unreadableFileException) {
                $fileUploadException = new FileUploadException($unreadableFileException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }
            catch (\Exception $generalException) {
                $fileUploadException = new FileUploadException($generalException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }



            if(strpos($createdPaths[0],$dirPath) !== false){
                $addPath = $dirPath.'/'.$fileName;
                $commitPath = $corpusProjectPath.'/'.$corpusPath.'/'.$uploadFolder;
                $pushPath = $corpusProjectPath.'/'.$corpus->directory_path.'/'.$uploadFolder;

                //add files
                $isAdded = $this->GitRepoService->addFiles($this->flysystem, $addPath);

                //git commit The files
                if($isAdded) {


                    if($corpusIsVersioned){
                        $returnPath = $this->GitRepoService->commitFiles($commitPath, "Adding files for" , $corpusId, $user->name, $user->email);

                        if (!empty($returnPath)) {

                            $setId = "CORPUS-DATA/".$this->GitRepoService->normalizeString($currentFormatName)."/".$fileName;

                            $corpusFile = new CorpusFile();
                            $corpusFile->file_name = $fileName;
                            $corpusFile->uid = $user->id;
                            $corpusFile->directory_path = $corpusPath;
                            $corpusFile->workflow_status = 0;
                            $corpusFile->format = $currentFormatId;
                            $corpusFile->gitlab_version_tag = $corpus->gitlab_version_tag;
                            $corpusFile->gitlab_commit_sha = $returnPath[$setId]['sha_string'];
                            $corpusFile->gitlab_commit_date = $returnPath[$setId]['date'];
                            $corpusFile->gitlab_commit_description = $returnPath[$setId]['message'];
                            $corpusFile->save();


                            $corpus->corpusfiles()->save($corpusFile);


                            $isPushedSucessfully = $this->GitRepoService->pushFiles($pushPath,$corpusId,$user);

                            //$isPushed = $this->GitRepoService->pushFiles($pushPath, $corpusId, $user);
                            if(is_array($isPushedSucessfully)){
                                $exception = $isPushedSucessfully["exception"];
                                $message = $exception->createUserFacingProcessJSONMessage($exception->getMessage());
                                $responsearray = array();
                                array_push($responsearray,$message);

                                $resetCorpusPath = "";
                                try {
                                    $isDeleted = $gitFunction->deleteFiles($dirPath."/".$fileName,$user->name,$user->email);
                                }
                                catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                                    $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());
                                    $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException->getMessage()." : directory path: ".$dirPath."/".$fileName,['file' => __FILE__, 'line' => __LINE__]);
                                    return back()->withError($message);
                                }

                                if($isDeleted) {
                                    $exception->logErrorMessage("Deleted uploaded file: ".$dirPath."/".$fileName." after unsuccessful push to git repository ",['file' => __FILE__, 'line' => __LINE__]);
                                }
                                return Response::json($responsearray, 400);
                            }
                            else{
                                $isPushed = true;
                            }

                            if($isPushed) {
                               // Log::info("CORPUSFILE UPLOAD: DELETING CACHE FOR CORPUS-DATA: ".$corpusProjectPath."/".$corpusPath."/CORPUS-DATA/, " .$corpusId);
                                //$this->LaudatioUtilService->emptyCorpusDataFileCache($corpusProjectPath.'/'.$corpusPath."/CORPUS-DATA/",$corpusId);
                                $contents = $this->flysystem->read($corpusProjectPath.'/'.$corpusPath."/CORPUS-DATA/");
                            }
                        }
                    }

                }

            }
        }

        return redirect()->route('corpus.edit', ['corpus' => $corpusId]);
    }


    /**
     *
     * UploadCorpusImage
     *
     *
     * @param Request $request
     * @return \Illuminate\Http\RedirectResponse
     */
    public function UploadCorpusImage(Request $request)
    {
        $user = \Auth::user();
        $dirPath = $request->directorypath;;
        $dirPathArray = explode("/",$dirPath);
        $corpusId = $request->corpusid;
        $corpusIsVersioned = null;

        $corpusProjectPath = $dirPathArray[0];
        $corpusPath = $dirPathArray[1];
        $uploadFolder = $dirPathArray[2];

        $corpus = null;

        if(isset($request->corpusid)) {
            $corpus = Corpus::find($corpusId);
            $corpusIsVersioned = $this->laudatioUtilsService->corpusIsVersioned($corpusId);


            $fileName = $request->formats->getClientOriginalName();
            $directoryPath = $this->laudatioUtilsService->getDirectoryPath(array($fileName),$fileName);

            $pathname = $request->formats->getPathName();
            $directoryPath = $this->laudatioUtilsService->getDirectoryPath(array($fileName),$fileName);

            $gitFunction = new GitFunction();

            $extension = substr($fileName,strrpos($fileName,"."));
            $newFileName = "corpusLogo".$extension;


            try {
                $createdPaths = $gitFunction->writeFiles($dirPath,array($newFileName), $this->flysystem,$request->formats->getRealPath(),$newFileName);
            }
            catch (FileExistsException $fileExistsException) {
                $fileUploadException = new FileUploadException($fileExistsException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FileNotFoundException $fileNotFoundException) {
                $fileUploadException = new FileUploadException($fileNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (FilesystemNotFoundException $filesystemNotFoundException) {
                $fileUploadException = new FileUploadException($filesystemNotFoundException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }
            catch (NotSupportedException $notSupportedException) {
                $fileUploadException = new FileUploadException($notSupportedException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);

            }
            catch (RootViolationException $rootViolationException) {
                $fileUploadException = new FileUploadException($rootViolationException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);


                return Response::json($responsearray, 400);

            }
            catch (UnreadableFileException $unreadableFileException) {
                $fileUploadException = new FileUploadException($unreadableFileException->getMessage(),0,null);

                $message = $fileUploadException->createUserFacingJSONMessage($fileUploadException->getMessage());
                $responsearray = array();
                array_push($responsearray,$message);

                return Response::json($responsearray, 400);
            }


            $addPath = $corpusProjectPath.'/'.$corpusPath.'/'.$uploadFolder.'/'.$newFileName;
            $commitPath = $corpusProjectPath.'/'.$corpusPath.'/'.$uploadFolder;
            $pushPath = $corpusProjectPath.'/'.$corpus->directory_path.'/'.$uploadFolder;


            if(strpos($createdPaths[0],$dirPath) !== false){

                //add files
                $isAdded = $this->GitRepoService->addFiles($this->flysystem, $addPath);

                //git commit The files
                if($isAdded) {
                    $corpus->corpus_logo = $newFileName;
                    $corpus->save();

                    if($corpusIsVersioned){
                        $returnPath = $this->GitRepoService->commitFiles($commitPath, "Adding files for " . $newFileName, $corpusId, $user->name, $user->email);
                        if (!empty($returnPath)) {
                            $isPushedSucessfully = $this->GitRepoService->pushFiles($pushPath,$corpusId,$user);
                            $isPushed = $this->GitRepoService->pushFiles($pushPath, $corpusId, $user);
                            if(is_array($isPushedSucessfully)){
                                $exception = $isPushedSucessfully["exception"];
                                $message = $exception->createUserFacingProcessJSONMessage($exception->getMessage());
                                $responsearray = array();
                                array_push($responsearray,$message);

                                $resetCorpusPath = "";
                                try {
                                    $isDeleted = $gitFunction->deleteFiles($addPath,$user->name,$user->email);
                                }
                                catch (CorpusRepositoryFileSystemException $corpusRepositoryFileSystemException) {
                                    $message = $corpusRepositoryFileSystemException->createUserFacingMessage($corpusRepositoryFileSystemException->getMessage());
                                    $corpusRepositoryFileSystemException->logErrorMessage($corpusRepositoryFileSystemException->getMessage()." : directory path: ".$addPath,['file' => __FILE__, 'line' => __LINE__]);
                                    return back()->withError($message);
                                }

                                if($isDeleted) {
                                    $exception->logErrorMessage("Deleted uploaded file: ".$addPath." after unsuccessful push to git repository ",['file' => __FILE__, 'line' => __LINE__]);
                                }
                                return Response::json($responsearray, 400);
                            }
                            else{
                                $isPushed = true;
                                $this->laudatioUtilsService->setCorpusLogoSymLink($this->flysystem,$addPath);
                            }
                        }
                    }
                }

            }
        }

        return redirect()->route('corpus.edit', ['corpus' => $corpusId]);
    }

    /**
     * @param $octal
     * @return string
     */
    public function convertOctalToCharacter($octal) {
        return chr(octdec($octal[1]));
    }

}