<?php
/**
 * Created by PhpStorm.
 * User: rolfguescini
 * Date: 06.07.17
 * Time: 16:44
 */

namespace App\Laudatio\Elasticsearch;

use App\Corpus;
use App\Custom\ElasticsearchInterface;
use App\Exceptions\LaudatioElasticsearchException;
use Elasticsearch\Common\Exceptions\NoNodesAvailableException;
use Elasticsearch\Common\Exceptions\Curl\CouldNotConnectToHost;
use Elasticsearch\Common\Exceptions\RequestTimeout408Exception;
use Elasticsearch\Common\Exceptions\Curl\CouldNotResolveHostException;
use Elasticsearch\Common\Exceptions\Curl\OperationTimeoutException;
use Elasticsearch\Common\Exceptions\Missing404Exception;
use Elasticsearch;
use Elasticsearch\Endpoints\DeleteByQuery;
use function GuzzleHttp\Psr7\_caseless_remove;
use Log;
use Cache;
use Response;
use Illuminate\Http\Request;

class ElasticService implements ElasticsearchInterface
{
    private $ELASTICSEARCH_HOST;
    private $ELASTICSEARCH_PORT;
    private $ELASTICSEARCH_SCHEME;
    private $ELASTICSEARCH_USER;
    private $ELASTICSEARCH_PASS;

    public function __construct()
    {
        $this->ELASTICSEARCH_HOST = config('laudatio.elasticsearch_host','depot1-5.cms.hu-berlin.de');
        $this->ELASTICSEARCH_PORT = config('laudatio.elasticsearch_port', 8080);
        $this->ELASTICSEARCH_SCHEME = config('laudatio.elasticsearch_scheme', 'http');
        $this->ELASTICSEARCH_USER = config('laudatio.elasticsearch_user', null);
        $this->ELASTICSEARCH_PASS = config('laudatio.elasticsearch_pass', null);
    }

    public function createIndex($params){

        return Elasticsearch::indices()->create($params);

    }

    public function reIndex($params) {
        return Elasticsearch::reindex($params);
    }

    public function deleteIndex($params){
        return Elasticsearch::indices()->delete($params);
    }


    public function createMappedIndex($indexMappingPath, $index_id) {
        $status = "";
        $result = array();

        //set mapping
        $mapping = json_decode(file_get_contents($indexMappingPath),true);

        $createIndexParams = array(
            'index' => $index_id,
            'body' => array(
                'settings' => $mapping['settings'],
                'mappings' => array(
                    'doc'=> array(
                        '_source' => array(
                            'enabled' => true
                        ),
                        'properties' => $mapping['mappings']['doc']['properties']
                    )
                )
            )
        );

        $indexResult = $this->createIndex($createIndexParams);

        /*
         * The Reindex API makes no effort to handle ID collisions. For such issues, the target index will remain valid,
         *  but it’s not easy to predict which document will survive because the iteration order isn’t well defined.
         *
         * setting new id to timestamp:now()_old_elasticsearch_id
         *
         */


        if($indexResult['acknowledged'] == 1
            && $indexResult['index'] == $index_id) {


            $status = "success";
            $result['create_mappedindex_response'] = "Success";

        }
        else{
            $status = "error";
            $result['publish_corpus_response'] = "There was a problem creating the Index. A message has been sent to the site administrator. Please try again later";
        }

        $response = array(
            'status' => $status,
            'message' => $result,
        );

        return $response;
    }

    public function postReindex($index, $postData) {
        $resultData = array();

        foreach($postData as $documentId => $data) {
            $params = array(
                "index" => $index,
                "body" => $data
            );

            $response = Elasticsearch::index($params);
            $resultData[$documentId] = $response;
        }
        return $resultData;
    }

    public function createReindex($index, $matchQuery, $new_corpus_id) {
        $status = "";
        $result = array();

        /*
         * The Reindex API makes no effort to handle ID collisions. For such issues, the target index will remain valid,
         *  but it’s not easy to predict which document will survive because the iteration order isn’t well defined.
         *
         * setting new id to timestamp:now()_old_elasticsearch_id
         *
         */

        if(strpos($index,"corpus") !== false){
            $indexParams = array(
                "body" => array(
                    "source" => array(
                        "index" => $index,
                        "query" => array(
                            "match" => $matchQuery
                        )
                    ),
                    "dest" => array(
                        "index" => $index
                    ),
                    "script" => array(
                        "source" => "ctx._source.publication_status = '0';ctx._source.is_latest_publication = 'false';ctx._source.laudatio_corpus_id = '".$new_corpus_id."'",
                        "lang" => "painless"
                    )
                )
            );
        }
        else{
            $indexParams = array(
                "body" => array(
                    "source" => array(
                        "index" => $index,
                        "query" => array(
                            "match" => $matchQuery
                        )
                    ),
                    "dest" => array(
                        "index" => $index
                    ),
                    "script" => array(
                        "source" => "ctx._source.publication_status = '0';ctx._source.is_latest_publication = 'false';ctx._source.laudatio_corpus_id = '".$new_corpus_id."'",
                        "lang" => "painless"
                    )
                )
            );
        }


        $reIndexResult = $this->reIndex($indexParams);

        $result['publish_response'] = "Success";
        $result['index_data'] = $reIndexResult;



        /*
        if($indexResult['acknowledged'] == 1
            && $indexResult['index'] == $new_index_id) {


            $reIndexResult = $this->reIndex($indexParams);


            $status = "success";
            $result['publish_response'] = "Success";
            $result['index_data'] = $reIndexResult;

        }
        else{
            $status = "error";
            $result['publish_response'] = "There was a problem publishing the Corpus. The error was: The corpus could not be published to git due to failed creation of index. A message has been sent to the site administrator. Please try again later";
            $result['index_data'] = array();
        }
        */

        $response = array(
            'status' => $status,
            'message' => $result,
        );

        return $response;
    }


    public function updateDocumentFieldsInAnnotation($new_annotation_index,$annotation_ids){
        //update the document ids in annotation documents
        $success = true;
        foreach($annotation_ids as $annotation_id => $document_ids) {
            $annotation_update_params = array(
                'index' => $new_annotation_index,
                
                'id' => $annotation_id,
                'body' => [
                    'doc' => [
                        'in_documents' => $document_ids
                    ]
                ]
            );
            $response = Elasticsearch::update($annotation_update_params);
            if(empty($response['_shards'])) {
                $success = false;
                break;
            }
        }
        return $success;
    }

    public function updatePublicationStatus($index,$ids){
        //update the document ids in annotation documents
        $success = true;
        foreach($ids as $oldAnnotationIndex => $elasticsearchids) {
            foreach ($elasticsearchids as $elasticsearchid) {
                if(!empty($elasticsearchid)) {
                    $update_params = array(
                        'index' => $oldAnnotationIndex,
                        
                        'id' => $elasticsearchid,
                        'body' => [
                            'doc' => [
                                'publication_status' => "1"
                            ]
                        ]
                    );
                    $response = Elasticsearch::update($update_params);

                    if(empty($response['_shards'])) {
                        $success = false;
                        break;
                    }
                }
            }

        }
        return $success;
    }

    /**
     * @param $index
     * @param $ids
     * @param $corpus_id
     */
    public function updateInCorpora($index, $ids, $corpus_id) {
        $success = true;

        foreach($ids as $elasticsearchid) {
            $update_params = array(
                'index' => $index,

                'id' => $elasticsearchid,
                'body' => [
                    'doc' => [
                        'in_corpora' => $corpus_id
                    ]
                ]
            );

            $response = Elasticsearch::update($update_params);

            if(empty($response['_shards'])) {
                $success = false;
                break;
            }
        }
    }




    public function postToIndex($params) {
        $response = null;

        try {
            $response = Elasticsearch::index($params);
        }
        catch (\Exception $e) {
            $response = array($e);
        }

        return $response;
    }

    public function setMapping($params){
        $response = array();
        try {
            $response = Elasticsearch::putMapping($params);
        }
        catch (\Exception $e) {
            $response = array($e);
        }

        return $response;
    }

    public function setCorpusToPublished($params){
        $response = array();
        try {
            $response = Elasticsearch::update($params);
        }
        catch (\Exception $e) {
            $response = array($e);
        }

        return $response;
    }

    public function unsetCorpusToPublished($params){
        $response = array();
        try {
            $response = Elasticsearch::update($params);
        }
        catch (\Exception $e) {
            $response = array($e);
        }

        return $response;
    }

    public function setDocumentAndAnnotationsToPublished($params){
        $response = array();
        try {
            foreach($params as $param) {
                $response = Elasticsearch::update($param);
            }
        }
        catch (\Exception $e) {
            $response = array($e);
        }

        return $response;
    }

    public function unsetDocumentAndAnnotationsToPublished($params){
        $response = array();
        try {
            foreach($params as $param) {
                $response = Elasticsearch::update($param);
            }
        }
        catch (\Exception $e) {
            $response = array($e);
        }

        return $response;
    }

    public function pingServer($wantsJSON = false){
        if(!Elasticsearch::ping()){
            throw new LaudatioElasticsearchException("Cluster is not available: ", 404, null,$wantsJSON);
        }
        else{
            return true;
        }
    }

    public function getPublishedCorpora(){
        $resultData = array();

        $current_corpus_index =  config("laudatio.elasticsearch_corpus_index","corpora");


        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        //$queryBody = $queryBuilder->buildMatchAllQuery(array());
        $queryBody = $queryBuilder->buildMustQuery(
            array(
                array(
                    "publication_status" => "1"
                )
            )
        );

        $params = [
            'size' => 10000,
            'index' => $current_corpus_index,
            'body' => $queryBody,
        ];

        $response = array();
        try {
            $response = Elasticsearch::search($params);
        } catch (NoNodesAvailableException $nodesAvailableException) {
            throw new LaudatioElasticsearchException("Cluster is not available: " . $nodesAvailableException->getMessage(), 404, null);
        } catch (CouldNotConnectToHost $couldNotConnectToHost) {
            throw new LaudatioElasticsearchException("Cluster is not available: " . $couldNotConnectToHost->getMessage(), 404, null);
        }
        catch (CouldNotResolveHostException $couldNotResolveHostException) {
            throw new LaudatioElasticsearchException("Cluster is not available: " . $couldNotResolveHostException->getMessage(), 404, null);
        }

        if(isset($response['hits']['hits'])){
            array_push($resultData,$response['hits']['hits']);
        }


        return array(
            'error' => false,
            'result' => $resultData
        );
    }


public function doSearch($params)
{

    $response = array();

    try {
        $response = Elasticsearch::search($params);
    } catch (NoNodesAvailableException $nodesAvailableException) {
        throw new LaudatioElasticsearchException("Cluster is not available: " . $nodesAvailableException->getMessage(), 404, null);
    } catch (CouldNotConnectToHost $couldNotConnectToHost) {
        throw new LaudatioElasticsearchException("Cluster is not available: " . $couldNotConnectToHost->getMessage(), 404, null);
    }
    catch (CouldNotResolveHostException $couldNotResolveHostException) {
        throw new LaudatioElasticsearchException("Cluster is not available: " . $couldNotResolveHostException->getMessage(), 404, null);
    }

    return $response;
}



    /**
     * @param $id
     * @param bool $full
     * @param $index
     * @return mixed
     */
    public function getCorpus($id,$full = true, $index){
        $returned_response = array();

        if(!$full){
            $params = [
                'index' => $index,

                'id' => $id,
                '_source' => ["document_title","document_publication_publishing_date","document_list_of_annotations_name","in_corpora","publication_version","publication_status", "has_annotations", "has_documents"]
            ];
        }
        else{
            $params = [
                'index' => $index,

                'id' => $id,
                '_source_excludes' => ['message']
            ];
        }

        $response = Elasticsearch::get($params);

        if(!$full){
            $returned_response = array(
                'status' => 'success',
                'result' => $response['_source']
            );
        }
        else{
            $returned_response = array(
                'status' => 'success',
                'found' => $response['found'],
                'result' => $response['_source']
            );
        }



        return Response::json($returned_response);
    }

    public function deleteCorpus($id){
        $result = array();
        $queryBuilder = new QueryBuilder();
        $queryBody = $queryBuilder->buildSingleMatchQuery(array(array('_id' => $id)));
        $params = [
            'size' => 1,
            'index' => 'corpus',
            
            'body' => $queryBody,
        ];

        $response = Elasticsearch::deletebyquery($params);
        if($response['deleted'] > 0){
            array_push($result,$response);
        }
        return array(
            'error' => false,
            'result' => $result
        );
    }

    /**
     * @param $id
     * @param bool $full
     * @param $index
     * @return mixed
     */
    public function getDocument($id,$full = true,$index){
        $returned_response = array();

        if(!$full){
            $params = [
                'index' => $index,

                'id' => $id,
                '_source' => ["document_title","document_publication_publishing_date","document_publication_place","document_list_of_annotations_name","in_corpora","document_size_extent"]
            ];
        }
        else{
            $params = [
                'index' => $index,

                'id' => $id,
                '_source_excludes' => ['message']
            ];
        }


        $response = Elasticsearch::get($params);

        if(!$full){
            $returned_response = array(
                'status' => 'success',
                'result' => $response['_source'],
                '_id' => $response['_id']
            );
        }
        else{
            $returned_response = array(
                'status' => 'success',
                'found' => $response['found'],
                'result' => $response['_source'],
                '_id' => $response['_id']
            );
        }


        return Response::json($returned_response);
    }


    /**
     * @param $id
     * @param bool $full
     * @return array
     */
    public function getAnnotation($id,$full = true, $index){
        $returned_response = array();

        if(!$full){
            $params = [
                'index' => $index,

                'id' => $id,
                '_source' => [
                    "preparation_encoding_annotation_group",
                    "preparation_annotation_id",
                    "_id",
                    "_index",
                    "preparation_title",
                    "in_documents",
                    "in_corpora",
                    "preparation_author_annotator_forename",
                    "preparation_author_annotator_surname",
                    "publication_status",
                    "laudatio_corpus_id"
                ]
            ];
        }
        else{
            $params = [
                'index' => $index,
                'id' => $id,
                '_source_excludes' => ['message']
            ];
        }

        $response = Elasticsearch::get($params);

        if(!$full){
            $returned_response = array(
                'status' => 'success',
                'result' => $response['_source'],
                '_id' => $response['_id']
            );
        }
        else{
            $returned_response = array(
                'status' => 'success',
                'found' => $response['found'],
                'result' => $response['_source'],
                '_id' => $response['_id']
            );
        }


        return Response::json($returned_response);

    }



    public function deleteIndexedObject($index,$args){
        $result = array();
        $error = false;

        $queryBody = null;
        $queryBuilder = new QueryBuilder();

        $queryBody = $queryBuilder->buildMustQuery($args);


        $params = [
            'size' => 1,
            'index' => $index,
            'body' => $queryBody,
        ];

        $response = Elasticsearch::deletebyquery($params);


        if($response['deleted'] > 0 && count($response['failures']) == 0){
            array_push($result,$response);
        }
        else {
            $error = true;
            array_push($result,$response['failures']);
        }

        return array(
            'error' => $error,
            'result' => $result
        );
    }

    public function setHasAnnotations($data, $corpusid, $corpusindex, $index, $deleted_id){
        $responseData = array();
        $corpusAnnotations = array();
        $corpus = Corpus::find($corpusid);
        $responseData['document_has_annotations'] = array();
        
        foreach($data as $documentElasticId => $annotationids){
            $updateData = [
                'index' => $index,
                'id' => $documentElasticId,
                'body' => array(
                    'script' => array(
                        "source" => "ctx._source.has_annotations = params.annotations",
                        "lang" => "painless",
                        "params" => array(
                            "annotations" => $annotationids
                        )
                    )
                )
            ];

            $documentresponse = Elasticsearch::update($updateData);
            array_push($responseData['document_has_annotations'],$documentresponse);
        }


        foreach($corpus->annotations()->get() as $corpusAnnotation) {
            if(!empty($corpusAnnotation->elasticsearch_id) && $corpusAnnotation->elasticsearch_id != $deleted_id) {
                array_push($corpusAnnotations, $corpusAnnotation->elasticsearch_id);
            }
        }

        $corpusUpdateData = [
            'index' => $corpusindex,
            'id' => $corpus->elasticsearch_id,
            'body' => [
                'doc' => [
                    'has_annotations' => $corpusAnnotations
                ]
            ]
        ];

        $corpusResponse = Elasticsearch::update($corpusUpdateData);
        $responseData['corpus_has_annotations'] = $corpusResponse;

        return $responseData;
    }


    public function setHasDocuments($data, $corpusid, $corpusindex, $index, $deleted_id){
        $responseData = array();
        $corpusDocuments = array();
        $corpus = Corpus::find($corpusid);

        $responseData['annotation_in_documents'] = array();
        foreach($data as $annotationElasticId => $documentids){

            $updateData = [
                'index' => $index,
                'id' => $annotationElasticId,
                'body' => array(
                    'script' => array(
                        "source" => "ctx._source.in_documents = params.documents",
                        "lang" => "painless",
                        "params" => array(
                            'documents' => $documentids
                        )
                    )
                )
            ];

            $annotationResponse = Elasticsearch::update($updateData);
            array_push($responseData['annotation_in_documents'],$annotationResponse);
        }


        foreach($corpus->documents()->get() as $corpusDocument) {
            if(!empty($corpusDocument->elasticsearch_id) && $corpusDocument->elasticsearch_id != $deleted_id){
                array_push($corpusDocuments, $corpusDocument->elasticsearch_id);
            }

        }


        $corpusUpdateData = [
            'index' => $corpusindex,
            'id' => $corpus->elasticsearch_id,
            'body' => [
                'doc' => [
                    'has_documents' => $corpusDocuments
                ]
            ]
        ];

        $corpusResponse = Elasticsearch::update($corpusUpdateData);
        $responseData['corpus_has_documents']  = $corpusResponse;

        return $responseData;
    }

    public function setInDocuments($data, $corpusid, $index){}


    public function getElasticIdByObjectId($index,$objectparams){
        $elasticIds = array();
        $queryBuilder = new QueryBuilder();
        foreach ($objectparams as $objectId => $objectparam){
            if(count($objectparam) == 1){
                $queryBody = $queryBuilder->buildSingleMatchQuery($objectparam);

            }
            else if(count($objectparam) >  1){
                $queryBody = $queryBuilder->buildMustQuery($objectparam);
            }

            $params = [
                'size' => 10,
                'index' => $index,
                
                'body' => $queryBody,
                '_source' => ["_id"]
            ];

            $response = Elasticsearch::search($params);

            $hits = isset($response['hits']['hits'][0]) ? $response['hits']['hits'][0] : false;
            if($hits){
                $elasticIds[$objectId] = array(
                    "elasticsearchid" => $response['hits']['hits'][0]['_id'],
                    "elasticsearchindex" => $response['hits']['hits'][0]['_index'],
                );
            }

        }

        return $elasticIds;
    }

    public function setWorkflowStatusByCorpusId($corpus_id){

    }


    public function getAnnotationByNameAndCorpusId($name, $corpusId, $fields,$index){
        $result = array();

        if (Cache::tags(['annotation_'.$name.'_'.$corpusId.'_'.$index])->has("getAnnotationByNameAndCorpusId_".$name."_".$corpusId."_".$index)) {
            $result = Cache::tags(['annotation_'.$name.'_'.$corpusId.'_'.$index])->get("getAnnotationByNameAndCorpusId_".$name."_".$corpusId."_".$index);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = $queryBuilder->buildMustQuery(array(
                array(
                    'preparation_annotation_id' => $name,
                    'in_corpora' => $corpusId
                )
            ));
            $params = [
                'size' => 1,
                'index' => $index,
                
                'body' => $queryBody,
                '_source' => $fields,
                'filter_path' => ['hits.hits']
            ];

            $response = Elasticsearch::search($params);

            if(count($response['hits']['hits']) > 0){
                array_push($result,$response['hits']['hits'][0]);
                Cache::tags(['annotation_'.$name.'_'.$corpusId.'_'.$index])->forever("getAnnotationByNameAndCorpusId_".$name."_".$corpusId."_".$index, $result);
            }

        }//end if cache


        return array(
            'error' => false,
            'result' => $result
        );
    }

    /**
     * @param $corpusId
     * @param $annotationsIds
     * @param $index
     * @param $fields
     * @return array
     */
    public function getAnnotationsByCorpus($corpusId, $annotationsIds, $index, $fields) {
        $resultData = array();

        foreach ($annotationsIds as $annotationId) {
            if(!empty($annotationId)) {
                $JSONDocument = $this->getAnnotation($annotationId, true, $index);
                $annotation = json_decode($JSONDocument->getContent(),true);
                if(isset($annotation['result'])){
                    $annotation['result']['_id'] = $annotation['_id'];
                    array_push($resultData,$annotation['result']);
                }
                else{
                    $resultData  = array();
                }
            }

        }

        return $resultData;
    }

    public function getAnnotationsByCorpusId($corpusId,$index,$fields){
        $result = array();
        $queryBuilder = new QueryBuilder();
        $queryBody = $queryBuilder->buildSingleMatchQuery(array(
            array(
                'in_corpora' => $corpusId
            )
        ));
        $params = [
            'size' => 10000,
            'index' => $index,

            'body' => $queryBody,
            '_source' => join(",",$fields),
            'filter_path' => ['hits.hits']
        ];

        $response = Elasticsearch::search($params);

        if(count($response['hits']['hits']) > 0){
            array_push($result,$response['hits']['hits']);

        }


        return array(
            'error' => false,
            'result' => $result
        );
    }

    public function getAnnotationByCorpus($searchData,$corpusData,$fields,$index){
        $resultData = array();
        $fieldString = join("_",$fields);

        if (Cache::tags(['annotation_'.$corpusData[0].'_'.$fieldString.'_'.$index])->has("getAnnotationByCorpus_".$corpusData[0].'_'.$fieldString.'_'.$index)) {
            $resultData = Cache::tags(['annotation_'.$corpusData[0].'_'.$fieldString.'_'.$index])->get("getAnnotationByCorpus_".$corpusData[0].'_'.$fieldString.'_'.$index);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $counter = 0;
            foreach($searchData as $queryData){
                $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));
                $params = [
                    'size' => 10000,
                    'index' => $index,
                    
                    'body' => $queryBody,
                    '_source' => $fields,
                ];

                $results = Elasticsearch::search($params);
                $termData = array_values($corpusData);

                if(count($results['hits']['hits']) > 0){
                    $resultData[$termData[$counter++]] = $results['hits']['hits'];
                }
                else{
                    $resultData[$counter++] = array();
                }
            }//end foreach queries
            Cache::tags(['annotation_'.$corpusData[0].'_'.$fieldString.'_'.$index])->forever("getAnnotationByCorpus_".$corpusData[0].'_'.$fieldString.'_'.$index, $resultData);
        }

        return $resultData;
    }

    /**
     * @param $documentIds
     * @param $corpusData
     * @param $fields
     * @param $index
     * @return array
     */
    public function getDocumentByCorpus($documentIds, $corpusData, $fields, $index)
    {
        $resultData = array();
        foreach ($documentIds as $documentId) {
            if(!empty($documentId)) {
                $JSONDocument = $this->getDocument($documentId, true, $index);
                $document = json_decode($JSONDocument->getContent(),true);
                if(isset($document['result'])){
                    $document['result']['_id'] = $document['_id'];
                    array_push($resultData,$document['result']);
                }
                else{
                    $resultData  = array();
                }
            }

        }
        return $resultData;
    }


    public function getCorpusByDocument($searchData,$documentData,$index){
        $resultData = array();

        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $counter = 0;
        foreach($searchData as $queryData){
            $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));

            $params = [
                'size' => 10000,
                'index' => $index,
                
                'body' => $queryBody,
                //'_source_excludes' => ['message'],
                '_source' => ["corpus_title","corpus_publication_publication_date","corpus_documents","annotation_name","corpus_publication_license_description","corpus_publication_publisher","corpus_encoding_project_homepage","corpus_editor_forename","corpus_editor_surname","corpus_version","publication_version","publication_status","corpus_publication_license"],
                'filter_path' => ['hits.hits']
            ];

            $results = Elasticsearch::search($params);
            $termData = array_values($documentData);

            if(count($results['hits']['hits']) > 0){
                $resultData[$termData[$counter++]] = $results['hits']['hits'];
            }
            else{
                $resultData[$counter++] = array();
            }

        }
        return $resultData;
    }

    public function getAnnotationsByDocumentId($documentId,$index,$fields) {
        $result = array();
        $queryBuilder = new QueryBuilder();
        $queryBody = $queryBuilder->buildSingleMatchQuery(array(
            array(
                'in_documents' => $documentId
            )
        ));
        $params = [
            'size' => 10000,
            'index' => $index,

            'body' => $queryBody,
            '_source' => join(",",$fields),
            'filter_path' => ['hits.hits']
        ];

        $response = Elasticsearch::search($params);

        if(count($response['hits']['hits']) > 0){
            array_push($result,$response['hits']['hits']);

        }


        return array(
            'error' => false,
            'result' => $result
        );
    }

    /**
     * @param $searchData
     * @param $documentData
     * @return array
     *
     * called by getAnnotationsByDocument in Elasticcontroller - to be deleted ?
     */
    public function getAnnotationByDocument($searchData,$documentData){
        $resultData = array();

        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $counter = 0;
        foreach($searchData as $queryData){
            $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));
            $params = [
                'size' => 10000,
                'index' => 'annotation',
                
                'body' => $queryBody,
                //'_source_excludes' => ['message'],
                '_source' => ["preparation_title", "in_corpora", "in_documents"],
                'filter_path' => ['hits.hits']
            ];

            $results = Elasticsearch::search($params);
            $termData = array_values($documentData);

            if(count($results['hits']['hits']) > 0){
                $resultData[$termData[$counter++]]['results'] = $results['hits']['hits'];
            }
            else{
                $resultData[$counter++]['results'] = array();
            }
        }//end foreach queries
        return $resultData;
    }

    /** GET search endpoint
     * @param $index
     * @param $field
     * @param $term
     * @return \Illuminate\Contracts\Routing\ResponseFactory^\Symfony\Component\HttpFoundation\Response
     */
    public function search($index,$field,$term)
    {
        $params = [
            'size' => 10000,
            'index' => $index,
            
            'body' => [
                'query' => [
                    'match' => [
                        ''.$field.'' => $term
                    ]
                ]
            ],
            '_source_excludes' => ['message']
        ];


        $results = Elasticsearch::search($params);
        $milliseconds = $results['took'];
        $maxScore     = $results['hits']['max_score'];

        return response(
            array(
                'error' => false,
                'milliseconds' => $milliseconds,
                'maxscore' => $maxScore,
                'results' => $results['hits']['hits']
            ),
            200
        );

    }




    /**
     * searchGeneral
     * @param $searchData
     * @return mixed
     */
    public function searchMain($searchData,$indexes,$source,$from,$size)
    {
        $resultData = array();

        try{
            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $queryBody = $queryBuilder->buildBoolMultiMatchFilteredHighlightQuery($searchData,$size);
            if($size < 10000) {
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $indexes,

                    'body' => $queryBody,
                    '_source' => $source
                ];
            }
            else{
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $indexes,

                    'body' => $queryBody,
                    '_source_includes' => 'total'
                ];
            }


            $resultData = Elasticsearch::search($params);
            //Log::info("SEARCHMAIN: ".print_r($params,1));
        } catch (NoNodesAvailableException $nodesAvailableException) {
            Log::info("NoNodesAvailableException: ".$nodesAvailableException->getMessage());
            throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
        } catch (CouldNotConnectToHost $couldNotConnectToHost) {
            Log::info("CouldNotConnectToHost: ".$couldNotConnectToHost->getMessage());
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
        }
        catch (CouldNotResolveHostException $couldNotResolveHostException) {
            Log::info("CouldNotResolveHostException: ".$couldNotResolveHostException->getMessage());
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
        }
        catch (\Exception $e) {
            Log::info("SomethingElseException: ".$e->getMessage());
        }

        return $resultData;
    }

    public function searchMainWildCards($searchData,$indexes,$source,$from,$size)
    {
        $resultData = array();

        try{
            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $queryBody = $queryBuilder->buildFilteredHighlightedWildCardQuery($searchData);
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $indexes,

                'body' => $queryBody,
                '_source' => $source
            ];
            //Log::info("searchMainWildCards: ".print_r($params,1));
            $resultData = Elasticsearch::search($params);

        } catch (NoNodesAvailableException $nodesAvailableException) {
            Log::info("searchMainWildCards: NoNodesAvailableException: ".$nodesAvailableException->getMessage());
            throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
        } catch (CouldNotConnectToHost $couldNotConnectToHost) {
            Log::info("searchMainWildCards: CouldNotConnectToHost: ".$couldNotConnectToHost->getMessage());
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
        }
        catch (CouldNotResolveHostException $couldNotResolveHostException) {
            Log::info("searchMainWildCards: CouldNotResolveHostException: ".$couldNotResolveHostException->getMessage());
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
        }
        catch (\Exception $e) {
            Log::info("searchMainWildCards: SomethingElseException: ".$e->getMessage());
        }

        return $resultData;
    }

    /**
     * @param $searchData
     * @param $fullTextQuery
     * @param $index
     * @param $from
     * @param $size
     * @param $source
     * @return array
     */
    public function filterSearch($searchData, $fullTextQuery,  $index, $from, $size, $source, $hasWildCards)
    {
        $resultData = array();

        try{

            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $fields = array();

            foreach ($searchData['filterData'] as $field => $value) {
                if(!empty($field) && !empty($value)) {
                    $fields[$field] = $value;
                }

            }
            //Log::info("filterSearch: ".print_r($searchData,1));
            $fields['fields'] = $searchData['fields'];

            $queryBody = $queryBuilder->buildFilteredHighlightedMustQuery(
                $fields,
                $fullTextQuery,
                $hasWildCards,
                array(
                    "is_latest_publication" => "true"
                )
            );

            if($size < 10000) {
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $index,
                    'body' => $queryBody,
                    '_source' => $source
                ];
            }
            else{
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $index,
                    'body' => $queryBody,
                    '_source_includes' => 'total'
                ];
            }

            //Log::info("filterSearch parms: ".print_r($params,1));
            $resultData = Elasticsearch::search($params);

        } catch (NoNodesAvailableException $nodesAvailableException) {
            throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
        } catch (CouldNotConnectToHost $couldNotConnectToHost) {
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
        }
        catch (CouldNotResolveHostException $couldNotResolveHostException) {
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
        }

        return $resultData;
    }

    /**
     * searchGeneral
     * @param $searchData
     * @return mixed
     */
    public function searchMain2($searchData,$indexes,$source,$from,$size)
    {
        $resultData = array();

        if (Cache::store('redis')->has('search_published_all_'.$indexes.'_'.$searchData['query'].'_'.$from.'_'.$size)) {
            $resultData = Cache::store('redis')->get('search_published_all_'.$indexes.'_'.$searchData['query'].'_'.$from.'_'.$size);
        }
        else {
            try{
                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildMultiMatchHighlightQuery($searchData);
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $indexes,
                    
                    'body' => $queryBody,
                    '_source' => $source
                ];
                $resultData = Elasticsearch::search($params);
                Cache::store('redis')->forever('search_published_all_'.$indexes.'_'.$searchData['query'].'_'.$from.'_'.$size, $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
        }

        return $resultData;
    }

    /**
     * listAllPublished
     * @param $searchData
     * @return mixed
     */
    public function listAllPublished($searchData,$randomIndexes)
    {
        $resultData = array();

        if (Cache::store('redis')->has('search_published_corpora_'.$randomIndexes)) {
            $resultData = Cache::store('redis')->get('search_published_corpora_'.$randomIndexes);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $queryBody = $queryBuilder->buildSingleMatchQuery(array(array("publication_status" => '1')));
            //$queryBody = $queryBuilder->buildMatchAllQuery(array());
            $params = [
                'size' => 10000,
                'index' => $randomIndexes,//$searchData['indices'],//trim($searchData['indices']),
                
                'body' => $queryBody,
                //'_source_excludes' => ['message']
                '_source' => $searchData['source']
            ];

            $resultData = Elasticsearch::search($params);
            Cache::store('redis')->forever('search_published_corpora_'.$randomIndexes, $resultData);
        }

        return $resultData;
    }



    public function getAllPublishedCorpora(){
        $resultData = array();
        if (Cache::store('redis')->has('search_published_corpora')) {
            $resultData = Cache::store('redis')->get('search_published_corpora');
        }
        else {
            try {

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildFilteredMustExistQuery(
                    array(
                        "field" => "corpus_title"
                    ),
                    array(
                        "is_latest_publication" => "true"
                    )
                );
                $params = [
                    'size' => 10000,
                    'index' => "_all",
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);
                Cache::store('redis')->forever('search_published_corpora', $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getLatestPublishedCorpora($indexes,$source,$from,$size){
        $resultData = array();

        try {

            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $queryBody = $queryBuilder->buildTermQuery(

                array(
                    "field" => "is_latest_publication",
                    "value" => "true"
                )
            );
            $params = [
                'from' => $from,
                'size' => $size,
                'index' => $indexes,
                
                'body' => $queryBody,
                '_source' => $source
            ];

            $resultData = Elasticsearch::search($params);

        } catch (NoNodesAvailableException $nodesAvailableException) {
            throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
        } catch (CouldNotConnectToHost $couldNotConnectToHost) {
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
        }
        catch (CouldNotResolveHostException $couldNotResolveHostException) {
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
        }

        return $resultData;
    }

    public function getPublishedCorpus($id,$index) {
        $resultData = array();
        if (Cache::store('redis')->has('search_published_corpora_'.$id.'_'.$index)) {
            $resultData = Cache::store('redis')->get('search_published_corpora_'.$id.'_'.$index);
        }
        else {
            try {

                $params = [
                    'index' => $index,
                    
                    'id' => $id,
                    '_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::get($params);
                Cache::store('redis')->forever('search_published_corpora', $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getAllPublishedDocumentsByCorpus($corpus_id,$document_index){
        $resultData = array();

        if (Cache::store('redis')->has('search_published_documents_by_corpus_'.$corpus_id.'_'.$document_index)) {
            $resultData = Cache::store('redis')->get('search_published_documents_by_corpus_'.$corpus_id.'_'.$document_index);
        }
        else {
            try {

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildFilteredMustQuery(
                    array(
                        "in_corpora" => $corpus_id
                    ),
                    array(
                        "is_latest_publication" => "true"
                    )
                );
                $params = [
                    'size' => 10000,
                    'index' => $document_index,
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);

                Cache::store('redis')->forever('search_published_documents_by_corpus_'.$corpus_id.'_'.$document_index, $resultData);
            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getAllPublishedDocuments(){
        $resultData = array();
        if (Cache::store('redis')->has('search_published_documents')) {
            $resultData = Cache::store('redis')->get('search_published_documents');
        }
        else {
            try {

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildFilteredMustExistQuery(
                    array(
                        "field" => "document_title"
                    ),
                    array(
                        "is_latest_publication" => "true"
                    )
                );
                $params = [
                    'size' => 10000,
                    'index' => "_all",
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);
                Cache::store('redis')->forever('search_published_documents', $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getLatestPublishedDocuments2($indexes,$source,$from,$size){
        $resultData = array();

        if (Cache::store('redis')->has('latest_published_documents_'.$from.'_'.$size)) {
            $resultData = Cache::store('redis')->get('latest_published_documents_'.$from.'_'.$size);
            Log::info("getLatestPublishedDocuments: GOT FROM CACHE with key: 'latest_published_documents_".$from."_".$size);
        }
        else {
            try {
                Log::info("getLatestPublishedDocuments: GETTING FROM server");
                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildTermsQuery(

                    array( array(
                        "field" => "publication_status",
                        "value" => "1"
                    ),
                        array(
                            "field" => "is_latest_publication",
                            "value" => true
                        ))
                );
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $indexes,

                    'body' => $queryBody,
                    '_source' => $source
                ];

                $resultData = Elasticsearch::search($params);
                Cache::store('redis')->forever('latest_published_documents_'.$from.'_'.$size, $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }

        }


        return $resultData;
    }

    public function getLatestPublishedDocuments($indexes,$source,$from,$size){
        $resultData = array();


        try {

            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $queryBody = $queryBuilder->buildTermsQuery(

                array( array(
                    "field" => "publication_status",
                    "value" => "1"
                ),
                    array(
                        "field" => "is_latest_publication",
                        "value" => true
                    ))
            );
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $indexes,

                'body' => $queryBody,
                '_source' => $source
            ];

            $resultData = Elasticsearch::search($params);

        } catch (NoNodesAvailableException $nodesAvailableException) {
            throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
        } catch (CouldNotConnectToHost $couldNotConnectToHost) {
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
        }
        catch (CouldNotResolveHostException $couldNotResolveHostException) {
            throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
        }

        return $resultData;
    }

    public function getPublishedDocument($id,$index) {
        $resultData = array();
        if (Cache::store('redis')->has('search_published_documents_'.$id.'_'.$index)) {
            $resultData = Cache::store('redis')->get('search_published_documents_'.$id.'_'.$index);
        }
        else {
            try {

                $params = [
                    'index' => $index,
                    
                    'id' => $id,
                    '_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::get($params);
                Cache::store('redis')->forever('search_published_documents_'.$id.'_'.$index, $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getAllPublishedAnnotationsByDocument($id,$index){
        $filteredResults = array();

        if (Cache::store('redis')->has('search_published_annotations_by_documents_'.$id.'_'.$index)) {
            $filteredResults = Cache::store('redis')->get('search_published_annotations_by_documents_'.$id.'_'.$index);
        }
        else {
            try {
                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildFilteredMustQuery(
                    array(
                        "in_documents" => $id
                    ),
                    array(
                        "is_latest_publication" => true
                    )
                );
                $params = [
                    'size' => 10000,
                    'index' => "_all",
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);
                $filteredResults = array();


                $first_underscore = strpos($index,"_");
                $postfix = substr($index, $first_underscore);
                $annotation_index = 'annotation'.$postfix;


                foreach($resultData['hits']['hits'] as $hit){
                    if($hit['_index'] == $annotation_index) {
                        array_push($filteredResults,$hit);
                    }
                }

                Cache::store('redis')->forever('search_published_annotations_by_documents_'.$id.'_'.$index, $filteredResults);
            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $filteredResults;
    }


    public function getAllPublishedDocumentsByAnnotation($id,$index){
        $filteredResults = array();

        if (Cache::store('redis')->has('search_published_documents_by_annotation_'.$id.'_'.$index)) {
            $filteredResults = Cache::store('redis')->get('search_published_documents_by_annotation_'.$id.'_'.$index);
        }
        else {
            try {

                $annotation  = $this->getPublishedAnnotation($id,$index);

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildIdsQuery($annotation['_source']['in_documents']);
                $params = [
                    'size' => 10000,
                    'index' => "_all",
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);
                $filteredResults = array();


                $first_underscore = strpos($index,"_");
                $postfix = substr($index, $first_underscore);
                $document_index = 'document'.$postfix;


                foreach($resultData['hits']['hits'] as $hit){
                    if($hit['_index'] == $document_index) {
                        array_push($filteredResults,$hit);
                    }
                }

                Cache::store('redis')->forever('search_published_documents_by_annotation_'.$id.'_'.$index, $filteredResults);
            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $filteredResults;
    }


    public function getAllPublishedAnnotationsByCorpus($corpus_id,$annotation_index){
        $resultData = array();

        if (Cache::store('redis')->has('search_published_annotations_by_corpus_'.$corpus_id.'_'.$annotation_index)) {
            $resultData = Cache::store('redis')->get('search_published_annotations_by_corpus_'.$corpus_id.'_'.$annotation_index);
        }
        else {
            try {

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildFilteredMustQuery(
                    array(
                        "in_corpora" => $corpus_id
                    ),
                    array(
                        "is_latest_publication" => true
                    )
                );
                $params = [
                    'size' => 10000,
                    'index' => $annotation_index,
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);

                Cache::store('redis')->forever('search_published_annotations_by_corpus_'.$corpus_id.'_'.$annotation_index, $resultData);
            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getAllPublishedAnnotations(){
        $resultData = array();
        if (Cache::store('redis')->has('search_published_annotations')) {
            $resultData = Cache::store('redis')->get('search_published_annotations');
        }
        else {
            try {

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildFilteredMustExistQuery(
                    array(
                        "field" => "preparation_title"
                    ),
                    array(
                        "is_latest_publication" => true
                    )
                );
                $params = [
                    'size' => 10000,
                    'index' => "_all",
                    
                    'body' => $queryBody,
                    //'_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::search($params);
                Cache::store('redis')->forever('search_published_annotations', $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }

    public function getAllPublishedAnnotationCount(){
        if (Cache::store('redis')->has('search_published_annotations_count')) {
            $resultData = Cache::store('redis')->get('search_published_annotations_count');
        }
        else {

        }
    }


    public function getLatestPublishedAnnotations2($indexes,$source,$from,$size){
        $resultData = array();

        if (Cache::store('redis')->has('latest_published_annotations_'.$from.'_'.$size)) {
            $resultData = Cache::store('redis')->get('latest_published_annotations_'.$from.'_'.$size);
            Log::info("getLatestPublishedAnnotations: GOT FROM CACHE with key: 'latest_published_annotations_".$from."_".$size);
        }
        else {
            try {
                Log::info("getLatestPublishedAnnotations: GETTING FROM server");
                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildTermsQuery(

                    array( array(
                        "field" => "publication_status",
                        "value" => "1"
                    ),
                        array(
                            "field" => "is_latest_publication",
                            "value" => true
                        ))
                );
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $indexes,

                    'body' => $queryBody,
                    '_source' => $source
                ];

                $resultData = Elasticsearch::search($params);
                Cache::store('redis')->forever('latest_published_annotations_'.$from.'_'.$size, $resultData);
            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }
        }

        return $resultData;
    }
    public function getLatestPublishedAnnotations($indexes,$source,$from,$size){
            $resultData = array();

            try {

                $queryBuilder = new QueryBuilder();
                $queryBody = null;
                $queryBody = $queryBuilder->buildTermsQuery(

                    array( array(
                        "field" => "publication_status",
                        "value" => "1"
                    ),
                        array(
                            "field" => "is_latest_publication",
                            "value" => true
                        ))
                );
                $params = [
                    'size' => $size,
                    'from' => $from,
                    'index' => $indexes,

                    'body' => $queryBody,
                    '_source' => $source
                ];

                $resultData = Elasticsearch::search($params);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 404, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 404, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 404, null, true);
            }

            return $resultData;
        }

    public function getPublishedAnnotation($id,$index) {
        $resultData = array();
        if (Cache::store('redis')->has('search_published_annotations_'.$id.'_'.$index)) {
            $resultData = Cache::store('redis')->get('search_published_annotations_'.$id.'_'.$index);
        }
        else {
            try {

                $params = [
                    'index' => $index,
                    
                    'id' => $id,
                    '_source_excludes' => ['message']
                ];

                $resultData = Elasticsearch::get($params);
                Cache::store('redis')->forever('search_published_annotations_'.$id.'_'.$index, $resultData);

            } catch (NoNodesAvailableException $nodesAvailableException) {
                throw new LaudatioElasticsearchException("Cluster is not available:NoNodesAvailableException: " . $nodesAvailableException->getMessage(), 503, null, true);
            } catch (CouldNotConnectToHost $couldNotConnectToHost) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotConnectToHost: " . $couldNotConnectToHost->getMessage(), 503, null, true);
            }
            catch (CouldNotResolveHostException $couldNotResolveHostException) {
                throw new LaudatioElasticsearchException("Cluster is not available:CouldNotResolveHostException " . $couldNotResolveHostException->getMessage(), 503, null, true);
            }
            catch(Missing404Exception $missing404Exception){
                throw new LaudatioElasticsearchException("Could not find an object with the given identificators " . $missing404Exception->getMessage(), 404, null, true);
            }
        }
        return $resultData;
    }


    /**
     * @param $searchData
     * @return array
     */
    public function searchCorpusIndex($searchData)
    {

        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        if(is_array($searchData)){
            if(count($searchData) > 1){
                $queryBody = $queryBuilder->buildMustQuery($searchData);
            }
            else{
                $queryBody = $queryBuilder->buildSingleMatchQuery($searchData);
            }
        }
        else{
            $queryBody = $queryBuilder->buildMustFilterRangeQuery($searchData->fields,$searchData->range);
        }
        $params = [
            'size' => 10000,
            'index' => 'corpus',
            
            'body' => $queryBody,
            '_source' => ["corpus_title","corpus_publication_publication_date","corpus_documents","annotation_name","corpus_publication_license_description","corpus_id"],
            //'_source_excludes' => ['message']
        ];


        return Elasticsearch::search($params);
    }

    public function documentDateRangeSearch($searchData, $fullTextSearch, $fullTextFields, $filters, $highlights, $from, $size, $hasWildCards) {
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $index = config("laudatio.elasticsearch_document_index","documents");
        $data = array();
        $filtered = false;

        if(count($filters) > 0){
            $filtered = true;
        }
        $obj = null;

        if(!empty($searchData['document_publication_publishing_date']) && empty($searchData['document_publication_publishing_date_to'])){

            $obj = app()->make('stdClass');
            $obj->field = 'document_history_not_before';
            $obj->from = $searchData['document_publication_publishing_date'];
            $obj->to = '';
            array_push($data,$obj);

            $obj = app()->make('stdClass');
            $obj->field = 'document_publication_publishing_date';
            $obj->from = $searchData['document_publication_publishing_date'];
            $obj->to = '';
            array_push($data,$obj);

        }
        else if(empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

            $obj = app()->make('stdClass');
            $obj->field = 'document_history_not_after';
            $obj->from = '';
            $obj->to = $searchData['document_publication_publishing_date_to'];
            array_push($data,$obj);

            $obj = app()->make('stdClass');
            $obj->field = 'document_publication_publishing_date';
            $obj->from = '';
            $obj->to = $searchData['document_publication_publishing_date_to'];
            array_push($data,$obj);
        }
        else if(!empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

            $obj = app()->make('stdClass');
            $obj->field = 'document_history_not_before';
            $obj->from = $searchData['document_publication_publishing_date'];
            $obj->to = '';
            array_push($data,$obj);

            $obj = app()->make('stdClass');
            $obj->field = 'document_history_not_after';
            $obj->from = '';
            $obj->to = $searchData['document_publication_publishing_date_to'];
            array_push($data,$obj);

            $obj = app()->make('stdClass');
            $obj->field = 'document_publication_publishing_date';
            $obj->from = $searchData['document_publication_publishing_date'];
            $obj->to = $searchData['document_publication_publishing_date_to'];
            array_push($data,$obj);
        }

        if($filtered) {
            $queryBody = $queryBuilder->buildHighlightedDateDocumentRangeQuery($data,$fullTextSearch, $fullTextFields, $filters, $highlights,$hasWildCards);
        }
        else{
            $queryBody = $queryBuilder->buildRangeQuery('corpus_publication_publication_date',$searchData['corpus_publication_publication_date'],$searchData['corpusYearTo']);
        }

        if($size < 10000) {
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody
            ];
        }
        else{
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_includes' => 'total'
            ];
        }

        return Elasticsearch::search($params);

    }

    public function documentRangeSearch($searchData, $fullTextSearch, $fullTextFields, $filters, $highlights, $from, $size, $hasWildCards){
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $index = config("laudatio.elasticsearch_document_index","documents");
        $data = array();
        $filtered = false;

        if(count($filters) > 0){
            $filtered = true;
        }
        $obj = null;

        /**
         * document_history_not_before
         * document_history_not_after
         *
         * document_publication_publishing_date (when)
         */

        if((!empty($searchData['document_publication_publishing_date']) || !empty($searchData['document_publication_publishing_date_to'])) &&
            (empty($searchData['document_size_extent']) && empty($searchData['document_size_extent_to']))){

        }
        else if((empty($searchData['document_publication_publishing_date']) && empty($searchData['document_publication_publishing_date_to'])) &&
            (!empty($searchData['document_size_extent']) || !empty($searchData['document_size_extent_to']))){
            if(!empty($searchData['document_size_extent']) && empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['document_size_extent']) && !empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = "";
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['document_size_extent']) && !empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
        }
        else if((!empty($searchData['document_publication_publishing_date']) || !empty($searchData['document_publication_publishing_date_to'])) &&
            (!empty($searchData['document_size_extent']) || !empty($searchData['document_size_extent_to']))) {
            if(!empty($searchData['document_publication_publishing_date']) && empty($searchData['document_publication_publishing_date_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before.string';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after.string';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date.string';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

            }
            else if(empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before.string';
                $obj->from = "";
                $obj->to = $searchData['document_publication_publishing_date_to'];;
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after.string';
                $obj->from = "";
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date.string';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before.string';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = $searchData['document_publication_publishing_date_to'];;
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after.string';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date.string';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

            }


            if(!empty($searchData['document_size_extent']) && empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = "";
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
        }

        if($filtered) {
            $queryBody = $queryBuilder->buildHighlightedDocumentRangeQuery($data,$fullTextSearch, $fullTextFields, $filters, $highlights,$hasWildCards);
        }
        else{
            $queryBody = $queryBuilder->buildRangeQuery('corpus_publication_publication_date',$searchData['corpus_publication_publication_date'],$searchData['corpusYearTo']);
        }

        Log::info("documentRangeSearch: size ".print_r($size,1));

        if($size < 10000) {
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody
            ];
        }
        else{
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_includes' => 'total'
            ];
        }


        return Elasticsearch::search($params);
    }


    public function corpusRangeSearch($searchData, $fullTextSearch, $fullTextFields,  $filters, $highlights,$from, $size,$hasWildCards)
    {
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $index = config("laudatio.elasticsearch_corpus_index","corpora");
        $data = array();

        $filtered = false;

        if(count($filters) > 0){
            $filtered = true;
        }
        $obj = null;

        if(
            (!empty($searchData['corpus_publication_publication_date']) || !empty($searchData['corpusYearTo'])) &&
            (empty($searchData['corpus_size_value']) && empty($searchData['corpusSizeTo']))){

            if(!empty($searchData['corpus_publication_publication_date']) && empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = '';
                array_push($data,$obj);

            }
            else if(empty($searchData['corpus_publication_publication_date']) && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = '';
                $obj->to = $searchData['corpusYearTo'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['corpus_publication_publication_date']) && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = $searchData['corpusYearTo'];
                array_push($data,$obj);

            }

        }
        else if(
            (empty($searchData['corpus_publication_publication_date']) && empty($searchData['corpusYearTo'])) &&
            (!empty($searchData['corpus_size_value']) || !empty($searchData['corpusSizeTo']))){

            if(!empty($searchData['corpus_size_value']) && empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = "";
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);

            }
            else if(!empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);
            }
        }
        else if(
            (!empty($searchData['corpus_publication_publication_date']) || !empty($searchData['corpusYearTo'])) &&
            (!empty($searchData['corpus_size_value']) || !empty($searchData['corpusSizeTo']))){
            if(!empty($searchData['corpus_publication_publication_date']) && empty($searchData['corpusYearTo'])){

                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['corpus_publication_publication_date']) && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = "";
                $obj->to = $searchData['corpusYearTo'];
                array_push($data,$obj);

            }
            else if(!empty($searchData['corpus_publication_publication_date'])  && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date.string';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = $searchData['corpusYearTo'];
                array_push($data,$obj);

            }


            if(!empty($searchData['corpus_size_value']) && empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = "";
                array_push($data,$obj);

            }
            else if(empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = "";
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);

            }
            else if(!empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);

            }
        }

        if($filtered) {
            $queryBody = $queryBuilder->buildHighlightedCorpusRangeQuery($data, $fullTextSearch, $fullTextFields, $filters, $highlights, $hasWildCards);
        }
        else{
            $queryBody = $queryBuilder->buildRangeQuery('corpus_publication_publication_date',$searchData['corpus_publication_publication_date'],$searchData['corpusYearTo']);
        }

        if($size < 10000) {
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_excludes' => ['message']
            ];
        }
        else{
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_includes' => 'total'
            ];
        }

        Log::info("corpusRangeSearch: params: ".print_r($params,1));
        return Elasticsearch::search($params);

    }


    public function corpusMixSearch($searchData, $fullTextQuery, $fulltextFields , $filters, $fields, $from, $size, $hasWildCards)
    {
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $index = config("laudatio.elasticsearch_corpus_index","corpora");
        $data = array();

        $filtered = false;

        if(count($filters) > 0){
            $filtered = true;
        }
        $obj = null;

        if((!empty($searchData['corpus_publication_publication_date']) || !empty($searchData['corpusYearTo'])) &&
            (empty($searchData['corpus_size_value']) && empty($searchData['corpusSizeTo']))){

            if(!empty($searchData['corpus_publication_publication_date']) && empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = '';
                $obj->value = '';
                array_push($data,$obj);

            }
            else if(empty($searchData['corpus_publication_publication_date']) && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = '';
                $obj->to = $searchData['corpusYearTo'];
                $obj->value = '';
                array_push($data,$obj);
            }
            else if(!empty($searchData['corpus_publication_publication_date']) && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = $searchData['corpusYearTo'];
                $obj->value = '';
                array_push($data,$obj);

            }

        }
        else if(
            (empty($searchData['corpus_publication_publication_date']) && empty($searchData['corpusYearTo'])) &&
            (!empty($searchData['corpus_size_value']) || !empty($searchData['corpusSizeTo']))
        ){


            if(!empty($searchData['corpus_size_value']) && empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = "";
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);

            }
            else if(!empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);
            }

        }
        else if((!empty($searchData['corpus_publication_publication_date']) || !empty($searchData['corpusYearTo'])) &&
            (!empty($searchData['corpus_size_value']) || !empty($searchData['corpusSizeTo']))){

            if(!empty($searchData['corpus_publication_publication_date']) && empty($searchData['corpusYearTo'])){

                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = "";
                array_push($data,$obj);

            }
            else if(empty($searchData['corpus_publication_publication_date']) && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = "";
                $obj->to = $searchData['corpusYearTo'];
                array_push($data,$obj);

            }
            else if(!empty($searchData['corpus_publication_publication_date'])  && !empty($searchData['corpusYearTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_publication_publication_date';
                $obj->from = $searchData['corpus_publication_publication_date'];
                $obj->to = $searchData['corpusYearTo'];
                array_push($data,$obj);

            }


            if(!empty($searchData['corpus_size_value']) && empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = "";
                array_push($data,$obj);

            }
            else if(empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = "";
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);

            }
            else if(!empty($searchData['corpus_size_value']) && !empty($searchData['corpusSizeTo'])){
                $obj = app()->make('stdClass');
                $obj->field = 'corpus_size_value';
                $obj->from = $searchData['corpus_size_value'];
                $obj->to = $searchData['corpusSizeTo'];
                array_push($data,$obj);

            }
        }//end if date or size


        if ((
            !empty($searchData['corpus_title']) ||
            !empty($searchData['corpus_publication_publisher']) ||
            !empty($searchData['corpus_merged_languages']) ||
            !empty($searchData['corpus_merged_formats']) ||
            !empty($searchData['corpus_publication_license'])
        )){
            foreach ($searchData as $field => $value) {
                if((!empty($field) && !empty($value)) &&
                    (
                        $field != 'corpus_publication_publication_date' &&
                        $field != 'corpusYearTo' &&
                        $field != 'corpus_size_value' &&
                        $field != 'corpusSizeTo'
                    )
                ) {

                    foreach($value as $valueitem) {
                        $obj = app()->make('stdClass');
                        $obj->field = $field;
                        $obj->from = '';
                        $obj->to = '';
                        $obj->value = $valueitem;
                        array_push($data,$obj);
                    }
                }
            }
        }//end if other fields


        $queryBody = $queryBuilder->buildMixedQuery($data, $fullTextQuery, $fulltextFields, array(
            "is_latest_publication" => true
        ), $fields,$hasWildCards);

        if($size < 10000) {
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_excludes' => ['message']
            ];
        }
        else{
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_includes' => 'total'
            ];
        }

        //Log::info("CORPUSMIXEDSEARCH: ".print_r($params,1));
        return  Elasticsearch::search($params);
    }


    public function documentMixSearch($searchData, $fullTextQuery, $fulltextFields, $filters, $fields, $from, $size, $hasWildCards){
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $index = config("laudatio.elasticsearch_document_index","documents");
        $data = array();
        $filtered = false;

        if(count($filters) > 0){
            $filtered = true;
        }
        $obj = null;

        /**
         * document_history_not_before
         * document_history_not_after
         *
         * document_publication_publishing_date (when)
         */

        if((!empty($searchData['document_publication_publishing_date']) || !empty($searchData['document_publication_publishing_date_to'])) &&
            (empty($searchData['document_size_extent']) && empty($searchData['document_size_extent_to']))){

            if(!empty($searchData['document_publication_publishing_date']) && empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

            }
            else if(empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

            }
        }
        else if((empty($searchData['document_publication_publishing_date']) && empty($searchData['document_publication_publishing_date_to'])) &&
            (!empty($searchData['document_size_extent']) || !empty($searchData['document_size_extent_to']))){

            if(!empty($searchData['document_size_extent']) && empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['document_size_extent']) && !empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = "";
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['document_size_extent']) && !empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
        }
        else if((!empty($searchData['document_publication_publishing_date']) || !empty($searchData['document_publication_publishing_date_to'])) &&
            (!empty($searchData['document_size_extent']) || !empty($searchData['document_size_extent_to']))){

            if(!empty($searchData['document_publication_publishing_date']) && empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);


            }
            else if(empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['document_publication_publishing_date']) && !empty($searchData['document_publication_publishing_date_to'])){

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_before';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = '';
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_history_not_after';
                $obj->from = '';
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

                $obj = app()->make('stdClass');
                $obj->field = 'document_publication_publishing_date';
                $obj->from = $searchData['document_publication_publishing_date'];
                $obj->to = $searchData['document_publication_publishing_date_to'];
                array_push($data,$obj);

            }


            if(!empty($searchData['document_size_extent']) && empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = "";
                array_push($data,$obj);
            }
            else if(empty($searchData['document_size_extent']) && !empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = "";
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
            else if(!empty($searchData['document_size_extent']) && !empty($searchData['document_size_extent_to'])){
                $obj = app()->make('stdClass');
                $obj->field = 'document_size_extent';
                $obj->from = $searchData['document_size_extent'];
                $obj->to = $searchData['document_size_extent_to'];
                array_push($data,$obj);
            }
        }


        if (
            !empty($searchData['document_title']) ||
            !empty($searchData['document_author_forename']) ||
            !empty($searchData['document_author_surname']) ||
            !empty($searchData['document_merged_authors']) ||
            !empty($searchData['document_publication_place']) ||
            !empty($searchData['document_merged_languages'])
        ){
            foreach ($searchData as $field => $value) {
                if((!empty($field) && !empty($value)) &&
                    (
                        $field != 'document_publication_publishing_date' &&
                        $field != 'document_publication_publishing_date_to' &&
                        $field != 'document_size_extent' &&
                        $field != 'document_size_extent_to'
                    )
                ) {
                    foreach($value as $valueitem) {
                        $obj = app()->make('stdClass');
                        $obj->field = $field;
                        $obj->from = '';
                        $obj->to = '';
                        $obj->value = $valueitem;
                        array_push($data,$obj);
                    }
                }
            }
        }//end if other fields


        $queryBody = $queryBuilder->buildDocumentMixedQuery($data, $fullTextQuery, $fulltextFields, array(
            "is_latest_publication" => true
        ), $fields, $hasWildCards);


        if($size < 10000) {
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_excludes' => ['message']
            ];
        }
        else{
            $params = [
                'size' => $size,
                'from' => $from,
                'index' => $index,
                'body' => $queryBody,
                '_source_includes' => 'total'
            ];
        }


        //Log::info("DICUFILTMIX: ".print_r($params,1));
        return  Elasticsearch::search($params);
    }

    /**
     * @param $searchData
     * @return array
     */
    public function searchDocumentIndex($searchData)
    {

        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        if(is_array($searchData)){
            if(count($searchData) > 1){
                $queryBody = $queryBuilder->buildMustQuery($searchData);
            }
            else{
                $queryBody = $queryBuilder->buildSingleMatchQuery($searchData);
            }
        }
        else {
            $queryBody = $queryBuilder->buildMustFilterRangeQuery($searchData->fields,$searchData->range);
        }





        $params = [
            'size' => 10000,
            'index' => 'document',
            
            'body' => $queryBody,
            '_source' => ["document_title","document_publication_publishing_date","document_list_of_annotations_name","in_corpora"],
            //'_source_excludes' => ['message']
        ];

        $results = Elasticsearch::search($params);
        return $results;
    }

    /**
     * @param Request $request
     * @return \Illuminate\Contracts\Routing\ResponseFactory^\Symfony\Component\HttpFoundation\Response
     */
    public function searchDocumentIndexWithParam(Request $request)
    {

        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        if(count($request->searchData) > 1){
            $queryBody = $queryBuilder->buildMustQuery($request->searchData);
        }
        else{
            $queryBody = $queryBuilder->buildSingleMatchQuery($request->searchData);
        }


        $filters = [];
        $params = [
            'size' => 10000,
            'index' => 'document',
            
            'body' => $queryBody,
            '_source' => ["document_title","document_publication_publishing_date","document_list_of_annotations_name","in_corpora"],
        ];

        if(count($request->params) > 0){
            foreach ($request->params as $paramkey => $paramvalues) {
                $filtervalues = [];
                foreach ($paramvalues as $paramvalue){
                    array_push($filtervalues,$paramvalue);
                }
                $params[$paramkey] = $filtervalues;
            }
        }

        $results = Elasticsearch::search($params);

        return $results;
    }


    /**
     * @param $searchData
     * @return array
     */
    public function getSearchTotal($searchData,$index)
    {
        $resultData = array();

        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        foreach($searchData as $queryData){
            $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));
            $params = [
                'size' => 10000,
                'index' => $index,
                
                'body' => $queryBody,
                'filter_path' => ['hits.total']
            ];
            $results = Elasticsearch::search($params);

            $termData = array_values($queryData);
            $resultData[$termData[0]] = $results['hits']['total'];
        }//end foreach queries

        return $resultData;
    }


    public function getCorpusTitlesByDocument($searchData,$documentData) {
        $resultData = array();

        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $counter = 0;
        foreach($searchData as $queryData){
            $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));
            $params = [
                'size' => 10000,
                'index' => 'corpus',
                
                'body' => $queryBody,
                '_source' => ["corpus_title"],
                'filter_path' => ['hits.hits']
            ];
            $results = Elasticsearch::search($params);
            $termData = array_values($documentData);

            if(count($results['hits']['hits']) > 0){
                $resultData[$termData[$counter++]] = $results['hits']['hits'][0]['_source'];
            }
            else{
                $resultData[$counter++] = array();
            }

        }//end foreach queries
        return $resultData;
    }



    public function getCorpusByAnnotation($searchData) {
        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        foreach($searchData as $queryData){
            $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));
            $params = [
                'size' => 10000,
                'index' => 'corpus',
                
                'body' => $queryBody,
                '_source' => ["corpus_title","corpus_publication_publication_date","corpus_documents","annotation_name","corpus_publication_license_description","corpus_publication_publisher","corpus_encoding_project_homepage","corpus_editor_forename","corpus_editor_surname"],
                'filter_path' => ['hits.hits']
            ];

            $results = Elasticsearch::search($params);
        }//end foreach queries
        return $results['hits']['hits'][0];
    }

    public function getDocumentsByAnnotationAndCorpusId($documentList,$corpusId,$index){
        $resultData = array();

        if (Cache::tags(['document_'.$corpusId.'_'.$index])->has("getDocumentsByAnnotationAndCorpusId_".$corpusId."_".$index)) {
            $resultData = Cache::tags(['document_'.$corpusId.'_'.$index])->get("getDocumentsByAnnotationAndCorpusId_".$corpusId."_".$index);
        }
        else {
            $documentList = is_array($documentList)? $documentList: array($documentList);
            foreach($documentList as $documentId) {
                $queryBuilder = new QueryBuilder();
                $queryBody = $queryBuilder->buildMustQuery(array(
                    array(
                        "laudatio_corpus_id" => $corpusId
                    ),
                    array(
                        "_id" => $documentId
                    ),
                ));

                $params = [
                    'index' => $index,
                    
                    'body' => $queryBody,
                    'size'=> 1000,
                    '_source' => ["document_title","document_size_extent","document_publication_publishing_date","document_id","document_publication_place","document_list_of_annotations_id","_id","_index", "has_annotations"]
                ];

                $result = Elasticsearch::search($params);

                if(count($result['hits']['hits']) > 0){
                    $document_indexid =  $result['hits']['hits'][0]['_id'];
                    $result['hits']['hits'][0]['_source']['document_indexid'] = $document_indexid;
                    $result['hits']['hits'][0]['_source']['document_index'] = $result['hits']['hits'][0]['_index'];
                    array_push($resultData,$result['hits']['hits'][0]['_source']);
                    Cache::tags(['document_'.$corpusId.'_'.$index])->forever("getDocumentsByAnnotationAndCorpusId_".$corpusId."_".$index, $resultData);
                }
            }
        }



        return array(
            'error' => false,
            'results' => $resultData
        );
    }

    /**
     * @todo: FOR DELETION
     */
    public function getDocumentsByDocumentId($documentids,$index) {
        $resultData = array();

        if (Cache::tags(['document_'.$index])->has("getDocumentsByDocumentId_".$index)) {
            $resultData = Cache::tags(['document_'.$index])->get("getDocumentsByDocumentId_".$index);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $documentIdParameters = array();
            $qs = "\r";
            $results = null;


            foreach($documentids as $queryData){
                if($queryData != ""){
                    $qs .= '{"query": {"match": {"document_id": "'.$queryData.'"}}, "size": 1000, "_source": ["document_title","document_size_extent","document_publication_publishing_date","document_id","document_history_original_place","document_list_of_annotations_id","_id"]}'."\n";

                }

            }

            $resultset = $this->curlRequest($qs,$index.'/_msearch');
            $results = $resultset['resultdata'];
            // dd($resultset);
            $results = json_decode($results,true);
            if(isset($results['responses'])){
                foreach ($results['responses'] as $result){
                    if(count($result['hits']['hits']) > 0){
                        array_push($resultData,$result['hits']['hits'][0]);
                    }
                }
            }
            Cache::tags(['document_'.$index])->forever("getDocumentsByDocumentId_".$index, $resultData);
        }


        return $resultData;
    }

    public function getSuggestions($searchData)
    {
        $resultdata = array(
            "corpus_suggest" => array(),
            "document_suggest" => array(),
            "annotation_suggest" => array()
        );

        $resultset = $this->curlRequest($searchData, '_msearch');
        $results = $resultset['resultdata'];

        $results = json_decode($results,true);

        if(isset($results['responses'])){
            foreach ($results['responses'] as $response){
                if(isset($response['suggest']['corpus-suggest']) && count($response['suggest']['corpus-suggest'][0]['options']) > 0 ){
                    //Log::info("CORPS: ".print_r($response['suggest']['corpus-suggest'][0],1));
                    foreach ($response['suggest']['corpus-suggest'][0]['options'] as $option){
                        array_push($resultdata['corpus_suggest'],$option);
                    }
                }
                else if(isset($response['suggest']['document-suggest']) && count($response['suggest']['document-suggest'][0]['options']) > 0 ){
                    //Log::info("DOCS: ".print_r($response['suggest']['document-suggest'][0],1));
                    foreach ($response['suggest']['document-suggest'][0]['options'] as $option){
                        array_push($resultdata['document_suggest'],$option);
                    }
                }
                else if(isset($response['suggest']['annotation-suggest']) && count($response['suggest']['annotation-suggest'][0]['options']) > 0 ){
                    //Log::info("ANNOS: ".print_r($response['suggest']['annotation-suggest'][0],1). " CUNT: ".count($response['suggest']['annotation-suggest'][0]['options']));
                    foreach ($response['suggest']['annotation-suggest'][0]['options'] as $option){
                        array_push($resultdata['annotation_suggest'],$option);
                    }
                }
            }
        }
        return $resultdata;
    }

    public function getSuggestions2($searchData)
    {
        $resultdata = array(
            "corpus_suggest" => array(),
            "document_suggest" => array(),
            "annotation_suggest" => array()
        );

        $resultset = $this->curlRequest($searchData, '_msearch');
        $results = $resultset['resultdata'];

        $results = json_decode($results,true);
        if(isset($results['responses'])){
            foreach ($results['responses'] as $result){
                if(isset($result['suggest']['corpus-suggest']) && count($result['suggest']['corpus-suggest'][0]['options']) > 0) {
                    foreach ($result['suggest']['corpus-suggest'][0]['options'] as $corpus_suggest) {
                        array_push($resultdata['corpus_suggest'],$corpus_suggest);
                    }
                }
                else if(isset($result['suggest']['document-suggest']) && count($result['suggest']['document-suggest'][0]['options']) > 0) {
                    foreach ($result['suggest']['document-suggest'][0]['options'] as $document_suggest) {
                        array_push($resultdata['document_suggest'],$document_suggest);
                    }
                }
                else if(isset($result['suggest']['annotation-suggest']) && count($result['suggest']['annotation-suggest'][0]['options']) > 0) {
                    foreach ($result['suggest']['annotation-suggest'][0]['options'] as $annotation_suggest) {
                        array_push($resultdata['annotation_suggest'],$annotation_suggest);
                    }
                }
            }
        }

        return $resultdata;
    }

    public function getDocumentsByAnnotation($searchData,$annotationData){
        $resultData = array();
        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        foreach ($searchData as $id => $annotationDatum) {
            $results = null;
            if(!isset($resultData[$id])){
                $resultData[$id] = array();
            }
            $qs = "\r";

            foreach($annotationDatum as $queryData){
                foreach ($queryData as $key => $value){
                    if($value != ""){
                        $qs .= '{"profile": true,"query": {"match": {"'.$key.'": "'.$value.'"}}, "size": 1000, "_source": ["document_title","document_publication_publishing_date","document_list_of_annotations_name","in_corpora"]}'."\n";

                    }
                }
            }

            $resultset = $this->curlRequest($qs,'document/_msearch');
            $results = $resultset['resultdata'];

            $results = json_decode($results,true);
            if(isset($results['responses'])){
                foreach ($results['responses'] as $result){
                    if(count($result['hits']['hits']) > 0){
                        array_push($resultData,$result['hits']['hits'][0]);
                    }
                }
            }
        }
        return $resultData;
    }

    public function getHitsCount($index,$query){
        $resultData = array();
        $resultData = $this->curlRequest($query,$index.'/_count');
        return $resultData;
    }

    public function getDocumentsByAnnotation_metrics($searchData,$annotationData){
        $resultData = array();
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $totaltime = 0;
        $metrics = array();
        $file = fopen("/Users/rolfguescini/source/phpelasticsearchlaudatio/storage/metrics.csv","w");
        //$file = fopen("/var/www/html/laravelaudatio/shared/storage/metrics.csv","w");
        fputcsv($file,array(
            'id',
            'EStime',
            'curlinfo_url',
            'curlinfo_http_code',
            'curlinfo_size_download',
            'curlinfo_size_upload',
            'curlinfo_starttransfer_time',
            'curlinfo_namelookup_time',
            'curlinfo_connect_time',
            'curlinfo_request',
            'total_time',
            'curlinfo_speed_upload',
            'curlinfo_speed_download',
            'query'


        ));
        foreach ($searchData as $id => $annotationDatum) {
            $results = null;
            if(!isset($resultData[$id])){
                $resultData[$id] = array();
            }
            $qs = "\r";
            $queries = array();
            foreach($annotationDatum as $queryData){
                $queryBody = $queryBuilder->buildSingleMatchQuery(array($queryData));
                array_push($queries,$queryBody);
                foreach ($queryData as $key => $value){
                    if($value != ""){
                        $qs .= '{"profile": true,"query": {"match": {"'.$key.'": "'.$value.'"}}, "size": 1000, "_source": ["document_title","document_publication_publishing_date","document_list_of_annotations_name","in_corpora"]}'."\n";

                    }
                }
            }

            $resultset = $this->curlRequest($qs,'document/_msearch');
            $results = $resultset['resultdata'];



            $results = json_decode($results,true);
            if(isset($results['responses'])){
                foreach ($results['responses'] as $result){
                    if(count($result['hits']['hits']) > 0){
                        array_push($resultData[$id],$result['hits']['hits'][0]);
                    }
                    else{
                        $resultData[$id] = array();
                    }

                    if(count($result['profile']['shards']) > 0){
                        foreach($result['profile']['shards'] as $shard) {
                            foreach($shard['searches'] as $shardsearch) {
                                foreach($shardsearch['query'] as $shardquery) {
                                    if(!array_key_exists($id,$metrics)){
                                        $metrics[$id] = array();
                                    }

                                    $metrics[$id]['id'] = $id;
                                    $totaltime += $shardquery['time'];
                                    $metrics[$id]['time'] = $shardquery['time'];
                                }

                            }
                        }
                    }
                }


                $metrics[$id]['curlinfo_url'] = $resultset['curlinfo']['url'];
                $metrics[$id]['curlinfo_http_code'] = $resultset['curlinfo']['http_code'];
                $metrics[$id]['curlinfo_size_download'] = $resultset['curlinfo']['size_download'];
                $metrics[$id]['curlinfo_size_upload'] = $resultset['curlinfo']['size_upload'];
                $metrics[$id]['curlinfo_starttransfer_time'] = $resultset['curlinfo']['starttransfer_time'];
                $metrics[$id]['curlinfo_namelookup_time'] = $resultset['curlinfo']['namelookup_time'];
                $metrics[$id]['curlinfo_connect_time'] = $resultset['curlinfo']['connect_time'];
                $metrics[$id]['curlinfo_request'] = $resultset['curlinfo']['pretransfer_time'];
                $metrics[$id]['total_time'] = $resultset['curlinfo']['total_time'];
                $metrics[$id]['curlinfo_speed_upload'] = $resultset['curlinfo']['speed_upload'];
                $metrics[$id]['curlinfo_speed_download'] = $resultset['curlinfo']['speed_download'];
            }
        }

        $allhits = array();
        foreach($metrics as $id => $data){
            array_push($data,'{"profile": true,"query": {"match": {"'.$key.'": "'.$value.'"}}, "size": 1000, "_source": ["document_title","document_publication_publishing_date","document_list_of_annotations_name","in_corpora"]}');
            fputcsv($file,$data);
        }
        fclose($file);

        $resultData['metrics'] = $metrics;
        $totaltime = floor($totaltime/60000).':'.floor(($totaltime%60000)/1000).':'.str_pad(floor($totaltime%1000),3,'0', STR_PAD_LEFT);
        $resultData['totaltime'] = $totaltime;


        return $resultData;
    }



    public function getCorporaByAnnotation($searchData,$annotationData,$index){
        $resultData = array();

        if (Cache::tags(['corpus_'.$searchData[0]['_id'].'_'.$index])->has('getCorporaByAnnotation_'.$searchData[0]['_id'].'_'.$annotationData[0].'_'.$index)) {
            $resultData =  Cache::tags(['corpus_'.$searchData[0]['_id'].'_'.$index])->get('getCorporaByAnnotation_'.$searchData[0]['_id'].'_'.$annotationData[0].'_'.$index);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = null;
            $counter = 0;
            $queryBody = $queryBuilder->buildSingleMatchQuery($searchData);
            $params = [
                'size' => 10000,
                'index' => $index,
                
                'body' => $queryBody,
                //'_source_excludes' => ['message'],
                '_source' => ["corpus_title","corpus_publication_publication_date","corpus_documents","annotation_name","corpus_publication_license_description","corpus_publication_publisher","corpus_encoding_project_homepage","corpus_editor_forename","corpus_editor_surname","corpus_version","publication_version","publication_status","corpus_publication_license"],
                'filter_path' => ['hits.hits']
            ];



            $results = Elasticsearch::search($params);
            $termData = array_values($annotationData);
            //dd($termData);
            if(count($results['hits']['hits']) > 0){
                $resultData[$termData[$counter++]] = $results['hits']['hits'];
            }
            else{
                $resultData[$counter++] = array();
            }

            if(count($resultData) > 0) {
                Cache::tags('corpus_'.$searchData[0]['_id'].'_'.$index)->forever('getCorporaByAnnotation_'.$searchData[0]['_id'].'_'.$annotationData[0].'_'.$index, $resultData);
            }
        }//end if cached


        return $resultData;
    }

    /**
     * @param $searchData
     * @return array
     */
    public function searchAnnotationIndex($searchData)
    {
        $queryBuilder = new QueryBuilder();
        $queryBody = null;

        if(count($searchData) > 1){
            $queryBody = $queryBuilder->buildMustQuery($searchData);
        }
        else{
            $queryBody = $queryBuilder->buildSingleMatchQuery($searchData);
        }

        $params = [
            'index' => 'annotation',
            
            'body' => $queryBody,
            //'_source_excludes' => ['message'],
            //'_source' => ["preparation_title", "in_corpora", "in_documents"],
            'size'=> 100
        ];

        $results = Elasticsearch::search($params);
        return $results;
    }

    /**
     * @param $index
     * @return mixed json
     */
    public function truncateIndex($index){
        $queryBuilder = new QueryBuilder();
        $queryBody = null;
        $searchData = array();
        $queryBody = $queryBuilder->buildMatchAllQuery($searchData);

        $params = [
            'index' => $index,
            
            'body' => $queryBody,
        ];

        $results = Elasticsearch::deleteByQuery($params);
        return $results;
    }

    /**
     * @return mixed
     * @todo: fix the function so it only aggregates within a given corpus
     */
    public function getAnnotationGroups($matchdata,$index){
        $results = null;
        $queryBuilder = new QueryBuilder();
        $queryBody = $queryBuilder->buildTermsAggregationQueryByMatchQuery(
            $matchdata,
            array(
                "name" => "annotations",
                "field" => "preparation_encoding_annotation_group.keyword"
            ));
        $params = [
            'index' => $index,

            'body' => $queryBody,
            'size'=> 100,
            'filter_path' => ['aggregations.annotations.buckets.key']
        ];

        return Elasticsearch::search($params);
    }

    public function getGuidelinesByCorpus($corpusId,$index){
        $results = array();
        if (Cache::tags(['guidelines_'.$corpusId])->has("getGuidelinesByCorpus_".$corpusId)) {
            $results = Cache::tags(['guidelines_'.$corpusId])->get("getGuidelinesByCorpus_".$corpusId);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = $queryBuilder->buildSingleMatchQuery(array(
                array(
                    "laudatio_corpus_id" => $corpusId
                )
            ));

            $params = [
                'index' => $index,

                'body' => $queryBody,
                'size'=> 10000,
                //'_source' => ["formats","in_annotations","id","desc"]
            ];

            $results = Elasticsearch::search($params);

            if(array_key_exists('result', $results) && count($results['result']['hits']['hits']) > 0) {
                Cache::tags(['guidelines_'.$corpusId])->forever("getGuidelinesByCorpus_".$corpusId, $results);
            }
        }

        return array(
            'error' => false,
            'result' => $results
        );
    }

    public function getGuidelinesByCorpusAndAnnotationId($corpusId,$annotationName,$index){
        $results = array();
        if (Cache::tags(['guidelines_'.$corpusId.'_'.$index])->has("getGuidelinesByCorpusAndAnnotationId_".$annotationName."_".$corpusId."_".$index)) {
            $results = Cache::tags(['guidelines_'.$corpusId.'_'.$index])->get("getGuidelinesByCorpusAndAnnotationId_".$annotationName."_".$corpusId."_".$index);
        }
        else {
            $queryBuilder = new QueryBuilder();
            $queryBody = $queryBuilder->buildMustQuery(array(
                array(
                    "in_corpora" => $corpusId
                ),
                array(
                    "in_annotations" => $annotationName
                ),
            ));

            $params = [
                'index' => $index,
                
                'body' => $queryBody,
                'size'=> 1000,
                '_source' => ["formats","in_annotations","id","desc"]
            ];

            $results = Elasticsearch::search($params);

            if(array_key_exists('result', $results) && count($results['result']['hits']['hits']) > 0) {
                Cache::tags(['guidelines_'.$corpusId.'_'.$index])->forever("getGuidelinesByCorpusAndAnnotationId_".$annotationName."_".$corpusId."_".$index, $results);
            }
        }


        return array(
            'error' => false,
            'result' => $results
        );
    }

    /**
     * getFormatsByCorpus
     * Fetches all format terms for a given Corpus
     * @param $corpusId
     * @return mixed
     */
    public function getFormatsByCorpus($corpusId,$index){
        //dd($corpusId);
        $results = array();
        if (Cache::tags(['formats_'.$corpusId.'_'.$index])->has("getFormatsByCorpus_".$corpusId."_".$index)){
            $results = Cache::tags(['formats_'.$corpusId.'_'.$index])->get("getFormatsByCorpus_".$corpusId."_".$index);
        }
        else {
            $queryBuilder = new QueryBuilder();

            $queryBody = $queryBuilder->buildTermsAggregationQueryByMatchQuery(
                array(
                    "field" => "in_corpora",
                    "value" => $corpusId
                ),
                array(
                    "name" => "formats",
                    "field" => "formats"
                )
            );

            $params = [
                'index' => $index,
                
                'body' => $queryBody,
                'size'=> 100,
                'filter_path' => ['aggregations.formats.buckets.key']
            ];

            $results = Elasticsearch::search($params);
            if(array_key_exists('result', $results) && count($results['result']['hits']['hits']) > 0) {
                Cache::tags(['formats_'.$corpusId.'_'.$index])->forever("getFormatsByCorpus_".$corpusId."_".$index, $results);
            }

        }

        return $results;
    }

    /**
     * Helpers
     */

    public function checkForKey($array, $key){
        $value = "";
        foreach($array as $item) {
            if(array_key_exists($key,$item)) {
                $value = $item[$key];
                break;
            }
        }
        return $value;
    }

    public function removeKey($array, $key){
        if(is_array($array)){
            for($i=0; $i < count($array); $i++){
                foreach($array[$i] as $itemkey => $param){
                    if($itemkey == $key){
                        unset($array[$i][$key]);
                        break;
                    }
                }
            }
        }
        else{
            foreach($array as $itemkey => $item){
                if($itemkey == $key){
                    unset($array->{$key});
                    break;
                }
            }
        }

        return $array;
    }

    public function curlRequest($queries,$path,$method = "POST"){
        $header = array(
            //"content-type: application/x-ndjson; charset=UTF-8"
            "content-type: application/json; charset=utf-8"
        );

        $url = $this->ELASTICSEARCH_SCHEME.'://'.$this->ELASTICSEARCH_USER.':'. $this->ELASTICSEARCH_PASS.'@'.$this->ELASTICSEARCH_HOST.':'.$this->ELASTICSEARCH_PORT.'/'.$path;

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl,CURLOPT_HTTPHEADER, $header);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        if($method == "POST") {
            curl_setopt($curl, CURLOPT_POSTFIELDS, $queries);
        }

        $results = curl_exec($curl);
        $info = curl_getinfo($curl);
        curl_close($curl);
        //Log::info("CURLINFO: ".print_r($info,1));
        //Log::info("CURLRES: ".print_r($results,1));
        $resultset = array(
            "curlinfo" => $info,
            "resultdata" => $results
        );
        return $resultset;
    }

    public function fixElasticsearchIds($corpusId,$corpusIndex,$documentIndex,$annotationIndex){
        $old_id_collection = array();
        $new_id_collection = array();
        $corpus = Corpus::findOrFail($corpusId);

        $document_elasticsearch_ids = array();
        $annotation_elasticsearch_ids = array();
        $queryBuilder = new QueryBuilder();
        $corpus_elasticsearch_id = null;

        if(null == $corpus->elasticsearch_id){
            $corpusparam = array(
                array(
                    "laudatio_corpus_id" => $corpusId
                )
            );

            $queryBody = $queryBuilder->buildMustQuery($corpusparam);

            $corpusparams = [
                'size' => 10,
                'index' => $corpusIndex,

                'body' => $queryBody,
                '_source' => ["_id"]
            ];

            $corpusresponse = Elasticsearch::search($corpusparams);

            $corpus->elasticsearch_id = $corpusresponse['hits']['hits'][0]['_id'];
            $corpus->save();
            $corpus_elasticsearch_ids[$corpus->id] = $corpusresponse['hits']['hits'][0]['_id'];
        }
        else{
            $corpus_elasticsearch_id = $corpus->elasticsearch_id;
        }


        foreach ($corpus->documents()->get() as $document){
            if(null == $document->elasticsearch_id){
                $documentparam = array(
                    array(
                        "laudatio_corpus_id" => $corpusId
                    ),
                    array(
                        "document_id" => $document->document_id
                    ),
                );

                $queryBody = $queryBuilder->buildMustQuery($documentparam);

                $documentparams = [
                    'size' => 10,
                    'index' => $documentIndex,

                    'body' => $queryBody,
                    '_source' => ["_id"]
                ];

                $documentresponse = Elasticsearch::search($documentparams);
                $document->elasticsearch_id = $documentresponse['hits']['hits'][0]['_id'];
                $document->save();
                $document_elasticsearch_ids[$document->document_id] = $documentresponse['hits']['hits'][0]['_id'];
            }
            else{
                $document_elasticsearch_ids[$document->document_id] = $document->elasticsearch_id;
            }

        }

        foreach ($corpus->annotations()->get() as $annotation){
            if(null == $annotation->elasticsearch_id){
                $annotationparam = array(
                    array(
                        "laudatio_corpus_id" => $corpusId
                    ),
                    array(
                        "preparation_annotation_id.keyword" => $annotation->annotation_id
                    ),
                );

                $queryBody = $queryBuilder->buildMustQuery($annotationparam);

                $annotationparams = [
                    'size' => 10,
                    'index' => $annotationIndex,
                    'body' => $queryBody,
                    '_source' => ["_id"]
                ];
                $annotationresponse = Elasticsearch::search($annotationparams);

                $annotation->elasticsearch_id = $annotationresponse['hits']['hits'][0]['_id'];
                $annotation->save();
                $annotation_elasticsearch_ids[$annotation->annotation_id] = $annotationresponse['hits']['hits'][0]['_id'];
            }
            else{
                $annotation_elasticsearch_ids[$annotation->annotation_id] = $annotation->elasticsearch_id;
            }
        }
        $elasticsearchIds['corpus'] = $corpus_elasticsearch_id;
        $elasticsearchIds['documents'] = $document_elasticsearch_ids;
        $elasticsearchIds['annotations'] = $annotation_elasticsearch_ids;

        return $elasticsearchIds;
    }

}