<?php
/**
 * Created by PhpStorm.
 * User: rolfguescini
 * Date: 25.04.17
 * Time: 15:33
 */

namespace App\Laudatio\GitLab;
use App\Exceptions\CorpusRepositoryFileSystemException;
use Symfony\Component\Process\Process;
use Symfony\Component\Process\Exception\ProcessFailedException;
use Symfony\Component\Process\Exception\RuntimeException;
use App\Exceptions\CorpusNameAlreadyExistsException;
use App\Exceptions\GitNetworkException;
use Response;
use App\Exceptions\GitInitException;
use DB;
use Log;

class GitFunction
{
    protected $repoId;
    protected $basePath;
    protected $scriptPath;
    protected $xPath;

    /**
     * GitFunction constructor.
     */
    public function __construct()
    {
        $this->repoId = config('laudatio.repoid');
        $this->basePath = config('laudatio.basePath');
        $this->scriptPath = config('laudatio.scriptPath');
        $this->xPath = config('laudatio.xPath');
    }

    /**
     *
     * getStatus: get git status
     *
     * @param $path
     * @return string
     */
    public function getStatus($path){
        $process = new Process("git status",$path);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }

        return $process->getOutput();
    }

    /**
     *
     * doAdd: do git add
     *
     * @param $path
     * @return string
     */
    public function doAdd($path){
        if(is_dir($path)){
            $process = new Process("git add .",$path);
        }
        else{
            $pathWithOutAFile = substr($path,0,strrpos($path,"/"));
            $file = substr($path,strrpos($path,"/")+1);
            $cwdPath = $this->basePath."/".$pathWithOutAFile;
            $process = new Process("git add $file",$cwdPath);
        }

        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }

        return $process->getOutput();
    }


    /**
     *
     * addGitHooks: add the githooks folder
     *
     * @param $path
     * @param $user
     * @param $email
     * @return bool
     */
    public function addGitHooks($path, $user, $email) {
        $isAdded = false;

        $process = new Process("git add githooks",$this->basePath."/".$path);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            $isAdded = true;
        }

        return $isAdded;
    }


    /**
     *
     * doAddFile: git add specific file
     *
     * @param $file
     * @param $path
     * @return string
     */
    public function doAddFile($file,$path){
        $process = new Process("git add \"$file\"",$path);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            Log::info("doAddFile: process not werking ".print_r($process->getErrorOutput(),1));
            Log::info("doAddFile: process not werking ".print_r($process->getExitCodeText(),1));
            throw new ProcessFailedException($process);
        }

        return $process->getOutput();
    }


    /**
     *
     * isUntracked: check if there are untracked files present
     *
     * @param $status
     * @return bool
     */
    public function isUntracked($status){
        return (strpos($status,"untracked files present") !== false || strpos($status,"Untracked files:") !== false);
    }


    /**
     *
     * isTracked: check if there are tracked files present
     *
     * @param $path
     * @return bool
     */
    public function isTracked($path){
        $isTracked = false;
        $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
        $folder = substr($path,strrpos($path,"/")+1);
        if(file_exists($pathWithOutAddedFolder) && file_exists($path)) {
            $process = new Process("git ls-files --exclude-standard \"$folder\"",$pathWithOutAddedFolder);
            $process->run();

            // executes after the command finishes
            if (!$process->isSuccessful()) {
                throw new ProcessFailedException($process);
            }


            if($process->getOutput() == ""){
                $isTracked = true;
            }
        }

        return $isTracked;
    }


    /**
     *
     * hasDiff: git diff
     *
     * @param $path
     * @return string|null
     */
    public function hasDiff($path){
        $hasDiff = null;

        $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
        $folder = substr($path,strrpos($path,"/")+1);

        $process = new Process("git diff --name-only",$pathWithOutAddedFolder);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }

        if($process->getOutput() != ""){
            $hasDiff = $process->getOutput();
        }

        return $hasDiff;
    }


    /**
     *
     * fileHasDiff: git check if specific file has diff
     *
     * @param $path
     * @param $file
     * @return bool
     */
    public function fileHasDiff($path, $file){
        $fileHasDiff = false;

        //$pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
        $folder = substr($path,strrpos($path,"/")+1);

        $process = new Process("git diff \"$file\"",$path);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }

        if($process->getOutput() != ""){
            $fileHasDiff = true;
        }

        return $fileHasDiff;
    }

    /**
     *
     * fileExists
     *
     * @param $path
     * @return bool
     */
    public function fileExists($path){
        return file_exists($this->basePath."/".$path);
    }

    /**
     *
     * isAdded: check if there are git added files
     *
     * @param $status
     * @return bool
     */
    public function isAdded($status){
        return (strpos($status,"Changes to be committed") !== false || strpos($status,"\"use git add\" and/or \"git commit -a\"") !== false);
    }


    /**
     *
     * isModified: git check if there modified, not staged files present
     *
     * @param $status
     * @return bool
     */
    public function isModified($status){
        return (strpos($status,"modified:") !== false && strpos($status, 'Changes not staged for commit') !== false);
    }


    /**
     *
     * isModifiedAndAdded: git check if there modified,  staged files present
     *
     * @param $status
     * @return bool
     */
    public function isModifiedAndAdded($status){
        return (strpos($status,"Changes to be committed") !== false && strpos($status,"modified:") !== false);
    }


    /**
     *
     * isCommitted: git check if something is committed
     *
     * @param $commitmessage
     * @param $logmessage
     * @return bool
     */
    public function isCommitted($commitmessage, $logmessage){
        return (strpos($logmessage,$commitmessage) !== false);
    }


    /**
     *
     * isReadyForPush: git check if there are elements to be pushed
     *
     * @param $status
     * @return bool
     */
    public function isReadyForPush($status){
        return (strpos($status,"use \"git push\" to publish your local commits") !== false);
    }


    /**
     *
     * isCleanWorkingTree: git check if all is committed and pushed
     *
     * @param $status
     * @return bool
     */
    public function isCleanWorkingTree($status){
        return(strpos($status,"working tree clean") !== false);
    }


    /**
     *
     * getLogMessage: git fetch commit log by number
     *
     * @param $nummer
     * @param $path
     * @return string
     */
    public function getLogMessage($nummer,$path){
        $process = new Process("git log -".$nummer." --pretty=%B",$path);
        $process->run();
        $logMessage = "";

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else {
            $logMessage = $process->getOutput();
        }
        return $logMessage;
    }


    /**
     *
     * addUntracked: git add untracked files
     *
     * @param $pathWithOutAddedFolder
     * @param string $folder
     * @return bool
     */
    public function addUntracked($pathWithOutAddedFolder, $folder = ""){
        $isAdded = false;
        $status = $this->getStatus($this->basePath."/".$pathWithOutAddedFolder);
        if($folder == ""){
            if($this->isUntracked($status)){
                $addResult = $this->doAdd($this->basePath."/".$pathWithOutAddedFolder);
                $addStatus = $this->getStatus($this->basePath."/".$pathWithOutAddedFolder);
                if($this->isAdded($addStatus)){
                    $isAdded = true;
                }
            }
        }
        else {
            if($this->isUntracked($status)){
                if (is_dir($folder)) {
                    $addResult = $this->doAdd($this->basePath."/".$pathWithOutAddedFolder."/".$folder);
                }
                else{
                    Log::info("GITTFUNK: doAddfile: ".$this->basePath."/".$pathWithOutAddedFolder. " FILE_ ".$folder);
                    $addResult = $this->doAddFile($folder,$this->basePath."/".$pathWithOutAddedFolder);
                }

                $addStatus = $this->getStatus($this->basePath."/".$pathWithOutAddedFolder);
                if($this->isAdded($addStatus)){
                    $isAdded = true;
                }
            }
        }

        return $isAdded;
    }


    /**
     *
     * addModified: git add modified files
     *
     * @param $pathWithOutAddedFolder
     * @param string $folder
     * @return bool
     */
    public function addModified($pathWithOutAddedFolder, $folder = ""){
        $isAdded = false;
        $status = $this->getStatus($this->basePath."/".$pathWithOutAddedFolder);

        if($folder == ""){
            /**
             * @todo
             */
        }
        else {
            $this->doAddFile($folder, $this->basePath."/".$pathWithOutAddedFolder);
            $addStatus = $this->getStatus($this->basePath."/".$pathWithOutAddedFolder);
            $isAdded = $this->isModifiedAndAdded($addStatus);
        }

        return $isAdded;
    }


    /**
     *
     * commitFiles: git commit files
     *
     * @param $path
     * @param $commitmessage
     * @param $user
     * @param $email
     * @return bool
     */
    public function commitFiles($path, $commitmessage, $user, $email){
        $isCommitted = false;

        $commitString = $commitmessage." by ".$user." (".$email.")";
        $commitString = escapeshellarg($commitString);

        //$process = new Process("git commit -m \"".$commitmessage." by ".$user." (".$email.") \" ",$path);
        $process = new Process("git commit -m ".$commitString,$path);
        $process->setTimeout(3600);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            if($this->isCommitted($commitmessage,$processOutput)){
                $isCommitted = true;
            }
        }

        return $isCommitted;
    }


    /**
     *
     * commitFile: git commit specific file
     *
     * @param $path
     * @param $file
     * @param $commitmessage
     * @param $user
     * @param $email
     * @return bool
     */
    public function commitFile($path,$file, $commitmessage, $user, $email){
        $isCommitted = false;
        Log::info("commitFile: data path: ".print_r($path,1)." FILE: ".print_r($file,1));
        /*
         * $process = new Process("git commit -m ".escapeshellarg($shellarg),$path);
         */

        $commitString = $commitmessage." ".$file. " by ".$user." (".$email.")";
        $commitString = escapeshellarg($commitString);
        Log::info("commitFile: commitstring: ".print_r($commitString,1));

        if(file_exists($path)) {
            //$process = new Process("git commit -m \"".$commitmessage." '".$file. "' by ".$user." (".$email.")\" -o"." \"$file\" ",$path);
            $process = new Process("git commit -m ".$commitString." -o ".escapeshellarg($file),$path);
            $process->setTimeout(3600);


            //$process->run();


            try {
                $process->mustRun();
                Log::info("commitFile: process not werking ".print_r($process->getErrorOutput(),1));
                Log::info("commitFile: process not werking ".print_r($process->getExitCodeText(),1));
                $processOutput = $process->getOutput();
                $elasticsearchId = $process->getExitCode();

                if($this->isCommitted($commitmessage,$processOutput)){
                    $isCommitted = true;
                }

            } catch (ProcessFailedException $exception) {
                echo $exception->getMessage();
            }

            // executes after the command finishes
            /*
            if (!$process->isSuccessful()) {
                //Log::info("gitFunction: commitFile: error: ".print_r($process->getErrorOutput(),1));
                throw new ProcessFailedException($process);
            }
            else{
                $processOutput = $process->getOutput();
                //Log::info("gitFunction: commitFile: success: ".print_r($process->getgetOutput(),1));
                if($this->isCommitted($commitmessage,$processOutput)){
                    $isCommitted = true;
                }
            }
            */

        }


        return $isCommitted;
    }

    public function getLastELasticsearchIndex($path, $tag) {
        $tagProcess = new Process("git show -s --format=%B ".$tag, $path);
        $tagProcess->setTimeout(3600);
        $indexes = array();

        try {
            $tagProcess->run();
            $tagOutput = $tagProcess->getOutput();
            $re = '/\{(?:[^{}]|(?R))*\}/';
            $tagJson = preg_match_all($re, $tagOutput, $matches, PREG_SET_ORDER, 0);

            $tagArray = json_decode($matches[0][0],true);

            if(isset($tagArray['lastId'])) {
                $indexes['lastIndexedId'] = array();
                foreach($tagArray['lastId'] as $item) {
                    foreach ($item as $filePath => $eId) {
                        $indexes['lastIndexedId'][$filePath] = $eId;
                    }
                }
            }

            if(isset($tagArray['corpusElasticId'])){
                $indexes['corpusElasticId'] = $tagArray['corpusElasticId'];
            }

        } catch (ProcessFailedException $tagexception) {
            echo $tagexception->getMessage();
        }
        return $indexes;
    }

    public function resetJSONIdArray($path, $tag) {
        $isReset = false;
        $pythonScript = $this->basePath."/".$path.'/.git/src/resetJSON.py';

        //try {
            $process = new Process("python3 ".$pythonScript." -i \"lastId\" -t  ".$tag." -p ".$this->basePath."/".$path);
            $process->run();

            if (!$process->isSuccessful()) {
                //  throw new ProcessFailedException($process);
            }
            else{
                $result = $process->getOutput();
                $re = '/\{(?:[^{}]|(?R))*\}/';
                $tagJson = preg_match_all($re, $result, $matches, PREG_SET_ORDER, 0);
                $result_decoded = json_decode($matches[0][0],true);
                if(isset($result_decoded->error)){
                    throw new CorpusRepositoryFileSystemException($result_decoded->error,404,null);
                }
                else{
                    if( !isset($result_decoded['lastId']) || count($result_decoded['lastId']) == 0) {
                        $isReset = true;
                    }
                }

            }

       // } catch (ProcessFailedException $tagexception) {
         //   echo $tagexception->getMessage();
        //}

        return $isReset;
    }


    /**
     *
     * setCorpusVersionTag: git set version tag
     *
     * @param $corpusPath
     * @param $tagmessage
     * @param $version
     * @param $user
     * @param $email
     * @param $blame
     * @return bool
     */
    public function setCorpusVersionTag($corpusPath, $tagmessage, $version, $user,$email, $blame) {
        $isTagged = false;
        $process = null;
        if($blame) {
            $process = new Process(" git tag -a v".$version." -m \"".$tagmessage." by ".$user." (".$email.") \" ",$this->basePath."/".$corpusPath);
        }
        else {
            $process = new Process(" git tag -a v".$version." -m \"".$tagmessage."\"",$this->basePath."/".$corpusPath);
        }

        $process->setTimeout(3600);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();

            if($this->isTagged($version,$this->basePath."/".$corpusPath)){

                $push_process = new Process(" git push origin v".$version,$this->basePath."/".$corpusPath);
                $push_process->setTimeout(3600);
                $push_process->run();

                if (!$push_process->isSuccessful()) {
                    throw new ProcessFailedException($push_process);
                }
                else{
                    $isTagged = true;
                }
            }
        }
        return $isTagged;
    }


    /**
     *
     * isTagged: git check if tagged by given version
     *
     * @param $version
     * @param $corpusPath
     * @return bool
     */
    public function isTagged($version,$corpusPath) {
        $isTagged = false;

        $process = new Process(" git tag",$corpusPath);
        $process->setTimeout(3600);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            if(strpos($processOutput, strval($version)   ) !== false){
                $isTagged = true;
            }
        }
        return $isTagged;
    }


    /**
     *
     * addRemote: git add remote repository
     *
     * @param $origin
     * @param $path
     * @return bool
     */
    public function addRemote($origin,$path) {
        $addedRemote = false;
        $shouldAdd = false;

        $askForOriginProcess = new Process("git remote -v",$this->basePath."/".$path);
        $askForOriginProcess->setTimeout(3600);
        $askForOriginProcess->run();


        if (!$askForOriginProcess->isSuccessful()) {
            throw new ProcessFailedException($askForOriginProcess);
        }
        else{
            $askForOriginProcessOutput = $askForOriginProcess->getOutput();
            if(empty($askForOriginProcessOutput)){
                $shouldAdd = true;
            }
        }



        if($shouldAdd){
            //Log::info("trying to: git remote add origin ".$origin." => ".$this->basePath."/".$path);
            $process = new Process("git remote add origin ".$origin,$this->basePath."/".$path);
            $process->setTimeout(3600);
            $process->run();
        }
        else{
            //Log::info("trying to: git remote set-url origin ".$origin." => ".$this->basePath."/".$path);
            $process = new Process("git remote set-url origin ".$origin,$this->basePath."/".$path);
            $process->setTimeout(3600);
            $process->run();
        }

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            $addedRemote = true;
        }
        return $addedRemote;
    }


    /**
     *
     * initialPush: git push initial corpusstructure
     *
     * @param $path
     * @param $user
     * @return bool
     */
    public function initialPush($path,$user){
        //git push --set-upstream origin master
        $isPushed = false;
        $process = new Process("git push --set-upstream origin master",$this->basePath."/".$path);
        $process->setTimeout(3600);
        $process->run();

        // executes after the command finishes
        /*
        if (!$process->isSuccessful()) {
            Log::info("initialPush:PROCESS ERROR: : ".print_r($process->getErrorOutput(),1));
            throw new ProcessFailedException($process);
        }
        else{
            Log::info("initialPush:PROCESS: : ".print_r($process->getOutput(),1));
            $processOutput = $process->getOutput();
            $isPushed = true;
        }
*/
        if (!$process->isSuccessful()) {
            $gitNetworkException = new GitNetworkException($process->getErrorOutput(),0,null);
            $responsearray = array("exception" => $gitNetworkException);
            $isPushed = $responsearray;
        }
        else{
            $processOutput = $process->getOutput();
            $isPushed = true;
        }

        return $isPushed;
    }


    /**
     *
     * resetAdd: git reset added files from the staging
     *
     * @param $path
     * @param $files
     * @return bool
     */
    public function resetAdd($path,$files){
        $isReset = false;
        foreach($files as $file) {
            //git reset -- main/*
            $process = null;
            if(is_dir($file)){
                $process = new Process("git reset -- ".$file."/*",$this->basePath."/".$path);
            }
            else{
                $process = new Process("git reset -- $file",$this->basePath."/".$path);
            }

            $process->setTimeout(3600);
            $process->run();

            // executes after the command finishes
            if (!$process->isSuccessful()) {
                throw new ProcessFailedException($process);
            }
            else{
                $processOutput = $process->getOutput();
                $isReset = true;
            }
        }
        return $isReset;
    }


    /**
     *
     * pushFiles: git push files
     *
     * @param $path
     * @param $corpusid
     * @return bool
     */
    public function pushFiles($path,$corpusid) {
        $isPushed = false;
        $process = new Process("git push",$this->basePath."/".$path);
        $process->setTimeout(3600);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            $gitNetworkException = new GitNetworkException($process->getErrorOutput(),0,null);
            $responsearray = array("exception" => $gitNetworkException);
            $isPushed = $responsearray;
        }
        else{
            $processOutput = $process->getOutput();
            $isPushed = true;
        }

        return $isPushed;
    }


    public function pushFiles2($path,$corpusid) {
        $isPushed = false;
        $process = new Process("git push",$this->basePath."/".$path);
        $process->setTimeout(3600);
        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            $isPushed = true;
        }
        return $isPushed;
    }


    /**
     *
     * addFileUpdate: git add modified files
     *
     * @param $path
     * @param $file
     * @return bool
     */
    public function addFileUpdate($path, $file){
        $isAdded = false;

        if($this->fileHasDiff($this->basePath."/".$path,$file)){
            $addStatus = $this->getStatus($this->basePath."/".$path);
            $modifiedStatus = $this->isModified($addStatus);
            if($modifiedStatus == 1){
                $isAdded = $this->addModified($path,$file);
            }
        }

        return $isAdded;
    }


    /**
     *
     * initiateRepository: git init
     *
     * @param $path
     * @return bool
     */
    public function initiateRepository($path){

        $isInitiated = false;
        $process = null;
        try {
            $process = new Process("git init",$this->basePath."/".$path);
            $process->setTimeout(3600);
            $process->run();

            $processOutput = $process->getOutput();
            $isInitiated = true;
        }
        catch (RuntimeException $runtimeException){
            throw new GitInitException($runtimeException->getMessage(),0,null);
        }
        catch(ProcessFailedException $processFailedException){
            throw new GitInitException($processFailedException->getMessage(),0,null);
        }

        return $isInitiated;
    }

    /**
     * 
     * @param String $path
     * @throws GitInitException
     * @return mixed
     */
    public function fetchUnPublishedGitArchive($path){
        
        $isArchived = "";
        $isFormat = false;
        $process = null;
        $formatPath = "";
        if(strpos($path,"CORPUS-DATA") !== false) {
            $formatPathArray = explode("/", $path);
            
            $fileName = $formatPathArray[3];
            $formatPath = $formatPathArray[0]."/".$formatPathArray[1]."/".$formatPathArray[2];
            $isFormat = true;
        }
        else{
            $corpusPath = substr($path,strpos($path,"/")+1);
            $corpusPathArray = explode("_",$corpusPath);
            $fileName = $corpusPathArray[0];
        }
        
        
        $storage_path = public_path();
        try {
            $process = null;
            
            //git archive -o mercurius-baumbank.zip --prefix=-mercurius-baumbank-version-1-0  v1.0 TEI-HEADERS
            if($isFormat){
                $process = new Process("git archive -o ".$storage_path."/".$fileName.".zip --prefix=".$fileName,$this->basePath."/".$formatPath);
            }
            else{
                $process = new Process("git archive -o ".$storage_path."/".$fileName.".zip --prefix=".$fileName." HEAD TEI-HEADER",$this->basePath."/".$path);
            }
            
            $process->setTimeout(3600);
            $process->run();
            
            $processOutput = $process->getOutput();
            $isArchived = $fileName;
            
        }
        catch (RuntimeException $runtimeException){
            throw new GitInitException($runtimeException->getMessage(),0,null);
        }
        catch(ProcessFailedException $processFailedException){
            throw new GitInitException($processFailedException->getMessage(),0,null);
        }
        
        return $isArchived;
    }

    public function fetchGitArchive($path,$version){

        $isArchived = "";
        $process = null;
        $corpusPath = substr($path,strpos($path,"/")+1);
        $corpusPathArray = explode("_",$corpusPath);
        $fileName = $corpusPathArray[0];

        $storage_path = public_path();
        try {
            //git archive -o mercurius-baumbank.zip --prefix=-mercurius-baumbank-version-1-0  v1.0 TEI-HEADERS
            $process = new Process("git archive -o ".$storage_path."/".$fileName.".zip --prefix=".$fileName."-v".$version." v".$version." TEI-HEADERS",$this->basePath."/".$path);
            $process->setTimeout(3600);
            $process->run();

            $processOutput = $process->getOutput();
            $isArchived = $fileName;
        }
        catch (RuntimeException $runtimeException){
            throw new GitInitException($runtimeException->getMessage(),0,null);
        }
        catch(ProcessFailedException $processFailedException){
            throw new GitInitException($processFailedException->getMessage(),0,null);
        }

        return $isArchived;
    }


    /**
     *
     * copyGitHooks: copy the githooks folder to the corpus
     *
     * @param $path
     * @return bool
     */
    public function copyGitHooks($path){
        $isCopied = false;
        /*
         * cp ../../../scripts/githooks/* .git
         */
        $hookProcess = new Process("cp ".$this->scriptPath."/githooks/pre-commit githooks",$this->basePath."/".$path);
        $hookProcess->run();
        // executes after the command finishes
        if (!$hookProcess->isSuccessful()) {
            throw new ProcessFailedException($hookProcess);
        }
        else{
            $processOutput = $hookProcess->getOutput();
            $isCopied = true;
        }

        return $isCopied;
    }


    /**
     *
     * setGitConfig
     *
     * @param $path
     * @param $configs
     * @return array
     */
    public function setGitConfig($path, $configs) {
        $listOfErrors = array();
        foreach ($configs as $config) {
            $process = new Process("git config ".$config,$this->basePath."/".$path);
            $process->run();
            // executes after the command finishes
            if (!$process->isSuccessful()) {
                throw new ProcessFailedException($process);
            }
            else{
                array_push($listOfErrors, $process->getOutput());

            }
        }
        return $listOfErrors;
    }


    /**
     *
     * setCoreHooksPath: set path of the githooks folder in git config
     *
     * @param $path
     * @return bool
     */
    public function setCoreHooksPath($path){
        $isSet = false;
        /*
         * cp ../../../scripts/githooks/* .git
         */
        $hookProcess = new Process('printf "\t hooksPath = githooks\n" >> .git/config',$this->basePath."/".$path);
        $hookProcess->run();
        // executes after the command finishes
        if (!$hookProcess->isSuccessful()) {
            throw new ProcessFailedException($hookProcess->getErrorOutput());
        }
        else{
            $processOutput = $hookProcess->getOutput();
            $isSet = true;
        }

        return $isSet;
    }


    public static function __callStatic($name, $arguments)
    {
        // TODO: Implement __callStatic() method.
    }


    /**
     *
     * copyScripts: copy the pipeline python scripts to the corpus
     *
     * @param $path
     * @return bool
     */
    public function copyScripts($path){
        $isCopied = false;

        $makeDirectoryProcess = new Process("mkdir .git/src",$this->basePath."/".$path);
        $makeDirectoryProcess->run();
        if (!$makeDirectoryProcess->isSuccessful()) {
            throw new ProcessFailedException($makeDirectoryProcess);
        }
        else{
            $scriptProcess = new Process("cp ".$this->scriptPath."/src/* .git/src",$this->basePath."/".$path);
            $scriptProcess->run();
            // executes after the command finishes
            if (!$scriptProcess->isSuccessful()) {
                throw new ProcessFailedException($scriptProcess);
            }
            else{
                $processOutput = $scriptProcess->getOutput();
                $isCopied = true;
            }
        }

        return $isCopied;
    }

    public function copyPipelineConfigurations($path){
        $isCopied = false;

        $makeDirectoryProcess = new Process("mkdir .git/xpath",$this->basePath."/".$path);
        $makeDirectoryProcess->run();
        if (!$makeDirectoryProcess->isSuccessful()) {
            throw new ProcessFailedException($makeDirectoryProcess);
        }
        else{
            Log::info("TRYING: ".$this->xPath."/*");
            $scriptProcess = new Process("cp ".$this->xPath."/* .git/xpath",$this->basePath."/".$path);
            $scriptProcess->run();
            // executes after the command finishes
            if (!$scriptProcess->isSuccessful()) {
                throw new ProcessFailedException($scriptProcess);
            }
            else{
                $processOutput = $scriptProcess->getOutput();
                $isCopied = true;
            }
        }

        return $isCopied;
    }

    /**
     *
     * writeFiles: write uploaded files to the corpus
     *
     * @param $dirPath
     * @param $fileDataArray
     * @param $flySystem
     * @param $fileTempPath
     * @param $theFilePath
     * @return array
     */
    public function writeFiles($dirPath, $fileDataArray,$flySystem,$fileTempPath,$theFilePath){
        $createdDirectoryPath = array();

        if(file_exists($dirPath)){
            throw new CorpusRepositoryFileSystemException("The filepath $dirPath already exists",0,null);
        }


        foreach($fileDataArray as $fileData) {
            $pathsArray = explode("/", $fileData);
            $fileInDirectory = array_pop($pathsArray);

            if($this->isDottedFile($fileInDirectory) === false){
                foreach ($pathsArray as $path) {
                    $created = "";
                    if (count($createdDirectoryPath) ==  0) {
                        $singlePath = $dirPath."/".$path;

                        if(!file_exists($this->basePath."/".$singlePath)){
                            $created = $this->makeDirectory($dirPath, $path);
                        }
                        else{
                            array_push($createdDirectoryPath,$path);
                        }
                    } else {
                        $combinedPath = $dirPath."/".implode("/", $createdDirectoryPath);
                        if(!file_exists($this->basePath."/".$combinedPath."/".$path)){
                            $created = $this->makeDirectory($combinedPath, $path);
                        }
                        else{
                            array_push($createdDirectoryPath,$path);
                        }
                    }


                    if ($created != "") {
                        array_push($createdDirectoryPath,$created);
                    }
                }
            }
        }

        $stream = null;
        $fileInDirectory = "";
        $writePath = "";

        if(strpos($theFilePath,'images') !== false) {
            $writePath = $theFilePath;
        }
        else if(strpos($theFilePath,'TEI-HEADERS/corpus') !== false) {
            //write file if exists
            $fileArray = explode("/", $theFilePath);

            if(count($fileArray) == 1){
                $fileInDirectory = array_pop($fileArray);
            }
            $writePath = $dirPath."/corpusHeader.xml";
        }
        else{
            //write file if exists
            $fileArray = explode("/", $theFilePath);

            if(count($fileArray) == 1){
                $fileInDirectory = array_pop($fileArray);
            }

            $writePath = $dirPath."/".$fileInDirectory;
        }

        $existingFile = $flySystem->has($dirPath."/".$fileInDirectory);

        if(!$existingFile){
            $stream = fopen($fileTempPath, 'r+');
            $flySystem->writeStream($writePath, $stream);
        }
        else{
            $stream = fopen($fileTempPath, 'r+');
            $flySystem->updateStream($writePath, $stream);
        }
        array_push($createdDirectoryPath,$this->basePath."/".$writePath);

        if (is_resource($stream)) {
            fclose($stream);
        }

        return $createdDirectoryPath;
    }


    /**
     *
     * makeDirectory: unix mkdir
     *
     * @param $path
     * @param $directory
     * @return string
     */
    public function makeDirectory($path,$directory){
        $createdDirectoryPath = "";
        $makeDirectoryProcess = new Process("mkdir $directory",$this->basePath."/".$path);
        $makeDirectoryProcess->run();
        if (!$makeDirectoryProcess->isSuccessful()) {
            throw new ProcessFailedException($makeDirectoryProcess);
        }
        else{
            $createdDirectoryPath = $directory;
        }
        return $createdDirectoryPath;
    }


    /**
     *
     * renameFile: unix mv
     *
     * @param $path
     * @param $oldname
     * @param $newname
     * @return string
     */
    public function renameFile($path,$oldname,$newname){
        $renamedPath = "";
        $makeDirectoryProcess = new Process("mv $oldname $newname",$this->basePath."/".$path);
        $makeDirectoryProcess->run();
        if (!$makeDirectoryProcess->isSuccessful()) {
            throw new ProcessFailedException($makeDirectoryProcess);
        }
        else{
            $renamedPath = $path."/".$newname;
        }
        return $renamedPath;
    }


    /**
     *
     * deleteFiles
     *
     * @param $path
     * @param $user
     * @param $email
     * @return bool
     */
    public function deleteFormatFiles($path,$user,$email){

        $isdeleted = false;
        $isFile = false;
        $pathWithOutAddedFolder = "";
        $folder = "";
        $cwdPath = "";


        if(!is_dir($this->basePath.'/'.$path)){
            $isFile = true;
        }

        if(strpos($path,"/") !== false){
            if($isFile) {
                $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
                $folder = substr($path,strrpos($path,"/")+1);
                $cwdPath = $this->basePath."/".$pathWithOutAddedFolder;
            }
            else{
                $folder = substr($path,strrpos($path,"/")+1);
                $cwdPath = $this->basePath."/".$pathWithOutAddedFolder;
            }

        }
        else{
            $cwdPath = $this->basePath;
            $folder = $path;
        }


        $process = null;
        $folder = str_replace(" ","\\ ",$folder);

        if($isFile){
            $process = new Process("git rm $folder",$cwdPath);
        }
        else{
            $process = new Process("git rm -rf *",$cwdPath);
        }


        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $addArray = explode("/",$cwdPath);
            $delete = array_pop($addArray);
            $addPath = join("/",$addArray);

            $addStatus = $this->getStatus($addPath);

            if($this->deletedIsAdded($addStatus)){

                $commit = $this->commitFiles($addPath,"deleting $folder",$user,$email);
                $commitstatus = $this->getStatus($addPath);

                if($this->isCleanWorkingTree($commitstatus)){
                    clearstatcache(TRUE);
                    $isdeleted = true;
                }
            }
        }

        return $isdeleted;
    }

    public function deletedIsAdded($status){
        return (strpos($status,"deleted:") !== false || strpos($status,"use \"git reset HEAD <file>...\" to unstage") !== false);
    }

    public function deleteFiles($path,$user,$email){

        $isdeleted = false;
        $isFile = false;
        $pathWithOutAddedFolder = "";
        $folder = "";
        $cwdPath = "";


        if(!is_dir($this->basePath.'/'.$path)){
            $isFile = true;
        }

        if(strpos($path,"/") !== false){
            if($isFile) {
                $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
                $folder = substr($path,strrpos($path,"/")+1);
                $cwdPath = $this->basePath."/".$pathWithOutAddedFolder;
            }
            else{
                $folder = substr($path,strrpos($path,"/")+1);
                $cwdPath = $this->basePath."/".$pathWithOutAddedFolder;
            }

        }
        else{
            $cwdPath = $this->basePath;
            $folder = $path;
        }


        $process = null;
        $folder = str_replace(" ","\\ ",$folder);

        if($isFile){
            $process = new Process("git rm $folder",$cwdPath);
        }
        else{
            $process = new Process("git rm -rf *",$cwdPath);
        }


        $process->run();

        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{

            $addStatus = $this->getStatus($cwdPath);
            if($this->isAdded($addStatus)){
                $commit = $this->commitFiles($cwdPath,"deleting $folder",$user,$email);
                $commitstatus = $this->getStatus($cwdPath);

                if($this->isCleanWorkingTree($commitstatus)){
                    clearstatcache(TRUE);
                    $isdeleted = true;
                }
            }
        }

        return $isdeleted;
    }


    public function deleteCorpusFiles($path,$user,$email){

        $isdeleted = false;

        $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
        $folder = substr($path,strrpos($path,"/")+1);
        $cwdPath = $this->basePath."/".$path;


        $process = new Process("git rm -rf *",$cwdPath);
        $process->run();


        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{

            $addStatus = $this->getStatus($this->basePath."/".$path);
            if($this->isAdded($addStatus)){
                $commit = $this->commitFiles($this->basePath."/".$path,"deleting $folder",$user,$email);
                $commitstatus = $this->getStatus($this->basePath."/".$path);

                if($this->isCleanWorkingTree($commitstatus)){
                    clearstatcache(TRUE);

                    $process2 = new Process("rm -rf $folder",$this->basePath."/".$pathWithOutAddedFolder);
                    $process2->run();
                    if (!$process2->isSuccessful()) {
                        throw new ProcessFailedException($process2);
                    }
                    else {
                        $isdeleted = true;
                    }
                }


                }
        }

        return $isdeleted;
    }


    /**
     *
     * deleteUntrackedDataFiles: delete corpus data files that are not git staged
     *
     * @param $path
     * @return bool
     */
    public function deleteUntrackedDataFiles($path){
        $isFile = false;
        $isDeleted = false;
        if(!is_dir($this->basePath.'/'.$path)){
            $isFile = true;
        }


        if(strpos($path,"/") !== false){
            $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
            $folder = substr($path,strrpos($path,"/")+1);
            $cwdPath = $this->basePath."/".$pathWithOutAddedFolder;
        }
        else{
            $cwdPath = $this->basePath;
            $folder = $path;
        }

        $process = null;
        if($isFile){
            $process = new Process("rm $folder",$cwdPath);
        }
        else{
            $process = new Process("rm -rf *",$cwdPath);
        }

        $process->run();


        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $isDeleted = true;
        }

        return $isDeleted;
    }


    /**
     *
     * deleteUntrackedFiles: delete corpusheader files not git staged
     *
     * @param $path
     * @param bool $isProject
     * @param bool $isCorpus
     * @return bool
     */
    public function deleteUntrackedFiles($path,$isProject = false,$isCorpus = false){

        $isdeleted = false;
        $isFile = false;
        $pathWithOutAddedFolder = "";
        $folder = "";
        $cwdPath = "";

        if(!is_dir($this->basePath.'/'.$path)){
            $isFile = true;
        }

        if($isFile){
            if(strpos($path,"/") !== false){
                $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
                $folder = substr($path,strrpos($path,"/")+1);
                $cwdPath = $this->basePath."/".$pathWithOutAddedFolder;
            }
            else{
                $cwdPath = $this->basePath;
                $folder = $path;
            }

            $process = null;
            $folder = str_replace(" ","\\ ",$folder);
            $process = new Process("rm -rf $folder",$cwdPath);
            $process->run();


            // executes after the command finishes
            if (!$process->isSuccessful()) {
                throw new ProcessFailedException($process);
            }
            else{
                $isdeleted = true;
            }
        }
        else{
            //we are deleting contents of a folder
            $dirArray = explode("/",$path);

            if(!$isCorpus && !$isProject){
                $type = $dirArray[3];
                $objects = null;
                switch ($type) {
                    case 'corpus':
                        $objects = DB::table('corpuses')->where('directory_path',$dirArray[1])->get();
                        break;
                    case 'document':
                        $objects = DB::table('documents')->where('directory_path',$dirArray[1])->get();
                        break;
                    case 'annotation':
                        $objects = DB::table('annotations')->where('directory_path',$dirArray[1])->get();
                        break;
                }
                foreach ($objects->toArray() as $object){
                    if($object->file_name){
                        $process = new Process("rm -rf $object->file_name",$this->basePath."/".$path);
                        $process->run();
                    }
                }
            }
            else{
                $pathForDeletion = "";
                if($isProject){
                    $process = new Process("rm -rf $dirArray[0]",$this->basePath);
                    $process->run();
                }
                else if($isCorpus){
                    $pathForDeletion = $this->basePath."/".$dirArray[0];
                    $process = new Process("rm -rf *",$pathForDeletion);
                    $process->run();
                }


            }


            $isdeleted = true;
        }

        return $isdeleted;
    }


    /**
     *
     * getCommitData: git show
     *
     * @param $path
     * @return array
     */
    public function getCommitData($path){
        $process = new Process("git show -s",$this->basePath.'/'.$path);
        $process->run();

        // executes after the command finishes
        $commitData = array();
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();

            //parse the output
            $stringArray = explode("\n",$processOutput);

            $sha_string_arr = explode(" ",$stringArray[0]);
            $sha_string = trim($sha_string_arr[1]);
            $commitData["sha_string"] = $sha_string;

            $author_string_arr = explode(":",$stringArray[1]);
            $author = trim($sha_string_arr[1]);
            $commitData["author"] = $author;

            $date_string_arr = explode(":",$stringArray[2]);
            $date = $date_string_arr[1];
            $commitData["date"] = $date;

            $message = trim($stringArray[4]);
            $commitData["message"] = $message;

        }
        return $commitData;
    }

    /**
     *
     * isDottedFile: determine if we have a .dotfile
     *
     * @param $file
     * @return bool
     */
    public function isDottedFile($file){
        return strpos($file,".") == 0;
    }


    /**
     *
     * getListOfStagedFiles: git status --porcelain
     *
     * @param $path
     * @return array
     */

    public function getListOfStagedFiles($path){
        $listOfFiles = array();
        $process = new Process("git diff --name-only --cached",$path);
        $process->run();


        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            $listOfFiles = explode("\n", $processOutput);
        }
        return array_filter($listOfFiles);
    }

    public function getListOfStagedFiles2($path){
        $listOfFiles = array();
        $process = new Process("git status --porcelain | sed s/^...//",$path);
        $process->run();


        // executes after the command finishes
        if (!$process->isSuccessful()) {
            throw new ProcessFailedException($process);
        }
        else{
            $processOutput = $process->getOutput();
            $listOfFiles = explode("\n", $processOutput);
        }
        return array_filter($listOfFiles);
    }


    /**
     *
     * checkForCorpusFiles: execute validateHeaders.py to check if the corpus has all files present
     *
     * @param $path
     * @return string|null
     */
    public function checkForCorpusFiles($path){
        $result = null;
        $pythonScript = $this->scriptPath.'/src/validateHeaders.py';

        $process = new Process("python3 ".$pythonScript." -p ".$path);
        $process->run();


        // executes after the command finishes

        if (!$process->isSuccessful()) {
            //  throw new ProcessFailedException($process);
        }
        else{
            $result = $process->getOutput();
            $result_decoded = json_decode($result);
            if(isset($result_decoded->error)){

                throw new CorpusRepositoryFileSystemException($result_decoded->error,404,null);
            }

        }

        return $result;
    }

}