<?php
/**
 * Created by PhpStorm.
 * User: rolfguescini
 * Date: 03.08.17
 * Time: 14:31
 */

namespace App\Laudatio\GitLab;

use App\Custom\GitRepoInterface;
use App\Laudatio\GitLab\GitFunction;
use App\Custom\GitLabInterface;
use App\Custom\LaudatioUtilsInterface;
use Illuminate\Http\Request;
use GrahamCampbell\Flysystem\FlysystemManager;
use League\Flysystem\Cached\CachedAdapter;
use League\Flysystem\Cached\Storage\Adapter;
use Carbon\Carbon;
use Log;
use App\Corpus;
use App\Document;
use App\Annotation;
use App\CorpusFile;
use App\User;
use App\CorpusProject;
use App\Exceptions\CorpusRepositoryFileSystemException;
use App\Exceptions\GitAddException;
use App\Exceptions\GitCommitException;
use App\Exceptions\GitPushException;
use App\Exceptions\GitDeleteException;
use App\Exceptions\GitInitException;
use App\Exceptions\FileUploadException;
use Symfony\Component\Process\Exception\ProcessFailedException;
use Cache;

/**
 * Class GitRepoService
 * @package App\Laudatio\GitLab
 */
class GitRepoService implements GitRepoInterface
{
    protected $basePath;
    protected $scriptPath;
    protected $laudatioUtilsService;
    protected $GitLabService;

    public function __construct(LaudatioUtilsInterface $laudatioUtilsService, GitLabInterface $GitLabService)
    {
        $this->basePath = config('laudatio.basePath');
        $this->scriptPath = config('laudatio.scriptPath');
        $this->laudatioUtilsService = $laudatioUtilsService;
        $this->GitLabService = $GitLabService;
    }

    /**
     *
     * createProjectFileStructure to contain GIT Corpus Repositories
     *
     * @param $flysystem
     * @param $projectName
     * @return false|mixed|string|string[]|null
     */
    public function createProjectFileStructure($flysystem,$projectName){
        $path = $this->normalizeString($projectName);
        if(!$flysystem->has($path)){
            $createResponse = $flysystem->createDir($path);
        }

        return $path;
    }


    /**
     *
     * createFormatDirectory : create the CORPUS-DATA folder
     * within the Corpus Repository file structure
     *
     * @param $flysystem
     * @param $corpusProjectPath
     * @param $corpusPath
     * @param $formatName
     * @return |null
     */
    public function createFormatDirectory($flysystem,$corpusProjectPath,$corpusPath,$formatName) {
        $formatPath = $this->normalizeString($formatName);
        $dirPath = $corpusProjectPath.'/'.$corpusPath."/CORPUS-DATA/".$formatPath;
        $realPath = realpath($this->basePath.'/'.$corpusProjectPath.'/'.$corpusPath."/CORPUS-DATA/".$formatPath);
        $createResponse = null;

        if(!$realPath){
            $createResponse = $flysystem->createDir($realPath);
        }
        else{
            throw new CorpusRepositoryFileSystemException("The format directory $formatPath already exists",0,null);
        }
        return $createResponse;
    }


    /**
     *
     * createCorpusFileStructure: create the Corpus GIT repository file structure
     *
     * @param $flysystem
     * @param $corpusProjectPath
     * @param $corpusName
     * @param $user
     * @return false|mixed|string|string[]|null
     */

    public function createCorpusFileStructure($flysystem,$corpusProjectPath,$corpusName,$user){
        $corpusPath = $this->normalizeString($corpusName);
        $dirPath = $corpusProjectPath.'/'.$corpusPath;

        if(!$flysystem->has($dirPath)){
            $flysystem->createDir($dirPath);
            $flysystem->createDir($dirPath."/githooks");
            $flysystem->createDir($dirPath."/.config");
            $flysystem->write($dirPath."/.config/laudatio.env","elasticuser=".config("laudatio.elasticsearch_user","elastic")."\nelasticpass=".config("laudatio.elasticsearch_pass","elastic"));
            $flysystem->createDir($dirPath."/TEI-HEADERS");
            $flysystem->createDir($dirPath."/TEI-HEADERS/corpus");
            $flysystem->write($dirPath."/TEI-HEADERS/corpus/README.md","# CORPUS HEADERS #\n This folder contains Corpus header meta data");
            $flysystem->createDir($dirPath."/TEI-HEADERS/document");
            $flysystem->write($dirPath."/TEI-HEADERS/document/README.md","# DOCUMENT HEADERS #\n This folder contains Document header meta data");
            $flysystem->createDir($dirPath."/TEI-HEADERS/annotation");
            $flysystem->write($dirPath."/TEI-HEADERS/annotation/README.md","# ANNOTATION HEADERS #\n This folder contains Annotation header meta data");
            $flysystem->createDir($dirPath."/CORPUS-DATA");
            $flysystem->write($dirPath."/CORPUS-DATA/README.md","# CORPUS DATA #\n This folder contains the Corpus Data itself");
            $flysystem->createDir($dirPath."/CORPUS-IMAGES");
            $flysystem->write($dirPath."/CORPUS-IMAGES/README.md","# CORPUS IMAGES #\n This folder contains any images associated with the corpus");
            $flysystem->write($dirPath."/.gitignore",".config/*\n.DS_Store\ndesktop.ini\n");

            $initiated = false;
            try {
                $initiated = $this->initiateRepository($dirPath);
            }
            catch (GitInitException $initException) {
                throw new GitInitException($initException->getMessage(),0,null);
            }


            if($initiated){
                $this->setGitConfig($dirPath,
                    array(
                        "core.precomposeUnicode false",
                        "core.quotepath false"
                    ));
                $this->copyGitHooks($dirPath);
                $this->setCoreHooksPath($dirPath);
                $this->copyScripts($dirPath);
                $this->copyPipelineConfigurations($dirPath);
            }

        }

        return $corpusPath;
    }



    /**
     *
     * hasCorpusFileStructure
     *
     * @param $flysystem
     * @param $corpusProjectPath
     * @param $corpusPath
     * @return mixed
     */
    public function hasCorpusFileStructure($flysystem , $corpusProjectPath, $corpusPath){
        return $flysystem->has($this->basePath."/".$corpusProjectPath."/".$corpusPath);
    }


    /**
     * checkForCorpusFiles
     * 
     * @param $path
     * @return string|null
     */
    public function checkForCorpusFiles($path) {
        $gitFunction = new GitFunction();
        $fullPath = $this->basePath.'/'.$path;
        return $gitFunction->checkForCorpusFiles($fullPath);
    }


    /**
     * 
     * {@inheritDoc}
     * @see \App\Custom\GitRepoInterface::archiveFiles()
     */
    public function archiveFiles($path,$version) {
        $gitFunction = new GitFunction();
        return $gitFunction->fetchGitArchive($path,$version);
    }
    
    /**
     * 
     * @param String $path
     * @return mixed
     */
    public function fetchUnPublishedArchive($path) {
        $gitFunction = new GitFunction();
        return $gitFunction->fetchUnPublishedGitArchive($path);
        
    }

    /**
     *
     * updateCorpusFileStructure
     * 
     * @param $flysystem
     * @param $corpusProjectPath
     * @param $oldCorpusPath
     * @param $corpusName
     * @return string
     */
    public function updateCorpusFileStructure($flysystem,$corpusProjectPath,$oldCorpusPath,$corpusName){
        $corpusPath = "";
        $normalizedCorpusName = $this->normalizeString($corpusName);
        $oldDirPath = $corpusProjectPath.'/'.$oldCorpusPath;
            if(!file_exists($this->basePath."/".$corpusProjectPath."/".$normalizedCorpusName)){
                if($flysystem->has($oldDirPath)){
                    $gitFunction = new GitFunction();
                    $corpusPath = $gitFunction->renameFile($corpusProjectPath,$oldCorpusPath,$normalizedCorpusName);
                }
                else {
                    throw new CorpusRepositoryFileSystemException("The Corpus with path ".$oldDirPath." could not be updated.",0,null);
                }
            }
            else{
                throw new CorpusRepositoryFileSystemException("The Corpus with path ".$corpusProjectPath."/".$normalizedCorpusName." and the same version is already taken within this Corpus project. Please create another project for the Corpus.",0,null);
            }


        return $corpusPath;
    }

    /**
     * commitStagedFiles
     * 
     * @param $corpusPath
     * @param $corpusId
     */
    public function commitStagedFiles($corpusPath,$corpusId) {
        $this->addFilesToRepository($corpusPath,"TEI-HEADERS");
        $gitFunction = new GitFunction();
        $stagedFiles = $gitFunction->getListOfStagedFiles($this->basePath."/".$corpusPath);
        $this->commitFilesToRepository($this->basePath.'/'.$corpusPath,"Created initial corpus file structure");
        foreach ($stagedFiles as $stagedFile){
            $dirArray = explode("/",trim($stagedFile));
            dd($dirArray);
            if($dirArray[0] != "CORPUS-DATA"){
                if(count($dirArray) > 1){
                    $fileName = $dirArray[2];
                    $this->laudatioUtilsService->setVersionMapping($fileName,$dirArray[1],false,$corpusId);
                }
            }
        }
    }

    /**
     * Delete a corpus file structure on the disk
     * @param $flysystem
     * @param $path
     * @return bool|null
     */
    public function deleteCorpusFileStructure($flysystem, $path,$username,$useremail){
        $deleted = false;
        $gitfunction = new GitFunction();

        $trackedResult = $gitfunction->deleteCorpusFiles($path,$username,$useremail);

        if(!$trackedResult){
            $deleted = $this->deleteUntrackedFile($flysystem,$path,false,true);
        }
        else{
            $deleted = $trackedResult;
        }
        return $deleted;
    }

    public function deleteUntrackedCorpusFileStructure($flysystem, $path){
        $deleted = false;

        if($flysystem->has($path)) {
            $response = $flysystem->deleteDir($path);
            if($response) {
                $deleted = true;
            }
        }
        return $deleted;
    }


    /**
     *
     * deleteProjectFileStructure
     *
     * @param $flysystem
     * @param $path
     * @return bool|null
     */
    public function deleteProjectFileStructure($flysystem, $path){
        $deleted = false;
        $trackedResult = $this->deleteFile($flysystem,$path);

        if(!$trackedResult){
            $deleted = $this->deleteUntrackedFile($flysystem,$path,true,false);
        }
        else{
            $deleted = $trackedResult;
        }
        return $deleted;
    }

    /**
     *
     * getCorpusFiles
     *
     * @param $flysystem
     * @param $corpusid
     * @param string $path
     * @return array|null
     */
    public function getCorpusFiles($flysystem,$corpusid,$path = ""){

        $fileData = null;
        if (Cache::has("getCorpusFiles_".$corpusid.$path)) {
            $fileData = Cache::get("getCorpusFiles_".$corpusid.$path);
        }
        else {
            $corpus = Corpus::find($corpusid);

            $corpusProjects = $corpus->corpusprojects()->get();
            $corpusProject_directory_path = '';

            $pathArray = explode("/",$path);
            end($pathArray);
            $last_id=key($pathArray);

            if(count($corpusProjects) == 1) {
                $corpusProject_directory_path = $corpusProjects->first()->directory_path;
            }


            $corpusPath = "";
            $corpus_directory_path = $corpus->directory_path;
            if($path == ""){
                $corpusPath = $corpusProject_directory_path."/".$corpus_directory_path;
            }
            else{
                $corpusPath = $path;
            }

            $folder = "";
            $folderType = "";
            $user_roles = array();

            $corpusBasePath = $pathArray[0]."/".$pathArray[1];
            if(strpos($corpusPath,"CORPUS-DATA") !== false && strpos($corpusPath,"TEI-HEADERS") === false){
                $corpusData = $this->getCorpusDataFiles($flysystem,$corpusPath,$corpusid);
                $headerData = $this->getCorpusFileInfo($flysystem,$corpusBasePath.'/TEI-HEADERS',$corpusid);
                $folderType = "CORPUS-DATA";

            }
            else if(strpos($corpusPath,"TEI-HEADERS") !== false && strpos($corpusPath,"CORPUS-DATA") === false){
                $corpusData = $this->getCorpusDataFiles($flysystem,$corpusBasePath.'/CORPUS-DATA',$corpusid);
                $headerData = $this->getCorpusFileInfo($flysystem, $corpusPath,$corpusid);
                $folderType = "TEI-HEADERS";
            }
            else{
                $corpusData = $this->getCorpusDataFiles($flysystem,$corpusPath.'/CORPUS-DATA',$corpusid);
                $headerData = $this->getCorpusFileInfo($flysystem,$corpusPath.'/TEI-HEADERS',$corpusid);
            }


            $corpusDataFolder = substr($corpusData['path'],strrpos($corpusData['path'],"/")+1);
            $headerDataFolder = substr($headerData['path'],strrpos($headerData['path'],"/")+1);
            $fileData = array(
                "corpusData" => $corpusData,
                "corpusDataFolder" => $corpusDataFolder,
                "headerData" => $headerData,
                "headerDataFolder" => $headerDataFolder,
                "folderType" => $folderType
            );

            Cache::forever("getCorpusFiles_".$corpusid.$path, $fileData);
        }

        return $fileData;
    }


    /**
     *
     * getUploader
     * 
     * @param $headerData
     * @param $headertype
     * @return mixed
     */
    public function getUploader($headerData,$headertype){
        for($i = 0; $i < count($headerData); $i++){
            $object = null;
            if(isset($headerData[$i]['headerObject'])){
                $object = $headerData[$i]['headerObject'];
                $uploader = User::find($object->uid);
                if($uploader){
                    $headerData[$i]['uploader_affiliation'] = $uploader->affiliation;
                    $headerData[$i]['uploader_name'] = $uploader->name;
                    $headerData[$i]['uploader_uid'] = $uploader->id;
                }
            }
            else{
                $headerData[$i]['uploader_affiliation'] = "NA";
                $headerData[$i]['uploader_name'] = "NA";
                $headerData[$i]['uploader_uid'] = "NA";
            }

        }

        return $headerData;

    }

    public function getFormatFileUploader($headerData){
        $newHeaderData = array();
        foreach($headerData as $formatName => $fileData) {
            if(!array_key_exists($formatName,$newHeaderData)) {
                $newHeaderData[$formatName] = array();
            }

            for($i = 0; $i < count($fileData[0]); $i++){
                $headerObject = null;
                if(array_key_exists('headerObject',$fileData[0][$i])) {
                    $headerObject = $fileData[0][$i]['headerObject'];
                }

                if(isset($headerObject) && $headerObject->uid != ""){
                    $uploader = User::find($headerObject->uid);
                    $fileData[0][$i]['uploader_affiliation'] = $uploader->affiliation;
                    $fileData[0][$i]['uploader_name'] = $uploader->name;
                    $fileData[0][$i]['uploader_uid'] = $uploader->id;
                }
                else{
                    $fileData[0][$i]['uploader_affiliation'] = "NA";
                    $fileData[0][$i]['uploader_name'] = "NA";
                    $fileData[0][$i]['uploader_uid'] = "NA";
                }
                array_push($newHeaderData[$formatName],$fileData[0][$i]);
            }
        }


        return $newHeaderData;

    }

    /**
     *
     * getCorpusFileInfo
     *
     * @param $flysystem
     * @param string $path
     * @param $corpusId
     * @return array|null
     */
    public function getCorpusFileInfo($flysystem, $path = "",$corpusId){
        $fileInfo = null;
        if (Cache::has("getCorpusFileInfo_".$corpusId.$path)) {
            $fileInfo = Cache::get("getCorpusFileInfo_".$corpusId.$path);
        }
        else {
            $gitFunction = new GitFunction();
            $hasDir = false;
            $projects = array();

            if($path == ""){
                $projects = $flysystem->listContents();
            }
            else{
                $projects = $flysystem->listContents($path);
            }


            $pathArray = explode("/",$path);
            end($pathArray);
            $last_id = key($pathArray);
            $elements = array();

            for ($i = 0; $i < count($projects);$i++){

                if(isset($projects[$i]['extension']) && $projects[$i]['extension'] !== "md") {
                    if(file_exists($this->basePath."/".$projects[$i]['path'])) {
                        $foldercount = count($flysystem->listContents($projects[$i]['path']));
                        $projects[$i]['foldercount'] = $foldercount;

                        if($gitFunction->isTracked($this->basePath."/".$projects[$i]['path'])){
                            $projects[$i]['tracked'] = "true";
                        }
                        else{
                            $projects[$i]['tracked'] = "false";
                        }


                        if(isset($projects[$i]['basename'])) {
                            $headerObject = $this->laudatioUtilsService->getModelByFileName($projects[$i]['basename'],$pathArray[$last_id],false,$corpusId);
                        }

                        if($headerObject && count($headerObject) > 0){
                            $projects[$i]['headerObject'] = $headerObject[0];
                        }

                        if(isset($projects[$i]['timestamp'])) {
                            $projects[$i]['lastupdated'] = Carbon::createFromTimestamp($projects[$i]['timestamp'])->toDateTimeString();
                        }


                        if($projects[$i]["type"] == "dir"){
                            $hasDir = true;
                        }


                        if(isset($projects[$i]['path'])){
                            $projects[$i]['filesize'] = $this->calculateFileSize(filesize($this->basePath."/".$projects[$i]['path']));
                            $hasDiff = $gitFunction->hasDiff($this->basePath."/".$projects[$i]['path']);
                        }



                        if($hasDiff){
                            $projects[$i]['hasDiff'] = "true";
                        }
                        else{
                            $projects[$i]['hasDiff'] = "false";
                        }

                        if($projects[$i]['hasDiff'] == "true"){
                            $projects[$i]['diffFiles'] = array();
                            array_push($projects[$i]['diffFiles'],$hasDiff);
                        }

                        array_push($elements,$projects[$i]);
                    }

                }

            }

            $patharray = explode("/",$path);
            $count = count($patharray);
            $projects = $this->filterDottedFiles($projects);

            $previouspath = "";
            $previouspath = substr($path,0,strrpos($path,"/"));

            $fileInfo =  array(
                "elements" => $elements,
                "headertype" => $pathArray[$last_id],
                "hasdir" => $hasDir,
                "pathcount" => $count,
                "path" => $path,
                "previouspath" => $previouspath,
            );
            Cache::forever("getCorpusFileInfo_".$corpusId.$path, $fileInfo);
        }

        return $fileInfo;
    }


    /**
     *
     * getCorpusDataFiles
     *
     *
     * @param $flysystem
     * @param string $path
     * @param $corpusId
     * @return array|null
     */

    public function getCorpusDataFilesTester($flysystem,$path = "",$corpusId){
        $gitFunction = new GitFunction();
        $projectArray = array();
        $projectDirs = $flysystem->listContents($path);
        $hasDir = false;
        $count = 0;
        for ($i = 0; $i < count($projectDirs);$i++){
            if($projectDirs[$i]['type'] == "dir") {
                if(!array_key_exists($projectDirs[$i]['basename'],$projectArray)){
                    $projectArray[$projectDirs[$i]['basename']] = array();
                }
                $projects = $flysystem->listContents($projectDirs[$i]['path']);
                for ($j = 0; $j < count($projects);$j++){
                    $foldercount = count($flysystem->listContents($projects[$j]['path']));
                    $projects[$j]['foldercount'] = $foldercount;

                    if($gitFunction->isTracked($this->basePath."/".$projects[$j]['path'])){
                        $projects[$j]['tracked'] = "true";
                    }
                    else{
                        $projects[$j]['tracked'] = "false";
                    }

                    $headerObject = $this->laudatioUtilsService->getModelByFileName($projects[$j]['basename'],'CORPUS-DATA',false,$corpusId);
                    if(count($headerObject) > 0){
                        $projects[$j]['headerObject'] = $headerObject[0];
                    }


                    $projects[$j]['lastupdated'] = "";
                    if(isset($projects[$j]['timestamp'])){
                        $projects[$j]['lastupdated'] = Carbon::createFromTimestamp($projects[$j]['timestamp'])->toDateTimeString();
                    }

                    //Carbon::parse($projects[$j]['timestamp'])->toDateTimeString()->format('H:i, M d');
                    if($projects[$j]["type"] == "dir"){
                        $hasDir = true;
                    }

                    $projects[$j]['filesize'] = "";
                    if($this->basePath."/".$projects[$j]['path'] != ""){
                        $projects[$j]['filesize'] = $this->calculateFileSize(filesize($this->basePath."/".$projects[$j]['path']));
                    }

                    $hasDiff = $gitFunction->hasDiff($this->basePath."/".$projects[$j]['path']);


                    if($hasDiff){
                        $projects[$j]['hasDiff'] = "true";
                    }
                    else{
                        $projects[$j]['hasDiff'] = "false";
                    }

                    if($projects[$j]['hasDiff'] == "true"){
                        $projects[$j]['diffFiles'] = array();
                        array_push($projects[$j]['diffFiles'],$hasDiff);
                    }
                    $count++;
                }//end for files
                $projects = $this->filterDottedFiles($projects);
                $projects = $this->filterMDFiles($projects);
                array_push($projectArray[$projectDirs[$i]['basename']],$projects);
                //Log::info("PROJECTS: ".print_r($projectArray,1));
            }//end if dir
        }//end for dirs
        $patharray = explode("/",$path);
        $previouspath = substr($path,0,strrpos($path,"/"));

        $fileData =  array(
            "projects" => $projectArray,
            "hasdir" => $hasDir,
            "filecount" => $count,
            "path" => $path,
            "previouspath" => $previouspath,
        );
        return $fileData;
    }

    public function getCorpusDataFiles($flysystem,$path = "",$corpusId){
        $fileData = null;
        if (Cache::has("getCorpusDataFiles_".$corpusId.$path)) {
            $fileData = Cache::get("getCorpusDataFiles_".$corpusId.$path);
        }
        else {
            $gitFunction = new GitFunction();
            $projectArray = array();
            $projectDirs = $flysystem->listContents($path);
            $hasDir = false;
            $count = 0;
            for ($i = 0; $i < count($projectDirs);$i++){
                if($projectDirs[$i]['type'] == "dir") {
                    if(!array_key_exists($projectDirs[$i]['basename'],$projectArray)){
                        $projectArray[$projectDirs[$i]['basename']] = array();
                    }
                    $projects = $flysystem->listContents($projectDirs[$i]['path']);
                    for ($j = 0; $j < count($projects);$j++){

                        if(file_exists($this->basePath."/".$projects[$j]['path'])) {
                            $foldercount = count($flysystem->listContents($projects[$j]['path']));
                            $projects[$j]['foldercount'] = $foldercount;

                            if($gitFunction->isTracked($this->basePath."/".$projects[$j]['path'])){
                                $projects[$j]['tracked'] = "true";
                            }
                            else{
                                $projects[$j]['tracked'] = "false";
                            }

                            $headerObject = $this->laudatioUtilsService->getModelByFileName($projects[$j]['basename'],'CORPUS-DATA',false,$corpusId);
                            if(count($headerObject) > 0 && $headerObject[0]->workflow_status == 0){
                                $projects[$j]['headerObject'] = $headerObject[0];

                                $projects[$j]['lastupdated'] = "";
                                if(isset($projects[$j]['timestamp'])){
                                    $projects[$j]['lastupdated'] = Carbon::createFromTimestamp($projects[$j]['timestamp'])->toDateTimeString();
                                }

                                //Carbon::parse($projects[$j]['timestamp'])->toDateTimeString()->format('H:i, M d');
                                if($projects[$j]["type"] == "dir"){
                                    $hasDir = true;
                                }

                                $projects[$j]['filesize'] = "";
                                if($this->basePath."/".$projects[$j]['path'] != ""){
                                    $projects[$j]['filesize'] = $this->calculateFileSize(filesize($this->basePath."/".$projects[$j]['path']));
                                }

                                $hasDiff = $gitFunction->hasDiff($this->basePath."/".$projects[$j]['path']);


                                if($hasDiff){
                                    $projects[$j]['hasDiff'] = "true";
                                }
                                else{
                                    $projects[$j]['hasDiff'] = "false";
                                }

                                if($projects[$j]['hasDiff'] == "true"){
                                    $projects[$j]['diffFiles'] = array();
                                    array_push($projects[$j]['diffFiles'],$hasDiff);
                                }

                                $count++;
                            }


                        }

                    }//end for files
                    $projects = $this->filterDottedFiles($projects);
                    $projects = $this->filterMDFiles($projects);
                    array_push($projectArray[$projectDirs[$i]['basename']],$projects);
                }//end if dir
            }//end for dirs
            $patharray = explode("/",$path);
            $previouspath = substr($path,0,strrpos($path,"/"));

            $fileData =  array(
                "projects" => $projectArray,
                "hasdir" => $hasDir,
                "filecount" => $count,
                "path" => $path,
                "previouspath" => $previouspath,
            );

            Cache::forever("getCorpusDataFiles_".$corpusId.$path, $fileData);
        }
        return $fileData;
    }


    /**
     *
     * calculateFileSize
     *
     * @param $size
     * @param int $accuracy
     * @return int|string
     */
    function calculateFileSize($size,$accuracy=2) {
        $output = 0;
        $units = array(' Bytes',' KB',' Mb',' Gb');
        foreach($units as $n=>$u) {
            $div = pow(1024,$n);
            if($size > $div) $output = number_format($size/$div,$accuracy).$u;
        }
        return $output;
    }


    /**
     *
     * deleteFile
     *
     * @param $flysystem
     * @param $path
     * @param $user
     * @param $email
     * @return bool|null
     */
    public function deleteFile($flysystem, $path, $user,$email){
        $result = null;
        if(file_exists($this->basePath."/".$path)){
            $gitFunction = new  GitFunction();
            $result = $gitFunction->deleteFiles($path,$user,$email);
        }
        return $result;
    }

    public function deleteFormatFile($flysystem, $path, $user,$email){
        $result = null;
        if($flysystem->has($path)){
            $flyresponse = $flysystem->delete($path);
            $gitFunction = new  GitFunction();
            $result = $gitFunction->deleteFormatFiles($path,$user,$email);
        }
        return $result;
    }


    /**
     *
     * deleteUntrackedFile
     *
     * @param $flysystem
     * @param $path
     * @param bool $isProject
     * @param bool $isCorpus
     * @return bool|null
     */
    public function deleteUntrackedFile($flysystem, $path,$isProject = false,$isCorpus = false){
        $result = null;
        if($flysystem->has($path)){
            $gitFunction = new  GitFunction();
            $result = $gitFunction->deleteUntrackedFiles($path,$isProject,$isCorpus);
        }
        return $result;
    }


    /**
     *
     * deleteUntrackedDataFile
     *
     * @param $flysystem
     * @param $path
     * @return bool|null
     */
    public function deleteUntrackedDataFile($flysystem,$path){

        $result = null;
        if($flysystem->has($path)){
            $gitFunction = new  GitFunction();
            $result = $gitFunction->deleteUntrackedDataFiles($path);
        }
        return $result;
    }


    /**
     *
     * addFilesToRepository
     *
     *
     * @param $path
     * @param $file
     * @return bool
     */
    public function addFilesToRepository($path,$file){
        $gitFunction = new  GitFunction();
        $isAdded = $gitFunction->addUntracked($path,$file);
        if(!$isAdded) {
            $isAdded = $gitFunction->addModified($path, $file);
        }
        return $isAdded;
    }


    /**
     *
     * addFiles
     *
     * @param $flysystem
     * @param $path
     * @return bool
     */
    public function addFiles($flysystem, $path) {
        $isAdded = false;
        try{
            if(is_dir($this->basePath."/".$path)) {
                $contents = $flysystem->listContents($path, true);
                foreach ($contents as $pathElm) {
                    if($pathElm['type'] == "file" && strpos($pathElm['path'], '.git') === false && strpos($pathElm['path'], 'githooks') === false && strpos($pathElm['path'], '.config') === false) {
                        $isAdded = $this->addFilesToRepository($pathElm['dirname'],$pathElm['basename']);
                    }
                }
            }
            else {
                $pathWithOutAddedFolder = substr($path,0,strrpos($path,"/"));
                $file = substr($path,strrpos($path,"/")+1);
                $isAdded = $this->addFilesToRepository($pathWithOutAddedFolder,$file);
            }
        }
        catch(Exception $e) {
            $isAdded = false;
            Log::info("addFiles: error: ".$e->getMessage());
        }

        return $isAdded;
    }

    /**
     *
     * addHooks
     *
     * @param $path
     * @param $user
     * @param $email
     * @return bool
     */
    public function addHooks($path, $user, $email){
        $gitFunction = new  GitFunction();
        return $gitFunction->addGitHooks($path, $user, $email);
    }


    /**
     *
     * pushFiles
     *
     *
     * @param $dirname
     * @param $corpusid
     * @param $user
     * @return bool
     */
    public function pushFiles($dirname,$corpusid,$user){
        $gitFunction = new  GitFunction();
        $result = $gitFunction->pushFiles($dirname,$corpusid);
        return $result;
    }


    /**
     *
     * initialPush
     *
     * @param $path
     * @param $user
     * @return bool
     */
    public function initialPush($path,$user) {
        $gitFunction = new  GitFunction();
        return $gitFunction->initialPush($path,$user);
    }

    public function resetAdd($path,$files){
        $gitFunction = new  GitFunction();
        return $gitFunction->resetAdd($path, $files);
    }

    /**
     * commitFiles by file path
     * @param string $dirname
     * @param $commitmessage
     * @param $corpusid
     * @param $user
     * @param $email
     * @return array|null
     */
    public function commitFiles($dirname = "", $commitmessage, $corpusid, $user, $email){
        $isCommited = false;
        $commitDataArray = array();
        $gitFunction = new  GitFunction();
        $pathWithOutAddedFolder = substr($dirname,0,strrpos($dirname,"/"));

        $stagedFiles = $gitFunction->getListOfStagedFiles($this->basePath."/".$dirname);

        foreach ($stagedFiles as $stagedFile){
            $stagedfileArray = explode("/",$stagedFile);

            if(isset($stagedfileArray[2])){
                $isCommited = $gitFunction->commitFile($this->basePath."/".$dirname."/".$stagedfileArray[1],$stagedfileArray[2],$commitmessage,$user,$email);
            }
            else {
                $dirPathArray = explode("/", $dirname);
                $newpath = $dirPathArray[0]."/".$dirPathArray[1];
                if(is_array($stagedfileArray[0]) && empty($stagedfileArray[1])) {
                    $isCommited = $gitFunction->commitFiles($this->basePath."/".$newpath,$commitmessage,$user,$email);
                }
                else {
                    if(!isset($stagedfileArray[1]) && $stagedfileArray[0] == ".gitignore") {
                        $isCommited = $gitFunction->commitFile($this->basePath."/".$newpath,$stagedfileArray[0],$commitmessage,$user,$email);
                    }
                    else{
                        $folder = str_replace("\"","",$stagedfileArray[0]);
                        $file = str_replace("\"","",$stagedfileArray[1]);
                        //@todo We need a better solution here, this happens because some files in some directories are not added beforehand, like after the githook commit
                        if(null != $file && $file != ""){
                            $isCommited = $gitFunction->commitFile($this->basePath."/".$newpath."/".$folder,$file,$commitmessage,$user,$email);
                        }
                    }

                }
            }

            if($isCommited){
                $commitData = $this->getCommitData($pathWithOutAddedFolder);
                $commitDataArray[$stagedFile] = $commitData;
            }
        }

        return $commitDataArray;
    }

    public function getLastELasticsearchIndex($dirname = "", $tag) {
        $path = $this->basePath."/".$dirname;
        $gitFunction = new  GitFunction();
        return $gitFunction->getLastELasticsearchIndex($path, $tag);
    }

    public function commitStructureFiles($dirname = "", $commitmessage, $corpusid, $user, $email){
        $commitDataArray = array();
        $gitFunction = new  GitFunction();

        $stagedFiles = $gitFunction->getListOfStagedFiles($this->basePath."/".$dirname);
        $isCommited = $gitFunction->commitFiles($this->basePath."/".$dirname,$commitmessage,$user,$email);
        if($isCommited){
            $commitData = $this->getCommitData($dirname);
            $commitDataArray[$dirname] = $commitData;
        }

        return $commitDataArray;
    }



    /**
     *
     * commitFile: commit singe file wrapper
     *
     * @param string $dirname
     * @param $file
     * @param $commitmessage
     * @param $corpusid
     * @param $user
     * @param $email
     * @return array|null
     */
    public function commitFile($dirname = "", $file, $commitmessage, $corpusid, $user, $email){
        $commitdata = null;

        $gitFunction = new  GitFunction();

        $pathWithOutAddedFolder = substr($dirname,0,strrpos($dirname,"/"));

        $isCommited = $gitFunction->commitFile($this->basePath."/".$dirname, $file,$commitmessage,$user,$email);

        if($isCommited){
            $commitdata = $this->getCommitData($pathWithOutAddedFolder);
        }

        return $commitdata;
    }


    /**
     *
     * initiateRepository
     *
     * @param $path
     * @return bool
     */
    public function initiateRepository($path){
        $gitFunction = new GitFunction();
        $isInitiated = false;
        try {
            $isInitiated = $gitFunction->initiateRepository($path);
        }
        catch (GitInitException $gitInitException) {
            throw new GitInitException($gitInitException->getMessage(),0,null);
        }

        return $isInitiated;
    }


    /**
     *
     * addRemote
     *
     * @param $origin
     * @param $path
     * @return bool
     */
    public function addRemote($origin,$path){
        $gitFunction = new GitFunction();
        //git@gitlab.com:Username/Project.git

        $isRemoteAdded = $gitFunction->addRemote($origin,$path);
        return $isRemoteAdded;
    }

    /**
     *
     * copyGitHooks
     *
     * @param $path
     * @return bool
     */
    public function copyGitHooks($path){
        $gitFunction = new  GitFunction();
        return $gitFunction->copyGitHooks($path);
    }

    /**
     *
     * setGitConfig
     *
     * @param $path
     * @param $configs
     * @return array
     */
    public function setGitConfig($path,$configs) {
        $gitFunction = new  GitFunction();
        return $gitFunction->setGitConfig($path,$configs);
    }

    /**
     *
     * setCoreHooksPath
     *
     * @param $path
     * @return bool
     */
    public function setCoreHooksPath($path){
        $gitFunction = new  GitFunction();
        return $gitFunction->setCoreHooksPath($path);
    }

    /**
     *
     * copyScripts
     *
     * @param $path
     * @return bool
     */
    public function copyScripts($path){
        $gitFunction = new  GitFunction();
        return $gitFunction->copyScripts($path);
    }

    /**
     * @param $path
     * @return bool
     */
    public function copyPipelineConfigurations($path) {
        $gitFunction = new  GitFunction();
        return $gitFunction->copyPipelineConfigurations($path);
    }


    /**
     *
     * commitFilesToRepository
     *
     * @param $path
     * @param $commitMessage
     * @return bool
     */
    public function commitFilesToRepository($path,$commitMessage){
        $isCommitted = false;
        $gitFunction = new  GitFunction();
        $commit = $gitFunction->commitFiles($path,$commitMessage);
        $commitStatus = $gitFunction->getStatus($path);

        if($gitFunction->isCleanWorkingTree($commitStatus)){
            $isCommitted = true;
        }
        return $isCommitted;
    }

    /**
     *
     * getCommitData
     *
     * @param $path
     * @return array
     */
    public function getCommitData($path){
        $gitFunction = new GitFunction();
        return $gitFunction->getCommitData($path);
    }


    /**
     *
     * setCorpusVersionTag
     *
     *
     * @param $corpusPath
     * @param $tagmessage
     * @param $version
     * @param $user
     * @param $email
     * @param bool $blame
     * @return bool
     */
    public function setCorpusVersionTag($corpusPath, $tagmessage, $version, $user,$email,$blame = true){
        $isTagged = false;
        $gitFunction = new  GitFunction();
        $tag = $gitFunction->setCorpusVersionTag($corpusPath,$tagmessage,$version,$user,$email, $blame);
        if($tag){
            $isTagged = true;
        }
        return $isTagged;
    }


    /**
     *
     * setInitialCorpusVersionTag
     *
     *
     * @param $corpusPath
     * @param $tagmessage
     * @param $version
     * @param $user
     * @param $email
     * @return bool
     */
    public function setInitialCorpusVersionTag($corpusPath, $tagmessage, $version, $user,$email){
        $isTagged = false;
        $gitFunction = new  GitFunction();
        $tag = $gitFunction->setInitialCorpusVersionTag($corpusPath,$tagmessage,$version,$user,$email);
        if($tag){
            $isTagged = true;
        }
        return $isTagged;
    }


    /**
     * HELPERS
     */


    /**
     *
     * filterDottedFiles
     *
     * @param $array
     * @return array
     */
    public function filterDottedFiles($array){
        $projects = array();
        foreach ($array as $item){
            $pos = strpos($item['basename'],".");
            if($pos === false || ($pos > 0)){
                array_push($projects,$item);
            }
        }

        return $projects;
    }


    public function updateCorpusGitLabVisibility($gitlab_group_id,$gitlab_project_id)
    {
        $result = array();
        $status = null;

        $params = array(
            'visibility' => 'internal'
        );

        try{
            $gitLabResponse = $this->GitLabService->updateGitLabGroup($gitlab_group_id,$params);
            if($gitLabResponse['visibility'] == 'internal'){
                $status = "success";
                $result['gitlab_group_response']  = "Project visibility was successfully set to internal";
            }


        }
        catch (\Exception $e) {
            $status = "error";
            $result['gitlab_group_response'] =  "There was a problem setting the project repository visibility to internal: ".$e->getMessage();
        }


        try{
            $gitLabResponse = $this->GitLabService->updateGitLabProject($gitlab_project_id,$params);

            if($gitLabResponse['visibility'] == 'internal'){
                $status = "success";
                $result['gitlab_repo_response']  = "Corpus visibility was successfully set to internal";
            }


        }
        catch (\Exception $e) {
            $status = "error";
            $result['gitlab_repo_response'] =  "There was a problem setting the Corpus repository visibility to internal: ".$e->getMessage();
        }


        $result['status'] = $status;
        return $result;
    }

    /**
     *
     * filterMDFiles
     *
     * @param $array
     * @return array
     */
    public function filterMDFiles($array){
        $projects = array();
        foreach ($array as $item){
            $pos = strpos($item['basename'],".md");
            if($pos === false){
                array_push($projects,$item);
            }
        }

        return $projects;
    }

    /**
     *
     * normalizeString
     *
     * @param string $str
     * @return false|mixed|string|string[]|null
     */
    public function normalizeString ($str = '')
    {
        $str = strip_tags($str);
        $str = preg_replace('/[\r\n\t ]+/', ' ', $str);
        $str = preg_replace('/[\"\*\/\:\<\>\?\'\|]+/', ' ', $str);
        $str = mb_strtolower($str);
        $str = html_entity_decode( $str, ENT_QUOTES, "utf-8" );
        $str = htmlentities($str, ENT_QUOTES, "utf-8");
        $str = preg_replace("/(&)([a-z])([a-z]+;)/i", '$2', $str);
        $str = str_replace(' ', '-', $str);
        $str = rawurlencode($str);
        $str = str_replace('%', '-', $str);
        $str = str_replace('.', '-', $str);
        $str = str_replace("corpus","",$str);
        $str = str_replace("document","",$str);
        $str = str_replace("annotation","",$str);
        return $str;
    }


    public function normalizeStringRetainCase ($str = '')
    {
        $str = strip_tags($str);
        $str = preg_replace('/[\r\n\t ]+/', ' ', $str);
        $str = preg_replace('/[\"\*\/\:\<\>\?\'\|]+/', ' ', $str);
        $str = html_entity_decode( $str, ENT_QUOTES, "utf-8" );
        $str = htmlentities($str, ENT_QUOTES, "utf-8");
        $str = preg_replace("/(&)([a-z])([a-z]+;)/i", '$2', $str);
        $str = str_replace(' ', '-', $str);
        $str = rawurlencode($str);
        $str = str_replace('%', '-', $str);
        $str = str_replace('.', '-', $str);
        return $str;
    }

    /**
     *
     * normalizeTitle
     *
     * @param string $str
     * @return string|string[]|null
     */
    public function normalizeTitle ($str = '')
    {
        $str = strip_tags($str);
        $str = preg_replace('/[\r\n\t ]+/', ' ', $str);
        $str = preg_replace('/[\"\*\/\:\<\>\?\'\|]+/', ' ', $str);
        $str = preg_replace('/\(|\)/','',$str);
        $str = html_entity_decode( $str, ENT_QUOTES, "utf-8" );
        $str = htmlentities($str, ENT_QUOTES, "utf-8");
        $str = preg_replace("/(&)([a-z])([a-z]+;)/i", '$2', $str);
        return $str;
    }

    public function truncateString($str, $count) {
        $clip = substr($str,0,$count);
        $clip = substr($clip,0, strrpos($clip,".")+1);
        return $clip." ...";
    }

}