<?php

/**
 * Created For Laudatio.
 * User: rolfguescini
 * Date: 06.12.17
 * Time: 10:50
 */
namespace App\Laudatio\Utils;

use App\Custom\LaudatioUtilsInterface;
use App\ExternalEnvironment;
use App\Format;
use Flow\JSONPath\JSONPath;


use App\CorpusProject;
use App\Corpus;
use App\Document;
use App\Annotation;
use App\CorpusFile;
use App\Role;
use App\User;
use App\Invite;
use Carbon\Carbon;
use Illuminate\Database\Eloquent\Collection;
use Illuminate\Pagination\LengthAwarePaginator;
use Symfony\Component\Process\Process;
use Symfony\Component\Process\Exception\ProcessFailedException;
use Log;
use Cache;
use DB;

/**
 * Class LaudatioUtilService
 * @package App\Laudatio\Utils
 */
class LaudatioUtilService implements LaudatioUtilsInterface
{


    protected $basePath;

    /**
     * LaudatioUtilService constructor.
     */
    public function __construct()
    {
        $this->basePath = config('laudatio.basePath');
    }

    /**
     *
     * parseXMLToJson: Parse xml to json
     *
     * @param $xml
     * @param $options
     * @return array
     */
    public function parseXMLToJson($xml, $options){

        $defaults = array(
            'namespaceSeparator' => ':',//you may want this to be something other than a colon
            'attributePrefix' => '',   //to distinguish between attributes and nodes with the same name
            'alwaysArray' => array(),   //array of xml tag names which should always become arrays
            'autoArray' => true,        //only create arrays for tags which appear more than once
            'textContent' => 'text',       //key used for the text content of elements
            'autoText' => true,         //skip textContent key if node has no attributes or child nodes
            'keySearch' => false,       //optional search and replace on tag and attribute names
            'keyReplace' => false       //replace values for above search values (as passed to str_replace())
        );

        $options = array_merge($defaults, $options);
        $namespaces = $xml->getDocNamespaces();
        $namespaces[''] = null; //add base (empty) namespace


        //get attributes from all namespaces
        $attributesArray = array();
        foreach ($namespaces as $prefix => $namespace) {
            foreach ($xml->attributes($namespace) as $attributeName => $attribute) {
                //replace characters in attribute name
                if ($options['keySearch']) $attributeName =
                    str_replace($options['keySearch'], $options['keyReplace'], $attributeName);
                $attributeKey = $options['attributePrefix']
                    . ($prefix ? $prefix . $options['namespaceSeparator'] : '')
                    . $attributeName;
                $attributesArray[$attributeKey] = (string)$attribute;
            }

            foreach ($xml->attributes('xml', TRUE) as $attributeName => $attribute) {
                //replace characters in attribute name
                if ($options['keySearch']) $attributeName =
                    str_replace($options['keySearch'], $options['keyReplace'], $attributeName);
                $attributeKey = $options['attributePrefix']
                    . ($prefix ? $prefix . $options['namespaceSeparator'] : '')
                    . $attributeName;
                $attributesArray[$attributeKey] = (string)$attribute;
            }


        }

        //get child nodes from all namespaces
        $tagsArray = array();
        foreach ($namespaces as $prefix => $namespace) {
            foreach ($xml->children($namespace) as $childXml) {
                //recurse into child nodes
                $childArray = $this->parseXMLToJson($childXml, $options);
                $childTagName = null;
                $childProperties = null;
                foreach($childArray as $tagName => $childProp) {
                    $childTagName = $tagName;
                    $childProperties = $childProp;
                }
                //list($childTagName, $childProperties) = each($childArray);

                //replace characters in tag name
                if ($options['keySearch']) $childTagName =
                    str_replace($options['keySearch'], $options['keyReplace'], $childTagName);
                //add namespace prefix, if any
                if ($prefix) $childTagName = $prefix . $options['namespaceSeparator'] . $childTagName;

                if (!isset($tagsArray[$childTagName])) {
                    //only entry with this key
                    //test if tags of this type should always be arrays, no matter the element count

                    $tagsArray[$childTagName] =
                        in_array($childTagName, $options['alwaysArray']) || !$options['autoArray']
                            ? array($childProperties) : $childProperties;
                    //? array($childProperties) : array($childProperties);
                } elseif (
                    is_array($tagsArray[$childTagName]) && array_keys($tagsArray[$childTagName])
                    === range(0, count($tagsArray[$childTagName]) - 1)
                ) {
                    //key already exists and is integer indexed array
                    $tagsArray[$childTagName][] = $childProperties;
                } else {
                    //key exists so convert to integer indexed array with previous value in position 0
                    $tagsArray[$childTagName] = array($tagsArray[$childTagName], $childProperties);
                }
            }
        }

        //get text content of node
        $textContentArray = array();
        $plainText = trim((string)$xml);
        $plainText = str_replace("\n","",$plainText);
        $plainText = str_replace("\r","",$plainText);
        $plainText = str_replace("\r\n","",$plainText);
        $plainText = str_replace(' +', ' ', $plainText);

        if ($plainText !== '') $textContentArray[$options['textContent']] = $plainText;

        //stick it all together
        $propertiesArray = !$options['autoText'] || $attributesArray || $tagsArray || ($plainText === '')
            ? array_merge($attributesArray, $tagsArray, $textContentArray) : $plainText;

        return array(
            $xml->getName() => $propertiesArray
        );
    }

    /**
     *
     * Populate Corpus  object with attributes from the Corpus Header
     *
     * @param $json
     * @param $params
     * @return mixed
     * @throws \Flow\JSONPath\JSONPathException
     */
    public function setCorpusAttributes($json,$params){
        $jsonPath = new JSONPath($json,JSONPath::ALLOW_MAGIC);
        $corpus_id = $params['corpus_id'];
        $corpusTitle = $jsonPath->find('$.TEI.teiHeader.fileDesc.titleStmt.title')->data();
        $corpusSizeType = $jsonPath->find('$.TEI.teiHeader.fileDesc.extent.type')->data();
        $corpusSizeValue = $jsonPath->find('$.TEI.teiHeader.fileDesc.extent.text')->data();

        $corpusDesc = $jsonPath->find('$.TEI.teiHeader.encodingDesc[0].projectDesc.p.text')->data();
        //there can be a more than one encodingdesc
        if(count($corpusDesc) < 1){
            $corpusDesc = $jsonPath->find('$.TEI.teiHeader.encodingDesc.projectDesc.p.text')->data();
        }


        $corpus = Corpus::find($params['corpusId']);
        if(count($corpusTitle) > 0 && count($corpusDesc) > 0  && count($corpusDesc) > 0 && count($corpusSizeType) > 0) {
            $corpus->update([
                "name" => $corpusTitle[0],
                "description" => $corpusDesc[0],
                "corpus_size_type" => $corpusSizeType[0],
                "corpus_size_value" => $corpusSizeValue[0],
                "publication_version" => $params['publication_version'],
                "workflow_status" => 0,
                'uid' => $params['uid'],
                'gitlab_group_id' => $params['gitlab_group_id'],
                'directory_path' => $params['corpus_path'],
                'corpus_id' => $corpus_id,
                'gitlab_id' => $params['gitlab_id'],
                'gitlab_web_url' => $params['gitlab_web_url'],
                'gitlab_ssh_url' => $params['gitlab_ssh_url'],
                'gitlab_namespace_path' => $params['gitlab_name_with_namespace'],
                "file_name" => $params['file_name'],
                "elasticsearch_index" => $params['elasticsearch_index'],
                "guidelines_elasticsearch_index" => $params['guidelines_elasticsearch_index'],
            ]);

        }

        return $corpus;
    }


    /**
     *
     * updateCorpusAttributes: update corpus object fields
     *
     * @param $params
     * @param $corpusId
     * @return mixed
     */
    public function updateCorpusAttributes($params, $corpusId){
        $corpus = Corpus::find($corpusId);
        $corpus->update($params);
        return $corpus;
    }


    /**
     *
     * updateCorpusFileAttributes: update corpus file object fields
     *
     * @param $corpus
     * @return bool
     */
    public function updateCorpusFileAttributes($corpus){
        $updated = true;

        try{
            foreach ($corpus->corpusfiles as $corpusfile) {
                $corpusfile->directory_path = $corpus->directory_path;
                $corpusfile->save();
            }
        }
        catch(\Exception $e) {
            $updated = false;
        }

        return $updated;
    }

    /**
     * @param $oldCorpus
     * @param $data
     * @return array
     */
    public function duplicateCorpus($oldCorpus, $data){
        $elasticsearchIndexes = array();

        $new_corpus = new Corpus();

        $new_corpus->name = $oldCorpus->name;
        $new_corpus->uid = $oldCorpus->uid;
        $new_corpus->vid = ($oldCorpus->vid +1);
        $new_corpus->description = $oldCorpus->description;
        $new_corpus->corpus_size_type = $oldCorpus->corpus_size_type;
        $new_corpus->corpus_size_value = $oldCorpus->corpus_size_value;
        $new_corpus->directory_path = $oldCorpus->directory_path;
        $new_corpus->file_name = $oldCorpus->file_name;
        //$new_corpus->elasticsearch_id = $data['corpus']['elasticsearchid'];
        //$new_corpus->guidelines_elasticsearch_index = $new_guideline_index;
        $new_corpus->elasticsearch_index = $data['corpus']['index'];
        $new_corpus->publication_version = "working_period";//Sie bearbeiten jetzt eine neuen Arbeitsversion. Sie sehen Daten der gerade publizierten version. Diese können sie für eine neue Version bearbeiten
        $new_corpus->gitlab_group_id = $oldCorpus->gitlab_group_id;
        $new_corpus->gitlab_id = $oldCorpus->gitlab_id;
        $new_corpus->gitlab_web_url = $oldCorpus->gitlab_web_url;
        $new_corpus->gitlab_ssh_url = $oldCorpus->gitlab_ssh_url;
        $new_corpus->gitlab_namespace_path = $oldCorpus->gitlab_namespace_path;
        $new_corpus->workflow_status = 0;
        $new_corpus->corpus_logo = $oldCorpus->corpus_logo;
        $new_corpus->save();

        $oldCorpus->workflow_status = 1;
        $oldCorpus->save();

        //attach user roles

        $corpusUser = User::find($new_corpus->uid);
        $corpusProjectAdminRole = Role::findById(2);
        $addProjectRole = false;
        if($corpusUser->roles->contains($corpusProjectAdminRole)){
            $addProjectRole = true;
        }


        $corpusAdminRole = Role::findById(3);
        $corpusUser->roles()->sync($corpusAdminRole);
        if($corpusUser) {
            if(!$corpusUser->roles->contains($corpusAdminRole)){
                $corpusUser->roles()->attach($corpusAdminRole);
            }

            $new_corpus->users()->save($corpusUser,['role_id' => 3]);
        }


        /* for each corpusproject attached to this corpus, detach the old corpus (?),
        and attach the new corpus to the corpus project for the next working period
        */


        $corpusUser->roles()->sync($corpusAdminRole);
        $corpusProjects = $oldCorpus->corpusprojects()->get();
        foreach($corpusProjects as $corpusProject) {
            //$corpusProject->corpora()->detach($corpus->id);
            $new_corpus->corpusprojects()->attach($corpusProject);
            if($addProjectRole) {
                $corpusUser->roles()->attach($corpusProjectAdminRole);
            }
        }


        /** GET INVITED CORPUS COLLABORATORS **/

        $corpus_invite_ids = DB::table('corpus_user')->select('user_id', 'role_id')->where([
                ['corpus_id','=',$oldCorpus->id]
            ]
        )->get();

        if(count($corpus_invite_ids->toArray()) > 0){
            foreach ($corpus_invite_ids->toArray() as $corpus_invite_object){
                DB::table('corpus_user')->insert(
                    [
                        'corpus_id' => $new_corpus->id,
                        'user_id' => $corpus_invite_object->user_id,
                        'role_id' => $corpus_invite_object->role_id,
                    ]
                );
            }
        }



        /** externallinks */
        /*
        foreach($oldCorpus->externalenvironments()->get() as $externalenvironment) {
            $newExternalEnvironment = new ExternalEnvironment();
            $newExternalEnvironment->uid = $externalenvironment->uid;
            $newExternalEnvironment->corpus_id = $new_corpus->id;
            $newExternalEnvironment->external_environment_name = $externalenvironment->external_environment_name;
            $newExternalEnvironment->external_environment_uri = $externalenvironment->external_environment_uri;
            $newExternalEnvironment->external_environment_description = $externalenvironment->external_environment_description;
            Log::info("SAVING externalenvironment: ".print_r($newExternalEnvironment,1));
            $newExternalEnvironment->save();
        }
*/
        $elasticsearchIndexes['corpus'] = array("new_corpus_id" => array());
        $elasticsearchIndexes['corpus']["new_corpus_id"][$oldCorpus->id] = $new_corpus->id;

        $elasticsearchIndexes['document'] = array("new_document_ids" => array());
        foreach ($oldCorpus->documents()->get() as $document) {
            $newDocument = new Document();
            $newDocument->title = $document->title;
            $newDocument->uid = $document->uid;
            $newDocument->vid = ($document->vid +1);
            $newDocument->file_name = $document->file_name;
            $newDocument->document_genre = $document->document_genre;
            $newDocument->document_size_type = $document->document_size_type;
            $newDocument->document_size_value = $document->document_size_value;
            $newDocument->document_id = $document->document_id;
            $newDocument->corpus_id = $new_corpus->id;
            $newDocument->publication_version = "working_period";
            //$newDocument->elasticsearch_id = $new_document_elasticsearch_id;
            $newDocument->elasticsearch_index = $data['document']['index'];
            $newDocument->publication_version = $document->publication_version;
            $newDocument->directory_path = $document->directory_path;
            $newDocument->workflow_status = 0;
            $newDocument->save();
            $new_corpus->documents()->save($newDocument);
            $document->workflow_status = 1;
            $document->save();
            $elasticsearchIndexes['document']['new_document_ids'][$document->id] = $newDocument->id;
        }


        $elasticsearchIndexes['annotation'] = array("new_annotation_ids" => array());
        foreach ($oldCorpus->annotations()->get() as $annotation) {
            $newAnnotation = new Annotation();
            $newAnnotation->uid = $annotation->uid;
            $newAnnotation->vid = ($annotation->vid +1);
            $newAnnotation->file_name = $annotation->file_name;
            $newAnnotation->annotation_id = $annotation->annotation_id;
            $newAnnotation->annotation_group = $annotation->annotation_group;
            $newAnnotation->annotation_size_type = $annotation->annotation_size_type;
            $newAnnotation->annotation_size_value = $annotation->annotation_size_value;
            $newAnnotation->corpus_id = $new_corpus->id;
            $newAnnotation->publication_version = "working_period";
            //$newAnnotation->elasticsearch_id = $new_annotation_elasticsearch_id;
            $newAnnotation->elasticsearch_index = $data['annotation']['index'];
            $newAnnotation->directory_path = $annotation->directory_path;
            $newAnnotation->workflow_status = 0;
            $newAnnotation->save();
            foreach ($annotation->documents()->get() as $annodocu){
                $newAnnotation->documents()->attach($annodocu);
            }
            $annotation->workflow_status = 1;
            $annotation->save();
            $elasticsearchIndexes['annotation']['new_annotation_ids'][$annotation->id] = $newAnnotation->id;
        }


        return $elasticsearchIndexes;
    }

    public function copyCorpusFiles($oldCorpus,$new_corpus_id,$newIndex) {
        //copy corpus files

        $newFileIds = array();
        $newCorpus = Corpus::find($new_corpus_id);
        foreach ($oldCorpus->corpusfiles as $corpusfile) {

            $newCorpusFile = new CorpusFile();
            $newCorpusFile->corpus_id = intval($newCorpus->id);
            $newCorpusFile->uid = intval($corpusfile->uid);
            $newCorpusFile->directory_path = $corpusfile->directory_path;
            $newCorpusFile->file_name = $corpusfile->file_name;
            $newCorpusFile->format = intval($corpusfile->format);
            $newCorpusFile->workflow_status = intval($corpusfile->workflow_status);
            $newCorpusFile->save();

            $newCorpus->corpusfiles()->save($newCorpusFile);
            array_push($newFileIds,$newCorpusFile->id);
        }
        return $newFileIds;
    }

    public function updateCorpusFiles($oldCorpus){
        $oldFileIds = array();

        foreach ($oldCorpus->corpusfiles as $corpusfile) {
            $corpusfile->workflow_status = 1;
            array_push($oldFileIds,$corpusfile->id);
        }

        return $oldFileIds;
    }

    /**
     * Checks if Corpus is versioned and
     * @param $corpusId
     * @return bool
     */
    public function corpusIsVersioned($corpusId){
        $isVersioned = false;
        $corpus = Corpus::findOrFail($corpusId);

        if($corpus->vid >= 1){
            $isVersioned = true;
        }
        return $isVersioned;
    }

    /**
     *
     * Populate Document object with attributes fom the Document header
     *
     * @param $json
     * @param $corpusId
     * @param $uid
     * @param $fileName
     * @param $isDir
     * @return Document|mixed|null
     * @throws \Flow\JSONPath\JSONPathException
     */
    public function setDocumentAttributes($json,$corpusId,$uid,$fileName,$isDir){
        $jsonPath = new JSONPath($json);

        $documentTitle = $jsonPath->find('$.TEI.teiHeader.fileDesc.titleStmt.title.text')->data();
        if(!$documentTitle){
            $documentTitle = $jsonPath->find('$.TEI.teiHeader.fileDesc.titleStmt.title')->data();
        }

        $document_id = $jsonPath->find('$.TEI.teiHeader.fileDesc.id')->data();

        $documentGenre = $jsonPath->find('$.TEI.teiHeader.style')->data();


        $documentSizeType = $jsonPath->find('$.TEI.teiHeader.fileDesc.extent.type')->data();


        $documentSizeValue = $jsonPath->find('$.TEI.teiHeader.fileDesc.extent.text')->data();

        $encodingDesc = $jsonPath->find('$.TEI.teiHeader.encodingDesc.schemaSpec.elementSpec[*]')->data();


        $document = null;
        $documentObject = $this->getModelByFileAndCorpus($fileName,'document',$isDir,$corpusId);
        $corpus = Corpus::find($corpusId);

        if(count($documentObject) > 0){
            $document = $documentObject[0];
            $document->title = $documentTitle[0];
            $document->document_genre = count($documentGenre) > 0 ? $documentGenre[0] : "";
            $document->document_size_type = count($documentSizeType) > 0 ? $documentSizeType[0]: "";
            $document->document_size_value = count($documentSizeValue) > 0 ? $documentSizeValue[0]: 0;
            $document->document_id = $document_id[0];
            $document->corpus_id = $corpusId;
            $document->uid = $uid;
            $document->file_name = $fileName;
            $document->directory_path = $corpus->directory_path;
            $document->marked_for_deletion = 0;
            $document->save();
        }
        else{
            $document = new Document;
            $document->title = $documentTitle[0];
            $document->document_genre = count($documentGenre) > 0 ? $documentGenre[0] : "";
            $document->document_size_type = count($documentSizeType) > 0 ? $documentSizeType[0]: "";
            $document->document_size_value = count($documentSizeValue) > 0 ? $documentSizeValue[0]: 0;
            $document->document_id = $document_id[0];
            $document->corpus_id = $corpusId;
            $document->uid = $uid;
            $document->directory_path = $corpus->directory_path;
            $document->file_name = $fileName;
            $document->marked_for_deletion = 0;
            $document->save();
        }


        /*
         * Populate database with list of annotations, with document_id and corpus_id as compound primary_key
         */

        foreach ($encodingDesc as $annotationJson) {

            $annotationPath = new JSONPath($annotationJson,JSONPath::ALLOW_MAGIC);
            $annotationGroup = $annotationPath->find('$.ident')->data();
            $annotations = $annotationPath->find('$.valList.valItem[*]')->data();
            foreach ($annotations as $annotation) {
                $annotationValue = "";
                if(is_array($annotation)){
                    /**
                     * @todo: Germanc-annotaiton-problem coresp / ident
                     */
                    $annotationValue = $annotation['corresp'];
                }
                else{
                    $annotationValue = $annotation;
                }

                $annotationsFromDB = Annotation::where(
                    [
                        ['annotation_id', '=', $annotationValue],
                        ['corpus_id', '=', $corpusId],
                    ]
                )->get();

                if(count($annotationsFromDB) == 0){
                    $annotationObject = new Annotation;
                    $annotationObject->annotation_id = $annotationValue;
                    $annotationObject->corpus_id = $corpusId;
                    $annotationObject->directory_path = $corpus->directory_path;
                    $annotationObject->annotation_group = $annotationGroup[0];
                    $annotationObject->marked_for_deletion = 0;
                    $annotationObject->save();
                    $annotationObject->documents()->attach($document);
                }
                else{
                    $annotationObject = $annotationsFromDB[0];

                    $annotationObject->annotation_id = $annotationValue;
                    $annotationObject->corpus_id = $corpusId;
                    $annotationObject->directory_path = $corpus->directory_path;
                    $annotationObject->annotation_group = $annotationGroup[0];
                    $annotationObject->marked_for_deletion = 0;
                    $annotationObject->save();
                    $annotationObject->documents()->attach($document);
                }//end if
            }//end foreach

        }//end foreach

        return $document;
    }

    /**
     * setCommitData: save git commit data on the object and persist to database
     *
     * @param $commitData
     * @param $corpusId
     * @return bool
     */
    public function setCommitData($commitData,$corpusId) {
        $setData = true;

        //try {
            foreach ($commitData as $path => $data) {
                $pathArray = explode("/",$path);
                $object = null;
                if(strrpos($path,"TEI-HEADERS/corpus") !== false) {

                    $fileName = $pathArray[2];
                    if(strpos($fileName,".md") === false){
                        $tempObject = $this->getModelByFileAndCorpus($fileName, "corpus", false, $corpusId);

                        if(count($tempObject) > 0) {
                            $object = Corpus::findOrFail($tempObject[0]->id);
                            $this->setVersionMapping($object, "corpus");
                        }
                    }
                }
                else if(strrpos($path,"TEI-HEADERS/document") !== false) {
                    $fileName = $pathArray[2];
                    if(strpos($fileName,".md") === false){
                        $tempObject = $this->getModelByFileAndCorpus($fileName, "document", false, $corpusId);
                        if(count($tempObject) > 0) {
                            $object = Document::findOrFail($tempObject[0]->id);
                            $this->setVersionMapping($object, "document");
                        }
                    }

                }
                else if(strrpos($path,"TEI-HEADERS/annotation") !== false) {
                    $fileName = $pathArray[2];
                    if(strpos($fileName,".md") === false){
                        $tempObject = $this->getModelByFileAndCorpus($fileName, "annotation", false, $corpusId);
                        if(count($tempObject) > 0) {
                            $object = Annotation::findOrFail($tempObject[0]->id);
                            $this->setVersionMapping($object,"annotation");
                        }

                    }

                }
                else if(strrpos($path,"CORPUS-DATA") !== false) {
                    if(count($pathArray) > 2) {
                        $fileName = $pathArray[2];
                    }
                    else {
                        $fileName = $pathArray[1];
                    }

                    if(strpos($fileName,".md") === false) {
                        $tempObject = $this->getModelByFileName($fileName, "CORPUS-DATA", false, $corpusId);
                        if(count($tempObject) > 0) {
                            $object = CorpusFile::findOrFail($tempObject->id);
                        }
                    }
                }



                $params = array(
                    "gitlab_commit_sha" => $data['sha_string'],
                    "gitlab_commit_date" => $data['date'],
                    "gitlab_commit_description" => $data['message'],
                    "elasticsearch_id" => isset($data['elasticsearch_id'])? $data['elasticsearch_id'] : ""
                );

                if(null != $object){
                    $object->update($params);
                }

            }
        //}
        //catch (\Exception $e) {
          //  $setData = false;
           // Log::info("setCommitData: error: ".$e->getMessage());
       // }



        return $setData;
    }

    /**
     *
     * updateDocumentAttributes: update the Document object fields
     *
     * @param $params
     * @param $documentId
     * @return mixed
     */
    public function updateDocumentAttributes($params,$documentId){
        $document = Document::find($documentId);
        $document->update($params);
        return $document;
    }


    /**
     *
     * documentIsVersioned: determine if a document object is versioned (internal vid)
     *
     * @param $documentId
     * @return bool
     */
    public function documentIsVersioned($documentId){
        $isVersioned = false;
        $document = Document::findOrFail($documentId);

        if($document->vid >= 1){
            $isVersioned = true;
        }
        return $isVersioned;
    }

    /**
     *
     * Populate Annotation object with attributes from the Annotation Header
     *
     * @param $json
     * @param $corpusId
     * @param $uid
     * @param $fileName
     * @param $isDir
     * @return Annotation
     * @throws \Flow\JSONPath\JSONPathException
     */
    public function setAnnotationAttributes($json,$corpusId,$uid,$fileName,$isDir){
        $jsonPath = new JSONPath($json);

        $annotationId = $jsonPath->find('$.TEI.teiHeader.fileDesc.titleStmt.title.corresp')->data();
        $annotationSizeType = $jsonPath->find('$.TEI.teiHeader.fileDesc.extent.type')->data();
        $annotationSizeValue = $jsonPath->find('$.TEI.teiHeader.fileDesc.extent.text')->data();

        $annotationsFromDB = Annotation::where(
            [
                ['annotation_id', '=', $annotationId[0]],
                ['corpus_id', '=', $corpusId],
            ]
        )->get();


        $corpus = Corpus::find($corpusId);

        if(count($annotationsFromDB) > 0){
            foreach($annotationsFromDB as $annotationFromDB)
                $annotationFromDB->update([
                    "annotation_size_type" => $annotationSizeType[0],
                    "annotation_size_value" => $annotationSizeValue[0],
                    "directory_path" => $corpus->directory_path,
                    "marked_for_deletion" => 0,
                    "file_name" => $fileName,
                    "uid" => $uid
                ]);

            $annotationFromDB->save();

            return $annotationFromDB;
        }
        else{
            $annotation = new Annotation;
            $annotation->annotation_id = $annotationId[0];
            $annotation->annotation_size_type = $annotationSizeType[0];
            $annotation->annotation_size_value = $annotationSizeValue[0];
            $annotation->corpus_id = $corpusId;
            $annotation->uid = $uid;
            $annotation->file_name = $fileName;
            $annotation->directory_path = $corpus->directory_path;
            $annotation->marked_for_deletion = 0;
            $annotation->save();
            return $annotation;
        }
    }


    /**
     *
     * setPreparationAttributes: not in use
     *
     * @param $json
     * @param $annotationId
     * @param $corpusId
     * @param $isDir
     * @throws \Flow\JSONPath\JSONPathException
     */
    public function setPreparationAttributes($json,$annotationId,$corpusId,$isDir){
        $jsonPath = new JSONPath($json);
        $preparationFromDB = Preparation::where([
            ['annotation_id', '=', $annotationId],
            ['corpus_id', '=', $corpusId],
        ])->get();


        if(count($preparationFromDB) > 0){

        }
        else{

            $preparationEncodingSteps = $jsonPath->find('$.TEI.teiHeader.encodingDesc[*]')->data();

            if(!is_array($preparationEncodingSteps[0])){
                $preparationEncodingSteps = array(
                    array(
                        'n' => $preparationEncodingSteps[0],
                        'style' => $preparationEncodingSteps[1],
                        'appInfo' => $preparationEncodingSteps[2],
                        'editorialDecl' => isset($preparationEncodingSteps[3]) ? $preparationEncodingSteps[3] : "",
                        'projectDesc' => isset($preparationEncodingSteps[4]) ? $preparationEncodingSteps[4] : "",
                    )

                );
            }

            foreach ($preparationEncodingSteps as $preparationEncodingStep) {
                $preparation = new Preparation;
                $preparation->preparation_encoding_step = $preparationEncodingStep['style'];
                $preparation->preparation_encoding_style = $preparationEncodingStep['appInfo']['style'];
                $preparation->preparation_encoding_tool =  $preparationEncodingStep['appInfo']['application']['ident'];
                $preparation->preparation_encoding_full_name = $preparationEncodingStep['appInfo']['application']['label'];
                $preparation->preparation_encoding_description = $preparationEncodingStep['appInfo']['application']['p'];
                $preparation->preparation_encoding_annotation_style = $preparationEncodingStep['appInfo']['application']['style'];
                $preparation->preparation_encoding_segmentation_style = isset($preparationEncodingStep['editorialDecl']['segmentation']) ? $preparationEncodingStep['editorialDecl']['segmentation']['style'] : "";
                if(isset($preparationEncodingStep['editorialDecl']['segmentation']['corresp'])){
                    $preparation->preparation_encoding_segmentation_type = $preparationEncodingStep['editorialDecl']['segmentation']['corresp'];
                }
                else{
                    $preparation->preparation_encoding_segmentation_type = "NA";
                }
                if(isset($preparationEncodingStep['editorialDecl']['segmentation']['p'])){
                    $preparation->preparation_encoding_segmentation_description = $preparationEncodingStep['editorialDecl']['segmentation']['p'];
                }
                else{
                    $preparation->preparation_encoding_segmentation_description = "NA";
                }

                $preparation->annotation_id = $annotationId;
                $preparation->corpus_id = $corpusId;
                $preparation->save();
            }
        }
    }

    /**
     *
     * updateAnnotationAttributes: update the Annotation model fields
     *
     * @param $params
     * @param $annotationId
     * @return mixed
     */
    public function updateAnnotationAttributes($params,$annotationId){
        $annotation = Annotation::find($annotationId);
        $annotation->update($params);
        return $annotation;
    }


    /**
     *
     * annotationIsVersioned: determine if an Annotation object is versioned (internal vid)
     *
     * @param $annotationId
     * @return bool
     */
    public function annotationIsVersioned($annotationId){
        $isVersioned = false;
        $annotation = Annotation::findOrFail($annotationId);

        if($annotation->vid >= 1){
            $isVersioned = true;
        }
        return $isVersioned;
    }


    /**
     *
     * Associate an array of Document objects to a Corpus
     *
     * @param $documents
     * @param $corpusId
     * @return mixed|static Corpus
     */
    public function associateDocumentsToCorpus($documents,$corpusId){
        $corpus = Corpus::find($corpusId);

        foreach($documents as $document){
            $corpus->documents()->save($document);
        }
        return $corpus;
    }

    /**
     *
     * Associate an array of Annotation objects to a Document
     *
     * @param $annotations
     * @param $documentId
     * @return mixed|static Document
     */
    public function associateAnnotationsToDocument($annotations,$documentId){
        $document = Document::find($documentId);

        foreach ($annotations as $annotation) {
            $document->annotations()->save($annotation);
        }

        return $document;
    }

    /**
     *
     * Associate an array of Annotation objects to a Corpus
     *
     * @param $annotations
     * @param $corpusId
     * @return mixed|static Corpus
     */
    public function associateAnnotationsToCorpus($annotations,$corpusId){
        $corpus = Corpus::find($corpusId);
        foreach ($annotations as $annotation) {
            $corpus->annotations()->save($annotation);
        }

        return $corpus;
    }

    /**
     *
     * Associate an array of Preparation objects to an Annotation: not in use
     *
     * @param $preparations
     * @param $annotationId
     * @return mixed|static Annotation
     */
    public function associatePreparationsToAnnotation($preparations,$annotationId){
        $annotation = Annotation::find($corpusId);

        foreach ($preparations as $preparation) {
            $annotation->preparations()->save($preparation);
        }

        return $annotation;
    }


    /**
     * setVersionMapping: set the internal database versioning (vid) for an object
     *
     * @param $object
     * @param $type
     * @return int
     */
    public function setVersionMapping($object,$type){

        if(!empty($object)){
            if(null != $object->vid){
                $object->vid++;
            }
            else{
                $object->vid = 1;
            }


            $object->save();

            $id_vid = DB::table('versions')->select('id', 'vid')->where([
                    ['id','=',$object->id],
                    ['type','=',$type],
                ]
            )->get();

            if(count($id_vid) > 0){
                DB::table('versions')->where('id',$object->id)->update(
                    ['vid' => $object->vid]
                );
            }
            else{
                DB::table('versions')->insert(
                    [
                        'id' => $object->id,
                        'vid' => $object->vid,
                        'type' => $type
                    ]
                );
            }
        }

        return $object->vid;

    }

    public function getDOIByCorpusDatabaseId($corpusDatabaseId) {
        $corpus = Corpus::findOrFail($corpusDatabaseId);
        $doiIdentifier = "";
        $doidata = null;

        if(!empty($corpus)) {

            $doidata = $corpus->doi()->get();

            if(!empty($doidata[0])){
                $doiIdentifier = $doidata[0]->doi_id;
            }
        }

        return $doiIdentifier;
    }

    public function getDOIByCorpusElasticsearchId($elasticsearchId,$corpus_index) {
        $doiIdentifier = "";

        $DBcorpus = Corpus::where([["elasticsearch_id","=",$elasticsearchId],["elasticsearch_index","=",$corpus_index]])->get();
        if(count($DBcorpus) > 0) {
            $corpusDatabaseId = $DBcorpus[0]->id;
            $corpus = Corpus::findOrFail($corpusDatabaseId);
            $doiIdentifier = "";
            $doidata = null;

            if(!empty($corpus)) {

                $doidata = $corpus->doi()->get();

                if(!empty($doidata[0])){
                    $doiIdentifier = $doidata[0]->doi_id;
                }
            }
        }

        return $doiIdentifier;
    }

    public function getDOIByCorpusId($corpusId,$corpus_index) {
        $doiIdentifier = "";

        $DBcorpus = Corpus::where([["corpus_id","=",$corpusId],["elasticsearch_index","=",$corpus_index]])->get();
        if(count($DBcorpus) > 0) {
            $corpusDatabaseId = $DBcorpus[0]->id;
            $corpus = Corpus::findOrFail($corpusDatabaseId);
            $doiIdentifier = "";
            $doidata = null;

            if(!empty($corpus)) {

                $doidata = $corpus->doi()->get();

                if(!empty($doidata[0])){
                    $doiIdentifier = $doidata[0]->doi_id;
                }
            }
        }

        return $doiIdentifier;
    }

    /**
     *
     * setCorpusLogoSymLink: set a symlink in the filesystem to corpus logo images
     *
     * @param $path
     * @return string
     */
    public function setCorpusLogoSymLink($flysystem, $path) {

        $pathArray = explode("/",$path);
        end($pathArray);
        $last_id = key($pathArray);

        $imagePath = public_path('images');
        if(!file_exists($imagePath."/corpuslogos")) {
            $mkdirprocess = new Process("mkdir ".$imagePath."/corpuslogos");
            $mkdirprocess->run();

            // executes after the command finishes
            if (!$mkdirprocess->isSuccessful()) {
                throw new ProcessFailedException($mkdirprocess);
            }
        }

        $toPath = $imagePath."/corpuslogos/".$pathArray[0]."_".$pathArray[1]."_".$pathArray[$last_id];

        if(!file_exists ($toPath)){

            $process = new Process("ln -s ".$this->basePath."/".$path." ".$toPath);
            $process->run();

            // executes after the command finishes
            if (!$process->isSuccessful()) {
                throw new ProcessFailedException($process);
            }

            return $process->getOutput();
        }

    }

    public function getCurrentPublishedCorpusFromDatabase($gitlab_id){
        $returnarray = array();

        $corpuscollectiondata = Corpus::where([
            ['gitlab_id', '=',$gitlab_id],
            ['workflow_status', '=', 1]
        ])->get();


        $data = array(
            "corpus_elasticsearch_id" => "",
            "document_elasticsearch_ids" => [],
            "annotation_elasticsearch_ids" => [],
        );

        if(count($corpuscollectiondata) > 0) {

            foreach ($corpuscollectiondata as $corpuscollection) {
                $corpus_id =  $corpuscollection->id;
                $corpus_elasticsearch_id =  $corpuscollection->elasticsearch_id;
                $data["corpus_elasticsearch_id"] = $corpus_elasticsearch_id;

                $documents = Document::where([
                    ['corpus_id', '=',$corpus_id]
                ])->get();

                foreach ($documents as $document){
                    array_push($data["document_elasticsearch_ids"],$document->elasticsearch_id);
                }

                $annotations = Annotation::where([
                    ['corpus_id', '=',$corpus_id]
                ])->get();

                foreach ($annotations as $annotation){
                    array_push($data["annotation_elasticsearch_ids"],$annotation->elasticsearch_id);
                }

                array_push($returnarray,$data);
            }


        }

        return $data;
    }

    /**
     *
     * getModelByType: Fetch eloquent model by id and type
     *
     * @param $id
     * @param $type
     * @return |null
     */
    public function getModelByType($id,$type){
        $object = null;
        switch ($type){
            case 'corpus':
                $object = Corpus::find($id);
                break;
            case 'document':
                $object = Document::find($id);
                break;
            case 'annotation':
                $object = Annotation::find($id);
                break;
        }
        return $object;
    }



    /**
     *
     * Fetch a Model by type and filename
     *
     * @param $fileName
     * @param $type
     * @return mixed
     * @todo: This is very brittle, we need some kind of GUID, and to add at least the corpus id to avoid ambiguity
     */
    public function getModelByFileName($fileName, $type, $isDir,$corpusId){
        $object = null;
        switch ($type){
            case 'corpus':
                if($isDir){
                    $object = Corpus::where([
                        ['directory_path', '=',$fileName],
                        ['id', '=',$corpusId]
                    ])->get();
                }
                else{
                    $object = Corpus::where([
                        ['file_name', '=',$fileName],
                        ['id', '=',$corpusId]
                    ])->get();
                }
                break;
            case 'document':
                if($isDir){
                    $object = Document::where([
                        ['directory_path', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }
                else{
                    $object = Document::where([
                        ['file_name', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }

                break;
            case 'annotation':
                if($isDir){
                    $object = Annotation::where([
                        ['directory_path', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }
                else{
                    $object = Annotation::where([
                        ['file_name', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }

                break;
            case 'CORPUS-DATA':
                $object = CorpusFile::where([
                    ['file_name', '=',$fileName],
                    ['corpus_id', '=',$corpusId]
                ])->get();

                if(count($object) > 0){
                    $formatObject = Format::find($object[0]->format);
                    $object[0]->formatName = $formatObject->name;
                }

                break;
        }

        return $object;
    }

    /**
     *
     * getModelByFileAndCorpus: fetch an eloquent model by its filename, type  and corpusid
     *
     * @param $fileName
     * @param $type
     * @param $isDir
     * @param $corpusId
     * @return \Illuminate\Support\Collection|null
     */
    public function getModelByFileAndCorpus($fileName, $type, $isDir, $corpusId){
        $object = null;
        switch ($type){
            case 'corpus':
                if($isDir){
                    $object = Corpus::where([
                        ['directory_path', '=',$fileName],
                        ['id', '=',$corpusId]
                    ])->get();
                }
                else{
                    $object = Corpus::where([
                        ['file_name', '=',$fileName],
                        ['id', '=',$corpusId]
                    ])->get();
                }
                break;
            case 'document':
                if($isDir){
                    $object = Document::where([
                        ['directory_path', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }
                else{
                    $object = Document::where([
                        ['file_name', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }

                break;
            case 'annotation':
                if($isDir){
                    $object = Annotation::where([
                        ['directory_path', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }
                else{
                    $object = Annotation::where([
                        ['file_name', '=',$fileName],
                        ['corpus_id', '=',$corpusId]
                    ])->get();
                }

                break;
        }
        return $object;
    }


    /**
     *
     * getElasticSearchIdByCorpusId: get elasticsearch document _id associated with an eloquent object
     *
     * @param $corpusid
     * @param $corpus_index
     * @return |null
     */
    public function getElasticSearchIdByCorpusId($corpusid)
    {
        $elasticSearchId = "";
        $corpus = Corpus::find($corpusid);
        if(!empty($corpus)) {
            $elasticSearchId = $corpus->elasticsearch_id;
        }
        return $elasticSearchId;
    }

    /**
     *
     * getElasticSearchIndexByCorpusId: get elasticsearch document index id associated with an eloquent object
     *
     * @param $corpusid
     * @return mixed
     */
    public function getElasticSearchIndexByCorpusId($corpusid)
    {
        $corpus = Corpus::find($corpusid);

        $elasticsearch_index = "";
        if(isset($corpus)){
            $elasticsearch_index = $corpus->elasticsearch_index;
        }
        return $elasticsearch_index;
    }


    /**
     *
     * getDatabaseIdByCorpusId: get a Corpus Eloquent objects database id
     *
     * @param $corpusid
     * @return mixed
     */
    public function getDatabaseIdByCorpusId($corpusid,$corpus_index)
    {
        $corpus = Corpus::where([["corpus_id","=",$corpusid],["elasticsearch_index","=",$corpus_index]])->get();

        $databaseId = "";
        if(isset($corpus[0])){
            $databaseId = $corpus[0]->id;
        }
        return $databaseId;
    }


    /**
     *
     * getCorpusLogoByCorpusId: get the logo image file path of a corpus Eloquent object
     *
     * @param $corpusid
     * @param $index
     * @return string
     */
    public function getCorpusLogoByCorpusId($corpusid,$index){
        $corpus_logo = "";
        $corpus = Corpus::where([["elasticsearch_id","=",$corpusid],["elasticsearch_index","=",$index]])->get();

        if(isset($corpus[0])){
            $corpus_logo = $corpus[0]->corpus_logo;
        }

        return $corpus_logo;
    }


    /**
     *
     * getCorpusNameByCorpusId
     *
     * @param $corpusid
     * @param $index
     * @return string
     */
    public function getCorpusNameByCorpusId($corpusid){
        $corpus_name = "";
        $corpus = Corpus::find($corpusid);

        if(isset($corpus)){
            $corpus_name = $corpus->name;
        }

        return $corpus_name;
    }


    /**
     *
     * getCorpusNameByObjectElasticsearchId: get the name of a corpus of any eloquent object by its elasticsearch id field
     *
     * @param $type
     * @param $objectId
     * @return string
     */
    public function getCorpusNameByObjectElasticsearchId($type,$objectId){
        $corpus_name = "";
        $object = null;
        switch($type) {
            case "corpus":
                $object = Corpus::where("elasticsearch_id",$objectId)->get();
                if(isset($object[0])){
                    $corpus_name = $object[0]->name;
                }
                break;
            case "document":
                $object = Document::where("elasticsearch_id",$objectId)->get();
                if(isset($object[0])){
                    $corpus = Corpus::find($object[0]->corpus_id);
                    if(isset( $corpus->name)) {
                        $corpus_name = $corpus->name;
                    }
                }
                break;
            case "annotation":
                $object = Annotation::where("elasticsearch_id",$objectId)->get();
                if(isset($object[0])){
                    $corpus = Corpus::find($object[0]->corpus_id);
                    if(isset( $corpus->name)) {
                        $corpus_name = $corpus->name;
                    }
                }
                break;
        }//end switch

        return $corpus_name;
    }

    /**
     *
     * getCurrentCorpusIndexByElasticsearchId
     *
     * @param $elasticSearchId
     * @return mixed
     */
    public function  getCurrentCorpusIndexByElasticsearchId($elasticSearchId) {
        $corpus = Corpus::where([
            ["elasticsearch_id","=",$elasticSearchId]
        ])->get();

        $elasticsearch_index = "";
        if(count($corpus) > 0){
            $elasticsearch_index = $corpus[0]->elasticsearch_index;
        }

        return $elasticsearch_index;
    }

    /**
     *
     * getCurrentDocumentIndexByElasticsearchId
     *
     * @param $elasticSearchId
     * @return mixed
     */
    public function getCurrentDocumentIndexByElasticsearchId($elasticSearchId){
        $document = Document::where([
            ["elasticsearch_id","=",$elasticSearchId]
        ])->get();


        $elasticsearch_index = "";
        if(count($document) > 0){
            $elasticsearch_index = $document[0]->elasticsearch_index;
        }

        return $elasticsearch_index;
    }

    /**
     *
     * getCurrentAnnotationIndexByElasticsearchId
     *
     * @param $elasticSearchId
     * @return mixed
     */
    public function getCurrentAnnotationIndexByElasticsearchId($elasticSearchId){
        $annotation = Annotation::where([
            ["elasticsearch_id","=",$elasticSearchId]
        ])->get();

        $elasticsearch_index = "";
        if(count($annotation) > 0){
            $elasticsearch_index = $annotation[0]->elasticsearch_index;
        }
        return $elasticsearch_index;
    }


    /**
     *
     * getCurrentCorpusIndexByAnnotationElasticsearchId
     *
     * @param $elasticSearchId
     * @return mixed
     */
    public function getCurrentCorpusIndexByAnnotationElasticsearchId($elasticSearchId){
        $annotation = Annotation::where([
            ["elasticsearch_id","=",$elasticSearchId]
        ])->get();



        $elasticsearch_index = "";
        $corpus = null;

        if(count($annotation) > 0){
            $corpus = $annotation[0]->corpus()->get();
            if(count($corpus) > 0){
                $elasticsearch_index = $corpus[0]->elasticsearch_index;
            }
        }

        return $elasticsearch_index;
    }

    public function getCurrentCorpusIndexByAnnotationElasticsearchIdAndIndex($elasticSearchId,$index){
        $annotation = Annotation::where([
            ["elasticsearch_id","=",$elasticSearchId],
            ["elasticsearch_index","=",$index]

        ])->get();



        $elasticsearch_index = "";
        $corpus = null;

        if(count($annotation) > 0){
            $corpus = $annotation[0]->corpus()->get();
            if(count($corpus) > 0){
                $elasticsearch_index = $corpus[0]->elasticsearch_index;
            }
        }

        return $elasticsearch_index;
    }


    /**
     *
     * getCurrentCorpusIndexByDocumentElasticsearchId
     *
     * @param $elasticSearchId
     * @return mixed
     */
    public function getCurrentCorpusIndexByDocumentElasticsearchId($elasticSearchId){
        $document = Document::where([
            ["elasticsearch_id","=",$elasticSearchId]
        ])->get();

        $elasticsearch_index = "";
        $corpus = null;

        if(count($document) > 0){
            $corpus = $document[0]->corpus()->get();

            if(count($corpus) > 0){
                $elasticsearch_index = $corpus[0]->elasticsearch_index;
            }
        }

        return $elasticsearch_index;
    }

    public function getCurrentCorpusIndexByDocumentElasticsearchIdAndIndex($elasticSearchId, $index){
        $document = Document::where([
            ["elasticsearch_id","=",$elasticSearchId],
            ["elasticsearch_index","=",$index]
        ])->get();

        $elasticsearch_index = "";
        $corpus = null;

        if(count($document) > 0){
            $corpus = $document[0]->corpus()->get();

            if(count($corpus) > 0){
                $elasticsearch_index = $corpus[0]->elasticsearch_index;
            }
        }

        return $elasticsearch_index;
    }


    /**
     *
     * getDocumentGenreByCorpusId: fetch a corpus' document genre by traversing all its document objects
     *
     * @param $corpusid
     * @param $index
     * @return string
     */
    public function getDocumentGenreByCorpusId($corpusid,$index)
    {
        $genre = "N/A";

        $corpus = Corpus::where([["corpus_id","=",$corpusid],["elasticsearch_index","=",$index]])->get();
        if(isset($corpus[0])){
            $documents = $corpus[0]->documents()->get();
            $genre = isset($documents[0]) ? $documents[0]->document_genre : "N/A";
        }

        return $genre;
    }

    /**
     *
     * getCorpusVersion: get the publication version of a corpus eloquent object
     *
     * @param $corpusId
     * @return mixed
     */
    public function getCorpusVersion($corpusId){
        $corpus = Corpus::find($corpusId);
        $publication_version = 0;
        if(isset($corpus)){
            $publication_version = $corpus->publication_version;
        }
        return $publication_version;
    }


    /**
     *
     * getWorkFlowStatus: get the workflow status of a corpus eloquent object
     *
     * @param $corpusId
     * @return mixed
     */
    public function getWorkFlowStatus($corpusId){
        $corpus = Corpus::find($corpusId);
        $workflow_status = 0;
        if(isset($corpus)){
            $workflow_status = $corpus->workflow_status;
        }
        return $workflow_status;
    }


    /**
     *
     * getInvitedByCorpus: get corpus collaboration invites
     *
     * @param $corpusId
     * @return array
     */
    public function getInvitedByCorpus($corpusId) {
        $invited = array();
        $invited = Invite::where([["corpus_id", "=", $corpusId],["accepted", "=", 0]])->get();
        return $invited;
    }

    public function getInvitedByCorpusProject($projectId) {
        $invited = array();
        $invited = Invite::where([["project_id", "=", $projectId],["accepted", "=", 0]])->get();
        return $invited;
    }


    /**
     *
     * getCorpusProjectPathByCorpusId: get the file path of a corpus' project
     *
     * @param $corpusId
     * @param $corpus_index
     * @return string
     */
    public function getCorpusProjectPathByCorpusId($corpusId,$corpus_index){
        $projectPath = "";
        $corpus = Corpus::where([["elasticsearch_id","=",$corpusId],["elasticsearch_index","=",$corpus_index]])->get();
        if(isset($corpus[0])){
            $corpusprojects = $corpus[0]->corpusprojects()->get();
            $project = $corpusprojects[0];
            $projectPath = $project->directory_path;
        }

        return $projectPath;
    }


    /**
     *
     * getCorpusAndProjectPathByCorpusId: get the file path of a corpus' project and the corpus
     *
     * @param $corpusid
     * @param $corpus_index
     * @return string
     */
    public function getCorpusAndProjectPathByCorpusId($corpusid){
        $corpusPath = "";
        $corpus = Corpus::find($corpusid);

        if(isset($corpus)){
            $corpusprojects = $corpus->corpusprojects()->get();
            $project = $corpusprojects[0];
            $corpusPath = $project->directory_path."/".$corpus->directory_path;
        }
        return $corpusPath;
    }


    /**
     *
     * getCorpusPathByCorpusId: get the file path of a corpus
     *
     * @param $corpusid
     * @param $corpus_index
     * @return string
     */
    public function getCorpusPathByCorpusId($corpusid,$corpus_index){
        $corpusPath = "";
        $corpus = Corpus::where([["elasticsearch_id","=",$corpusid],["elasticsearch_index","=",$corpus_index]])->get();
        if(isset($corpus[0])){
            $corpusPath = $corpus[0]->directory_path;
        }

        return $corpusPath;
    }


    public function getCorpusTokensByCorpusId($corpusid,$corpus_index){
        $tokens = 0;
        $corpus = Corpus::where([["corpus_id","=",$corpusid],["elasticsearch_index","=",$corpus_index]])->get();
        if(isset($corpus[0])){
            $tokens = $corpus[0]->corpus_size_value;
        }

        return $tokens;
    }

    public function getCorpusGitLabIdByElasticsearchId($elasticsearchId,$corpus_index){
        $gitlabid = 0;
        $corpus = Corpus::where([["elasticsearch_id","=",$elasticsearchId],["elasticsearch_index","=",$corpus_index]])->get();
        if(isset($corpus[0])){
            $gitlabid = $corpus[0]->gitlab_id;
        }

        return $gitlabid;
    }

    /**
     *
     * deleteModels: delete eloquent models by file path
     *
     * @param $path
     */
    public function deleteModels($path){
        $dirArray = explode("/",$path);
        $type = $dirArray[3];
        $objects = null;
        switch ($type) {
            case 'corpus':
                $objects = DB::table('corpuses')->where('directory_path',$dirArray[1])->get();
                break;
            case 'document':
                $objects = DB::table('documents')->where('directory_path',$dirArray[1])->get();
                break;
            case 'annotation':
                $objects = DB::table('annotations')->where('directory_path',$dirArray[1])->get();
                break;
        }
        foreach ($objects->toArray() as $object){
            $this->deleteModel($type, $object->id);
        }
    }


    /**
     *
     * deleteModel: delete eloquent model by type and database id
     *
     * @param $type
     * @param $id
     */
    public function deleteModel($type,$id){
        $object = null;
        switch ($type) {
            case 'corpus':
                $object = Corpus::find($id);
                break;
            case 'document':
                $object = Document::find($id);
                if(count($object->annotations()) > 0) {
                    $object->annotations()->detach();
                }

                $object->delete();
                break;
            case 'annotation':
                $object = Annotation::find($id);
                if(count($object->documents()) > 0) {
                    $object->documents()->detach();
                }
                $object->preparations()->delete();
                $object->delete();
                break;
        }
    }


    /**
     *
     * updateDirectoryPaths: update Directory Paths for Document and Annotation eloquent objects
     *
     * @param $directory_path
     * @param $corpusId
     */
    public function updateDirectoryPaths($directory_path,$corpusId){
        DB::update("update documents set directory_path = ? where corpus_id = ?", [$directory_path,$corpusId]);
        DB::update("update annotations set directory_path = ? where corpus_id = ?", [$directory_path,$corpusId]);
    }


    /**
     *
     * getDirectoryPath: get directorypaths by collection of paths and filename
     *
     * @param $paths
     * @param $fileName
     * @return string
     */
    public function getDirectoryPath($paths,$fileName){
        $directoryPath = "";
        foreach ($paths as $path) {
            if (strpos($path,$fileName) !== false){
                $directoryPath = $path;
                break;
            }
        }
        return $directoryPath;
    }


    /**
     * @param $corpusresponses
     * @param $elasticService
     * @param $perPage
     * @param $sortKriterium
     * @param $currentPage
     * @return array
     */
    public function getPublishedCorpusData($corpusresponses, $elasticService, $perPage ,$sortKriterium, $currentPage){

        $responseArray = array(
            "entries" => array(),
            "totalcount" => 0,
            "perPageArray" => array(),
            "perPage" => $perPage,
        );

        $corpusdata = array();
        $entries = null;
        $perPageArray = array();
        $sortedCollection = array();

        $current_corpus_index =  config("laudatio.elasticsearch_corpus_index","corpora");
        $current_document_index = config("laudatio.elasticsearch_document_index","documents");
        $current_annotation_index =  config("laudatio.elasticsearch_annotation_index","annotations");

        $documentcount = 0;
        $annotationcount = 0;

        if(count($corpusresponses['result']) > 0){
            $document_range = "";
            //dd($corpusresponses['result']);
            foreach($corpusresponses['result'][0] as $publicationresponse){
                $current_git_repo_id = "";
                //dd($publicationresponse);
                if(isset($publicationresponse['_source']['has_documents'])) {
                    $documentcount = count($publicationresponse['_source']['has_documents']);
                }

                if(isset($publicationresponse['_source']['has_annotations'])) {
                    $annotationcount = count($publicationresponse['_source']['has_annotations']);
                }


                if(isset($publicationresponse['_source']['laudatio_corpus_gitlab_id'])) {
                    $current_git_repo_id = $publicationresponse['_source']['laudatio_corpus_gitlab_id'];
                }



                $documentResult = $elasticService->getDocumentByCorpus(
                    $publicationresponse['_source']['has_documents'],
                    array($publicationresponse['_id']),
                    array("document_title","document_publication_publishing_date","document_publication_place","document_list_of_annotations_name","in_corpora","document_size_extent"),
                    $current_document_index
                );


                if(!empty($current_corpus_index)){

                   // if(!array_key_exists($publicationresponse['_source']['corpus'],$corpusdata)){


                        $publishedCorpusid = $publicationresponse['_id'];//$this->getElasticSearchIdByCorpusId($publicationresponse['_id']['corpus'],$current_corpus_index);

                        if(!empty($publishedCorpusid)){
                            $apiData = $elasticService->getCorpus($publishedCorpusid,true,$current_corpus_index);
                            $corpusresponse = json_decode($apiData->getContent(), true);
                            $document_range = $this->getDocumentRange($corpusresponse,$documentResult);

                            $authors = "";
                            for($i = 0; $i < count($corpusresponse['result']['corpus_editor_forename']); $i++){
                                $authors .= $corpusresponse['result']['corpus_editor_surname'][$i].", ".$corpusresponse['result']['corpus_editor_forename'][$i]."; ";
                            }

                            $authors = substr($authors,0,strrpos($authors,";"));

                            $corpus_publication_date = "1900";
                            for($j = 0; $j < count($corpusresponse['result']['corpus_publication_publication_date']); $j++) {
                                $date =   $corpusresponse['result']['corpus_publication_publication_date'][$j];
                                if(strpos($date,"-") !== false) {
                                    $datearr = explode("-",$date);
                                    $date = $datearr[0];
                                }

                                if($date > $corpus_publication_date) {
                                    $corpus_publication_date = $date;
                                }
                            }

                            $corpusDatabaseId = $publicationresponse['_source']['laudatio_corpus_id'];

                            $corpus_path = $this->getCorpusAndProjectPathByCorpusId($corpusDatabaseId);
                            $corpusPathArray = explode("/",$corpus_path);

                            $corpusLogo = "";
                            $corpusLogobyCorpusid = $this->getCorpusLogoByCorpusId($publicationresponse['_id'],$current_corpus_index);

                            if(null != $corpusLogobyCorpusid) {
                                $corpusLogo = $corpusLogobyCorpusid;
                            }
                            $data['result']['corpus_logo'] = $corpusLogo;


                            $corpusFiles = $this->getFormatFilesByCorpusId($publicationresponse['_source']['laudatio_corpus_id']);

                            $downloadPath = "";

                            /*
                            if(file_exists($this->basePath."/".$corpusPathArray[0]."/".$corpusPathArray[1])){
                                $downloadPath = $corpusPathArray[0]."/".$corpusPathArray[1];
                            }
                            */
                            $corpusPathArray = array();
                            $project_path_string = "";
                            $corpus_path_string = "";
                            if(!empty($corpus_path)){
                                $corpusPathArray = explode("/",$corpus_path);
                                if(count($corpusPathArray) >= 2 && file_exists($this->basePath."/".$corpusPathArray[0]."/".$corpusPathArray[1])){
                                    $downloadPath = $corpusPathArray[0]."/".$corpusPathArray[1];
                                    $project_path_string = $corpusPathArray[0];
                                    $corpus_path_string = $corpusPathArray[1];
                                }
                            }


                            //count($publicationresponse['_source']['corpus_encoding_project_description']) -1
                            $corpusdata[$publicationresponse['_id']] = array(
                                'corpus_id' => $corpusDatabaseId,
                                'corpus_index' => $current_corpus_index,
                                'corpus_title' => $publicationresponse['_source']['corpus_title'][0],
                                'is_latest_publication' => $publicationresponse['_source']['is_latest_publication'],
                                'publication_status' => $publicationresponse['_source']['publication_status'],
                                'corpus_version' => $publicationresponse['_source']['publication_version'],
                                'laudatio_corpus_gitlab_id' => $current_git_repo_id,
                                'corpus_logo' => $corpusLogo,
                                //'corpus_project_directorypath' => $corpusPathArray[0],
                                //'corpus_directorypath' => $corpusPathArray[1],
                                'corpus_project_directorypath' => $project_path_string,
                                'corpus_directorypath' => $corpus_path_string,
                                'tei_download_path' => $downloadPath,
                                'corpus_path' => $corpus_path,
                                'corpus_formatdata' => $corpusFiles,
                                'authors' => $authors,
                                'corpus_languages_language' => $corpusresponse['result']['corpus_languages_language'][0],
                                'corpus_size_value' => str_replace(array(',','.'),'',$corpusresponse['result']['corpus_size_value'][0]),
                                'corpus_publication_date' => $corpus_publication_date,
                                'corpus_publication_license' => $corpusresponse['result']['corpus_publication_license'][0],
                                'corpus_encoding_project_description' => $this->pickCorpusDescription($publicationresponse['_source']['corpus_encoding_project_description']),
                                'corpus_document_genre' => isset($publicationresponse['_source']['corpus_document_genre']) ? $publicationresponse['_source']['corpus_document_genre'][0]: "NA",
                                'document_publication_range' => $document_range,
                                'documentcount' => $documentcount,
                                'annotationcount' => $annotationcount,
                                'elasticid' => $publishedCorpusid
                            );
                        }
                    //}
                }
            }

            $latestPublishedCorpora = array();

            foreach($corpusdata as $corpusid => $corpusItem) {
                if($corpusItem['is_latest_publication'] == "true" && $corpusItem['publication_status'] == 1) {
                    $latestPublishedCorpora[$corpusid] = $corpusItem;
                }
            }

            $latestCollection = new Collection($latestPublishedCorpora);

            $kriterium = null;
            $sortKriteria = array(
                "1" => "corpus_title",
                "2" => "corpus_size_value",
                "3_desc" => "corpus_size_value",
                "4" => "corpus_publication_date",
                "5_desc" => "corpus_publication_date",
                "6" => "document_publication_range",
                "7_desc" => "document_publication_range"
            );

            if(!isset($sortKriterium)) {
                $kriterium = "corpus_title";
            }
            else{
                switch($sortKriterium){
                    case 3:
                    case 5:
                    case 7:
                        $sortKriterium .= "_desc";
                        break;
                }
                $kriterium = $sortKriteria[$sortKriterium];
            }


            if(!isset($perPage)) {
                $perPage  = 4;
            }

            $sortedCollection = null;
            $latestSortedCollection = null;

            if(strpos($sortKriterium, "desc") !== false) {
                $latestSortedCollection = $latestCollection->sortByDesc($kriterium, SORT_NATURAL|SORT_FLAG_CASE);
            }
            else{
                $latestSortedCollection = $latestCollection->sortBy($kriterium, SORT_NATURAL|SORT_FLAG_CASE);
            }



            $perPageArray = array(
                $perPage => "",
                "4" => "",
                "6" => "",
                "12" => "",
                "18" => "",
                "all" => ""
            );

            if($perPage == count($latestSortedCollection)){
                $perPageArray['all'] = "selected";
            }
            else{
                $perPageArray[$perPage] = "selected";
            }

            if($perPage == "all") {
                $perPage = count($latestSortedCollection) > 0 ? count($latestSortedCollection) : 4;
            }

            $currentPageSearchResults = $latestSortedCollection->slice(($currentPage - 1) * $perPage, $perPage)->all();
            $entries = new LengthAwarePaginator($currentPageSearchResults, count($latestSortedCollection), $perPage, $currentPage,['path' => LengthAwarePaginator::resolveCurrentPath()] );

            $responseArray = array(
                "entries" => $entries,
                "allpublished" => $sortedCollection,
                "totalcount" => count($latestSortedCollection),
                "perPageArray" => $perPageArray,
                "perPage" => $perPage
            );
        }

        return $responseArray;
    }

    public function pickCorpusDescription($descriptions) {
        $returned_description = "";
        foreach ($descriptions as $description) {
            if(strlen($description) > 20) {
                $returned_description = $description;
                break;
            }
        }

        return $returned_description;
    }

    /**
     *
     * buildCiteFormat creates citations for a given piece of data
     *
     * @param $data
     * @return array
     */
    public function buildCiteFormat($data){
        $cite = "";
        $citedauthors = "";
        $authorstring = "";

        foreach ($data['authors'] as $authorData) {
            $namearray = explode(" ", $authorData);
            $citedauthors .= $namearray[1].", ";
            $authorstring .= $namearray[1]." ".$namearray[0].", ";
        }


        $homepage = $data['corpus_encoding_project_homepage'];
        $handle = $data['published_handle'];

        $citedauthors = substr($citedauthors,0,strrpos($citedauthors,","));
        $authorstring = substr($authorstring,0,strrpos($authorstring,","));
        $authorstring .= ' ('. $data['publishing_year'].')';
        $citedauthors .= ' ('. $data['publishing_year'].')';

        $APAcite = $namearray[1].", ".
            substr($namearray[0],0,1).
            ". (".$data['publishing_year']."). ".
            $data['title']." (".$data['version']."). ".
            $data['publishing_institution'].".".
            " Homepage: ".$homepage.".";
            " DOI: ".$handle.".";


        $CHIAGOcite = $namearray[1].", ".
            $namearray[0]."."."´".$data['title']." (".$data['version'].")´ ".$data['publishing_institution'].", ".$data['publishing_year']."."." Homepage: ".$homepage."."." DOI: ".$handle.".";


        $HARVARDcite = $namearray[1].", ".
            substr($namearray[0],0,1).
            ". (".$data['publishing_year']."). ".
            "´".$data['title']." (".$data['version'].")´ ".
            $data['publishing_institution'].
            " Homepage: ".$homepage.".".
            " DOI: ".$handle.".";



        $BibTexcite = "";
        $BibTexcite .= "@Misc{".$citedauthors.",\n";
        $BibTexcite .= "\t author \t = {".$authorstring."}, \n";
        $BibTexcite .= "\t title \t = {{".$data['title']." (Version ".$data['version'].")}}, \n";
        $BibTexcite .= "\t year \t = {".$data['publishing_year']."}, \n";
        $BibTexcite .= "\t note \t = {".$data['publishing_institution']."}, \n";
        $BibTexcite .= "\t url \t = {".$handle."}, \n";
        $BibTexcite .= "}";

        $TXTcite = "";
        //$TXTcite .= $citedauthors."\n";
        //$TXTcite .= $authorstring."\n";
        $TXTcite .= $authorstring."; ";
        //$TXTcite .= $data['title']." (Version ".$data['version'].")\n";
        $TXTcite .= $data['title']." (Version ".$data['version']."); ";
        //$TXTcite .= $data['publishing_year']."\n";
        //$TXTcite .= $data['publishing_institution']."\n";
        $TXTcite .= $data['publishing_institution']."; ";
        $TXTcite .= " Homepage: ".$homepage."; ";
        $TXTcite .= " DOI: ".$handle."\n";

        $citations = array();
        $citations['apa'] = $APAcite;
        $citations['chicago'] = $CHIAGOcite;
        $citations['harvard'] = $HARVARDcite;
        $citations['bibtex'] = $BibTexcite;
        $citations['txt'] = $TXTcite;

        return $citations;
    }


    /**
     *
     * getDocumentRange
     *
     * @param $data
     * @param $documentResult
     * @return mixed|string
     */
    public function getDocumentRange($data, $documentResult)
    {
        $document_dates = array();
        $document_range = "";

        for($d = 0; $d < count($documentResult); $d++) {
            $doc = $documentResult[$d];
            if(isset($doc['document_publication_publishing_date'][0])){
                array_push($document_dates, $doc['document_publication_publishing_date'][0]);
            }
        }//end for

        sort($document_dates);

        if($document_dates[count($document_dates) -1] > $document_range = $document_dates[0]) {
            $document_range = $document_dates[0]." - ".$document_dates[count($document_dates) -1];
        }
        else{
            $document_range = $document_dates[0];
        }

        return $document_range;

    }


    /**
     *
     * emptyCorpusCache: empty the corpus redis cache
     * @param $corpusId
     * @param $index
     * @return int
     */
    public function emptyCorpusCache($corpusId, $index){
        $result = 0;
        Cache::tags(['corpus_'.$corpusId.'_'.$index])->flush();

        Cache::tags(['formats_'.$corpusId.'_'.$index])->flush();

        $indexes = $this->transformElasticSearchIndexes($index);
        $guidelineIndex = $indexes["guidelineIndex"];
        Cache::tags(['guidelines_'.$corpusId.'_'.$guidelineIndex])->flush();

        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $corpora =  $redis->keys('*corpora*');

        foreach ($corpora as $lkey){
            $result = $redis->del($lkey);
        }

        return $result;
    }
    public function emptyDocumentCache(){
        $result = 0;

        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $documents =  $redis->keys('*document*');

        foreach ($documents as $dkey){
            $result = $redis->del($dkey);
        }

        return $result;
    }

    public function emptyAnnotationCache(){
        $result = 0;

        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $documents =  $redis->keys('*annotation*');

        foreach ($documents as $akey){
            $result = $redis->del($akey);
        }

        return $result;
    }

    /**
     * emptySearchCache
     */
    public function emptySearchCache() {
        //empty published search cache
        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');

        $publishedCorpora = $redis->keys('*latest_published_corpora*');
        foreach ($publishedCorpora as $pkey){
            Log::info("emptySearchCache: Deleting latest_published_corpora: ".$pkey);
            $redis->del($pkey);
        }

        $publishedDocuments = $redis->keys('*latest_published_documents_*');
        foreach ($publishedDocuments as $dkey){
            Log::info("emptySearchCache: Deleting latest_published_documents_: ".$dkey);
            $redis->del($dkey);
        }

        $publishedAnnotations = $redis->keys('*latest_published_annotations_*');
        foreach ($publishedAnnotations as $akey){
            Log::info("emptySearchCache: Deleting latest_published_annotations_: ".$akey);
            $redis->del($akey);
        }

    }


    /**
     * emptyCorpusProjectCache
     * @return int
     */
    public function emptyCorpusProjectCache() {
        $result = 0;
        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $listAllCorpusProjects = $redis->keys('*corpusproject*');

        foreach ($listAllCorpusProjects as $lkey){
            $result = $redis->del($lkey);
        }

        return $result;
    }

    /**
     *
     * emptyCorpusDataFileCache: empty redis corpus data file cache
     *
     * @param $path
     * @param $corpusId
     */
    public function emptyCorpusDataFileCache($path,$corpusId){
        Log::info("TRYING FOR CORPUS-DATA: getCorpusDataFiles_".$corpusId.$path);
        Cache::forget("getCorpusDataFiles_".$corpusId.$path);
        /*
        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $listAllCorpusDatafiles = $redis->keys('*getCorpusDataFiles*');
        Log::info("emptyCorpusDataFileCache: listAllCorpusDatafiles: ".print_r($listAllCorpusDatafiles,1));
        foreach ($listAllCorpusDatafiles as $lkey){
            if(strrpos($lkey,'getCorpusDataFiles_'.$corpusId) > -1){
                Log::info("STRPOS FOUND ".'getCorpusDataFiles_'.$corpusId." EMPTYING: ".$lkey);
                $result = $redis->del($lkey);
            }
        }
        */
    }

    public function emptyFlySystemFileCache(){
        $result = 0;
        $redis = Cache::getRedis();
        $redis->connect('xxxxxx', 6379); // use your Host from Redis desktop Manager - connection
        $redis->select('1');
        $listAllCorpusDatafiles = $redis->keys('*flysystem*');

        foreach ($listAllCorpusDatafiles as $lkey){
            $result = $redis->del($lkey);
        }
        return $result;
    }


    /**
     *
     * emptyCorpusFileCache: empty redis corpus header file cache
     *
     * @param $path
     * @param $corpusId
     */
    public function emptyCorpusFileCache($path,$corpusId){
        Cache::forget("getCorpusFiles_".$corpusId.$path);
    }



    public function emptyFileCaches($corpusId){

        $corpusPath = $this->getCorpusAndProjectPathByCorpusId($corpusId);

        $corpusFilePath = $corpusPath."/TEI-HEADERS/corpus";
        $this->emptyCorpusFileInfoCache($corpusFilePath,$corpusId);
        $this->emptyCorpusFileCache($corpusFilePath,$corpusId);

        $documentFilePath = $corpusPath."/TEI-HEADERS/document";
        $this->emptyCorpusFileInfoCache($documentFilePath,$corpusId);
        $this->emptyCorpusFileCache($documentFilePath,$corpusId);

        $annotationFilePath = $corpusPath."/TEI-HEADERS/annotation";
        $this->emptyCorpusFileInfoCache($annotationFilePath,$corpusId);
        $this->emptyCorpusFileCache($annotationFilePath,$corpusId);

        $dataFilePath = $corpusPath."/CORPUS-DATA";
        $this->emptyCorpusDataFileCache($dataFilePath,$corpusId);

        $teiHeaderPath = $corpusPath."/TEI-HEADERS";
        $this->emptyCorpusFileInfoCache($teiHeaderPath,$corpusId);
        $this->emptyCorpusFileCache($teiHeaderPath,$corpusId);
    }

    /**
     *
     * emptyCorpusFileInfoCache: empty redis corpus header file info cache
     *
     * @param $path
     * @param $corpusId
     */
    public function emptyCorpusFileInfoCache($path,$corpusId){
        Cache::forget("getCorpusFileInfo_".$corpusId.$path);
    }


    /**
     *
     * emptyDocumentCacheByCorpusId: empty redis eloquent document object cache by corpus id
     *
     * @param $corpusId
     * @param $index
     */
    public function emptyDocumentCacheByCorpusId($corpusId,$index){

        Cache::tags(['document_'.$corpusId.'_'.$index])->flush();

        // @todo: flushes too all docs, and not only the relevant ones ?

    }

    /**
     *
     * emptyDocumentCacheByDocumentIndex: empty redis eloquent document object cache
     *
     * @param $documentIndex
     */
    public function emptyDocumentCacheByDocumentIndex($documentIndex){
        Cache::tags(['document_'.$documentIndex])->flush();
    }

    /**
     *
     * emptyDocumentCacheByDocumentElasticsearchId
     *
     * @param $documentId
     * @param $documentIndex
     */
    public function emptyDocumentCacheByDocumentElasticsearchId($documentId, $documentIndex){
        Cache::tags(['document_'.$documentId.'_'.$documentIndex])->flush();
    }

    /**
     *
     * emptyAnnotationCacheByCorpusId
     *
     * @param $corpusId
     * @param $index
     */
    public function emptyAnnotationCacheByCorpusId($corpusId, $index){
        Cache::tags(['annotation_'.$corpusId.'_'.$index])->flush();
    }


    /**
     *
     * emptyAnnotationCacheByAnnotationIndex
     *
     * @param $index
     */
    public function emptyAnnotationCacheByAnnotationIndex($index){
        Cache::tags(['annotation_'.$index])->flush();
    }

    /**
     *
     * emptyAnnotationCacheByNameAndCorpusId
     *
     * @param $name
     * @param $corpusId
     * @param $index
     */
    public function emptyAnnotationCacheByNameAndCorpusId($name, $corpusId, $index) {
        Cache::tags([ 'annotation_'.$name.'_'.$corpusId.'_'.$index])->flush();
    }


    /**
     *
     * emptyAnnotationGroupCacheByAnnotationAndCorpusId
     *
     * @param $annotationId
     * @param $corpusId
     * @param $index
     */
    public function emptyAnnotationGroupCacheByAnnotationAndCorpusId($annotationId,$corpusId, $index){
        Cache::tags(['annotationgroup_'.$corpusId.'_'.$index])->flush();
    }


    /**
     * checkForDuplicateCorpusName checks the path-normalized name of a corpus if it already exists with in a project
     *
     * @param $corpus_name_path
     * @param $corpus_project_id
     * @return mixed
     */
    public function checkForDuplicateCorpusName($corpus_name, $corpus_name_path, $corpus_project_id) {
        $isDuplicate = false;
        $corpusProject = CorpusProject::findOrFail($corpus_project_id);
        foreach ($corpusProject->corpora()->get() as $projectCorpus) {
            if($projectCorpus->directory_path == $corpus_name_path) {

                throw new CorpusNameAlreadyExistsException("The Corpus name $corpus_name is already taken within this Corpus project. Please select another title for the Corpus.",0,null);
            }
        }
        
        return $isDuplicate;
    }

    /**
     * determineAdminRole
     *
     * @param $admin_roles
     * @return array
     */
    public function determineAdminRole($admin_roles){
     $admin_role_filtered = array();

     $lastRoleId = 1000;
     foreach ($admin_roles as $id => $admin_role_array) {
         foreach ($admin_role_array as  $admin_role) {
             if ($admin_role['role_id'] < $lastRoleId) {
                 $admin_role_filtered = $admin_role;
             }
             $lastRoleId = $admin_role['role_id'];
         }

     }

     return $admin_role_filtered;
    }

    public function getRoleMappingByUser($user){
        $role_mapping = array(
            "corpus_project_roles" => array(),
            "corpus_roles" => array()
        );
        $corpus_project_role_data = DB::table('corpus_project_user')->select('corpus_project_id', 'role_id')->where([
                ['user_id','=',$user->id],
            ]
        )->get();

        foreach($corpus_project_role_data as $corpus_project_role_datum) {
            $role_mapping['corpus_project_roles'][$corpus_project_role_datum->corpus_project_id] = $corpus_project_role_datum->role_id;
        }

        $corpus_role_data = DB::table('corpus_user')->select('corpus_id', 'role_id')->where([
                ['user_id','=',$user->id],
            ]
        )->get();

        foreach($corpus_role_data as $corpus_role_datum) {
            $role_mapping['corpus_roles'][$corpus_role_datum->corpus_id] = $corpus_role_datum->role_id;
        }

        return $role_mapping;
    }

    /**
     *
     * determineUserAdminRole
     *
     * @param $user_roles
     * @return array
     */
    public function determineUserAdminRole($user_roles){
        $user_role_filtered = array();


        foreach ($user_roles as $id => $user_role_array) {
            $lastRoleId = 1000;
            if(!isset($user_role_filtered[$id])){
                $user_role_filtered[$id] = array();
            }
            foreach ($user_role_array as  $user_role) {
                if ($user_role['role_id'] < $lastRoleId) {
                    array_push($user_role_filtered[$id],$user_role);
                }
                $lastRoleId = $user_role['role_id'];
            }

        }

        return $user_role_filtered;
    }


    /**
     *
     * removeMergedDuplicates: remove duplicates from comma separated strings generated from merged arrays
     *
     * @param $dataString
     * @return string
     */
    public function removeMergedDuplicates($dataString){
        $stringArray = explode(",",$dataString);
        return join(",",array_unique($stringArray));
    }

    /**
     *
     * removeMergedNA: remove NA-entries from comma separated strings generated from merged arrays
     *
     * @param $dataString
     * @return string
     */
    public function removeMergedNA($dataString){
        $stringArray = explode(",",$dataString);
        return join(",",array_diff($stringArray, array("NA")));
    }

    public function getFormatFilesByCorpusId($databaseId){
        $foundFormats = array();
        $corpusFiles = CorpusFile::where("corpus_id",$databaseId)->get();
        $corpusPath = $this->getCorpusAndProjectPathByCorpusId($databaseId);
        foreach($corpusFiles as $corpusFile) {
            if(!array_key_exists($corpusFile->format,$foundFormats)) {
                $format = Format::find($corpusFile->format);
                if(isset($format->name) && isset($format->id)) {
                    $foundFormats[$corpusFile->format] = array(
                        "format_name" => $format->name,
                        "format_id" => $format->id,
                        "format_files" => array()
                    );
                }

            }
            if(!empty($foundFormats[$corpusFile->format])) {
                array_push($foundFormats[$corpusFile->format]['format_files'],$corpusPath."/CORPUS-DATA/".$foundFormats[$corpusFile->format]['format_name']."/".$corpusFile->file_name);
            }

        }
        return $foundFormats;
    }

    public function stripQuotes($text) {
        return preg_replace('/^(\'(.*)\'|"(.*)")$/', '$2$3', $text);
    }

    public function json_validate($string) {
        if (is_string($string)) {
            @json_decode($string);
            return (json_last_error() === JSON_ERROR_NONE);
        }
        return false;
    }

    public function json_validator($string)
    {
        // decode the JSON data
        $result = json_decode($string);

        // switch and check possible JSON errors
        switch (json_last_error()) {
            case JSON_ERROR_NONE:
                $error = ''; // JSON is valid // No error has occurred
                break;
            case JSON_ERROR_DEPTH:
                $error = 'The maximum stack depth has been exceeded.';
                break;
            case JSON_ERROR_STATE_MISMATCH:
                $error = 'Invalid or malformed JSON.';
                break;
            case JSON_ERROR_CTRL_CHAR:
                $error = 'Control character error, possibly incorrectly encoded.';
                break;
            case JSON_ERROR_SYNTAX:
                $error = 'Syntax error, malformed JSON.';
                break;
            // PHP >= 5.3.3
            case JSON_ERROR_UTF8:
                $error = 'Malformed UTF-8 characters, possibly incorrectly encoded.';
                break;
            // PHP >= 5.5.0
            case JSON_ERROR_RECURSION:
                $error = 'One or more recursive references in the value to be encoded.';
                break;
            // PHP >= 5.5.0
            case JSON_ERROR_INF_OR_NAN:
                $error = 'One or more NAN or INF values in the value to be encoded.';
                break;
            case JSON_ERROR_UNSUPPORTED_TYPE:
                $error = 'A value of a type that cannot be encoded was given.';
                break;
            default:
                $error = 'Unknown JSON error occured.';
                break;
        }

        if ($error !== '') {
            // throw the Exception or exit // or whatever :)
            exit($error);
        }

        // everything is OK
        return $result;
    }

    public function transformElasticSearchIndexes($corpusIndex){
        $indexes = array(
            "corpusIndex" => $corpusIndex
        );
        $first_underscore = strpos($corpusIndex,"_");
        $postfix = substr($corpusIndex,$first_underscore,strlen($corpusIndex));
        $indexes["documentIndex"] = "document".$postfix;
        $indexes["annotationIndex"] = "annotation".$postfix;
        $indexes["guidelineIndex"] = "guideline".$postfix;
        return $indexes;
    }

    public function checkIfElasticsearchIdExists($header, $elasticSearchId,$elasticsearchIndex) {
        $idExists = false;
        if($header == "corpus"){
            $object = DB::table('corpuses')->select('id')->where([
                    ['elasticsearch_id','=',$elasticSearchId],
                    ['elasticsearch_index','=',$elasticsearchIndex],
                ]
            )->get();
        }
        else if($header == "document"){
            $object = DB::table('documents')->select('id')->where([
                    ['elasticsearch_id','=',$elasticSearchId],
                    ['elasticsearch_index','=',$elasticsearchIndex],
                ]
            )->get();
        }
        else if($header == "annotation"){
            $object = DB::table('annotations')->select('id')->where([
                    ['elasticsearch_id','=',$elasticSearchId],
                    ['elasticsearch_index','=',$elasticsearchIndex],
                ]
            )->get();
        }

        if(count($object) > 0 && isset($object[0]->id)){
            $idExists = true;
        }

        return $idExists;
    }


    public function hasHigherCorpusVersion($gitRepoId, $version){
        $publishedVersions = array();

        $versionBiggerObjects = DB::table('corpuses')->select('publication_version')->where([
                ['gitlab_id','=',$gitRepoId],['workflow_status', '=', 1],
                ['publication_version', '>', floatval($version)]
            ]
        )->get();

        foreach($versionBiggerObjects as $versionObject) {
            if(floatval($versionObject->publication_version) > floatval($version)) {
                array_push($publishedVersions,$versionObject->publication_version);
            }

        }

        return $publishedVersions;
    }

    public function hasHigherCorpusVersionByGitRepoId($indicesByCorpus,$gitrepoid){
        $latestIndexes = array();
        $lastversion = 0.0;
        foreach($indicesByCorpus as $corpusIndex => $corpusIndexData){
                if($corpusIndexData['git_repo_id'] == $gitrepoid) {
                    if(floatval($corpusIndexData['publication_version']) > $lastversion){
                        $corpusIndexData['corpus_index'] = $corpusIndex;
                        $latestIndexes = $corpusIndexData;
                    }
                    $lastversion = floatval($corpusIndexData['publication_version']);
                }

        }

        return $latestIndexes;
    }

    public function getGitRepoId($id, $index, $type)
    {
        $object = null;
        $repoId = "";
        if($type == "corpus"){
            $object = Corpus::where([["elasticsearch_id","=",$id],["elasticsearch_index","=",$index]])->get();
        }
        else if($type == "document"){
            $object = Document::where([["elasticsearch_id","=",$id],["elasticsearch_index","=",$index]])->get();
        }
        else if($type == "annotation"){
            $object = Annotation::where([["elasticsearch_id","=",$id],["elasticsearch_index","=",$index]])->get();
        }



        if(count($object) > 0){
            if($type == "corpus") {
                $repoId = $object[0]->gitlab_id;
            }
            else{
                $corpus = $object[0]->corpus()->get();
                if(count($corpus) > 0) {
                    $repoId = $corpus[0]->gitlab_id;
                }

            }
        }

        return $repoId;
    }

    /**
     * @param $id
     * @param $index
     * @param $type
     * @return |null
     */
    public function getModelByTypeElasticserchIdAndIndex($id, $index, $type){
        $object = null;
        if($header == "corpus"){
            $object = DB::table('corpuses')->select('id')->where([
                    ['elasticsearch_id','=',$elasticSearchId],
                    ['elasticsearch_index','=',$elasticsearchIndex],
                ]
            )->get();
            $object = Corpus::where([["elasticsearch_id","=",$id],["elasticsearch_index","=",$index]])->get();
        }
        else if($header == "document"){
            $object = DB::table('documents')->select('id')->where([
                    ['elasticsearch_id','=',$elasticSearchId],
                    ['elasticsearch_index','=',$elasticsearchIndex],
                ]
            )->get();
        }
        else if($header == "annotation"){
            $object = DB::table('annotations')->select('id')->where([
                    ['elasticsearch_id','=',$elasticSearchId],
                    ['elasticsearch_index','=',$elasticsearchIndex],
                ]
            )->get();
        }
        return $object;
    }


    public function pickRandomCorpora(){
        $lastCorpora = array();
        $randomIndexes = array();

        $corpora = Corpus::all()->where("workflow_status", "==", "1");
        foreach ($corpora as $corpus){
            if(!array_key_exists($corpus->gitlab_id,$lastCorpora)){
                $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
            }
            else{
                if(floatval($corpus->publication_version) > floatval($lastCorpora[$corpus->gitlab_id])) {
                    $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
                }
            }
        }
        $rand_keys = array_rand($lastCorpora, 5);

        foreach ($rand_keys as $rand_key) {
            $corpusindex = $lastCorpora[$rand_key]['elasticsearch_index'];
            array_push($randomIndexes,$corpusindex);
            $postFix = substr($corpusindex,strpos($corpusindex,"_"));
            $documentindex = "document".$postFix;
            array_push($randomIndexes,$documentindex);
            $annotationindex = "annotation".$postFix;
            array_push($randomIndexes,$annotationindex);

        }

        return join(",",$randomIndexes);
    }

    public function pickLatestCorpora(){
        $lastCorpora = array();
        $indexes = array();

        $corpora = Corpus::all()->where("workflow_status", "==", "1");
        foreach ($corpora as $corpus){
            if(!array_key_exists($corpus->gitlab_id,$lastCorpora)){
                $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
            }
            else{
                if(floatval($corpus->publication_version) > floatval($lastCorpora[$corpus->gitlab_id])) {
                    $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
                }
            }
        }


        foreach ($lastCorpora as $key => $data) {
            $corpusindex = $lastCorpora[$key]['elasticsearch_index'];
            array_push($indexes,$corpusindex);
        }
        return join(",",$indexes);
    }

    public function pickLatestDocuments(){
        $lastCorpora = array();
        $indexes = array();

        $corpora = Corpus::all()->where("workflow_status", "==", "1");
        foreach ($corpora as $corpus){
            if(!array_key_exists($corpus->gitlab_id,$lastCorpora)){
                $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
            }
            else{
                if(floatval($corpus->publication_version) > floatval($lastCorpora[$corpus->gitlab_id])) {
                    $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
                }
            }
        }


        foreach ($lastCorpora as $key => $data) {
            $corpusindex = $lastCorpora[$key]['elasticsearch_index'];
            $postFix = substr($corpusindex,strpos($corpusindex,"_"));
            $documentindex = "document".$postFix;
            array_push($indexes,$documentindex);

        }
        return join(",",$indexes);
    }

    public function getLatestCorpusTagByCorpusAndGitlabId($corpusId, $gitLabId){
        $gitTag = '';
        $corpus = Corpus::where([["id","=",$corpusId],["gitlab_id","=",$gitLabId]])->get();
        if(count($corpus) > 0) {
            $gitTag = $corpus[0]->gitlab_version_tag;
        }
        return $gitTag;
    }

    public function pickLatestAnnotations(){
        $lastCorpora = array();
        $indexes = array();

        $corpora = Corpus::all()->where("workflow_status", "==", "1");
        foreach ($corpora as $corpus){
            if(!array_key_exists($corpus->gitlab_id,$lastCorpora)){
                $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
            }
            else{
                if(floatval($corpus->publication_version) > floatval($lastCorpora[$corpus->gitlab_id])) {
                    $lastCorpora[$corpus->gitlab_id] = array("version" => $corpus->publication_version, "elasticsearch_index" => $corpus->elasticsearch_index);
                }
            }
        }


        foreach ($lastCorpora as $key => $data) {
            $corpusindex = $lastCorpora[$key]['elasticsearch_index'];
            $postFix = substr($corpusindex,strpos($corpusindex,"_"));
            $annotationindex = "annotation".$postFix;
            array_push($indexes,$annotationindex);

        }
        return join(",",$indexes);
    }

    public function convert_date_format($strs)
    {
        $dates = array();
        //container to store possible date matches from strings
        $possible_dates = array();

        //prep months - long and short forms, to be used in matching
        $date_prefix_words = array('between', 'on', 'at', 'during', 'and');
        $months = array('January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December');
        $months_short = array_map(function($month) { return substr($month, 0, 3); }, $months);

        //iterate over and search strings - convert times like 2pm to 14:00, as, if they appear before the date, e.g. string 3, it doesn't get parsed
        foreach($strs as $str) {
            $str = preg_replace_callback('/\b\d{1,2}[ap]m\b/', function($time) { return date('H:i', strtotime($time[0])); }, $str);
            preg_match_all('/(?<=\b'.implode('\b( |:)|\b', $date_prefix_words).'\b( |:))(\d|am|pm| |,|\'|:|'.implode('|', $months).'|'.implode('|', $months_short).')+/i', $str, $matches);
            if (count($matches)) $possible_dates = array_merge($possible_dates, $matches[0]);
        }

        //output before and after results
        foreach($possible_dates as &$pd) {
            $pd = preg_replace('/, ?$/', '', $pd);
            array_push($dates,date('Y-m-d', strtotime($pd)));
            //echo '<p>Before: '.$pd.'<br />After: '.date('Y-m-d', strtotime($pd)).'</p>';
        }
        return $dates;
    }

    public function setMarkForDeletion($type,$id,$mark){
        switch($type) {
            case "document":
                $document = Document::find($id);
                $document->marked_for_deletion = $mark;
                $document->save();
                break;
            case "annotation":
                $annotation = Annotation::find($id);
                $annotation->marked_for_deletion = $mark;
                $annotation->save();
                break;
        }
    }

    public function strPosInArray($string, $array) {
        $foundIndex = -1;
        for($i = 0; $i < count($array); $i++){
            if(strpos(strtolower($array[$i]),strtolower($string)) > -1) {
                $foundIndex = $i;
                break;
            }
        }
        return $foundIndex;
    }

    public function importAnnisLinks() {
        $json = '{
    "externalenvironments": [
        {
            "corpus": "anselm",
            "version": "1.0",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://www.linguistics.rub.de/annis/annis3/anselm#_c=QW5zZWxt"
        },
        {
            "corpus": "deutsch-diachron-digital---referenzkorpus-altdeutsch_0-1",
            "version": "0.1",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/ddd"
        },
        {
            "corpus": "deutsch-diachron-digital---referenzkorpus-altdeutsch_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/ddd"
        },
        {
            "corpus": "deutsch-diachron-digital---referenzkorpus-altdeutsch_1-1",
            "version": "1.1",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/ddd"
        },
        {
            "corpus": "deutsch-diachron-digital---referenzkorpus-altdeutsch_0-1",
            "version": "0.1",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/ddd"
        },
        {
            "corpus": "deutsch-diachron-digital---referenzkorpus-altdeutsch_0-1",
            "version": "1.0",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/ddd"
        },
        {
            "corpus": "deutsch-diachron-digital---referenzkorpus-altdeutsch_0-1",
            "version": "1.1",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/ddd"
        },
        {
            "corpus": "deutsche-diachrone-baumbank_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "description": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=RERCLkFIRA"
        },
        {
            "corpus": "deutsche-diachrone-baumbank_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=RERCLkZOSEQ"
        },
        {
            "corpus": "deutsche-diachrone-baumbank_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=RERCLk1IRA"
        },
        {
            "corpus": "fuerstinnenkorrespondenz_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=ZnVlcnN0aW5uZW5rb3JyZXNwb25kZW56"
        },
        {
            "corpus": "fuerstinnenkorrespondenz_1-0",
            "version": "1.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=RnVlcnN0aW5uZW5rb3JyZXNwb25kZW56MS4x"
        },
        {
            "corpus": "fuerstinnenkorrespondenz_1-1",
            "version": "1.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=RnVlcnN0aW5uZW5rb3JyZXNwb25kZW56MS4x"
        },
        {
            "corpus": "heliand_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=YjQuaGVsaWFuZA"
        },
        {
            "corpus": "hipkon--historisches-predigtenkorpus-zum-nachfeld_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=SElQS09O"
        },
        {
            "corpus": "historische-syntax-des-jiddischen_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=SFNKLUJhc2VsZXJfTWFlc2VidWNo"
        },
        {
            "corpus": "historische-syntax-des-jiddischen_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=SFNKLUJyaWVmZQ"
        },
        {
            "corpus": "historische-syntax-des-jiddischen_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=SFNKLU1hZXNlX05pc3NpbWE"
        },
        {
            "corpus": "historische-syntax-des-jiddischen_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=SFNKLVNjaGV2ZXRfSmVodWRl"
        },
        {
            "corpus": "historische-syntax-des-jiddischen_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=SFNKLVZhcmlh"
        },
        {
            "corpus": "kasseler-junktionskorpus_1-1",
            "version": "1.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=S0FKVUs"
        },
        {
            "corpus": "ludolf_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=YjQubHVkb2xm"
        },
        {
            "corpus": "mercurius-baumbank_1-0",
            "version": "1.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=TWVyY3VyaXVz"
        },
        {
            "corpus": "mercurius-baumbank_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=TWVyY3VyaXVz"
        },
        {
            "corpus": "muspilli_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=YjQubXVzcGlsbGk"
        },
        {
            "corpus": "marchenkorpus-version_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=TWFlcmNoZW5rb3JwdXM"
        },
        {
            "corpus": "reference--of-middle-high-german--281050n1350-29_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://linguistics.rub.de/annis/annis3/REM/"
        },
        {
            "corpus": "ridges-ocr-version-1-0_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UmlkZ2VzT0NSXzEuMA"
        },
        {
            "corpus": "sachsiche-weltchronik_1-0",
            "version": "1.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=YjQuc2FlY2hzaWNoZV93ZWx0Y2hyb25paw"
        },
        {
            "corpus": "shenoute-a-22_1-1",
            "version": "1.1",
            "name": "ANNIS",
            "uri": "https://corpling.uis.georgetown.edu/annis/scriptorium#_c=c2hlbm91dGUuYTIy"
        },
        {
            "corpus": "tatian--of-deviating-examples_2-1",
            "version": "2.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=YjQudGF0aWFu"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "2.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UmlkZ2VzX0hlcmJvbG9neV9WZXJzaW9uXzIuMA"
        },
        {
            "corpus": "ridges-herbology-version_3-0",
            "version": "3.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uMy4w"
        },
        {
            "corpus": "ridges-herbology-version_4-0",
            "version": "4.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNC4w"
        },
        {
            "corpus": "ridges-herbology-version_4-1",
            "version": "4.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNC4x"
        },
        {
            "corpus": "ridges-herbology-version_5-0",
            "version": "5.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNS4w"
        },
        {
            "corpus": "ridges-herbology-version_6-0",
            "version": "6.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNi4w"
        },
        {
            "corpus": "ridges-herbology-version_7-0",
            "version": "7.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNy4w"
        },
        {
            "corpus": "ridges-herbology-version_8-0",
            "version": "8.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uOC4w"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "2.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UmlkZ2VzX0hlcmJvbG9neV9WZXJzaW9uXzIuMA"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "3.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uMy4w"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "4.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNC4w"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "4.1",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNC4x"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "5.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNS4w"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "6.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNi4w"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "7.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uNy4w"
        },
        {
            "corpus": "ridges-herbology-version_2-0",
            "version": "8.0",
            "name": "ANNIS",
            "uri": "https://korpling.org/annis3/#_c=UklER0VTX0hlcmJvbG9neV9WZXJzaW9uOC4w"
        }
    ]
}';
    DB::statement('truncate table external_environments');
    $externaldata = json_decode($json,true);
        foreach($externaldata['externalenvironments'] as $environment) {
            $res = DB::table('corpuses')->select('id')->where([
                    ['directory_path','=',$environment['corpus']],
                    ['publication_version','=',$environment['version']]
            ])->get();
            if(!empty($res[0])) {
                DB::table('external_environments')->insert(
                    [
                        'uid' => '1',
                        'corpus_id' => $res[0]->id,
                        'external_environment_name' => $environment['name'],
                        'external_environment_uri' => $environment['uri'],
                        'external_environment_description' => $environment['name']
                    ]
                );
            }
        }
    }

    public function checkForElasticsearchIds($corpusId) {
        $elasticsearchIds = array();

        $corpus = Corpus::findOrFail($corpusId);
        $corpus_elasticsearch_id = $corpus->elasticsearch_id;
        $document_elasticsearch_ids = array();
        $annotation_elasticsearch_ids = array();
        foreach ($corpus->documents()->get() as $document){
            $document_elasticsearch_ids[$document->document_id] = $document->elasticsearch_id;
        }
        foreach ($corpus->annotations()->get() as $annotation){
            $annotation_elasticsearch_ids[$annotation->annotation_id] = $annotation->elasticsearch_id;
        }

        $elasticsearchIds['corpus'] = $corpus_elasticsearch_id;
        $elasticsearchIds['documents'] = $document_elasticsearch_ids;
        $elasticsearchIds['annotations'] = $annotation_elasticsearch_ids;

        return $elasticsearchIds;
    }

}
